/* 2014-08-28: File changed by Sony Corporation */
/*
 *Copyright 2012 Sony Corporation
 *This program is free software; you can redistribute it and/or modify
 *it under the terms of the GNU General Public License, version 2, as
 *published by the Free Software Foundation.
 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *GNU General Public License for more details.
 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>

#include <asm/uaccess.h>

#include <mach/hardware.h>

#include "expcd_debug.h"
#include "expcd_ioctl.h"
#include "expcd_drv.h"
#include "expcd_local.h"
#define EXPCD_NAME                      "expcd"

//------------------------------------------------------------------------------
// char device related
extern int  fasync_setup(int fd, struct file * filp, int on);
extern void fasync_sighandler_call(int sig, int async);
extern void fasync_sighandler_pluginout(int sig, int async);
extern void fasync_sighandler_done(void);

extern int aldControl(int slotno, int color, int on);
extern int aldRedControl(int slotno, int on);
extern int aldRedControl_now(int slot, int on);
extern int aldTimingSet(int time);
extern int aldVersionGet(void);
extern int aldSelfDiagnose (int slot);
struct expcd_sig_param_struct {
	void * pointer;
	int size;
};
static struct expcd_sig_param_struct expcd_sig_param;
DECLARE_MUTEX(expcd_fasync_sigcall_lock);

void expcd_cdev_fasync_sigcall(int sig, void *param, int size)
{
	down(&expcd_fasync_sigcall_lock);
	expcd_sig_param.pointer = param;
	expcd_sig_param.size = size;
	fasync_sighandler_call(sig, 0);
   up(&expcd_fasync_sigcall_lock);
}
EXPORT_SYMBOL(expcd_cdev_fasync_sigcall);
void expcd_cdev_fasync_inout_sigcall(int sig, void *param, int size)
{
	expcd_sig_param.pointer = param;
	expcd_sig_param.size = size;
	fasync_sighandler_call(sig, 1);
}
EXPORT_SYMBOL(expcd_cdev_fasync_inout_sigcall);


static int expcd_chardev_open(struct inode * node, struct file * f)
{
	expcd_debugl("");
	return 0;
}

static int expcd_chardev_release(struct inode * node, struct file * f)
{
	expcd_errorl("If you see this message, cxd9225gg.bin has been killed!");
	(void)fasync_setup(-1, f, 0);
	return 0;
}

//XXX to be deleted, only for test.
static ssize_t expcd_chardev_write(struct file * f, const char __user * buf, size_t nbytes, loff_t * pos)
{
    int cmd = 0;
    char str[256] = {0};
    char *p = &str[0];

	nbytes = (nbytes > 256) ? 256 : nbytes;
    cmd = copy_from_user(str, buf, nbytes);

	if (str[0]>='0' && str[0]<='9') {
		int end = 0;
		for (;!end; p++) {
			switch (*p) {
			case '0' ... '9':
				cmd = 10*cmd+(*p-'0');
				break;
			default:
				end = 1;
				break;
			}
		}
	} else {
		cmd = -1;
	}

	if (cmd == 2) {
		int slotno = 0;
		expcdForcedDetach(slotno);
	}

	return nbytes;
}

extern unsigned int usbConnectStatusGet(void);
extern void cxd9225gg_udc_set_mass_flag(int flag);

static int expcd_chardev_ioctl(struct inode * node, struct file * f, unsigned int cmd, unsigned long arg)
{
	int ret = -1;
	void * __user puser = (void*)arg;
	int slotno, color, on;

	expcd_debugl("cmd=%u, arg=0x%08lx", cmd, arg);

	switch(cmd) {
	case EXPCD_IOCTL_SigHandlerDone:
		/* arg is ignored */

		/* Listeners must call this command to notify this driver when the
		 * signal handler has been finished. */
		fasync_sighandler_done();
		ret = 0;
		break;

	case EXPCD_IOCTL_GetSigParam:
		/* arg is a pointer to a memory region */
		if (NULL != expcd_sig_param.pointer && 0 != expcd_sig_param.size) {
			ret = copy_to_user(puser, expcd_sig_param.pointer, expcd_sig_param.size);
		} else {
			ret = -1; //no param data
		}
		break;

    /* added for udc vbus detection... */
	case EXPCD_IOCTL_GetConnectStatus:
		/* arg is a pointer to a memory region */
      ret = -1;      		
		if (puser) {
		
		    USBCONNECTSTATUS usbConnectStatus;
          usbConnectStatus = usbConnectStatusGet();
		    if(2==usbConnectStatus)
		    {
            expcd_errorl("==usbConnectStatus==2 !!\n");
		    }
			
			ret = copy_to_user(puser, &usbConnectStatus, sizeof(&usbConnectStatus));
		}

		break;
	case EXPCD_IOCTL_SetMassOn:
        ret = -1;
		if (puser) {
			int masson;

			ret = copy_from_user(&masson, puser, sizeof(masson));
			//if(sizeof(masson) != ret) { ret = -1; break; }
			//printk("==masson==[%d] !!\n",masson);
			cxd9225gg_udc_set_mass_flag(masson);

			ret = 0;
		}
		break;
	case EXPCD_IOCTL_CardStatusGet:
		/* arg is a pointer to (EXPCDCARDSTATUS x 2) */

		ret = -1;
		if (puser) {
			EXPCDCARDSTATUS status[EXPCD_SLOT_MAX];
			expcdCardStatusGet(status);
			ret = copy_to_user(puser, status, sizeof(status));
		}
		break;

	case EXPCD_IOCTL_CardPowCtrl:
		/* arg is a pointer to (EXPCDCARDPOWCTRL x 1) */

		if (puser) {
			EXPCDCARDPOWCTRL powctrl;

			ret = copy_from_user(&powctrl, puser, sizeof(powctrl));
			if(sizeof(powctrl) != ret) { ret = -1; break; }

			powctrl.result = expcdCardPowCtrl(powctrl.slotno, powctrl.pow);

			ret = copy_to_user(puser, &powctrl, sizeof(powctrl));
			if(sizeof(powctrl) != ret) { ret = -1; break; }

			ret = 0;
		}
		break;

	case EXPCD_IOCTL_ForcedDetach:
		expcdForcedDetach(arg);
		ret = 0;
		break;

	case EXPCD_IOCTL_ErrorForceDetach:
		expcdErrorForceDetach(arg);
		ret = 0;
		break;

	case EXPCD_IOCTL_aldControl:
		slotno = (arg & EXPCD_LED_SLOT_MASK)>>EXPCD_LED_SLOT_SHIFT;
		color = (arg & EXPCD_LED_COLOR_MASK)>>EXPCD_LED_COLOR_SHIFT;
		on = (arg & EXPCD_LED_ON_MASK)>>EXPCD_LED_ON_SHIFT;
		ret = aldControl(slotno, color, on);
		break;
		
	case EXPCD_IOCTL_aldRedControl:
		slotno = (arg & EXPCD_LED_SLOT_MASK)>>EXPCD_LED_SLOT_SHIFT;
		on = (arg & EXPCD_LED_ON_MASK)>>EXPCD_LED_ON_SHIFT;
		ret = aldRedControl(slotno, on);
		break;
	case EXPCD_IOCTL_aldRedControl_now:
		slotno = (arg & EXPCD_LED_SLOT_MASK)>>EXPCD_LED_SLOT_SHIFT;
		on = (arg & EXPCD_LED_ON_MASK)>>EXPCD_LED_ON_SHIFT;
		ret = aldRedControl_now(slotno, on);
		break;
	case EXPCD_IOCTL_aldTimingSet:
		ret = aldTimingSet(arg);
		break;
		
	case EXPCD_IOCTL_aldVersionGet:
		ret = aldVersionGet();
		break;

	case EXPCD_IOCTL_aldSelfDiagnose:
		slotno = (arg & EXPCD_LED_SLOT_MASK)>>EXPCD_LED_SLOT_SHIFT;
		ret = aldSelfDiagnose(slotno);
		break;
	case EXPCD_IOCTL_expcd_slot_set_led:
		slotno = (arg & EXPCD_LED_SLOT_MASK)>>EXPCD_LED_SLOT_SHIFT;
		color = (arg & EXPCD_LED_COLOR_MASK)>>EXPCD_LED_COLOR_SHIFT;
		on = (arg & EXPCD_LED_ON_MASK)>>EXPCD_LED_ON_SHIFT;
		expcd_slot_set_led(slotno, on, color);
		ret =0;
		break;
	default:
		expcd_errorl("Invalid ioctl command %d", cmd);
		break;
	}
	expcd_debugl("cmd=%u, arg=0x%08lx end", cmd, arg);
	return ret;
}

static int expcd_major = CXD9225GG_MAJOR_ID_EXPCD;
static struct cdev expcd_cdev;
static struct file_operations expcd_fops ={
	.owner   = THIS_MODULE,
	.open    = expcd_chardev_open,
	.release = expcd_chardev_release,
	.ioctl   = expcd_chardev_ioctl,
	.fasync  = fasync_setup,
	//XXX just for test
	.write   = expcd_chardev_write,
};

int expcd_cdev_register(void)
{
	int ret = -1;
	dev_t dev_id;

	/* register and init char device */
    dev_id = MKDEV(expcd_major, 0);
    ret = register_chrdev_region(dev_id, 1, EXPCD_NAME);
    if (ret) {
        expcd_errorl("register_chrdev_region() failed !!");
        return ret;
    }
	cdev_init(&expcd_cdev, &expcd_fops);
	ret = cdev_add(&expcd_cdev, dev_id, 1);
	if (ret) {
        expcd_errorl("cdev_add() failed !!");
        return ret;
	}

	return 0;
}

void expcd_cdev_unregister(void)
{
	dev_t dev_id;

	dev_id = MKDEV(expcd_major, 0);	
	cdev_del(&expcd_cdev);
	unregister_chrdev_region(dev_id, 1);
}

