/* 2014-08-28: File changed by Sony Corporation */
/*
 *Copyright 2012 Sony Corporation
 *This program is free software; you can redistribute it and/or modify
 *it under the terms of the GNU General Public License, version 2, as
 *published by the Free Software Foundation.
 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *GNU General Public License for more details.
 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef _EXPCD_DEBUG_H
#define _EXPCD_DEBUG_H

#include <linux/kernel.h>
#include <linux/sched.h>

#define PROMPT  "[EXPCD] "

//==============================================================================
/*
 * Define a static variable in header, to make sure each .c file be able to
 * store its own basename. I store the basename instead of parse it every
 * time, just for performance.
*/
static __used char * __f_ = NULL;

/*
 * Parse and get the basename from a full-path filename.
 * E.g:
 * given "/where/you/put/yourfile.c", return "yourfile.c"
 * given "yourfiletoo.c", return "yourfiletoo.c"
*/
static __used char * __basename_(char * str) {
	char *p;
	for(p=str; *p && ((*p=='/')? (str=p+1, 1): 1); p++);
	return str;
}

#define cmn_printk(level, format, arg...) \
	printk(level PROMPT "%4d | %s %s() @%d :> " format, current->pid, \
		(__f_? __f_: (__f_=__basename_(__FILE__))), \
		__FUNCTION__, __LINE__, ## arg)
//==============================================================================

#define CONFIG_EXPCD_DEBUG
#ifdef  CONFIG_EXPCD_DEBUG

#define expcd_debugl(format, arg...) \
	cmn_printk(KERN_DEBUG, format "\n" , ## arg)/*KERN_INFO*/

#define expcd_infol(format, arg...) \
	cmn_printk(KERN_NOTICE, format "\n" , ## arg)/*KERN_NOTICE*/

#define expcd_warningl(format, arg...) \
	cmn_printk(KERN_WARNING, format "\n" , ## arg)

#define expcd_errorl(format, arg...) \
	cmn_printk(KERN_ERR, format "\n" , ## arg)

#else /* CONFIG_EXPCD_DEBUG */

#define expcd_debugl(format, arg...)

#define expcd_infol(format, arg...)

#define expcd_warningl(format, arg...)

#define expcd_errorl(format, arg...) \
	cmn_printk(KERN_ERR, format "\n" , ## arg)

#endif /* CONFIG_EXPCD_DEBUG */

#endif /* _EXPCD_DEBUG_H */
