/* 2014-08-28: File changed by Sony Corporation */
/*
 *Copyright 2012 Sony Corporation
 *This program is free software; you can redistribute it and/or modify
 *it under the terms of the GNU General Public License, version 2, as
 *published by the Free Software Foundation.
 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *GNU General Public License for more details.
 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <linux/module.h>

#include <mach/gpio.h>
#include <asm/io.h>

#include "expcd_debug.h"
#include "expcd_local.h"

/*
* Get bonding Port of the given Slot Number.
* Note: If in Exchange Mode, Port != Slot
*/
#define EXPCD_SLOT_TO_PORT(slot) ((0)?(((slot)+1)&1):(slot))

/* get gpio status of ExpressCard. return 0 on success. */
int expcd_slot_get_status(unsigned int * pstatus)
{
	unsigned int data;
	*pstatus = 0;

	/* PCI-Exress Card #0 */
	data = cxd9225gg_get_gpio_datain(CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,1));
	*pstatus |= (data << EXPCD_GPIO_SLOT0_CPPE_SHIFT);
	/* USB Card #0 */
	data = cxd9225gg_get_gpio_datain(CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,0));
	*pstatus |= (data << EXPCD_GPIO_SLOT0_CPUSB_SHIFT);
	/* STBY #0 */
	data = cxd9225gg_get_gpio_datain(CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_10,1));
	*pstatus |= (data << EXPCD_GPIO_SLOT0_STBY_SHIFT);
	/* SHDN #0 */
	data = cxd9225gg_get_gpio_datain(CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_10,2));
	*pstatus |= (data << EXPCD_GPIO_SLOT0_SHDN_SHIFT);
	/* PCI-Exress Card #1 */
	data = cxd9225gg_get_gpio_datain(CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,5));
	*pstatus |= (data << EXPCD_GPIO_SLOT1_CPPE_SHIFT);
	/* USB Card #1 */
	data = cxd9225gg_get_gpio_datain(CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,4));
	*pstatus |= (data << EXPCD_GPIO_SLOT1_CPUSB_SHIFT);
	/* STBY #1 */
	data = cxd9225gg_get_gpio_datain(CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_11,1));
	*pstatus |= (data << EXPCD_GPIO_SLOT1_STBY_SHIFT);
	/* SHDN #1 */
	data = cxd9225gg_get_gpio_datain(CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_11,2));
	*pstatus |= (data << EXPCD_GPIO_SLOT1_SHDN_SHIFT);
	/* OVC #0 */
	data = cxd9225gg_get_gpio_datain(CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,2));
	*pstatus |= (data << EXPCD_GPIO_SLOT0_OC_SHIFT);
	/* OVC #1 */
	data = cxd9225gg_get_gpio_datain(CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,6));
	*pstatus |= (data << EXPCD_GPIO_SLOT1_OC_SHIFT);
	return 0;
}

/* iface : 0 is USB, 1 is PCIE
 * return 1 if card is in the slot */
int expcd_slot_check(int slotno, int iface)
{
	int pin = 0;
	if(slotno == 0) {
		pin = !cxd9225gg_get_gpio_datain(EXPCD_INT_S0_USB);
		pin |= !cxd9225gg_get_gpio_datain(EXPCD_INT_S0_PCIE);
	} else {
		pin = !cxd9225gg_get_gpio_datain(EXPCD_INT_S1_USB);
		pin |= !cxd9225gg_get_gpio_datain(EXPCD_INT_S1_PCIE);
	}
	
	return pin;
}
EXPORT_SYMBOL(expcd_slot_check);

/* pwr_up : 1 means power up, 0 means power down. */
void expcd_slot_set_power(int slotno, int pwr_up)
{
	int gpio_port = (slotno==0) ? CXD9225GG_GPIO_PORT_10 : CXD9225GG_GPIO_PORT_11;
	cxd9225gg_set_gpio_direction(CXD9225GG_GPIO_ID(gpio_port, 2), 0);
	cxd9225gg_set_gpio_dataout(CXD9225GG_GPIO_ID(gpio_port,2),pwr_up);
}

/* return 1 in case the power is on, 0 in case the power is off. */
int expcd_slot_get_power(int slotno)
{
	int gpio_port = (slotno==0) ? CXD9225GG_GPIO_PORT_10 : CXD9225GG_GPIO_PORT_11;
	//: do not change gpio direction for read !!
	//cxd9225gg_set_gpio_direction(CXD9225GG_GPIO_ID(gpio_port, 2), 1);
	return cxd9225gg_get_gpio_datain(CXD9225GG_GPIO_ID(gpio_port,2));
}
EXPORT_SYMBOL(expcd_slot_get_power);

static __used void print_gpio(void)
{
	//GPIO_10
	unsigned int bar = CXD9225GG_APB_REG_VIRT_BASE
				+CXD9225GG_GPIO_REG_OFFSET_OF_APB+0x100;
	unsigned int data = 0;
	int i;
	for (i=0; i<0x80; i+=0x20) {
		printk("0x%02x  :", i);
		data = readl(bar+i) & 0xff;
		printk(" %1d%1d%1d%1d%1d%1d%1d%1d,",
			(data>>7)&1, (data>>6)&1, (data>>5)&1, (data>>4)&1,
			(data>>3)&1, (data>>2)&1, (data>>1)&1, (data>>0)&1);
		data = readl(bar+i+4) & 0xff;
		printk(" %1d%1d%1d%1d%1d%1d%1d%1d,",
			(data>>7)&1, (data>>6)&1, (data>>5)&1, (data>>4)&1,
			(data>>3)&1, (data>>2)&1, (data>>1)&1, (data>>0)&1);
		data = readl(bar+i+0x10) & 0xff;
		printk(" %1d%1d%1d%1d%1d%1d%1d%1d\n",
			(data>>7)&1, (data>>6)&1, (data>>5)&1, (data>>4)&1,
			(data>>3)&1, (data>>2)&1, (data>>1)&1, (data>>0)&1);
	}
}

/* set slot power on/off, and attach/detach the device */
void expcd_power_ctrl(int slotno, int is_powon, int is_pcie)
{
	int pow = is_powon ? 1 : 0;
	if (is_pcie) {
		extern int cxd9225gg_pcie_port_attach(int);
		extern int cxd9225gg_pcie_port_detach(int);
		
		if (pow) {
			/* first: power up the slot */
			expcd_slot_set_power(slotno, pow);
			/* then: attach pcie device */
			cxd9225gg_pcie_port_attach(EXPCD_SLOT_TO_PORT(slotno));
		} else {
			/* first: detach pcie device */
			cxd9225gg_pcie_port_detach(EXPCD_SLOT_TO_PORT(slotno));
			/* then: power down the slot */
			expcd_slot_set_power(slotno, pow);
		}
	} else { /* is_usb */
		expcd_slot_set_power(slotno, pow);
	}
}

#define EXPCD_LED_SLOT_0_GREEN          CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_10,4)
#define EXPCD_LED_SLOT_0_RED            CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_10,3)
#define EXPCD_LED_SLOT_1_GREEN          CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_11,4)
#define EXPCD_LED_SLOT_1_RED            CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_11,3)

enum {
	EXPCD_LED_CLR_GREEN,
	EXPCD_LED_CLR_RED,
	EXPCD_LED_CLR_YELLOW,
};
/* 
 * Set expcd LED on or off.
 * Each slot has 2 LEDs, a Red and a Green, which make up 3 colors:
 * Red, Green and Yellow(Red+Green).
 *
 * Params:
 * slotno (0,1)  - slot number.
 * on (0,1)      - if it is 0, set led off, so color is ignored.
 * color (0,1,2) - 0=GREED, 1=RED, 2=YELLOW
*/
void expcd_slot_set_led(int slotno, int on, int color)
{
	cxd9225gg_gpio_pin_name_t gpio_red;
	cxd9225gg_gpio_pin_name_t gpio_grn;

	unsigned int led_polarity = 0;
	int data_red = 0;
	int data_grn = 0;

	/* decide led gpio */
	gpio_red = (slotno==0)? EXPCD_LED_SLOT_0_RED  : EXPCD_LED_SLOT_1_RED;
	gpio_grn = (slotno==0)? EXPCD_LED_SLOT_0_GREEN: EXPCD_LED_SLOT_1_GREEN;

	/* decide led state */
	if (on) {
		if (color==EXPCD_LED_CLR_YELLOW) {
			data_red = 0;
			data_grn = 0;
		} else
		if (color==EXPCD_LED_CLR_RED){
			data_red = 0;
			data_grn = 1;
		} else
		if (color==EXPCD_LED_CLR_GREEN) {
			data_red = 1;
			data_grn = 0;
		}
	} else {
		data_red = 1;
		data_grn = 1;
	}


	led_polarity = 0;
	if (led_polarity) {
		data_red = (~data_red)&0x1;
		data_grn = (~data_grn)&0x1;
	}

	/* really set the leds */
	cxd9225gg_set_gpio_direction(gpio_red, 0);
	cxd9225gg_set_gpio_dataout(gpio_red, data_red);
	cxd9225gg_set_gpio_direction(gpio_grn, 0);
	cxd9225gg_set_gpio_dataout(gpio_grn, data_grn);
}
EXPORT_SYMBOL(expcd_slot_set_led);

