/* 2014-08-28: File changed by Sony Corporation */
/*
 *Copyright 2012 Sony Corporation
 *This program is free software; you can redistribute it and/or modify
 *it under the terms of the GNU General Public License, version 2, as
 *published by the Free Software Foundation.
 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *GNU General Public License for more details.
 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include<linux/module.h>
#include<linux/slab.h>
#include<linux/timer.h> //timer_list
#include<linux/jiffies.h>
#include<linux/spinlock.h>
#include"expcd_debug.h"

#define EXPCD_LED_SUCCESS	0x00000000
#define EXPCD_LED_ERROR		0xc0000000
#define DELAYED_TIME 		(2*HZ)
extern void expcd_slot_set_led(int slotno, int on, int color);

/*	xxx = red1 green red2,
	o=on,
	f=off,
	d=delay off for red LED or delay on for green LED

	red1 is the status of red LED controlled by aldControl()
	red2 is the status of red LED controlled by aldRedControl() and aldRedControl_now()
*/
enum status {
	fff = 0,
	off,
	fof,
	ffo,
	dff,
	odf,
	ofo,
	fdo,
	ffd,
	ddf,
	dfo,
	odo,
	ofd,
	fdd,
	ddo,
	dfd,
	odd,
	ddd,
};
enum {
	slot0 = 0,
	slot1,
};
enum {
	green = 0,
	red,
	yellow,
};
enum {
	loff = 0,
	lon = 1,
};
enum command {
	red1_on = 0,
	red1_off,
	grn_on,
	grn_off,
	red2_on,
	red2_off,
};
static enum status slot0_status = fff;
static enum status slot1_status = fff;
static struct timer_list * slot0_red1_timer;
static struct timer_list * slot0_red2_timer;
static struct timer_list * slot1_red1_timer;
static struct timer_list * slot1_red2_timer;
static struct timer_list * slot0_diagnose_timer;
static struct timer_list * slot1_diagnose_timer;
static unsigned long delayedtime = DELAYED_TIME;
static spinlock_t slot0_lock;
static spinlock_t slot1_lock;
void diagnose_time_up(unsigned long time)
{
	unsigned long j;
	switch(time)
	{
	case 3:
		expcd_slot_set_led(slot0, lon, green);
		time--;
		break;
	case 2:
		expcd_slot_set_led(slot0, lon, yellow);
		time--;
		break;
	case 1:
		expcd_slot_set_led(slot0, loff, yellow);
		time--;
		break;
	case 7:
		expcd_slot_set_led(slot1, lon, green);
		time--;
		break;
	case 6:
		expcd_slot_set_led(slot1, lon, yellow);
		time--;
		break;
	case 5:
		expcd_slot_set_led(slot1, loff, yellow);
		time--;
		break;
	}
	if(time == 0 || time == 4)
		return ;
	j = jiffies;
	if(time > 4 && time < 8)
	{
		slot1_diagnose_timer->expires = j + HZ;
		slot1_diagnose_timer->data = time;
		add_timer(slot1_diagnose_timer);
	}
	else if(time >0 && time <4)
	{
		slot0_diagnose_timer->expires = j + HZ;
		slot0_diagnose_timer->data = time;
		add_timer(slot0_diagnose_timer);
	}
	return ;
}
void time_up(unsigned long slot);
/*
	slot = 0x00, 0x01, 0x02, 0x03
	0x00 : slot0, red1_timer
	0x01 : slot1, red1_timer
	0x02 : slot0, red2_timer
	0x03 : slot1, red2_timer
*/

int expcd_led_init(void)
{
	int result = 0;
	if(!slot0_red1_timer)
	{
		slot0_red1_timer = kmalloc(sizeof(struct timer_list), GFP_KERNEL);
		if(!slot0_red1_timer)
		{
			result = -ENOMEM;
			goto slot0red1_fail;
		}
		init_timer(slot0_red1_timer);
		slot0_red1_timer->function = time_up;
		slot0_red1_timer->data = 0x00;
	}
	if(!slot0_red2_timer)
	{
		slot0_red2_timer = kmalloc(sizeof(struct timer_list), GFP_KERNEL);
		if(!slot0_red2_timer)
		{
			result = -ENOMEM;
			goto slot0red2_fail;
		}
		init_timer(slot0_red2_timer);
		slot0_red2_timer->function = time_up;
		slot0_red2_timer->data = 0x02;
	}
	if(!slot1_red1_timer)
	{
		slot1_red1_timer = kmalloc(sizeof(struct timer_list), GFP_KERNEL);
		if(!slot1_red1_timer)
		{
			result = -ENOMEM;
			goto slot1red1_fail;
		}
		init_timer(slot1_red1_timer);
		slot1_red1_timer->function = time_up;
		slot1_red1_timer->data = 0x01;
	}
	if(!slot1_red2_timer)
	{
		slot1_red2_timer = kmalloc(sizeof(struct timer_list), GFP_KERNEL);
		if(!slot1_red2_timer)
		{
			result = -ENOMEM;
			goto slot1red2_fail;
		}
		init_timer(slot1_red2_timer);
		slot1_red2_timer->function = time_up;
		slot1_red2_timer->data = 0x03;
	}
	if(!slot0_diagnose_timer)
	{
		slot0_diagnose_timer = kmalloc(sizeof(struct timer_list), GFP_KERNEL);
		if(!slot0_diagnose_timer)
		{
			result = -ENOMEM;
			goto slot0diag_fail;
		}
		init_timer(slot0_diagnose_timer);
		slot0_diagnose_timer->function = diagnose_time_up;
		slot0_diagnose_timer->data = 3;
	}
	if(!slot1_diagnose_timer)
	{
		slot1_diagnose_timer = kmalloc(sizeof(struct timer_list), GFP_KERNEL);
		if(!slot1_diagnose_timer)
		{
			result = -ENOMEM;
			goto slot1diag_fail;
		}
		init_timer(slot1_diagnose_timer);
		slot1_diagnose_timer->function = diagnose_time_up;
		slot1_diagnose_timer->data = 7;
	}
	spin_lock_init(&slot0_lock);
	spin_lock_init(&slot1_lock);
	return result;
	slot1diag_fail:
		kfree(slot0_diagnose_timer);
	slot0diag_fail:
		kfree(slot1_red2_timer);
	slot1red2_fail:
		kfree(slot1_red1_timer);
	slot1red1_fail:
		kfree(slot0_red2_timer);
	slot0red2_fail:
		kfree(slot0_red1_timer);
	slot0red1_fail:
		return result;
}
EXPORT_SYMBOL(expcd_led_init);

void expcd_led_exit(void)
{
	kfree(slot0_red1_timer);
	kfree(slot0_red2_timer);
	kfree(slot1_red1_timer);
	kfree(slot1_red2_timer);
	kfree(slot0_diagnose_timer);
	kfree(slot1_diagnose_timer);
	return ;
}
EXPORT_SYMBOL(expcd_led_exit);

void time_up(unsigned long slot)
{
	if((slot & 0x01) == 0)
	{
		spin_lock(&slot0_lock);
		switch(slot0_status)
		{
		case dff:
		case ffd:
			expcd_slot_set_led(slot0, loff, red);
			slot0_status = fff;
			break;
		case ddf:
		case fdd:
			expcd_slot_set_led(slot0, lon, green);
			slot0_status = fof;
			break;
		case dfo:
			slot0_status = ffo;
			break;
		case ofd:
			slot0_status = off;
			break;
		case ddo:
			slot0_status = fdo;
			break;
		case dfd:
			if(slot & 0x02)
			{
				slot0_status = dff;
			}
			else
			{
				slot0_status = ffd;
			}
			break;
		case odd:
			slot0_status = odf;
			break;
		case ddd:
			if(slot & 0x02)
			{
				slot0_status = ddf;
			}
			else
			{
				slot0_status = fdd;
			}
			break;
		default :

			expcd_warningl("expcd_led : wrong timer");
		}
		spin_unlock(&slot0_lock);
	}
	else if((slot & 0x01) == 1)
	{
		spin_lock(&slot1_lock);
		switch(slot1_status)
		{
		case dff:
		case ffd:
			expcd_slot_set_led(slot1, loff, red);
			slot1_status = fff;
			break;
		case ddf:
		case fdd:
			expcd_slot_set_led(slot1, lon, green);
			slot1_status = fof;
			break;
		case dfo:
			slot1_status = ffo;
			break;
		case ofd:
			slot1_status = off;
			break;
		case ddo:
			slot1_status = fdo;
			break;
		case dfd:
			if(slot & 0x02)
			{
				slot1_status = dff;
			}
			else
			{
				slot1_status = ffd;
			}
			break;
		case odd:
			slot1_status = odf;
			break;
		case ddd:
			if(slot & 0x02)
			{
				slot1_status = ddf;
			}
			else
			{
				slot1_status = fdd;
			}
			break;
		default :
			expcd_warningl("expcd_led : wrong timer");
		}
		spin_unlock(&slot1_lock);
	}
	else
		expcd_warningl("expcd_led timer : arg error");
	return ;
}
void addledtimer(struct timer_list * timer)
{
	unsigned long j = jiffies;
	timer->expires = j + delayedtime;
	add_timer(timer);
	return;
}

void resetledtimer(struct timer_list * timer)
{
	del_timer(timer);
	addledtimer(timer);
	return;
}
/*
slot(0, 1)	slot number
color(0, 1, 2)	0=GREEN, 1=RED, 2=YELLOW
on(0, 1)	0=off, 1=on
*/

int aldControl(int slot, int color, int on)
{
	enum command cmd;
	if(on == 1)
		if(color == 1)
			cmd = red1_on;
		else if(color == 0)
			cmd = grn_on;
		else
			return EXPCD_LED_ERROR;
	else if(on == 0)
		if(color == 0)
			cmd = grn_off;
		else if(color == 1)
			cmd = red1_off;
		else
			return EXPCD_LED_ERROR;
	else
		return EXPCD_LED_ERROR;
	if(slot == slot0)
	{
		spin_lock(&slot0_lock);
		switch(slot0_status)
		{
		case fff:
			switch(cmd)
			{
			case red1_on:
				expcd_slot_set_led(slot0, lon, red);
				slot0_status = off;
				break;
			case grn_on:
				expcd_slot_set_led(slot0, lon, green);
				slot0_status = fof;
				break;
			default :
				break;
			}
			break;
		case off:
			switch(cmd)
			{
			case red1_off:
				addledtimer(slot0_red1_timer);
				slot0_status = dff;
				break;
			case grn_on:
				slot0_status = odf;
				break;
			default :
				break;
			}
			break;
		case fof:
			switch(cmd)
			{
			case red1_on:
				expcd_slot_set_led(slot0, lon, red);
				slot0_status = odf;
				break;
			case grn_off:
				expcd_slot_set_led(slot0, loff, green);
				slot0_status = fff;
				break;
			default :
				break;
			}
			break;
		case ffo:
			switch(cmd)
			{
			case red1_on:
				slot0_status = ofo;
				break;
			case grn_on:
				slot0_status = fdo;
				break;
			default:
				break;
			}
			break;
		case dff:
			switch(cmd)
			{
			case red1_on:
				del_timer(slot0_red1_timer);
				slot0_status = off;
				break;
			case red1_off:
				resetledtimer(slot0_red1_timer);
				break;
			case grn_on:
				slot0_status = ddf;
				break;
			default:
				break;
			}
			break;
		case odf:
			switch(cmd)
			{
			case red1_off:
				addledtimer(slot0_red1_timer);
				slot0_status = ddf;
				break;
			case grn_off:
				slot0_status = off;
				break;
			default:
				break;
			}
			break;
		case ofo:
			switch(cmd)
			{
			case red1_off:
				addledtimer(slot0_red1_timer);
				slot0_status = dfo;
				break;
			case grn_on:
				slot0_status = odo;
				break;
			default:
				break;
			}
			break;
		case fdo:
			switch(cmd)
			{
			case red1_on:
				slot0_status = odo;
				break;
			case grn_off:
				slot0_status = ffo;
				break;
			default:
				break;
			}
			break;
		case ffd:
			switch(cmd)
			{
			case red1_on:
				slot0_status = ofd;
				break;
			case grn_on:
				slot0_status = fdd;
				break;
			default:
				break;
			}
			break;
		case ddf:
			switch(cmd)
			{
			case red1_on:
				del_timer(slot0_red1_timer);
				slot0_status = odf;
				break;
			case red1_off:
				resetledtimer(slot0_red1_timer);
				break;
			case grn_off:
				slot0_status = dff;
				break;
			default:
				break;
			}
			break;
		case dfo:
			switch(cmd)
			{
			case red1_on:
				del_timer(slot0_red1_timer);
				slot0_status = ofo;
				break;
			case red1_off:
				resetledtimer(slot0_red1_timer);
				break;
			case grn_on:
				slot0_status = ddo;
				break;
			default:
				break;
			}
			break;
		case odo:
			switch(cmd)
			{
			case red1_off:
				addledtimer(slot0_red1_timer);
				slot0_status = ddo;
				break;
			case grn_off:
				slot0_status = ofo;
				break;
			default:
				break;
			}
			break;
		case ofd:
			switch(cmd)
			{
			case red1_off:
				addledtimer(slot0_red1_timer);
				slot0_status = dfd;
				break;
			case grn_on:
				slot0_status = odd;
				break;
			default:
				break;
			}
			break;
		case fdd:
			switch(cmd)
			{
			case red1_on:
				slot0_status = odd;
				break;
			case grn_off:
				slot0_status = ffd;
				break;
			default:
				break;
			}
			break;
		case ddo:
			switch(cmd)
			{
			case red1_on:
				del_timer(slot0_red1_timer);
				slot0_status = odo;
				break;
			case red1_off:
				resetledtimer(slot0_red1_timer);
				break;
			case grn_off:
				slot0_status = dfo;
				break;
			default:
				break;
			}
			break;
		case dfd:
			switch(cmd)
			{
			case red1_on:
				del_timer(slot0_red1_timer);
				slot0_status = ofd;
				break;
			case red1_off:
				resetledtimer(slot0_red1_timer);
				break;
			case grn_on:
				slot0_status = ddd;
				break;
			default:
				break;
			}
			break;
		case odd:
			switch(cmd)
			{
			case red1_off:
				addledtimer(slot0_red1_timer);
				slot0_status = ddd;
				break;
			case grn_off:
				slot0_status = ofd;
				break;
			default:
				break;
			}
			break;
		case ddd:
			switch(cmd)
			{
			case red1_on:
				del_timer(slot0_red1_timer);
				slot0_status = odd;
				break;
			case red1_off:
				resetledtimer(slot0_red1_timer);
				break;
			case grn_off:
				slot0_status = dfd;
				break;
			default:
				break;
			}
			break;
		}
		spin_unlock(&slot0_lock);
	}
	else if(slot == slot1)
	{
		spin_lock(&slot1_lock);
		switch(slot1_status)
		{
		case fff:
			switch(cmd)
			{
			case red1_on:
				expcd_slot_set_led(slot1, lon, red);
				slot1_status = off;
				break;
			case grn_on:
				expcd_slot_set_led(slot1, lon, green);
				slot1_status = fof;
				break;
			default :
				break;
			}
			break;
		case off:
			switch(cmd)
			{
			case red1_off:
				addledtimer(slot1_red1_timer);
				slot1_status = dff;
				break;
			case grn_on:
				slot1_status = odf;
				break;
			default :
				break;
			}
			break;
		case fof:
			switch(cmd)
			{
			case red1_on:
				expcd_slot_set_led(slot1, lon, red);
				slot1_status = odf;
				break;
			case grn_off:
				expcd_slot_set_led(slot1, loff, green);
				slot1_status = fff;
				break;
			default :
				break;
			}
			break;
		case ffo:
			switch(cmd)
			{
			case red1_on:
				slot1_status = ofo;
				break;
			case grn_on:
				slot1_status = fdo;
				break;
			default:
				break;
			}
			break;
		case dff:
			switch(cmd)
			{
			case red1_on:
				del_timer(slot1_red1_timer);
				slot1_status = off;
				break;
			case red1_off:
				resetledtimer(slot1_red1_timer);
				break;
			case grn_on:
				slot1_status = ddf;
				break;
			default:
				break;
			}
			break;
		case odf:
			switch(cmd)
			{
			case red1_off:
				addledtimer(slot1_red1_timer);
				slot1_status = ddf;
				break;
			case grn_off:
				slot1_status = off;
				break;
			default:
				break;
			}
			break;
		case ofo:
			switch(cmd)
			{
			case red1_off:
				addledtimer(slot1_red1_timer);
				slot1_status = dfo;
				break;
			case grn_on:
				slot1_status = odo;
				break;
			default:
				break;
			}
			break;
		case fdo:
			switch(cmd)
			{
			case red1_on:
				slot1_status = odo;
				break;
			case grn_off:
				slot1_status = ffo;
				break;
			default:
				break;
			}
			break;
		case ffd:
			switch(cmd)
			{
			case red1_on:
				slot1_status = ofd;
				break;
			case grn_on:
				slot1_status = fdd;
				break;
			default:
				break;
			}
			break;
		case ddf:
			switch(cmd)
			{
			case red1_on:
				del_timer(slot1_red1_timer);
				slot1_status = odf;
				break;
			case red1_off:
				resetledtimer(slot1_red1_timer);
				break;
			case grn_off:
				slot1_status = dff;
				break;
			default:
				break;
			}
			break;
		case dfo:
			switch(cmd)
			{
			case red1_on:
				del_timer(slot1_red1_timer);
				slot1_status = ofo;
				break;
			case red1_off:
				resetledtimer(slot1_red1_timer);
				break;
			case grn_on:
				slot1_status = ddo;
				break;
			default:
				break;
			}
			break;
		case odo:
			switch(cmd)
			{
			case red1_off:
				addledtimer(slot1_red1_timer);
				slot1_status = ddo;
				break;
			case grn_off:
				slot1_status = ofo;
				break;
			default:
				break;
			}
			break;
		case ofd:
			switch(cmd)
			{
			case red1_off:
				addledtimer(slot1_red1_timer);
				slot1_status = dfd;
				break;
			case grn_on:
				slot1_status = odd;
				break;
			default:
				break;
			}
			break;
		case fdd:
			switch(cmd)
			{
			case red1_on:
				slot1_status = odd;
				break;
			case grn_off:
				slot1_status = ffd;
				break;
			default:
				break;
			}
			break;
		case ddo:
			switch(cmd)
			{
			case red1_on:
				del_timer(slot1_red1_timer);
				slot1_status = odo;
				break;
			case red1_off:
				resetledtimer(slot1_red1_timer);
				break;
			case grn_off:
				slot1_status = dfo;
				break;
			default:
				break;
			}
			break;
		case dfd:
			switch(cmd)
			{
			case red1_on:
				del_timer(slot1_red1_timer);
				slot1_status = ofd;
				break;
			case red1_off:
				resetledtimer(slot1_red1_timer);
				break;
			case grn_on:
				slot1_status = ddd;
				break;
			default:
				break;
			}
			break;
		case odd:
			switch(cmd)
			{
			case red1_off:
				addledtimer(slot1_red1_timer);
				slot1_status = ddd;
				break;
			case grn_off:
				slot1_status = ofd;
				break;
			default:
				break;
			}
			break;
		case ddd:
			switch(cmd)
			{
			case red1_on:
				del_timer(slot1_red1_timer);
				slot1_status = odd;
				break;
			case red1_off:
				resetledtimer(slot1_red1_timer);
				break;
			case grn_off:
				slot1_status = dfd;
				break;
			default:
				break;
			}
			break;
		}
		spin_unlock(&slot1_lock);
	}
	else
		return EXPCD_LED_ERROR;
	return EXPCD_LED_SUCCESS;
}

EXPORT_SYMBOL(aldControl);


int aldRedControl(int slot, int on)
{
	enum command cmd;
	if(on == 1)
		cmd = red2_on;
	else if(on == 0)
		cmd = red2_off;
	else
		return EXPCD_LED_ERROR;
	if(slot == slot0)
	{
		spin_lock(&slot0_lock);
		switch(slot0_status)
		{
		case fff:
			switch(cmd)
			{
			case red2_on:
				expcd_slot_set_led(slot0, lon, red);
				slot0_status = ffo;
				break;
			default :
				break;
			}
			break;
		case off:
			switch(cmd)
			{
			case red2_on:
				slot0_status = ofo;
				break;
			default :
				break;
			}
			break;
		case fof:
			switch(cmd)
			{
			case red2_on:
				expcd_slot_set_led(slot0, lon, red);
				slot0_status = fdo;
				break;
			default :
				break;
			}
			break;
		case ffo:
			switch(cmd)
			{
			case red2_off:
				addledtimer(slot0_red2_timer);
				slot0_status = ffd;
				break;
			default:
				break;
			}
			break;
		case dff:
			switch(cmd)
			{
			case red2_on:
				slot0_status = dfo;
				break;
			default:
				break;
			}
			break;
		case odf:
			switch(cmd)
			{
			case red2_on:
				slot0_status = odo;
				break;
			default:
				break;
			}
			break;
		case ofo:
			switch(cmd)
			{
			case red2_off:
				addledtimer(slot0_red2_timer);
				slot0_status = ofd;
				break;
			default:
				break;
			}
			break;
		case fdo:
			switch(cmd)
			{
			case red2_off:
				addledtimer(slot0_red2_timer);
				slot0_status = fdd;
				break;
			default:
				break;
			}
			break;
		case ffd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot0_red2_timer);
				slot0_status = ffo;
				break;
			case red2_off:
				resetledtimer(slot0_red2_timer);
				break;
			default:
				break;
			}
			break;
		case ddf:
			switch(cmd)
			{
			case red2_on:
				slot0_status = ddo;
				break;
			default:
				break;
			}
			break;
		case dfo:
			switch(cmd)
			{
			case red2_off:
				addledtimer(slot0_red2_timer);
				slot0_status = dfd;
				break;
			default:
				break;
			}
			break;
		case odo:
			switch(cmd)
			{
			case red2_off:
				addledtimer(slot0_red2_timer);
				slot0_status = odd;
				break;
			default:
				break;
			}
			break;
		case ofd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot0_red2_timer);
				slot0_status = ofo;
				break;
			case red2_off:
				resetledtimer(slot0_red2_timer);
				break;
			default:
				break;
			}
			break;
		case fdd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot0_red2_timer);
				slot0_status = fdo;
				break;
			case red2_off:
				resetledtimer(slot0_red2_timer);
				break;
			default:
				break;
			}
			break;
		case ddo:
			switch(cmd)
			{
			case red2_off:
				addledtimer(slot0_red2_timer);
				slot0_status = ddd;
				break;
			default:
				break;
			}
			break;
		case dfd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot0_red2_timer);
				slot0_status = dfo;
				break;
			case red2_off:
				resetledtimer(slot0_red2_timer);
				break;
			default:
				break;
			}
			break;
		case odd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot0_red2_timer);
				slot0_status = odo;
				break;
			case red2_off:
				resetledtimer(slot0_red2_timer);
				break;
			default:
				break;
			}
			break;
		case ddd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot0_red2_timer);
				slot0_status = ddo;
				break;
			case red2_off:
				resetledtimer(slot0_red2_timer);
				break;
			default:
				break;
			}
			break;
		}
		spin_unlock(&slot0_lock);
	}
	else if(slot == slot1)
	{
		spin_lock(&slot1_lock);
		switch(slot1_status)
		{
		case fff:
			switch(cmd)
			{
			case red2_on:
				expcd_slot_set_led(slot1, lon, red);
				slot1_status = ffo;
				break;
			default :
				break;
			}
			break;
		case off:
			switch(cmd)
			{
			case red2_on:
				slot1_status = ofo;
				break;
			default :
				break;
			}
			break;
		case fof:
			switch(cmd)
			{
			case red2_on:
				expcd_slot_set_led(slot1, lon, red);
				slot1_status = fdo;
				break;
			default :
				break;
			}
			break;
		case ffo:
			switch(cmd)
			{
			case red2_off:
				addledtimer(slot1_red2_timer);
				slot1_status = ffd;
				break;
			default:
				break;
			}
			break;
		case dff:
			switch(cmd)
			{
			case red2_on:
				slot1_status = dfo;
				break;
			default:
				break;
			}
			break;
		case odf:
			switch(cmd)
			{
			case red2_on:
				slot1_status = odo;
				break;
			default:
				break;
			}
			break;
		case ofo:
			switch(cmd)
			{
			case red2_off:
				addledtimer(slot1_red2_timer);
				slot1_status = ofd;
				break;
			default:
				break;
			}
			break;
		case fdo:
			switch(cmd)
			{
			case red2_off:
				addledtimer(slot1_red2_timer);
				slot1_status = fdd;
				break;
			default:
				break;
			}
			break;
		case ffd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot1_red2_timer);
				slot1_status = ffo;
				break;
			case red2_off:
				resetledtimer(slot1_red2_timer);
				break;
			default:
				break;
			}
			break;
		case ddf:
			switch(cmd)
			{
			case red2_on:
				slot1_status = ddo;
				break;
			default:
				break;
			}
			break;
		case dfo:
			switch(cmd)
			{
			case red2_off:
				addledtimer(slot1_red2_timer);
				slot1_status = dfd;
				break;
			default:
				break;
			}
			break;
		case odo:
			switch(cmd)
			{
			case red2_off:
				addledtimer(slot1_red2_timer);
				slot1_status = odd;
				break;
			default:
				break;
			}
			break;
		case ofd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot1_red2_timer);
				slot1_status = ofo;
				break;
			case red2_off:
				resetledtimer(slot1_red2_timer);
				break;
			default:
				break;
			}
			break;
		case fdd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot1_red2_timer);
				slot1_status = fdo;
				break;
			case red2_off:
				resetledtimer(slot1_red2_timer);
				break;
			default:
				break;
			}
			break;
		case ddo:
			switch(cmd)
			{
			case red2_off:
				addledtimer(slot1_red2_timer);
				slot1_status = ddd;
				break;
			default:
				break;
			}
			break;
		case dfd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot1_red2_timer);
				slot1_status = dfo;
				break;
			case red2_off:
				resetledtimer(slot1_red2_timer);
				break;
			default:
				break;
			}
			break;
		case odd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot1_red2_timer);
				slot1_status = odo;
				break;
			case red2_off:
				resetledtimer(slot1_red2_timer);
				break;
			default:
				break;
			}
			break;
		case ddd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot1_red2_timer);
				slot1_status = ddo;
				break;
			case red2_off:
				resetledtimer(slot1_red2_timer);
				break;
			default:
				break;
			}
			break;
		}
		spin_unlock(&slot1_lock);
	}
	else
		return EXPCD_LED_ERROR;
	return EXPCD_LED_SUCCESS;
}

EXPORT_SYMBOL(aldRedControl);
//lignt on same as aldRedControl
//lignt off immediately
int aldRedControl_now(int slot, int on)
{
	enum command cmd;
	if(on == 1)
		cmd = red2_on;
	else if(on == 0)
		cmd = red2_off;
	else
		return EXPCD_LED_ERROR;
	if(slot == slot0)
	{
		spin_lock(&slot0_lock);
		switch(slot0_status)
		{
		case fff:
			switch(cmd)
			{
			case red2_on:
				expcd_slot_set_led(slot0, lon, red);
				slot0_status = ffo;
				break;
			default :
				break;
			}
			break;
		case off:
			switch(cmd)
			{
			case red2_on:
				slot0_status = ofo;
				break;
			case red2_off:
				expcd_slot_set_led(slot0, loff, red);
				slot0_status = fff;
			default :
				break;
			}
			break;
		case fof:
			switch(cmd)
			{
			case red2_on:
				expcd_slot_set_led(slot0, lon, red);
				slot0_status = fdo;
				break;
			default :
				break;
			}
			break;
		case ffo:
			switch(cmd)
			{
			case red2_off:
				expcd_slot_set_led(slot0, loff, red);
				slot0_status = fff;
				break;
			default:
				break;
			}
			break;
		case dff:
			switch(cmd)
			{
			case red2_on:
				slot0_status = dfo;
				break;
			case red2_off:
				del_timer(slot0_red1_timer);
				expcd_slot_set_led(slot0, loff, red);
				slot0_status = fff;
				break;
			default:
				break;
			}
			break;
		case odf:
			switch(cmd)
			{
			case red2_on:
				slot0_status = odo;
				break;
			case red2_off:
				expcd_slot_set_led(slot0, lon, green);
				slot0_status = fof;
				break;
			default:
				break;
			}
			break;
		case ofo:
			switch(cmd)
			{
			case red2_off:
				expcd_slot_set_led(slot0, loff, red);
				slot0_status = fff;
				break;
			default:
				break;
			}
			break;
		case fdo:
			switch(cmd)
			{
			case red2_off:
				expcd_slot_set_led(slot0, lon, green);
				slot0_status = fof;
				break;
			default:
				break;
			}
			break;
		case ffd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot0_red2_timer);
				slot0_status = ffo;
				break;
			case red2_off:
				del_timer(slot0_red2_timer);
				expcd_slot_set_led(slot0, loff, red);
				slot0_status = fff;
				break;
			default:
				break;
			}
			break;
		case ddf:
			switch(cmd)
			{
			case red2_on:
				slot0_status = ddo;
				break;
			case red2_off:
				del_timer(slot0_red1_timer);
				expcd_slot_set_led(slot0, lon, green);
				slot0_status = fof;
				break;
			default:
				break;
			}
			break;
		case dfo:
			switch(cmd)
			{
			case red2_off:
				del_timer(slot0_red1_timer);
				expcd_slot_set_led(slot0, loff, red);
				slot0_status = fff;
				break;
			default:
				break;
			}
			break;
		case odo:
			switch(cmd)
			{
			case red2_off:
				expcd_slot_set_led(slot0, lon, green);
				slot0_status = fof;
				break;
			default:
				break;
			}
			break;
		case ofd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot0_red2_timer);
				slot0_status = ofo;
				break;
			case red2_off:
				del_timer(slot0_red2_timer);
				expcd_slot_set_led(slot0, loff, red);
				slot0_status = fff;
				break;
			default:
				break;
			}
			break;
		case fdd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot0_red2_timer);
				slot0_status = fdo;
				break;
			case red2_off:
				del_timer(slot0_red2_timer);
				expcd_slot_set_led(slot0, lon, green);
				slot0_status = fof;
				break;
			default:
				break;
			}
			break;
		case ddo:
			switch(cmd)
			{
			case red2_off:
				del_timer(slot0_red1_timer);
				expcd_slot_set_led(slot0, lon, green);
				slot0_status = fof;
				break;
			default:
				break;
			}
			break;
		case dfd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot0_red2_timer);
				slot0_status = dfo;
				break;
			case red2_off:
				del_timer(slot0_red1_timer);
				del_timer(slot0_red2_timer);
				expcd_slot_set_led(slot0, loff, red);
				slot0_status = fff;
				break;
			default:
				break;
			}
			break;
		case odd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot0_red2_timer);
				slot0_status = odo;
				break;
			case red2_off:
				del_timer(slot0_red2_timer);
				expcd_slot_set_led(slot0, lon, green);
				slot0_status = fof;
				break;
			default:
				break;
			}
			break;
		case ddd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot0_red2_timer);
				slot0_status = ddo;
				break;
			case red2_off:
				del_timer(slot0_red1_timer);
				del_timer(slot0_red2_timer);
				expcd_slot_set_led(slot0, lon, green);
				slot0_status = fof;
				break;
			default:
				break;
			}
			break;
		}
		spin_unlock(&slot0_lock);
	}
	else if(slot == slot1)
	{
		spin_lock(&slot1_lock);
		switch(slot1_status)
		{
		case fff:
			switch(cmd)
			{
			case red2_on:
				expcd_slot_set_led(slot1, lon, red);
				slot1_status = ffo;
				break;
			default :
				break;
			}
			break;
		case off:
			switch(cmd)
			{
			case red2_on:
				slot1_status = ofo;
				break;
			case red2_off:
				expcd_slot_set_led(slot1, loff, red);
				slot1_status = fff;
			default :
				break;
			}
			break;
		case fof:
			switch(cmd)
			{
			case red2_on:
				expcd_slot_set_led(slot1, lon, red);
				slot1_status = fdo;
				break;
			default :
				break;
			}
			break;
		case ffo:
			switch(cmd)
			{
			case red2_off:
				expcd_slot_set_led(slot1, loff, red);
				slot1_status = fff;
				break;
			default:
				break;
			}
			break;
		case dff:
			switch(cmd)
			{
			case red2_on:
				slot1_status = dfo;
				break;
			case red2_off:
				del_timer(slot1_red1_timer);
				expcd_slot_set_led(slot1, loff, red);
				slot1_status = fff;
				break;
			default:
				break;
			}
			break;
		case odf:
			switch(cmd)
			{
			case red2_on:
				slot1_status = odo;
				break;
			case red2_off:
				expcd_slot_set_led(slot1, lon, green);
				slot1_status = fof;
				break;
			default:
				break;
			}
			break;
		case ofo:
			switch(cmd)
			{
			case red2_off:
				expcd_slot_set_led(slot1, loff, red);
				slot1_status = fff;
				break;
			default:
				break;
			}
			break;
		case fdo:
			switch(cmd)
			{
			case red2_off:
				expcd_slot_set_led(slot1, lon, green);
				slot1_status = fof;
				break;
			default:
				break;
			}
			break;
		case ffd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot1_red2_timer);
				slot1_status = ffo;
				break;
			case red2_off:
				del_timer(slot1_red2_timer);
				expcd_slot_set_led(slot1, loff, red);
				slot1_status = fff;
				break;
			default:
				break;
			}
			break;
		case ddf:
			switch(cmd)
			{
			case red2_on:
				slot1_status = ddo;
				break;
			case red2_off:
				del_timer(slot1_red1_timer);
				expcd_slot_set_led(slot1, lon, green);
				slot1_status = fof;
				break;
			default:
				break;
			}
			break;
		case dfo:
			switch(cmd)
			{
			case red2_off:
				del_timer(slot1_red1_timer);
				expcd_slot_set_led(slot1, loff, red);
				slot1_status = fff;
				break;
			default:
				break;
			}
			break;
		case odo:
			switch(cmd)
			{
			case red2_off:
				expcd_slot_set_led(slot1, lon, green);
				slot1_status = fof;
				break;
			default:
				break;
			}
			break;
		case ofd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot1_red2_timer);
				slot1_status = ofo;
				break;
			case red2_off:
				del_timer(slot1_red2_timer);
				expcd_slot_set_led(slot1, loff, red);
				slot1_status = fff;
				break;
			default:
				break;
			}
			break;
		case fdd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot1_red2_timer);
				slot1_status = fdo;
				break;
			case red2_off:
				del_timer(slot1_red2_timer);
				expcd_slot_set_led(slot1, lon, green);
				slot1_status = fof;
				break;
			default:
				break;
			}
			break;
		case ddo:
			switch(cmd)
			{
			case red2_off:
				del_timer(slot1_red1_timer);
				expcd_slot_set_led(slot1, lon, green);
				slot1_status = fof;
				break;
			default:
				break;
			}
			break;
		case dfd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot1_red2_timer);
				slot1_status = dfo;
				break;
			case red2_off:
				del_timer(slot1_red1_timer);
				del_timer(slot1_red2_timer);
				expcd_slot_set_led(slot1, loff, red);
				slot1_status = fff;
				break;
			default:
				break;
			}
			break;
		case odd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot1_red2_timer);
				slot1_status = odo;
				break;
			case red2_off:
				del_timer(slot1_red2_timer);
				expcd_slot_set_led(slot1, lon, green);
				slot1_status = fof;
				break;
			default:
				break;
			}
			break;
		case ddd:
			switch(cmd)
			{
			case red2_on:
				del_timer(slot1_red2_timer);
				slot1_status = ddo;
				break;
			case red2_off:
				del_timer(slot1_red1_timer);
				del_timer(slot1_red2_timer);
				expcd_slot_set_led(slot1, lon, green);
				slot1_status = fof;
				break;
			default:
				break;
			}
			break;
		}
		spin_unlock(&slot1_lock);
	}
	else
		return EXPCD_LED_ERROR;
	return EXPCD_LED_SUCCESS;
}

EXPORT_SYMBOL(aldRedControl_now);
int aldTimingSet(int time)
{
	if(time<0 || time >70000000)
		return EXPCD_LED_ERROR;
	delayedtime = time/1000*HZ;
	return EXPCD_LED_SUCCESS;
}
EXPORT_SYMBOL(aldTimingSet);

int aldVersionGet(void)
{
	return EXPCD_LED_SUCCESS;
}
EXPORT_SYMBOL(aldVersionGet);

int aldSelfDiagnose (int slot)
{
	unsigned long j = jiffies;
	if(slot == slot0)
	{
		expcd_slot_set_led(slot0, lon, red);
		slot0_diagnose_timer->expires = j + HZ;
		slot0_diagnose_timer->data = 3;
		add_timer(slot0_diagnose_timer);
	}
	else if(slot == slot1)
	{
		expcd_slot_set_led(slot1, lon, red);
		slot1_diagnose_timer->expires = j+HZ;
		add_timer(slot1_diagnose_timer);
		slot1_diagnose_timer->data = 7;
	}
	else
		return EXPCD_LED_ERROR;
	return EXPCD_LED_SUCCESS;
}
EXPORT_SYMBOL(aldSelfDiagnose);
