/* 2014-08-28: File changed by Sony Corporation */
/*
 *Copyright 2012 Sony Corporation
 *This program is free software; you can redistribute it and/or modify
 *it under the terms of the GNU General Public License, version 2, as
 *published by the Free Software Foundation.
 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *GNU General Public License for more details.
 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

//==============================================================================
// headers inclusion

#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>

#include <asm/io.h>
#include <asm/atomic.h>

#include <mach/gpio.h>

#include "expcd_debug.h"
#include "expcd_local.h"
#include "expcd_drv.h"
#include "expcd_thread.h"
#include <mach/expcd_hss.h>

/* HSS releated */
#ifdef CONFIG_SNSC_HSS
#include <linux/snsc_hss.h>
#endif /* #ifdef CONFIG_SNSC_HSS */

//==============================================================================
// file-global and extern members declaration

/* expcd_force_op is used by software interrupts which may pretend to be
 * the real HotPlug irqs.
 * if expcd_force_op is set, we have to accept the forced operation instead
 * of what-is-detected.
 * NOTE: it is initiated as (~0), not (0). */
atomic_t expcd_force_op = ATOMIC_INIT(~0);

extern void expcd_cdev_fasync_sigcall(int sig, void *param, int size);
extern void expcd_cdev_fasync_inout_sigcall(int sig, void *param, int size);
extern int  expcd_cdev_register(void);
extern void expcd_cdev_unregister(void);

extern void expcd_oc_callback(void);
extern void expcd_hp_callback(void);

extern int expcd_led_init(void);
extern void expcd_led_exit(void);

static  int g_slot_status[2]={0,0};

static  int g_initialize_flag = 0;	/* Initialize completion */

int expcd_slot_out_get(int slot_no)
{
    if((slot_no>1) || (slot_no<0)) return 1;
       return g_slot_status[slot_no];
}
EXPORT_SYMBOL(expcd_slot_out_get);

extern int cxd9225gg_udc_get_mass_flag(void);

#ifdef CONFIG_SNSC_HSS
static int expcd_hss_node_suspend(struct hss_node *node);
static int expcd_hss_node_resume(struct hss_node *node);
static int expcd_hss_node_init(void);
static void expcd_hss_node_exit(void);

static struct hss_expcd_info {
    struct hss_node *node;
};

struct hss_expcd_info *p_hss_expcd_info;
#endif /* #ifdef CONFIG_SNSC_HSS */

//==============================================================================
// irq handler related
/* notify PlugIn Signal to app if any PCIe slot has SxS Card in it. */
void expcd_pcie_plugin_notify(void)
{
	int i;
	for (i=0; i<2; i++) {
		if (expcd_slot_check(i, 1)) {
			/* make a permanent room for async access */
			static int hp_sig_param = EXPCD_IF_PCIE|EXPCD_OP_PLUGIN;
			switch(i) {
			case 0: hp_sig_param|=EXPCD_SLOT_0; break;
			case 1: hp_sig_param|=EXPCD_SLOT_1; break;
			default: break;
			}
			expcd_cdev_fasync_inout_sigcall(SIG_EXPCD_HP, 
				&hp_sig_param, sizeof(hp_sig_param));
		}
	}
}
EXPORT_SYMBOL(expcd_pcie_plugin_notify);

/* really do the hotplug work */
void expcd_hotplug_work_func(unsigned long data)
{
	int slotno, is_pcie, is_plugin, is_plugin_ag, irq;
	static int total_delay=0;

	/* make a permanent room */
	static unsigned long hp_sig_param;

	/* get hotplug information. */
	slotno    = (data & EXPCD_SLOT_MASK) >> EXPCD_SLOT_SHIFT;
	is_pcie   = (data & EXPCD_IF_PCIE) >> EXPCD_IF_SHIFT;
	is_plugin = (data & EXPCD_OP_PLUGIN) >> EXPCD_OP_SHIFT;
	irq       = data >> EXPCD_IRQ_SHIFT;

	/* corespond to mistake */
	is_plugin_ag = expcd_slot_check(slotno, is_pcie);
	if( is_plugin != is_plugin_ag ) {
		mdelay(4);
		is_plugin = expcd_slot_check(slotno, is_pcie);
		if( is_plugin == 1 ) {
			data |= EXPCD_OP_PLUGIN;
			g_slot_status[slotno] = 0;
		} 
		else {
			data &= ~EXPCD_OP_PLUGIN;
			g_slot_status[slotno] = 1;
		}
		expcd_errorl("change status=0x%08x, is_plugin=%d", data, is_plugin);
	}

	/* make sure the interrupt is stable */
	if(is_plugin==1) {
		int loop=5;
		while(loop--) {
			mdelay(16);
         total_delay+=16;
			is_plugin_ag = expcd_slot_check(slotno, is_pcie);
         if( is_plugin_ag != 1 )
         {
            if(total_delay>120)
            {
                total_delay=0;
                expcd_errorl("not find media slot=[%d] is_pcie=[%d]\n",slotno,is_pcie);
                return;
            }
            else
            {
                loop=5;
            }
         }
		}
	}
    else {
    	int loop=3;
    	while(loop--) {
    		mdelay(16);
    		total_delay+=16;
    		is_plugin_ag = expcd_slot_check(slotno, is_pcie);
    		if( is_plugin_ag != 0 ) {
    			if(total_delay>50) {
    				total_delay=0;
    				expcd_debugl("Slot[%d] is pulled out after attach failure.",slotno);
            		return;
    			}
    			else {
    				loop=3;
    			}
    		}
    	}
    }
	total_delay=0;

	/* realy do the hotplug work. this func is heavy. */
	/*
	 * If the interrupt says the card is pulled out, 
	 * we trust it immediately.
	 * But if it says the card is pluged in, we have to wait for a while
	 * to see whether the card is still in the slot.
	*/
	if((!is_plugin)||(is_plugin_ag==is_plugin)) {
		int mass_on = cxd9225gg_udc_get_mass_flag();
		/* print some logs */
		expcd_infol("%s ExpressCard is %s Slot %d. [Mass%s]",
					is_pcie?("PCIE"):("USB"),
					is_plugin?("pluged into"):("pulled from"), 
					slotno, mass_on?"On":"Off");

		/* notify all the listeners in kernel */
		expcd_hp_callback();

		/* notify all the listeners in userspace by emiting a signal. */
		if(!mass_on) {
			if (is_plugin) {
				/* power up the port */
				expcd_power_ctrl(slotno, is_plugin, is_pcie);
				/* notify userspace */
				hp_sig_param = data;
				expcd_cdev_fasync_sigcall(SIG_EXPCD_HP, &hp_sig_param, sizeof(hp_sig_param));
			} else {
				/* notify userspace */
				hp_sig_param = data;
				expcd_cdev_fasync_sigcall(SIG_EXPCD_HP, &hp_sig_param, sizeof(hp_sig_param));
				/* power off the port */
				expcd_power_ctrl(slotno, is_plugin, is_pcie);
			}
		} else {
			extern int cxd9225gg_udc_find_media(int findno, int ispcie);
			extern int cxd9225gg_udc_remove_media(int findno);
			if(is_plugin) {
				expcd_power_ctrl(slotno, is_plugin, is_pcie);
				expcd_errorl("cxd9225gg_udc_find_media slot=[%d] is_pcie=[%d]\n",slotno,is_pcie);
				cxd9225gg_udc_find_media(slotno, is_pcie);
			}
			else {
				expcd_errorl("cxd9225gg_udc_remove_media slot=[%d] is_pcie=[%d]\n",slotno,is_pcie);
				cxd9225gg_udc_remove_media(slotno);
				expcd_power_ctrl(slotno, is_plugin, is_pcie);
			}
		}

		expcd_debugl("work is over !!");
	}
}

/* realy do the overcurrent work */
void expcd_overcurrent_work_func(unsigned long data)
{
	int slotno = data;

	/* make a permanent room */
	static int oc_sig_param;

	expcd_errorl("ExpressCard Slot %d is OverCurrent !!", slotno);

     /* return 1 if card is in the slot */
	//notify all OverCurrent listeners.
	expcd_oc_callback();

	expcd_slot_set_power(slotno,0);
	//emit the signal to all listeners.
	if(1==expcd_slot_check(slotno, 1))
        {
	oc_sig_param = slotno;
	expcd_cdev_fasync_sigcall(SIG_EXPCD_OC, &oc_sig_param, sizeof(oc_sig_param));
       }
}

static struct expcd_task_list multiloophead_expcd;

/* usb slot hotplug isr */
static irqreturn_t expcd_hotplug_isr(int irq, void *dev_id)
{
	int slotno, is_pcie, is_plugin, status = (int)dev_id;

	if(g_initialize_flag == 0) {
		/* Handler call at the time of the initialization non-completion */
		return IRQ_NONE;
	} else {
		slotno    = (status & EXPCD_SLOT_MASK) >> EXPCD_SLOT_SHIFT;
		is_pcie   = (status & EXPCD_IF_PCIE) >> EXPCD_IF_SHIFT;

		is_plugin = expcd_slot_check(slotno, is_pcie);
		expcd_errorl("status=0x%06x, is_plugin=%d", status, is_plugin);
		
	    if(is_plugin==1) {
	        g_slot_status[slotno] = 0;
	    } else {
	        g_slot_status[slotno] = 1;
		}
		if (atomic_read(&expcd_force_op) == ~0) {
			/* expcd_force_op is not set, normal irq */
			status |= (is_plugin ? EXPCD_OP_PLUGIN : EXPCD_OP_PULLOUT);
		} else {
			/* if expcd_force_op is set, this irq would be a sw irq,
			 * we have to accept the forced operation */
			status |= atomic_read(&expcd_force_op);
			
			/* since it is a sw irq, here we clear swi related things. */
			atomic_set(&expcd_force_op, ~0);
		}

		/* record the irq that we are answering. */
		status |= irq << EXPCD_IRQ_SHIFT;


		/* add a task to expcd thread */
		{
		expcd_add_task(&multiloophead_expcd, expcd_hotplug_work_func, (unsigned long)status);
		}

	    return IRQ_HANDLED;
	}
}

/* over current isr */
static irqreturn_t expcd_ovc_isr(int irq, void *dev_id)
{
	int slotno = (int)dev_id;

	if(g_initialize_flag == 0) {
		/* Handler call at the time of the initialization non-completion */
		return IRQ_NONE;
	} else {
		expcd_add_task(&multiloophead_expcd, expcd_overcurrent_work_func, (unsigned long)slotno);

		expcd_warningl("Detected OverCurrent on Slot %d !!", slotno);

		return IRQ_HANDLED;
	}
}

/* hotplug irqactions */
static struct irqaction expcd_irqact_slot0_usb = {
	.name      = "expcd_irq_slot0_usb",
	.flags     = IRQF_DISABLED,
	.handler   = expcd_hotplug_isr,
	.dev_id    = (void*)(EXPCD_IF_USB|EXPCD_SLOT_0),
};
static struct irqaction expcd_irqact_slot0_pcie = {
	.name      = "expcd_irq_slot0_pcie",
	.flags     = IRQF_DISABLED,
	.handler   = expcd_hotplug_isr,
	.dev_id    = (void*)(EXPCD_IF_PCIE|EXPCD_SLOT_0),
};
static struct irqaction expcd_irqact_slot1_usb = {
	.name      = "expcd_irq_slot1_usb",
	.flags     = IRQF_DISABLED,
	.handler   = expcd_hotplug_isr,
	.dev_id    = (void*)(EXPCD_IF_USB|EXPCD_SLOT_1),
};
static struct irqaction expcd_irqact_slot1_pcie = {
	.name      = "expcd_irq_slot1_pcie",
	.flags     = IRQF_DISABLED,
	.handler   = expcd_hotplug_isr,
	.dev_id    = (void*)(EXPCD_IF_PCIE|EXPCD_SLOT_1),
};

/* overcurrent irqactions */
static struct irqaction expcd_irqact_slot0_ovc = {
	.name      = "expcd_irq_slot0_ovc",
	.flags     = IRQF_DISABLED,
	.handler   = expcd_ovc_isr,
	.dev_id    = (void*)EXPCD_SLOT_0,
};
static struct irqaction expcd_irqact_slot1_ovc = {
	.name      = "expcd_irq_slot1_ovc",
	.flags     = IRQF_DISABLED,
	.handler   = expcd_ovc_isr,
	.dev_id    = (void*)EXPCD_SLOT_1,
};

/* request all irqs */
int expcd_irq_request(void)
{
	int ret = 0;

	/* request hotplug irqs */
	ret |= cxd9225gg_gpio_irq_request(EXPCD_INT_S0_USB, &expcd_irqact_slot0_usb);
	ret |= cxd9225gg_gpio_irq_request(EXPCD_INT_S0_PCIE,&expcd_irqact_slot0_pcie);
	ret |= cxd9225gg_gpio_irq_request(EXPCD_INT_S1_USB, &expcd_irqact_slot1_usb);
	ret |= cxd9225gg_gpio_irq_request(EXPCD_INT_S1_PCIE,&expcd_irqact_slot1_pcie);

	/* request overcurrent irqs */
	ret |= cxd9225gg_gpio_irq_request(EXPCD_INT_S0_OVC, &expcd_irqact_slot0_ovc);
	ret |= cxd9225gg_gpio_irq_request(EXPCD_INT_S1_OVC, &expcd_irqact_slot1_ovc);

	return (ret ? -1 : 0);
}

/* free all hotplug irqs */
void expcd_hotplug_irq_free(void)
{
	/* free hotplug irqs */
	cxd9225gg_gpio_irq_free(EXPCD_INT_S0_USB, &expcd_irqact_slot0_usb);
	cxd9225gg_gpio_irq_free(EXPCD_INT_S0_PCIE,&expcd_irqact_slot0_pcie);
	cxd9225gg_gpio_irq_free(EXPCD_INT_S1_USB, &expcd_irqact_slot1_usb);
	cxd9225gg_gpio_irq_free(EXPCD_INT_S1_PCIE,&expcd_irqact_slot1_pcie);

	/* free overcurrent irqs */
	cxd9225gg_gpio_irq_free(EXPCD_INT_S0_OVC, &expcd_irqact_slot0_ovc);
	cxd9225gg_gpio_irq_free(EXPCD_INT_S1_OVC, &expcd_irqact_slot1_ovc);
}


//==============================================================================
// module init related

void expcd_preset_slots(void)
{
	int i;
	for (i=0; i<EXPCD_SLOT_MAX; i++) {
		if(expcd_slot_check(i, 0) || expcd_slot_check(i, 1)) {
			//power up the slot.
			expcd_slot_set_power(i, 1);
		}
	}
}

#ifdef CONFIG_SNSC_HSS
static struct hss_node_ops hss_expcd_node_ops = {
    .suspend = expcd_hss_node_suspend,
    .resume = expcd_hss_node_resume,
};

static int expcd_hss_node_suspend(struct hss_node *node)
{
	/* free irqs */
	expcd_hotplug_irq_free();
    return 0;
}

static int expcd_hss_node_resume(struct hss_node *node)
{
	/* request interrupts */
	if (unlikely(expcd_irq_request())) {
		expcd_errorl("expcd_irq_request() failed on resume !!");
		expcd_hotplug_irq_free();
		return -1;
	}

	/* preset slots */
	expcd_preset_slots();

	/* turn off expcd slot leds */
	expcd_slot_set_led(0, 0, 0);
	expcd_slot_set_led(1, 0, 0);

    return 0;
}

static int expcd_hss_node_init(void)
{
    struct hss_node *node;
    int err = 0;

    /* alloc for csi info */
    p_hss_expcd_info = kzalloc(sizeof(*p_hss_expcd_info), GFP_KERNEL);
    if (p_hss_expcd_info == NULL) {
        err = -ENOMEM;
        return err;
    }

    /*
     * alloc the HSS node
     */
    node = hss_alloc_node(EXPCD_D_HSS_NODE_NAME);
    if (node == NULL) {
        kfree(p_hss_expcd_info);
        err = -ENOMEM;
        return err;
    }

    /*
     * save struct to HSS
     */
    hss_set_private(node, p_hss_expcd_info);
    p_hss_expcd_info->node = node;

    /*
     * Register the node.  If you want to get the callbacks to be called
     * before the specified node's one, specify the third parameter with
     * the name of the parent node. NULL means for no parent needed.
     */

    /* register HSS Platform for suspend/resume */
    err = hss_register_node(node, &hss_expcd_node_ops, EXPCD_D_HSS_NODE_PARENT);

    if (err != 0) {
        kfree(p_hss_expcd_info);
        return err;
    }

    return err;
}

static void expcd_hss_node_exit(void)
{
    hss_unregister_node(p_hss_expcd_info->node);
    hss_free_node(p_hss_expcd_info->node);
    kfree(p_hss_expcd_info);
}
#endif /* #ifdef CONFIG_SNSC_HSS */

/* module entry */
static int __init expcd_init(void)
{
	int ret=0;

	expcd_debugl("expcd module start !!");

	/* create a thread for running tasks */
	expcd_create_thread(&multiloophead_expcd);
	/* request interrupts */
	if (unlikely(expcd_irq_request())) {
		expcd_errorl("expcd_irq_request() failed !!");
		expcd_hotplug_irq_free();
		return -1;
	}

	(void)expcd_cdev_register();

	expcd_preset_slots();
	if(expcd_led_init() < 0)
	{
		expcd_errorl("expcd_led_init: fail");
		return -1;
	}

	/* turn off expcd slot leds */
	expcd_slot_set_led(0, 0, 0);
	expcd_slot_set_led(1, 0, 0);
    {
	//add for LSM
	extern int usbConnectionNotifyRegister(void* func);
	usbConnectionNotifyRegister(expcd_cdev_fasync_inout_sigcall);
    }

#ifdef CONFIG_SNSC_HSS
    /* register hss platform driver function */
    ret = expcd_hss_node_init();
    if(ret < 0){
        expcd_errorl("expcd hss init error.\n");
    }
#endif /* #ifdef CONFIG_SNSC_HSS */

	/* Completion flag set */
	g_initialize_flag = 1;

	return 0;
}

/* module exit */
static void __exit expcd_exit(void)
{
	expcd_debugl("Exit expcd module.");

#ifdef CONFIG_SNSC_HSS
    expcd_hss_node_exit();
#endif /* #ifdef CONFIG_SNSC_HSS */

	/* stop the thread created in expcd_init */
	expcd_stop_thread(&multiloophead_expcd);

	expcd_led_exit();

	/* free irqs */
	expcd_hotplug_irq_free();

	/* unregister char device */
	expcd_cdev_unregister();
	{
	//add for LSM
	extern int usbConnectionNotifyUnRegister(void);
	usbConnectionNotifyUnRegister();
	}
}

#ifdef CONFIG_SNSC_DEFERRED_INITCALLS_CXD9225GG_PCIE
deferred_initcall(expcd_init, CONFIG_SNSC_DEFERRED_INITCALLS_GROUP_CXD9225GG_PCIE);
#else
late_initcall(expcd_init);
#endif
module_exit(expcd_exit);
