/* 2014-08-28: File changed by Sony Corporation */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/list.h>
#include <linux/kthread.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/spinlock.h>

#include <expcd_thread.h>
#include "expcd_debug.h"

/* thread function */
int expcd_thread_func(void *data) {
	struct expcd_task_list *list_head = NULL;
	struct expcd_task_list *list_entry = NULL;
	struct list_head *plist = NULL;
	struct list_head *tmp = NULL;
	unsigned long irq_flags = 0;
	long timeout = 0;

	/* change my task name. */
	daemonize("%s", "expcd_work_thrd");

	/* run the task. */
	list_head = (struct expcd_task_list*)data;
	complete(&list_head->thread_comp);
	while(1) {
#ifdef EXPCD_THD_USE_UDIF
		if (list_head->should_stop)
#else
		if (kthread_should_stop())
#endif
		{
			if (list_empty(&list_head->list)) {
				expcd_warningl("task list of thread is empty !");
				
				break;
			} else {
				/* if task list is not empty, continue ro run task function */
			}
		} else {
			/* I do not recieve kthread_stop() */
		}
		timeout = wait_event_interruptible_timeout(list_head->wait_queue,
			list_head != list_first_entry(&list_head->list, struct expcd_task_list, list),
			HZ);
		if (0 == timeout) {
			/* we should run kthread_end() */
		} else {
			/* I think this loop must be locked. because expcd_add_task() is called from 
			   interrupt handler etc.
			*/
			list_entry = NULL;
			spin_lock_irqsave(&list_head->lock, irq_flags);
			list_for_each_safe(plist, tmp, &list_head->list) {
				list_entry = list_entry(plist, struct expcd_task_list, list);
				list_del(plist);
				
				/* break , because I would like to do func in safety */
				break;
			}
			spin_unlock_irqrestore(&list_head->lock, irq_flags);

			if (list_entry != NULL) {
				list_entry->func(list_entry->data);
				kfree(list_entry);
			}
		}
	}

	return 0;
}

/* create a thread */
void expcd_create_thread(struct expcd_task_list *list_head) {
	INIT_LIST_HEAD(&list_head->list);
	spin_lock_init(&list_head->lock);
	init_waitqueue_head(&list_head->wait_queue);
	init_completion(&list_head->thread_comp);

#ifdef EXPCD_THD_USE_UDIF
	list_head->should_stop = 0; /* maybe not necessary, just for safety. */
	list_head->pid = kernel_thread(expcd_thread_func, (void*)list_head, 0);
#else /* EXPCD_THD_USE_UDIF */
	list_head->task = kthread_create(expcd_thread_func, (void*)list_head, "expcd_thread");
	if(IS_ERR(list_head->task)) {
		expcd_errorl("kthread_create error");
		
		return;
	}
	wake_up_process(list_head->task);
#endif /* EXPCD_THD_USE_UDIF */

	wait_for_completion(&list_head->thread_comp);
		
	return;
}

/* end thread */
void expcd_stop_thread(struct expcd_task_list *list_head) {
#ifdef EXPCD_THD_USE_UDIF
	list_head->should_stop = 1;
#else
	kthread_stop(list_head->task);
#endif
}

/* add task to thread */
void expcd_add_task(struct expcd_task_list *list_head, 
    void (*func)(unsigned long data), 
	unsigned long data) {

	struct expcd_task_list *list_entry = NULL;
	unsigned long irq_flags = 0;

	list_entry = (struct expcd_task_list*)kmalloc(sizeof(struct expcd_task_list), GFP_ATOMIC);
	if (NULL == list_entry) {
		expcd_errorl("kmalloc error");
		
		return;
	}
	list_entry->func = func;
	list_entry->data = data;

	spin_lock_irqsave(&list_head->lock, irq_flags);

	/* I would like to execute as first in first out */
	list_add_tail(&list_entry->list, &list_head->list);

	spin_unlock_irqrestore(&list_head->lock, irq_flags);

	wake_up_interruptible(&list_head->wait_queue);
}

