#ifndef _UMSC_H
#define _UMSC_H

#include <linux/usb.h>

#define UMSC_STR_SIZE  32

struct usbcard_lun
{
	unsigned long nsec;  /* Linux size - 512 byte sectors */
	unsigned int bsize;  /* Linux hardsect_size */
	unsigned int bshift; /* Shift between 512 and hard sects */
};

struct usbcard_info
{
	int devnum;             /* Address on USB bus */
	u8  portnum;            /* Parent port number (origin 1) */
	u8  level;              /* Number of USB hub ancestors */
	u8  state;              /* configured, not attached, etc */
	u8  speed;              /* high/full/low (or error) */

	char product[UMSC_STR_SIZE];      /* iProduct string, if present */
	char manufacturer[UMSC_STR_SIZE]; /* iManufacturer string, if present */
	char serial[UMSC_STR_SIZE];       /* iSerialNumber string, if present */

	struct usb_device_descriptor desc_dev;    /* the device descriptor */
	struct usb_config_descriptor desc_cfg;    /* the active configuration desc */
	struct usb_interface_descriptor	desc_inf; /* the current interface desc */

	int maxlun;            /* Max LUN the device supports */
	struct usbcard_lun luns[1]; /* only the first LUN is supported */
};

#define UMSC_CBW_CBD_MAX_LEN  16
#define UMSC_SENSE_SIZE       18

#define UMSC_FLG_PHYS_ADDR  0x100  /* data_buf is physical address */
#define UMSC_FLG_KERN_ADDR  0x200  /* data_buf is logical address */
#define UMSC_FLG_SHMM_OFFS  0x400  /* data_buf is shmem offset */
#define UMSC_FLG_USER_ADDR  0x800  /* data_buf is userspace address */

struct usbcard_cmd
{
/* Input */
	int  cbd_len;
	char cbd[UMSC_CBW_CBD_MAX_LEN]; /* Command Block Data */
	int  data_dir;
	char *data_buf;
	int  data_len;
	unsigned int data_flg;        /* UMSC_FLG_PHYS_ADDR or ... */

/* Output */
	int  error;                     /* Error Code */
	int  sense_len;
	char sense[UMSC_SENSE_SIZE];    /* Sense Buffer */
};

struct usbcard_cap {
	unsigned long nsec;		/* Linux size - 512 byte sectors */
	unsigned int bsize;		/* Linux hardsect_size */
	unsigned int bshift;		/* Shift between 512 and hard sects */
};

/* struct size */
struct usbcard_stinfo {
	unsigned int st_usbcard_info_size;
	unsigned int st_usbcard_cmd_size;
	unsigned int st_usbcard_cap_size;
};

#define UMSC_MAGIC              'U'

#define UMSC_IOCTL_GETBASEINFO  _IOWR(UMSC_MAGIC, 0x00, struct usbcard_stinfo)

/* get generic information of the device */
#define UMSC_IOCTL_GETDEVINFO   _IOWR(UMSC_MAGIC, 0x01, struct usbcard_info)

/* get capacity of the device */
#define UMSC_IOCTL_GETCAPACITY  _IOWR(UMSC_MAGIC, 0x02, struct usbcard_cap)

/* make kernel space driver do the command */
#define UMSC_IOCTL_CMD          _IOWR(UMSC_MAGIC, 0x03, struct usbcard_cmd)

/* get shared memory lenght */
#define UMSC_IOCTL_GETSHMEMLEN  _IOWR(UMSC_MAGIC, 0x04, unsigned long)

/* get shared memory physical address */
#define UMSC_IOCTL_GETSHMEMPHY  _IOWR(UMSC_MAGIC, 0x05, unsigned long)

#endif //_UMSC_H

