/* 2014-08-28: File changed by Sony Corporation */
/*
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <linux/spi/spi.h>
#include <linux/spi/flash.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <mach/plat_flash.h>
#include <mach/hardware.h>
#include <mach/gpio.h>

static struct mtd_partition nor_parts0[] = {
	{
		.name		= "bootloader",
		.offset		= 0,
		.size		= SZ_1M,
		.mask_flags	= 0,
	},
	{
		.name		= "r_nup",
		.offset		= MTDPART_OFS_APPEND,
		.size		= SZ_512K,
		.mask_flags	= 0,
	},
	{
		.name		= "r_vmlinux",
		.offset		= MTDPART_OFS_APPEND,
		.size		= 3 * SZ_1M,
		.mask_flags	= 0,
	},
	{
		.name		= "r_rootfs",
		.offset		= MTDPART_OFS_APPEND,
		.size		= 8 * SZ_1M,
		.mask_flags	= 0,
	},
	{
		.name		= "nvm0",
		.offset		= MTDPART_OFS_APPEND,
		.size		= SZ_64K,
		.mask_flags	= 0,
	},
	{
		.name		= "nvm1",
		.offset		= MTDPART_OFS_APPEND,
		.size		= SZ_64K,
		.mask_flags	= 0,
	},
	{
		.name		= "nvm2",
		.offset		= MTDPART_OFS_APPEND,
		.size		= SZ_64K,
		.mask_flags	= 0,
	},
	{
		.name		= "nvm3",
		.offset		= MTDPART_OFS_APPEND,
		.size		= SZ_64K,
		.mask_flags	= 0,
	},
	{
		.name		= "nvm4",
		.offset		= MTDPART_OFS_APPEND,
		.size		= SZ_64K,
		.mask_flags	= 0,
	},
	{
		.name		= "nvm5",
		.offset		= MTDPART_OFS_APPEND,
		.size		= SZ_64K,
		.mask_flags	= 0,
	},
	{
		.name		= "nvm6",
		.offset		= MTDPART_OFS_APPEND,
		.size		= SZ_64K,
		.mask_flags	= 0,
	},
	{
		.name		= "nvm7",
		.offset		= MTDPART_OFS_APPEND,
		.size		= SZ_64K,
		.mask_flags	= 0,
	},
	{
		.name		= "user",
		.offset		= MTDPART_OFS_APPEND,
		.size		= (49 * SZ_1M) + SZ_512K,
		.mask_flags	= 0,
	},
	{
		.name		= "data",
		.offset		= MTDPART_OFS_APPEND,
		.size		= MTDPART_SIZ_FULL,
		.mask_flags	= 0,
	},
};

static struct mtd_partition nor_parts1[] = {
	{
		.name		= "flashbackup",
		.offset		= 0,
		.size		= 20 * SZ_1M,
		.mask_flags	= 0,
	},
	{
		.name		= "backup log",
		.offset		= MTDPART_OFS_APPEND,
		.size		= SZ_32M,
		.mask_flags	= 0,
	},
	{
		.name		= "reserve",
		.offset		= MTDPART_OFS_APPEND,
		.size		= MTDPART_SIZ_FULL,
		.mask_flags	= 0,
	},
};

static struct flash_platform_data nor_data0 = {
	.name		= "spi_flash0",
	.parts		= nor_parts0,
	.nr_parts	= ARRAY_SIZE(nor_parts0),
	.type		= "mx66l51235f",
};

static struct flash_platform_data nor_data1 = {
	.name		= "spi_flash1",
	.parts		= nor_parts1,
	.nr_parts	= ARRAY_SIZE(nor_parts1),
	.type		= "mx66l51235f",
};

/* must invoke atomic */
static int flash_chip_select(u32 cs)
{
	if (cs >= SPI_CHIP_NUM) {
		printk(KERN_ERR "%s: invalid chip select %d\n", __func__, cs);
		return -EINVAL;
	}

	cxd9225gg_set_gpio_dataout(CXD9225GG_BGPIO_05, cs);
	cxd9225gg_set_gpio_dataout(CXD9225GG_BGPIO_21, !cs);

	return 0;
}

static struct spi_board_info spi_board[SPI_CHIP_NUM] __initdata = {
	{
		.modalias		= "m25p80",
		.bus_num		= SPI_BUS_NUM,
		.chip_select		= 0,
		.max_speed_hz		= 1000000,
		.controller_data	= flash_chip_select,
		.irq			= -1,
		.platform_data		= &nor_data0,
	},
	{
		.modalias		= "m25p80",
		.bus_num		= SPI_BUS_NUM,
		.chip_select		= 1,
		.max_speed_hz		= 1000000,
		.controller_data	= &flash_chip_select,
		.irq			= -1,
		.platform_data		= &nor_data1,
	}
};

int __init platform_flash_init(void)
{
	return spi_register_board_info(spi_board, ARRAY_SIZE(spi_board));
}
