/* 2014-08-28: File changed by Sony Corporation */
/*
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/serial_8250.h>
#include <linux/serial_reg.h>
#include <linux/io.h>

#include <mach/hardware.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <asm/mach/map.h>

#include <mach/uart.h>
#include <mach/hardware.h>
#include <mach/cxd9225gg_Chipset.h>
#include <mach/plat_flash.h>


#include "generic.h"
#include "uart.h"


static struct map_desc cxd9225gg_io_desc[] __initdata = {
	/* 1 M section #2, maps the top half of AXI */
	{
		.virtual = CXD9225GG_1M_MAP_2_VIRT_BASE,
		.pfn = __phys_to_pfn(CXD9225GG_1M_MAP_2_PHYS_BASE),
		.length = CXD9225GG_1M_MAP_2_SIZE,
		.type = MT_DEVICE,
	},
	/* 1 M section #1, maps the bottom half of AXI */
	{
		.virtual = CXD9225GG_1M_MAP_1_VIRT_BASE,
		.pfn = __phys_to_pfn(CXD9225GG_1M_MAP_1_PHYS_BASE),
		.length = CXD9225GG_1M_MAP_1_SIZE,
		.type = MT_DEVICE,
	},
	/* PCI-E bus controller */
	{
		.virtual = CXD9225GG_PCIE_BUSC_REG_VIRT_BASE,
		.pfn = __phys_to_pfn(CXD9225GG_PCIE_BUSC_REG_PHYS_BASE),
		.length = CXD9225GG_PCIE_BUSC_REG_SIZE,
		.type = MT_DEVICE,
	},
	{
		.virtual = CXD9225GG_PCIE_1X_0_VIRT_BASE,
		.pfn = __phys_to_pfn(CXD9225GG_PCIE_1X_0_PHYS_BASE),
		.length = CXD9225GG_PCIE_1X_0_SIZE,
		.type = MT_DEVICE,
	},

	{
		.virtual = CXD9225GG_PCIE_1X_1_VIRT_BASE,
		.pfn = __phys_to_pfn(CXD9225GG_PCIE_1X_1_PHYS_BASE),
		.length = CXD9225GG_PCIE_1X_1_SIZE,
		.type = MT_DEVICE,
	},
	{
		.virtual = CXD9225GG_PCIE_4X_VIRT_BASE,
		.pfn = __phys_to_pfn(CXD9225GG_PCIE_4X_PHYS_BASE),
		.length = CXD9225GG_PCIE_4X_SIZE,
		.type = MT_DEVICE,
	},


	/* PCI bus controller */
	{
		.virtual = CXD9225GG_PCI_BUSC_REG_VIRT_BASE,
		.pfn = __phys_to_pfn(CXD9225GG_PCI_BUSC_REG_PHYS_BASE),
		.length = CXD9225GG_PCI_BUSC_REG_SIZE,
		.type = MT_DEVICE,
	},
	/* work RAM */
	{
		.virtual = CXD9225GG_WORKRAM_VIRT_BASE,
		.pfn = __phys_to_pfn(CXD9225GG_WORKRAM_PHYS_BASE),
		.length = CXD9225GG_WORKRAM_SIZE,
		.type = MT_DEVICE, /* MT_DEVICE_CACHED, */
	},
	/* PCI window #1 (IO/Configuration space) */
	{
		.virtual = CXD9225GG_PCI_WIN_1_VIRT_BASE,
		.pfn = __phys_to_pfn(CXD9225GG_PCI_WIN_1_PHYS_BASE),
		.length = CXD9225GG_PCI_WIN_1_SIZE,
		.type = MT_DEVICE,
	},
	/* 1M section map #0, DDR2 controller, INTC and all APB registers */
	{
		.virtual = CXD9225GG_1M_MAP_0_VIRT_BASE,
		.pfn = __phys_to_pfn(CXD9225GG_1M_MAP_0_PHYS_BASE),
		.length = CXD9225GG_1M_MAP_0_SIZE,
		.type = MT_DEVICE,
	},
	/* SPI NOR */
	{
		.virtual = CXD9225GG_NOR_VIRT_BASE,
		.pfn = __phys_to_pfn(CXD9225GG_NOR_PHYS_BASE),
		.length = CXD9225GG_NOR_SIZE,
		.type = MT_ROM,
	},

    {
        .virtual = CXD9225GG_USBFUNCC_REG_VIRT_BASE,
        .pfn = __phys_to_pfn(CXD9225GG_USBFUNCC_REG_PHY_BASE),
        .length = CXD9225GG_USBFUNCC_REG_SIZE,
        .type = MT_DEVICE,
    },

};

#define CXD9225GG_SFMEMC_REG_PHYS_BASE (CXD9225GG_AHB_REG_PHYS_BASE + 0x10000)
#define CXD9225GG_SFMEMC_REG_SIZE      0x400

static struct resource sfmemc_reg_area[] = {
	{
		.start	= CXD9225GG_SFMEMC_REG_PHYS_BASE,
		.end	=
			CXD9225GG_SFMEMC_REG_PHYS_BASE +
			CXD9225GG_SFMEMC_REG_SIZE - 1,
		.flags	= IORESOURCE_MEM,
	},
};

static struct platform_device sfmemc_device = {
	.name		= "sfmemc",
	.id		= 0,
	.resource	= sfmemc_reg_area,
	.num_resources	= ARRAY_SIZE(sfmemc_reg_area),
};

struct platform_device avp13_async_device = {
        .name = "avp13_sync",
        .id = 9,
        .dev = {
                .platform_data = NULL,
        },
};


static struct platform_device *platform_devs[] __initdata = {
	&cxd9225gg_uart0_device,
	&avp13_async_device,
	&sfmemc_device,
};

void __init cxd9225gg_map_io(void)
{
	iotable_init(cxd9225gg_io_desc, ARRAY_SIZE(cxd9225gg_io_desc));
//	safe_to_call_sched_clock = 1;
}

void __init cxd9225gg_init(void)
{
	int err;

	err = platform_add_devices(platform_devs, ARRAY_SIZE(platform_devs));
	if (err)
		panic("failed platform_add_devices: %d\n", err);

	err = platform_flash_init();
	if (err)
		panic("failed platform_flash_init: %d\n", err);
}

/* sysHwVer - return the HW version number
 *
 * RETURNS: The HW version number.
 *               [15:8] - Original version
 *               [7:4]  - Major version
 *               [3:0]  - Minor version
 */
unsigned short sysHwVer (int unit)
{
    unsigned short HWVer;

    /*  */
    HWVer = readw(CXD9225GG_SYSC_VERX + (unit * 2));

    return (HWVer);
}
EXPORT_SYMBOL(sysHwVer);


