/* 2014-08-28: File changed by Sony Corporation */
 /*
 *  Copyright 2012 Sony Corporation
  *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
  *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
   *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
   */

/*
 * Implementation based on cxd9225gg gpio.c
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/sysdev.h>
#include <linux/ptrace.h>
#include <linux/fs.h>		/* everything... */
#include <linux/cdev.h>
#include <mach/hardware.h>
#include <asm/irq.h>
#include <mach/irqs.h>
#include <asm/mach/irq.h>
#include <asm/io.h>
#include <asm/uaccess.h>	/* copy_*_user */

#include <mach/gpio.h>

/**
 * @file mach-cxd9225gg/gpio.c
 *
 * @brief This file contains the GPIO implementation details.
 *
 * @ingroup GPIO
 */

#define CXD9225GG_GPIO_PORT_NUM 12

static u32 gpio_port_offset[CXD9225GG_GPIO_PORT_NUM] = {
	0x00,    /* CXD9225GG_GPIO_PORT_00 */
	0x01,    /* CXD9225GG_GPIO_PORT_01 */
	0x02,    /* CXD9225GG_GPIO_PORT_02 */
	0x03,    /* CXD9225GG_GPIO_PORT_03 */
	0x04,    /* CXD9225GG_GPIO_PORT_04 */
	0x05,    /* CXD9225GG_GPIO_PORT_05 */
	0x08,    /* CXD9225GG_GPIO_PORT_10 */
	0x09,    /* CXD9225GG_GPIO_PORT_11 */
	0x0A,    /* CXD9225GG_GPIO_PORT_12 */
	0x0B,    /* CXD9225GG_GPIO_PORT_13 */
	0x0C,    /* CXD9225GG_GPIO_PORT_14 */
	0x0D     /* CXD9225GG_GPIO_PORT_15 */
};

#ifndef CONFIG_DPR_SUPPORT
static u32 gpio_pmcr_default_value[6] = {
	RB_CXD9225GG_PMC00_DEFAULT,
	RB_CXD9225GG_PMC01_DEFAULT,
	RB_CXD9225GG_PMC02_DEFAULT,
	RB_CXD9225GG_PMC03_DEFAULT,
	RB_CXD9225GG_PMC04_DEFAULT,
	RB_CXD9225GG_PMC05_DEFAULT
};

static u32 gpio_pmr_default_value[CXD9225GG_GPIO_PORT_NUM] = {
	RB_CXD9225GG_PM00_DEFAULT,
	RB_CXD9225GG_PM01_DEFAULT,
	RB_CXD9225GG_PM02_DEFAULT,
	RB_CXD9225GG_PM03_DEFAULT,
	RB_CXD9225GG_PM04_DEFAULT,
	RB_CXD9225GG_PM05_DEFAULT,
	RB_CXD9225GG_PM10_DEFAULT,
	RB_CXD9225GG_PM11_DEFAULT,
	RB_CXD9225GG_PM12_DEFAULT,
	RB_CXD9225GG_PM13_DEFAULT,
	RB_CXD9225GG_PM14_DEFAULT,
	RB_CXD9225GG_PM15_DEFAULT
};

static u32 gpio_pr_default_value[CXD9225GG_GPIO_PORT_NUM] = {
	RB_CXD9225GG_PORT00_DEFAULT,
	RB_CXD9225GG_PORT01_DEFAULT,
	RB_CXD9225GG_PORT02_DEFAULT,
	RB_CXD9225GG_PORT03_DEFAULT,
	RB_CXD9225GG_PORT04_DEFAULT,
	RB_CXD9225GG_PORT05_DEFAULT,
	RB_CXD9225GG_PORT10_DEFAULT2,
	RB_CXD9225GG_PORT11_DEFAULT,
	RB_CXD9225GG_PORT12_DEFAULT,
	RB_CXD9225GG_PORT13_DEFAULT,
	RB_CXD9225GG_PORT14_DEFAULT,
	RB_CXD9225GG_PORT15_DEFAULT
};
#endif	/* !CONFIG_DPR_SUPPORT */

/* trace function of GPIO module. */
/* #define CXD9225GG_GPIO_DEBUG */
#ifdef CXD9225GG_GPIO_DEBUG
	#define GPIO_DEBUG(fmt, args...)	\
		printk(KERN_DEBUG "%s: " fmt, __func__, ## args)
	#define GPIO_DEBUG_PRINT_ALLREG()    _debug_print_allreg()
	#define GPIO_DEBUG_PRINT_REGOF(p)    _debug_print_regof(p)
#else
	#define GPIO_DEBUG(fmt, args...)
	#define GPIO_DEBUG_PRINT_ALLREG()
	#define GPIO_DEBUG_PRINT_REGOF(p)
#endif

/* #define CXD9225GG_GPIO_ERROR */
#ifdef CXD9225GG_GPIO_ERROR
	#define GPIO_WARNING(fmt, args...)	\
		printk(KERN_ERR "%s: " fmt, __func__, ## args)
#else
	#define GPIO_WARNING(fmt, args...)
#endif


/********************************************************
        Local Variables
********************************************************/

static cxd9225gg_gpio_port_t gpio_port[CXD9225GG_GPIO_PORT_NUM];

/*
 * defines a spinlock to protected the accessing to gpio pin.
 */
DEFINE_SPINLOCK(gpio_mux_lock);

/********************************************************
        GPIO High/Low Control Driver
********************************************************/
#define DEVICE_NAME  "gpio_hlc"

enum GPIO_HLC_CTRL
{
	 GPIO_HLC_IOC_LOW = 0
	,GPIO_HLC_IOC_HIGH = 1
};

static int gpio_hlc_major = CXD9225GG_MAJOR_ID_GPIO;
static int gpio_hlc_minor = 0;
static dev_t gpio_hlc_devid = 0;
struct cdev *pgpio_hlc_dev = NULL;

/********************************************************
        Functions
********************************************************/

#ifdef CXD9225GG_GPIO_DEBUG
static void _debug_print_allreg(void)
{
	int i = 0;
	u32 portAddress = 0;
	u32 reg = 0;
	cxd9225gg_gpio_port_t *port;

	printk( KERN_DEBUG "------------------------------------------\n");
	
	/* PORT MODE CONTROLLER Register of Shared IO 00-05 */
	for (i = 0; i < 6; i++) {
		port = get_gpio_port(i);
		portAddress = port->base + CXD9225GG_GPIO_PORT_MODE_CONTROLLER_REG;
		reg = readl(portAddress);
		printk( KERN_DEBUG "GPIO num: %d, PMCr: 0x%X.\n", i, reg);
	}

	/* PORT MODE Register of Shared IO 00-05 & GPIO 10-15 */
	for (i = 0; i < 12; i++) {
		port = get_gpio_port(i);
		portAddress = port->base + CXD9225GG_GPIO_PORT_MODE_REG;
		reg = readl(portAddress);
		printk( KERN_DEBUG "GPIO num: %d, PMr: 0x%X.\n", i, reg);
	}

	/* PORT Register of GPIO 00-05 & GPIO 10-15 */
	for (i = 0; i < 12; i++) {
		port = get_gpio_port(i);
		portAddress = port->base + CXD9225GG_GPIO_PORT_REG;
		reg = readl(portAddress);
		printk( KERN_DEBUG "GPIO num: %d, Pr: 0x%X.\n", i, reg);
	}	

	/* PORTIN Register of GPIO 00-05 & GPIO 10-15 */
	for (i = 0; i < 12; i++) {
		port = get_gpio_port(i);
		portAddress = port->base + CXD9225GG_GPIO_PORT_IN_REG;
		reg = readl(portAddress);
		printk( KERN_DEBUG "GPIO num: %d, PINr: 0x%X.\n", i, reg);
	}

	printk( KERN_DEBUG "------------------------------------------\n");
	
}


static void _debug_print_regof(cxd9225gg_gpio_port_t *port)
{
	u32 portAddress = 0;
	u32 reg = 0;
	
	printk( KERN_DEBUG "******************************************\n");
	
	portAddress = port->base + CXD9225GG_GPIO_PORT_MODE_CONTROLLER_REG;
	reg = readl(portAddress);
	printk( KERN_DEBUG "PMCr: 0x%X.\n", reg);
	
	portAddress = port->base + CXD9225GG_GPIO_PORT_MODE_REG;
	reg = readl(portAddress);
	printk( KERN_DEBUG "PMr: 0x%X.\n", reg);
	
	portAddress = port->base + CXD9225GG_GPIO_PORT_REG;
	reg = readl(portAddress);
	printk( KERN_DEBUG "Pr: 0x%X.\n", reg);
	
	portAddress = port->base + CXD9225GG_GPIO_PORT_IN_REG;
	reg = readl(portAddress);
	printk( KERN_DEBUG "PINr: 0x%X.\n", reg);
	
	printk( KERN_DEBUG "******************************************\n");

}
#endif


/*
 * Check if the gpio index and the pin number are valid.
 * @param gpio		a gpio port index
 * @param pin		a pin number
 * @return		0 if the gpio index and pin number are valid; -1 otherwise
 */
static int check_gpio(u32 gpio, int pin)
{
	if (gpio > CXD9225GG_GPIO_PORT_15) {
		GPIO_WARNING( "cxd9225gg-gpio: invalid GPIO %d\n", gpio);
		return -1;
	}
	if (pin < 0 || pin > 7) {
		GPIO_WARNING( "cxd9225gg-gpio: invalid pin %d\n", pin);
		return -1;
	}
	if (CXD9225GG_GPIO_PORT_00 == gpio || CXD9225GG_GPIO_PORT_04 == gpio)
		if (pin < 0 || pin > 2) {
			GPIO_WARNING( "cxd9225gg-gpio: valid gpio:%d, invalid pin %d\n", gpio, pin);
			return -1;
		}
	if (CXD9225GG_GPIO_PORT_01 == gpio || CXD9225GG_GPIO_PORT_02 == gpio
		|| CXD9225GG_GPIO_PORT_03 == gpio || CXD9225GG_GPIO_PORT_05 == gpio)
		if (pin < 0 || pin > 1) {
			GPIO_WARNING( "cxd9225gg-gpio: valid gpio:%d, invalid pin %d\n", gpio, pin);
			return -1;
		}
	if (CXD9225GG_GPIO_PORT_15 == gpio)
		if (pin < 0 || pin > 3) {
			GPIO_WARNING( "cxd9225gg-gpio: valid gpio:%d, invalid pin %d\n", gpio, pin);
			return -1;
		}

	return 0;
}

/**
 * Exported function to set a GPIO pin's direction
 * @param pin		a name defined by \b cxd9225gg_gpio_pin_name_t
 * @param is_input	1 (or non-zero) for input; 0 for output
 */
void cxd9225gg_set_gpio_direction(cxd9225gg_gpio_pin_name_t pin, int is_input)
{
	u32 pmr = 0;
	u32 portAddress = 0;
	u32 gpio = 0;
	u32 index = 0;
	cxd9225gg_gpio_port_t *port = NULL;

	gpio = CXD9225GG_GPIO_ID_TO_PORT(pin);
	index = CXD9225GG_GPIO_ID_TO_INDEX(pin);
	GPIO_DEBUG( "gpio num: %d, pin: %d.\n", gpio, index);

	if (check_gpio(gpio, index) < 0)
		return;

	port = &gpio_port[gpio];
	spin_lock(&port->lock);

	portAddress = port->base + CXD9225GG_GPIO_PORT_MODE_REG;
	GPIO_DEBUG( "gpio portAddress: 0x%X.\n", portAddress);
	pmr = readl(portAddress);
	GPIO_DEBUG( "gpio get pmr: 0x%X.\n", pmr);

	/*
	 * 0 : ouput port
	 * 1 : input port
	*/
	if (is_input) {
		pmr |= (1 << index);
	}
	else {
		pmr &= (~(1 << index));
	}
	writel(pmr, portAddress);
	GPIO_DEBUG( "gpio put pmr: 0x%X.\n", pmr);

	GPIO_DEBUG_PRINT_REGOF(port);

	spin_unlock(&port->lock);
}


/*
 * Exported function to set a GPIO pin's data output
 * @param pin		a name defined by \b cxd9225gg_gpio_pin_name_t
 * @param data		value to be set (only 0 or 1 is valid)
 */
void cxd9225gg_set_gpio_dataout(cxd9225gg_gpio_pin_name_t pin, u32 data)
{
	u32 pr = 0; /* port register of GPIO port */
	u32 portAddress = 0;
	u32 gpio = 0;
	u32 index = 0;
	cxd9225gg_gpio_port_t *port = NULL;

	gpio = CXD9225GG_GPIO_ID_TO_PORT(pin);
	index = CXD9225GG_GPIO_ID_TO_INDEX(pin);
	GPIO_DEBUG( "gpio num: %d, pin: %d.\n", gpio, index);

	if (check_gpio(gpio, index) < 0)
		return;

	port = &gpio_port[gpio];
	spin_lock(&port->lock);

	portAddress = port->base + CXD9225GG_GPIO_PORT_REG;
	GPIO_DEBUG( "gpio portAddress: 0x%X.\n", portAddress);

	pr = readl(portAddress);
	GPIO_DEBUG( "gpio get pr: 0x%X.\n", pr);

	if (data == 0) {
		pr &= (~(1 << index));
	}
	else if (data == 1) {
		pr |= (1 << index);
	}
	writel(pr, portAddress);
	GPIO_DEBUG( "gpio put pr: 0x%X.\n", pr);

	GPIO_DEBUG_PRINT_REGOF(port);

	spin_unlock(&port->lock);

}


/**
 * Return the data value of a GPIO signal.
 * @param pin	a name defined by \b cxd9225gg_gpio_pin_name_t
 * @return 	value (0 or 1) of the GPIO signal; -EINVAL if pass in invalid pin
 */
int cxd9225gg_get_gpio_datain(cxd9225gg_gpio_pin_name_t pin)
{
	u32 pir = 0; /* port in register of GPIO port */
	int ret = -1;
	u32 portAddress = 0;
	u32 gpio = 0;
	u32 index = 0;
	cxd9225gg_gpio_port_t *port = NULL;

	gpio = CXD9225GG_GPIO_ID_TO_PORT(pin);
	index = CXD9225GG_GPIO_ID_TO_INDEX(pin);
	GPIO_DEBUG( "gpio num: %d, pin: %d.\n", gpio, index);

	if (check_gpio(gpio, index) < 0)
		return -EINVAL;

	port = &gpio_port[gpio];
	spin_lock(&port->lock);

	portAddress = port->base + CXD9225GG_GPIO_PORT_IN_REG;
	GPIO_DEBUG( "gpio portAddress: 0x%X.\n", portAddress);

	pir = readl(portAddress);
	GPIO_DEBUG( "gpio get pir: 0x%X.\n", pir);

	pir &= (1 << index);

	if ( pir > 0 ) {
		ret = 1;
	}
	else {
		ret = 0;
	}

	GPIO_DEBUG_PRINT_REGOF(port);

	spin_unlock(&port->lock);

	return ret;

}


/**
 * Request ownership for a GPIO pin. The caller has to check the return value
 * of this function to make sure it returns 0 before make use of that pin.
 * @param pin		a name defined by \b cxd9225gg_gpio_pin_name_t
 * @return		0 if successful; Non-zero otherwise
 */
int cxd9225gg_request_gpio(cxd9225gg_gpio_pin_name_t pin)
{
	u32 gpio = 0;
	u32 index = 0;
	cxd9225gg_gpio_port_t *port = NULL;

	gpio = CXD9225GG_GPIO_ID_TO_PORT(pin);
	index = CXD9225GG_GPIO_ID_TO_INDEX(pin);

	GPIO_DEBUG( "gpio num: %d, pin: %d.\n", gpio, index);

	if (check_gpio(gpio, index) < 0)
		return -EINVAL;

	port = get_gpio_port(gpio);
	spin_lock(&port->lock);

	if (port->reserved_map & (1 << index)) {
		GPIO_WARNING(
			   "GPIO port %d (0-based), pin %d is already reserved!\n",
			   gpio, index);
		spin_unlock(&port->lock);
		return -EBUSY;
	}
	port->reserved_map |= (1 << index);

	/* 00 ~ 05 are shared IO for GPIOs */
	if ( gpio <= CXD9225GG_GPIO_PORT_05 ){
		if (cxd9225gg_gpio_config_mux(pin, GPIO_MUX_GPIO) < 0) {
			spin_unlock(&port->lock);
			return -EINVAL;
		}
	}

	spin_unlock(&port->lock);

	return 0;
}

/**
 * Release ownership for a GPIO pin
 * @param pin		a name defined by \b cxd9225gg_gpio_pin_name_t
 */
void cxd9225gg_free_gpio(cxd9225gg_gpio_pin_name_t pin)
{
	u32 gpio = 0;
	u32 index = 0;
	cxd9225gg_gpio_port_t *port = NULL;

	gpio = CXD9225GG_GPIO_ID_TO_PORT(pin);
	index = CXD9225GG_GPIO_ID_TO_INDEX(pin);

	if (check_gpio(gpio, index) < 0)
		return;

	port = get_gpio_port(gpio);

	spin_lock(&port->lock);
	if ((!(port->reserved_map & (1 << index)))) {
		GPIO_WARNING( "GPIO port %d, pin %d wasn't reserved!\n",
			   gpio, index);
		spin_unlock(&port->lock);
		return;
	}
	port->reserved_map &= ~(1 << index);
	spin_unlock(&port->lock);

}


/*
 * Check if the gpio index and the pin number are valid.
 * @param gpio		a gpio port index
 * @param pin		a pin number
 * @return		0 if the gpio index and pin number are valid; -1 otherwise
 */
static int check_gpio_irq(u32 gpio, int pin)
{
	if (gpio < CXD9225GG_GPIO_PORT_13 || gpio > CXD9225GG_GPIO_PORT_15) {
		GPIO_WARNING( "cxd9225gg-gpio: invalid GPIO %d\n", gpio);
		return -1;
	}
	if (pin < 0 || pin > 7) {
		GPIO_WARNING( "cxd9225gg-gpio: invalid pin %d\n", pin);
		return -1;
	}

	if (CXD9225GG_GPIO_PORT_15 == gpio)
		if (pin < 0 || pin > 3) {
			GPIO_WARNING( "cxd9225gg-gpio: valid gpio:%d, invalid pin %d\n", gpio, pin);
			return -1;
		}


	return 0;
}

/*
 * This function is responsible for registering a GPIO signal's ISR.
 * @param pin		a name defined by \b cxd9225gg_gpio_pin_name_t
 * @param desc		an irqaction assigned to the pin's ISR
 * @return 0 			if successful; non-zero otherwise.
 */
int cxd9225gg_gpio_irq_request(cxd9225gg_gpio_pin_name_t pin, struct irqaction *desc)
{
	int irq;
	u32 gpio = 0;
	u32 index = 0;
	int ret = -EINVAL;
	cxd9225gg_gpio_port_t *port = NULL;

	if ( desc == NULL )
		return ret;
	
	ret = cxd9225gg_request_gpio(pin);
	if (ret < 0)
		return ret;

	gpio = CXD9225GG_GPIO_ID_TO_PORT(pin);
	index = CXD9225GG_GPIO_ID_TO_INDEX(pin);

	if (check_gpio_irq(gpio,index) < 0)
		return ret;

	GPIO_DEBUG( "gpio num: %d, pin: %d.\n", gpio, index);
	
	port = get_gpio_port(gpio);
	irq = CXD9225GG_GPIO_TO_IRQ(gpio, index);
	GPIO_DEBUG( "irq num: %d.\n", irq);
	
	ret = request_irq(irq, desc->handler, desc->flags, desc->name, desc->dev_id);
	if (ret == 0) {
		port->irq_is_requested_map |= (1 << index);	
		GPIO_DEBUG( "gpio %d interrupt enabled.\n", irq);
	}
		
	return ret;
}

/*
 * Free a GPIO signal's ISR.
 * @param pin		a name defined by \b cxd9225gg_gpio_pin_name_t
 * @param desc		an irqaction assigned to the pin's ISR
 */
void cxd9225gg_gpio_irq_free(cxd9225gg_gpio_pin_name_t pin, struct irqaction *desc)
{
	int irq;
	u32 gpio = 0;
	u32 index = 0;
	cxd9225gg_gpio_port_t *port = NULL;

	if (desc == NULL)
		return;
	
	gpio = CXD9225GG_GPIO_ID_TO_PORT(pin);
	index = CXD9225GG_GPIO_ID_TO_INDEX(pin);
	if (check_gpio_irq(gpio,index) < 0)
		return;
	
	port = get_gpio_port(gpio);
	if ((!(port->irq_is_requested_map & (1 << index)))) {
		GPIO_WARNING( "GPIO port %d, pin %d 's irq wasn't requested!\n",
			   gpio, index);		
		return;
	}
	
	irq = CXD9225GG_GPIO_TO_IRQ(gpio, index);
	free_irq(irq, desc->dev_id);
	port->irq_is_requested_map &= ~(1 << index);

	cxd9225gg_free_gpio(pin);

	GPIO_DEBUG( "Free irq num: %d.\n", irq);
}


/*
 * Enable/disable a GPIO signal's interrupt.
 * @param pin         a name defined by \b cxd9225gg_gpio_pin_name_t
 * @param enable        \b #true for enabling the interrupt; \b #false otherwise
 */
void cxd9225gg_gpio_irq_enable(cxd9225gg_gpio_pin_name_t pin, bool enable)
{
	int irqno = 0;
	u32 gpio = 0;
	u32 index = 0;
	cxd9225gg_gpio_port_t *port = NULL;

	gpio = CXD9225GG_GPIO_ID_TO_PORT(pin);
	index = CXD9225GG_GPIO_ID_TO_INDEX(pin);
	if (check_gpio_irq(gpio, index) <0)
		return ;

	port = get_gpio_port(gpio);
	if ((!(port->irq_is_requested_map & (1 << index)))) {
		GPIO_WARNING( "GPIO port %d, pin %d 's irq wasn't requested!\n",
			   gpio, index);		
		return;
	}
	
	irqno = CXD9225GG_GPIO_TO_IRQ(gpio,index);
	if (enable)
		enable_irq(irqno);
	else
		disable_irq(irqno);

	
}

/*
 * Set a GPIO pin's interrupt edge
 * @param pin         a name defined by \b cxd9225gg_gpio_pin_name_t
 * @param icr           one of the values defined in \b gpio_edge_t
 *                      to indicate how to generate an interrupt
 */
int cxd9225gg_gpio_irq_set_type(cxd9225gg_gpio_pin_name_t pin, gpio_int_cfg_t cfg_type)
{
	int irqno = 0;
	u32 gpio = 0;
	u32 index = 0;
	int ret = -EINVAL;
	cxd9225gg_gpio_port_t *port = NULL;

	gpio = CXD9225GG_GPIO_ID_TO_PORT(pin);
	index = CXD9225GG_GPIO_ID_TO_INDEX(pin);
	if (check_gpio_irq(gpio, index) <0)
		return -EINVAL;;

	port = get_gpio_port(gpio);
	if ((!(port->irq_is_requested_map & (1 << index)))) {
		GPIO_WARNING( "GPIO port %d, pin %d 's irq wasn't requested!\n",
			   gpio, index);		
		return -EINVAL;
	}
	
	irqno = CXD9225GG_GPIO_TO_IRQ(gpio,index);
	GPIO_DEBUG( "irq num: %d.\n", irqno);

	switch (cfg_type) {
	case GPIO_INT_LOW_LEV:
		ret = set_irq_type(irqno, IRQ_TYPE_LEVEL_LOW);
		break;
	case GPIO_INT_HIGH_LEV:
		ret = set_irq_type(irqno, IRQ_TYPE_LEVEL_HIGH);
		break;
	case GPIO_INT_FALL_EDGE:
		ret = set_irq_type(irqno, IRQ_TYPE_EDGE_FALLING);
		break;
	case GPIO_INT_RISE_EDGE:
		ret = set_irq_type(irqno, IRQ_TYPE_EDGE_RISING);
		break;
	case GPIO_INT_BOTH_EDGE:
		ret = set_irq_type(irqno, IRQ_TYPE_EDGE_BOTH);
		break;
	case GPIO_INT_NONE:
		ret = set_irq_type(irqno, IRQ_TYPE_NONE);
		break;

	default:
		return -EINVAL;
	}

	return ret;
}

/*
 * Check if the gpio index and the pin number are valid.
 * @param gpio		a gpio port index
 * @param pin		a pin number
 * @return		0 if the gpio index and pin number are valid; -1 otherwise
 */
static int check_gpio_mux(u32 gpio, int pin)
{
	if (gpio > CXD9225GG_GPIO_PORT_05) {
		GPIO_WARNING( "cxd9225gg-gpio: invalid GPIO %d\n", gpio);
		return -1;
	}
	if (pin < 0 || pin > 7) {
		GPIO_WARNING( "cxd9225gg-gpio: invalid pin %d\n", pin);
		return -1;
	}

	if (CXD9225GG_GPIO_PORT_00 == gpio || CXD9225GG_GPIO_PORT_04 == gpio)
		if (pin < 0 || pin > 2) {
			GPIO_WARNING( "cxd9225gg-gpio: valid gpio:%d, invalid pin %d\n", gpio, pin);
			return -1;
		}
	if (CXD9225GG_GPIO_PORT_01 == gpio || CXD9225GG_GPIO_PORT_02 == gpio
		|| CXD9225GG_GPIO_PORT_03 == gpio || CXD9225GG_GPIO_PORT_05 == gpio)
		if (pin < 0 || pin > 1) {
			GPIO_WARNING( "cxd9225gg-gpio: valid gpio:%d, invalid pin %d\n", gpio, pin);
			return -1;
		}

	return 0;
}


/**
 *@brief cxd9225gg_gpio_config_mux - just configure the mode of the gpio pin.
 *@param pin   a pin name as defined in \b #cxd9225gg_gpio_pin_name_t
 *@param mode  a module as define in \b #cxd9225gg_gpio_mux_mode_t;
 *	GPIO_MUX_GPIO set pin to work as GPIO port
 *	GPIO_MUX_UART set pin to work as UART
 *	GPIO_MUX_CSI set pin to work as CSI
 *@return      0 if successful, Non-zero otherwise
 */
int cxd9225gg_gpio_config_mux(cxd9225gg_gpio_pin_name_t pin, cxd9225gg_gpio_mux_mode_t mode)
{
	int pmcr = 0; /* port mode control register */
	int portAddress = 0;
	unsigned long lock_flags;
	u32 gpio = 0;
	u32 index = 0;
	cxd9225gg_gpio_port_t *port = NULL;

	gpio = CXD9225GG_GPIO_ID_TO_PORT(pin);
	index = CXD9225GG_GPIO_ID_TO_INDEX(pin);

	GPIO_DEBUG( "gpio num: %d, pin: %d.\n", gpio, index);
	
	/* 00 ~ 05 are shared IO for GPIOs */
	if (check_gpio_mux(gpio, index) < 0)
		return -EINVAL;

	port = get_gpio_port(gpio);

	spin_lock_irqsave(&gpio_mux_lock, lock_flags);

	portAddress = port->base + CXD9225GG_GPIO_PORT_MODE_CONTROLLER_REG;
	pmcr = readl(portAddress);
	GPIO_DEBUG( "gpio get pmcr: 0x%X.\n", pmcr);

	switch (mode) {
	case GPIO_MUX_GPIO:
		pmcr &= (~(1 << index));
		break;
	case GPIO_MUX_UART:
		pmcr |= (1 << index);
		break;
	case GPIO_MUX_CSI:
		pmcr |= (1 << index);
		break;
	default:
		spin_unlock_irqrestore(&gpio_mux_lock, lock_flags);
		return -EINVAL;
	}
	writel(pmcr, portAddress);

	GPIO_DEBUG_PRINT_REGOF(port);
	
	spin_unlock_irqrestore(&gpio_mux_lock, lock_flags);
	return 0;
}




/**
 * This function is just used to request a pin and configure it.
 * @param pin	a pin name as defined in \b #cxd9225gg_gpio_pin_name_t
 * @param mode	a module as define in \b #cxd9225gg_gpio_mux_mode_t;
 * @return	0 if successful, Non-zero otherwise
 */
int cxd9225gg_gpio_request_mux(cxd9225gg_gpio_pin_name_t pin, cxd9225gg_gpio_mux_mode_t mode)
{
	int ret = 0;
	ret = cxd9225gg_request_gpio(pin);
	if (ret == 0) {
		ret = cxd9225gg_gpio_config_mux(pin, mode);
		if (ret != 0) {
			cxd9225gg_free_gpio(pin);
		}
	}
	return ret;
}


/**
 * This function is just used to release a pin.
 * @param pin	a pin name as defined in \b #cxd9225gg_gpio_pin_name_t
 * @return	none	
 */
void cxd9225gg_gpio_free_mux(cxd9225gg_gpio_pin_name_t pin)
{
	cxd9225gg_free_gpio(pin);
}

/********************************************************
        GPIO High/Low Control Driver
********************************************************/
static int gpio_hlc_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int ret = 0;
	void __user *argp = (void __user *)arg;
	unsigned long gpioId;
	if(copy_from_user(&gpioId, argp, sizeof(unsigned long)) != 0 )
	{
		GPIO_DEBUG("copy_from_user error! line=%d\n",__LINE__);
		ret = -EFAULT;
	}

	switch(cmd)
	{
	case GPIO_HLC_IOC_LOW:
        cxd9225gg_set_gpio_direction(gpioId, false);
		cxd9225gg_set_gpio_dataout(gpioId, 0);
		break;
	case GPIO_HLC_IOC_HIGH:
		cxd9225gg_set_gpio_direction(gpioId, false);
		cxd9225gg_set_gpio_dataout(gpioId, 1);
		break;
	default :
		GPIO_DEBUG("Ioctol: no type run [ --kernel-- ]\n");
		ret = -1;
	}
	return ret;
}


static int gpio_hlc_open(struct inode *inode, struct file *file)
{
	return 0;
}

static int gpio_hlc_release(struct inode *inode, struct file *file)
{
	return 0;
}

/* file operations */
static struct file_operations gpio_hlc_fops = {
      .open = gpio_hlc_open,
      .release = gpio_hlc_release,
      .ioctl = gpio_hlc_ioctl,
};

static int __init gpio_hlc_init(void)
{
	int ret = -1;

	GPIO_DEBUG("Hello gpio_hlc.\n\n");

    /* register a char device driver */
	if (gpio_hlc_major)
	{
		gpio_hlc_devid = MKDEV(gpio_hlc_major, gpio_hlc_minor);
		ret = register_chrdev_region(gpio_hlc_devid, 1, DEVICE_NAME);
	}
	else
	{
		ret = alloc_chrdev_region(&gpio_hlc_devid, gpio_hlc_minor, 1, DEVICE_NAME);
		gpio_hlc_major = MAJOR(gpio_hlc_devid);
	}
	if (ret < 0)
	{
		GPIO_DEBUG("gpio_hlc: can't get major %d\n", gpio_hlc_major);
		return ret;
	}

	GPIO_DEBUG("maior_number =%d\n", gpio_hlc_major);

	pgpio_hlc_dev = cdev_alloc();
	pgpio_hlc_dev->owner = THIS_MODULE;
	pgpio_hlc_dev->ops = &gpio_hlc_fops;
	ret = cdev_add(pgpio_hlc_dev, gpio_hlc_devid, 1);
	if(ret != 0)
    {
		GPIO_DEBUG("Err occurs when adding cdev into kernel!\n");
		return ret;
    }
	return ret;
}

static void __exit gpio_hlc_exit(void)
{
	GPIO_DEBUG("Goodbye gpio_hlc.\n");
	cdev_del(pgpio_hlc_dev);
	unregister_chrdev_region(gpio_hlc_devid, 1);
}


#ifdef CONFIG_PM
/**
 * This function puts the GPIO in low-power mode/state.
 * All the interrupts that are enabled are first saved.
 * Only those interrupts which registers as a wake source by calling
 * enable_irq_wake are enabled. All other interrupts are disabled.
 *
 * @param   dev  the system device structure used to give information
 *                on GPIO to suspend
 * @param   mesg the power state the device is entering
 *
 * @return  The function always returns 0.
 */
static int cxd9225gg_gpio_suspend(struct sys_device *dev, pm_message_t mesg)
{
	return 0;
}

/**
 * This function brings the GPIO back from low-power state.
 * All the interrupts enabled before suspension are re-enabled from
 * the saved information.
 *
 * @param   dev  the system device structure used to give information
 *                on GPIO to resume
 *
 * @return  The function always returns 0.
 */
static int cxd9225gg_gpio_resume(struct sys_device *dev)
{
	return 0;
}
#else
#define cxd9225gg_gpio_suspend  NULL
#define cxd9225gg_gpio_resume   NULL
#endif				/* CONFIG_PM */

/**
 * This structure contains pointers to the power management callback functions.
 */
static struct sysdev_class cxd9225gg_gpio_sysclass = {
	.name = "cxd9225gg_gpio",
	.suspend = cxd9225gg_gpio_suspend,
	.resume = cxd9225gg_gpio_resume,
};

/**
 * This structure represents GPIO as a system device.
 * System devices follow a slightly different driver model.
 * They don't need to do dynammic driver binding, can't be probed,
 * and don't reside on any type of peripheral bus.
 * So, it is represented and treated a little differently.
 */
static struct sys_device cxd9225gg_gpio_device = {
	.id = 0,
	.cls = &cxd9225gg_gpio_sysclass,
};

/**
 * This function initializes the GPIO hardware.
 */
static int __init _cxd9225gg_gpio_init(void)
{
	int i = 0;
	u32 reg = 0;
	u32 portAddress = 0;
	cxd9225gg_gpio_port_t *port = NULL;

	GPIO_DEBUG("CXD9225GG GPIO hardware\n");

	for (i = 0; i < CXD9225GG_GPIO_PORT_NUM; i++) {
		port = &gpio_port[i];
		port->base = CXD9225GG_GPIO_REG_VIRT(gpio_port_offset[i], CXD9225GG_GPIO_PORT_REG);
		port->reserved_map = 0;
		port->irq_is_requested_map = 0;
		spin_lock_init(&port->lock);
	}

#ifdef CONFIG_DPR_SUPPORT
	/* SPI Flash Init */
	portAddress = gpio_port[CXD9225GG_GPIO_PORT_10].base + CXD9225GG_GPIO_PORT_REG;
	reg = readl(portAddress);
	reg |= 0x20;
	writel(reg, portAddress);
	portAddress = gpio_port[CXD9225GG_GPIO_PORT_12].base + CXD9225GG_GPIO_PORT_REG;
	reg = readl(portAddress);
	reg |= 0x20;
	writel(reg, portAddress);
#else
	/*----------------------------------------------------------*
	 * GPIO register setup                                      *
	 *----------------------------------------------------------*/

	/* Setup PORT MODE CONTROLLER Register of Shared IO 00-05 */
	for (i = 0; i < 6; i++) {
		portAddress = gpio_port[i].base + CXD9225GG_GPIO_PORT_MODE_CONTROLLER_REG;
		writel(gpio_pmcr_default_value[i], portAddress);
	}

	/* Setup PORT MODE Register of Shared IO 00-05 & GPIO 10-15 */
	for (i = 0; i < 12; i++) {
		portAddress = gpio_port[i].base + CXD9225GG_GPIO_PORT_MODE_REG;
		writel(gpio_pmr_default_value[i], portAddress);
	}

	/* Setup PORT Register of GPIO 10-15 */
	for (i = 6; i < 12; i++) {
		if(i !=7 ){
			portAddress = gpio_port[i].base + CXD9225GG_GPIO_PORT_REG;
			writel(gpio_pr_default_value[i], portAddress);
		}
	}

	GPIO_DEBUG_PRINT_ALLREG();
	cxd9225gg_set_gpio_dataout(CXD9225GG_BGPIO_08,1);
	cxd9225gg_set_gpio_dataout(CXD9225GG_BGPIO_09,1);
	cxd9225gg_set_gpio_dataout(CXD9225GG_BGPIO_10,0);
	cxd9225gg_set_gpio_dataout(CXD9225GG_BGPIO_11,1);
	cxd9225gg_set_gpio_dataout(CXD9225GG_BGPIO_12,1);

/*
	REG_GPIO_PMC02 |= 0x03;
	REG_GPIO_PM02 |=  0x02;
	REG_GPIO_PM02 &= ~0x01;
*/

	portAddress = gpio_port[CXD9225GG_GPIO_PORT_02].base + CXD9225GG_GPIO_PORT_MODE_CONTROLLER_REG;
	reg = readl(portAddress);
	reg |= 0x03;
	writel(reg, portAddress);

	portAddress = gpio_port[CXD9225GG_GPIO_PORT_02].base + CXD9225GG_GPIO_PORT_MODE_REG;
	reg = readl(portAddress);
	reg |= 0x02;
	reg &= ~0x01;
	writel(reg, portAddress);
	GPIO_DEBUG_PRINT_ALLREG();

	cxd9225gg_set_gpio_direction(CXD9225GG_GPINT_11, 1);
#endif	/* CONFIG_DPR_SUPPORT */

	return 0;
}

/**
 * This function registers GPIO hardware as a system device and
 * intializes all the GPIO ports if not already done.
 * System devices will only be suspended with interrupts disabled, and
 * after all other devices have been suspended. On resume, they will be
 * resumed before any other devices, and also with interrupts disabled.
 * This may get called early from board specific init
 *
 * @return       This function returns 0 on success.
 */
int __init cxd9225gg_gpio_init(void)
{
	int ret = 0;

	ret = _cxd9225gg_gpio_init();

	if (ret == 0) {
		ret = sysdev_class_register(&cxd9225gg_gpio_sysclass);
		if (ret == 0)
			ret = sysdev_register(&cxd9225gg_gpio_device);
	}

	ret = gpio_hlc_init();

	return ret;
}

cxd9225gg_gpio_port_t *get_gpio_port(u32 gpio) {
	return &gpio_port[gpio];
}

EXPORT_SYMBOL(cxd9225gg_request_gpio);
EXPORT_SYMBOL(cxd9225gg_free_gpio);
EXPORT_SYMBOL(cxd9225gg_set_gpio_direction);
EXPORT_SYMBOL(cxd9225gg_set_gpio_dataout);
EXPORT_SYMBOL(cxd9225gg_get_gpio_datain);
EXPORT_SYMBOL(get_gpio_port);

EXPORT_SYMBOL(cxd9225gg_gpio_config_mux);
EXPORT_SYMBOL(cxd9225gg_gpio_request_mux);
EXPORT_SYMBOL(cxd9225gg_gpio_free_mux);

EXPORT_SYMBOL(cxd9225gg_gpio_irq_request);
EXPORT_SYMBOL(cxd9225gg_gpio_irq_free);
EXPORT_SYMBOL(cxd9225gg_gpio_irq_enable);
EXPORT_SYMBOL(cxd9225gg_gpio_irq_set_type);
