/* 2014-08-28: File changed by Sony Corporation */
/*
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */
#ifndef _BIO_DATA_POOL_H_
#define _BIO_DATA_POOL_H_

#ifdef __KERNEL__
#include<linux/types.h>
#else
#include<stdint.h>
#include<sys/ioctl.h>
#endif


/* major driver number for bio data pool device. */
#define BIO_DATA_POOL_MAJOR         289

/* device-name for bio data pool device. */
#define BIO_DATA_DEV_NAME       "bio_pool"


/* use get_free_pages or ioremap. */
#define BIO_POOL_MALLOC


/*  */
#ifdef BIO_POOL_MALLOC

/* total size of reserved memory for fs module.
 * bio-pool:          (10 * 1024 *1024)
 * gap-size:          (16 * 1024)
 * FS_MEM_POOL_SIZE:  (54 * 1024 * 1024 - 32 * 1024)
 * * gap-size:        (16 * 1024)
 */
#define BIO_DATA_POOL_SIZE      ( (48 * 1024* 1024) + 0 )

/* when we use __get_free_pages,
 * since max_order==15, so could only get 64M-bytes at once.
 * MIFMC_WORKMEM_MAXSIZE:  (16 * 1024 * 1024)
 */
#define MIFMC_MEM_POOL_SIZE      ( (32 * 1024* 1024) + 0 )

#define BIO_DATA_POOL_START_ADDR    (unsigned long)phys_to_virt(0x48000000)

#else

/* total size of reserved memory for fs module.
 * bio-pool:          (10 * 1024 *1024)
 * gap-size:          (16 * 1024)
 * FS_MEM_POOL_SIZE:  (54 * 1024 * 1024 - 32 * 1024)
 * * gap-size:        (16 * 1024)
 * MIFMC_WORKMEM_MAXSIZE:  (16 * 1024 * 1024).
 */
#define BIO_DATA_POOL_SIZE      ( (80 * 1024* 1024) + 0 )

#endif



/*  */
typedef struct tagBufDesc
{
    unsigned long phy_addr;
    unsigned long size;

}bio_buf_desc_t;


/* ioctl magic commands */
#define BIO_IO_MAGIC             'B'

/*  */
#define BIO_IO_CMD_PHY_ADDR         _IOWR( BIO_IO_MAGIC,  0, unsigned long)

/*  */
#define BIO_IO_CMD_MEM_SIZE         _IOWR( BIO_IO_MAGIC,  1, unsigned long)

/*  */
#define BIO_IO_CMD_FLUSH            _IOWR( BIO_IO_MAGIC,  2, bio_buf_desc_t)

/* when we use __get_free_pages,
 * since max_order==15, so could only get 64M-bytes at once.
 */
#ifdef BIO_POOL_MALLOC

/*  */
#define BIO_IO_CMD_PHY_ADDR_MIF     _IOWR( BIO_IO_MAGIC,  3, unsigned long)

/*  */
#define BIO_IO_CMD_MEM_SIZE_MIF     _IOWR( BIO_IO_MAGIC,  4, unsigned long)

/* Change this when new command appended. */
#define BIO_IO_MAX_NR       5

#else

/* Change this when new command appended. */
#define BIO_IO_MAX_NR       3

#endif

/* HSS releated */
#ifdef CONFIG_SNSC_HSS
#define BIO_D_HSS_NODE_NAME     "bio_d_hss_node_name"
#define BIO_D_HSS_NODE_PARENT   NULL
#endif  /* #ifdef CONFIG_SNSC_HSS */


#endif /* end of _BIO_DATA_POOL_H_. */

