/* 2014-08-28: File changed by Sony Corporation */
/*
 * Copyright (C) 2012 Sony Corporation
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef __ASM_ARCH_CXD9225GG_GPIO_H__
#define __ASM_ARCH_CXD9225GG_GPIO_H__

#include <linux/interrupt.h>
#include <asm/sizes.h>
#include <mach/hardware.h>
#include <asm/irq.h>


/* GPIO0 Offset from APB area top */
#define CXD9225GG_GPIO_REG_OFFSET_OF_APB	0x00008000

/* Offset address from GPIO0 area base */
#define CXD9225GG_GPIO_REG_OFFSET(n, r)   (r + 0x0020 * n)
/* Physical address */
#define CXD9225GG_GPIO_REG_PHYS(n, r)     (CXD9225GG_APB_REG_PHYS_BASE + \
                                         CXD9225GG_GPIO_REG_OFFSET_OF_APB + \
                                         CXD9225GG_GPIO_REG_OFFSET(n, r))
/* virtual address */
#define CXD9225GG_GPIO_REG_VIRT(n, r)     (CXD9225GG_APB_REG_VIRT_BASE + \
                                         CXD9225GG_GPIO_REG_OFFSET_OF_APB + \
                                         CXD9225GG_GPIO_REG_OFFSET(n, r))

#define CXD9225GG_GPIO_ID(n, r)            (((n) << 8) + (r))
#define CXD9225GG_GPIO_ID_TO_PORT(n)       (((n) & 0x0F00) >> 8)
#define CXD9225GG_GPIO_ID_TO_INDEX(n)      ((n) & 0xF)
#define CXD9225GG_GPIO_TO_IRQ(b,x)    (INT_GPIO0 + 8 * ((b) - CXD9225GG_GPIO_PORT_13) + (x))

typedef enum cxd9225gg_gpio_reg {
    CXD9225GG_GPIO_PORT_REG = 0x00,
    CXD9225GG_GPIO_PORT_MODE_REG = 0x04,
    CXD9225GG_GPIO_PORT_MODE_CONTROLLER_REG =0x08,
    CXD9225GG_GPIO_PORT_IN_REG = 0x10,
}cxd9225gg_gpio_reg_t;

typedef struct cxd9225gg_gpio_port {
	u32 base;
	u32 reserved_map;
	u32 irq_is_requested_map;
	spinlock_t lock;
}cxd9225gg_gpio_port_t;

/* index of gpio_port[] */
typedef enum cxd9225gg_gpio_port_index {
    CXD9225GG_GPIO_PORT_00 = 0,
    CXD9225GG_GPIO_PORT_01 = 1,
    CXD9225GG_GPIO_PORT_02 = 2,
    CXD9225GG_GPIO_PORT_03 = 3,
    CXD9225GG_GPIO_PORT_04 = 4,
    CXD9225GG_GPIO_PORT_05 = 5,
    CXD9225GG_GPIO_PORT_10 = 6,
    CXD9225GG_GPIO_PORT_11 = 7,
    CXD9225GG_GPIO_PORT_12 = 8,
    CXD9225GG_GPIO_PORT_13 = 9,
    CXD9225GG_GPIO_PORT_14 = 10,
    CXD9225GG_GPIO_PORT_15 = 11
}cxd9225gg_gpio_port_index_t;


typedef enum cxd9225gg_gpio_pins {
/* Shared IO port00 3 pins */
    CXD9225GG_SHIO_00 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_00,0),
    CXD9225GG_SHIO_01 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_00,1),
    CXD9225GG_SHIO_02 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_00,2),
/* Shared IO port01 2 pins */
    CXD9225GG_SHIO_03 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_01,0),
    CXD9225GG_SHIO_04 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_01,1),
/* Shared IO port02 2 pins */
    CXD9225GG_SHIO_05 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_02,0),
    CXD9225GG_SHIO_06 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_02,1),
/* Shared IO port03 2 pins */
    CXD9225GG_SHIO_07 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_03,0),
    CXD9225GG_SHIO_08 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_03,1),
/* Shared IO port04 3 pins */
    CXD9225GG_SHIO_09 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_04,0),
    CXD9225GG_SHIO_10 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_04,1),
    CXD9225GG_SHIO_11 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_04,2),
/* Shared IO port05 2 pins */
    CXD9225GG_SHIO_12 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_05,0),
    CXD9225GG_SHIO_13 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_05,1),
	
/* GPIO port10 8 pins */
    CXD9225GG_BGPIO_00 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_10,0),
    CXD9225GG_BGPIO_01 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_10,1),
    CXD9225GG_BGPIO_02 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_10,2),
    CXD9225GG_BGPIO_03 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_10,3),
    CXD9225GG_BGPIO_04 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_10,4),
    CXD9225GG_BGPIO_05 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_10,5),
    CXD9225GG_BGPIO_06 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_10,6),
    CXD9225GG_BGPIO_07 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_10,7),
/* GPIO port11 8 pins */
    CXD9225GG_BGPIO_08 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_11,0),
    CXD9225GG_BGPIO_09 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_11,1),
    CXD9225GG_BGPIO_10 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_11,2),
    CXD9225GG_BGPIO_11 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_11,3),
    CXD9225GG_BGPIO_12 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_11,4),
    CXD9225GG_BGPIO_13 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_11,5),
    CXD9225GG_BGPIO_14 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_11,6),
    CXD9225GG_BGPIO_15 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_11,7),
/* GPIO port12 8 pins */
    CXD9225GG_BGPIO_16 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_12,0),
    CXD9225GG_BGPIO_17 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_12,1),
    CXD9225GG_BGPIO_18 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_12,2),
    CXD9225GG_BGPIO_19 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_12,3),
    CXD9225GG_BGPIO_20 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_12,4),
    CXD9225GG_BGPIO_21 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_12,5),
    CXD9225GG_BGPIO_22 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_12,6),
    CXD9225GG_BGPIO_23 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_12,7),
/* GPIO shared with INT, port13 8 pins */
    CXD9225GG_GPINT_00 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,0),
    CXD9225GG_GPINT_01 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,1),
    CXD9225GG_GPINT_02 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,2),
    CXD9225GG_GPINT_03 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,3),
    CXD9225GG_GPINT_04 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,4),
    CXD9225GG_GPINT_05 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,5),
    CXD9225GG_GPINT_06 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,6),
    CXD9225GG_GPINT_07 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,7),
/* GPIO shared with INT, port14 8 pins */
    CXD9225GG_GPINT_08 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_14,0),
    CXD9225GG_GPINT_09 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_14,1),
    CXD9225GG_GPINT_10 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_14,2),
    CXD9225GG_GPINT_11 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_14,3),
    CXD9225GG_GPINT_12 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_14,4),
    CXD9225GG_GPINT_13 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_14,5),
    CXD9225GG_GPINT_14 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_14,6),
    CXD9225GG_GPINT_15 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_14,7),
/* GPIO shared with INT, port15 4 pins */
    CXD9225GG_GPINT_16 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_15,0),
    CXD9225GG_GPINT_17 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_15,1),
    CXD9225GG_GPINT_18 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_15,2),
    CXD9225GG_GPINT_19 =	CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_15,3)
	
} cxd9225gg_gpio_pin_name_t;


/* gpio related defines */

/* GPIO Register default value */
#define RB_CXD9225GG_PORT00_DEFAULT	0x00000000
#define RB_CXD9225GG_PORT01_DEFAULT	0x00000000
#define RB_CXD9225GG_PORT02_DEFAULT	0x00000000
#define RB_CXD9225GG_PORT03_DEFAULT	0x00000000
#define RB_CXD9225GG_PORT04_DEFAULT	0x00000000
#define RB_CXD9225GG_PORT05_DEFAULT	0x00000000
#define RB_CXD9225GG_PORT10_DEFAULT	0x00000073
#define RB_CXD9225GG_PORT10_DEFAULT2	0x000000Fb
#define RB_CXD9225GG_PORT11_DEFAULT	0x000000Fb
#define RB_CXD9225GG_PORT12_DEFAULT	0x000000F0
#define RB_CXD9225GG_PORT13_DEFAULT	0x00000000
#define RB_CXD9225GG_PORT14_DEFAULT	0x000000C9
#define RB_CXD9225GG_PORT15_DEFAULT	0x00000000
#define RB_CXD9225GG_PM00_DEFAULT		0x00000002
#define RB_CXD9225GG_PM01_DEFAULT		0x00000001
#define RB_CXD9225GG_PM02_DEFAULT		0x00000002
#define RB_CXD9225GG_PM03_DEFAULT		0x00000001
#define RB_CXD9225GG_PM04_DEFAULT		0x00000002
#define RB_CXD9225GG_PM05_DEFAULT		0x000000FF
#define RB_CXD9225GG_PM10_DEFAULT		0x00000000
#define RB_CXD9225GG_PM11_DEFAULT		0x00000000
#define RB_CXD9225GG_PM12_DEFAULT		0x0000000F
#define RB_CXD9225GG_PM13_DEFAULT		0x000000FF
#define RB_CXD9225GG_PM14_DEFAULT		0x0000003E
#define RB_CXD9225GG_PM15_DEFAULT		0x000000F7
#define RB_CXD9225GG_PMC00_DEFAULT		0x00000003
#define RB_CXD9225GG_PMC01_DEFAULT		0x00000003
#define RB_CXD9225GG_PMC02_DEFAULT		0x00000003
#define RB_CXD9225GG_PMC03_DEFAULT		0x00000003
#define RB_CXD9225GG_PMC04_DEFAULT		0x00000007
#define RB_CXD9225GG_PMC05_DEFAULT		0x00000003


/*!
 * There are two queues for registered GPIO ISRs. One is for high priority and
 * the other is for low priority. The ISRs in the high priority queue will be
 * called first before the low priority queue if more than one GPIO interrupt
 * occurs at the same time.
 */
enum gpio_prio {
    GPIO_HIGH_PRIO = 0,	/*!< high priority queue */
    GPIO_LOW_PRIO		/*!< low priority queue */
};

/*!
 * This enumeration data type defines various different ways for interrupting
 * the ARM core from GPIO signals. The way to interrupt the core is dictated
 * by the external hardware.
 */
typedef enum gpio_int_cfg {
    GPIO_INT_LOW_LEV = 0x0,	    /*!< low level sensitive */
    GPIO_INT_HIGH_LEV = 0x1,	/*!< high level sensitive */
    GPIO_INT_FALL_EDGE = 0x2,	/*!< rising edge sensitive */
    GPIO_INT_RISE_EDGE = 0x3,	/*!< falling edge sensitive */
    GPIO_INT_BOTH_EDGE = 0x04,  /*!< both edge sensitive */
    GPIO_INT_NONE = 0x5	        /*!< No interrupt */
} gpio_int_cfg_t;

/*!
 * This enumeration data type defines the modes of the pin .
 *	GPIO_MUX_GPIO is the PORT mode.
 *	GPIO_MUX_UART is the UART mode.
 *	GPIO_MUX_CSI is the CSI mode.
 */
typedef enum {
    GPIO_MUX_GPIO,
    GPIO_MUX_UART,
    GPIO_MUX_CSI
} cxd9225gg_gpio_mux_mode_t;


typedef irqreturn_t(*gpio_irq_handler) (int, void *);

/*!
 * Request ownership for a GPIO pin. The caller has to check the return value
 * of this function to make sure it returns 0 before make use of that pin.
 * @param pin		a name defined by \b cxd9225gg_gpio_pin_name_t
 * @return		0 if successful; Non-zero otherwise
 */

int cxd9225gg_request_gpio(cxd9225gg_gpio_pin_name_t pin);

/*!
 * Exported function to set a GPIO pin's direction
 * @param pin		a name defined by \b cxd9225gg_gpio_pin_name_t
 * @param is_input	1 (or non-zero) for input; 0 for output
 */
void cxd9225gg_set_gpio_direction(cxd9225gg_gpio_pin_name_t pin, int is_input);



/*
 * Exported function to set a GPIO pin's data output
 * @param pin		a name defined by \b cxd9225gg_gpio_pin_name_t
 * @param data		value to be set (only 0 or 1 is valid)
 */
void cxd9225gg_set_gpio_dataout(cxd9225gg_gpio_pin_name_t pin, u32 data);



/*!
 * Return the data value of a GPIO signal.
 * @param pin	a name defined by \b cxd9225gg_gpio_pin_name_t
 * @return 	value (0 or 1) of the GPIO signal; -1 if pass in invalid pin
 */
int cxd9225gg_get_gpio_datain(cxd9225gg_gpio_pin_name_t pin);
cxd9225gg_gpio_port_t *get_gpio_port(u32 gpio);

/*!
 * Release ownership for a GPIO pin
 * @param pin		a name defined by \b cxd9225gg_gpio_pin_name_t
 */
void cxd9225gg_free_gpio(cxd9225gg_gpio_pin_name_t pin);


/*!
 * GPIO driver initialization
 * @return    always 0
 */
int cxd9225gg_gpio_init(void);

/*
 * This function is responsible for registering a GPIO signal's ISR.
 * @param pin		a name defined by \b cxd9225gg_gpio_pin_name_t
 * @param desc		an irqaction assigned to the pin's ISR
 * @return 0 if successful; non-zero otherwise.
 */
int cxd9225gg_gpio_irq_request(cxd9225gg_gpio_pin_name_t pin, struct irqaction *desc);

/*
 * Free a GPIO signal's ISR.
 * @param pin		a name defined by \b cxd9225gg_gpio_pin_name_t
 * @param desc		an irqaction assigned to the pin's ISR
 */
void cxd9225gg_gpio_irq_free(cxd9225gg_gpio_pin_name_t pin, struct irqaction *desc);


/*
 * Enable/disable a GPIO signal's interrupt.
 * @param pin         a name defined by \b cxd9225gg_gpio_pin_name_t
 * @param enable        \b #true for enabling the interrupt; \b #false otherwise
 */
void cxd9225gg_gpio_irq_enable(cxd9225gg_gpio_pin_name_t pin, bool enable);

/*
 * Set a GPIO pin's interrupt edge
 * @param pin         a name defined by \b cxd9225gg_gpio_pin_name_t
 * @param icr           one of the values defined in \b gpio_edge_t
 *                      to indicate how to generate an interrupt
 */
int cxd9225gg_gpio_irq_set_type(cxd9225gg_gpio_pin_name_t pin, gpio_int_cfg_t cfg);


/*!
 *@brief cxd9225gg_gpio_config_mux - just configure the mode of the gpio pin.
 *@param pin   a pin name as defined in \b #cxd9225gg_gpio_pin_name_t
 *@param mode  a module as define in \b #cxd9225gg_gpio_mux_mode_t;
 *	GPIO_MUX_GPIO set pin to work as GPIO port
 *	GPIO_MUX_UART set pin to work as UART
 *	GPIO_MUX_CSI set pin to work as CSI
 *@return      0 if successful, Non-zero otherwise
 */
int cxd9225gg_gpio_config_mux(cxd9225gg_gpio_pin_name_t pin, cxd9225gg_gpio_mux_mode_t mode);

/*!
 * This function is just used to request a pin and configure it.
 * @param pin	a pin name as defined in \b #cxd9225gg_gpio_pin_name_t
 * @param mode	a module as define in \b #cxd9225gg_gpio_mux_mode_t;
 * @return	0 if successful, Non-zero otherwise
 */
int cxd9225gg_gpio_request_mux(cxd9225gg_gpio_pin_name_t pin, cxd9225gg_gpio_mux_mode_t mode);


/*!
 * This function is just used to release a pin.
 * @param pin	a pin name as defined in \b #cxd9225gg_gpio_pin_name_t
 * @return	none	
 */
void cxd9225gg_gpio_free_mux(cxd9225gg_gpio_pin_name_t pin);

#endif				/* __ASM_ARCH_CXD9225GG_GPIO_H__ */
