/* 2014-08-28: File changed by Sony Corporation */
/*
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef __ASM_ARCH_CXD9225GG_IO_H
#define __ASM_ARCH_CXD9225GG_IO_H

#include <mach/hardware.h>
#include <mach/cxd9225gg_Chipset.h>

#define IO_SPACE_LIMIT	(PCIBIOS_MIN_IO + CXD9225GG_PCI_WIN_1_SIZE - 1)

#define __is_pcie_io_address(p) \
	((((unsigned long)(p)) & ~(CXD9225GG_PCIE_4X_SIZE - 1)) == CXD9225GG_PCIE_4X_PHYS_BASE)

/*
 * Translate IO bus address to virtual MMIO address
 * We use busaddr == cpuaddr(phy) mapping
 */
static inline void __iomem *__io(unsigned long addr)
{
	if (__is_pcie_io_address(addr)) {
		return __typesafe_io(addr);
	} else {
		return (void __iomem *)(addr - PCIBIOS_MIN_IO +
				CXD9225GG_PCI_WIN_1_VIRT_BASE);
	}
}
/* for #ifdef __io */
#define __io(a)		__io(a)

/* always use MMU, so the address returned by ioremap() is sufficient */
#define __mem_pci(a)	(a)

/*
 * cxd9225gg will read/write PCIE io space with the following steps
 *  1. If read/write io space, Set Read Byte Enable Register
 *  2. Set PCI I/O Issue Address Register
 *  3. Write to PCI I/O Issue Data Register
 */
#define pcie_raw_readl(reg) __raw_readl(CXD9225GG_PCIE_4X_BUSC_REG_VIRT_BASE + (reg))
#define pcie_raw_readw(reg) __raw_readw(CXD9225GG_PCIE_4X_BUSC_REG_VIRT_BASE + (reg))
#define pcie_raw_readb(reg) __raw_readb(CXD9225GG_PCIE_4X_BUSC_REG_VIRT_BASE + (reg))

#define pcie_raw_writel(val, reg) __raw_writel((val), CXD9225GG_PCIE_4X_BUSC_REG_VIRT_BASE + (reg))
#define pcie_raw_writew(val, reg) __raw_writew((val), CXD9225GG_PCIE_4X_BUSC_REG_VIRT_BASE + (reg))
#define pcie_raw_writeb(val, reg) __raw_writeb((val), CXD9225GG_PCIE_4X_BUSC_REG_VIRT_BASE + (reg))

static inline unsigned int
__cxd9225gg_ioread8(const void __iomem *p)
{
	u8 byte;
	u32 addr = (unsigned long __force)p;

	byte = addr & 0x3;
	addr = addr & (~0x3);

	pcie_raw_writel(0xF, CXD9225GG_PCIEX_READ_BYTE_ENABLE);
	pcie_raw_writel(addr | CXD9225GG_PCIEX_PCIE_IO_ISSUE_ADR_IDX_IO_EN,
			CXD9225GG_PCIEX_PCIE_IO_ISSUE_ADR);
	return(pcie_raw_readb(CXD9225GG_PCIEX_PCIE_IO_ISSUE_DATA + byte));
}

static inline unsigned int
__cxd9225gg_ioread16(const void __iomem *p)
{
	u32 addr = (unsigned long __force)p;

	pcie_raw_writel(0x3, CXD9225GG_PCIEX_READ_BYTE_ENABLE);
	pcie_raw_writel(addr | CXD9225GG_PCIEX_PCIE_IO_ISSUE_ADR_IDX_IO_EN,
			CXD9225GG_PCIEX_PCIE_IO_ISSUE_ADR);
	return(pcie_raw_readw(CXD9225GG_PCIEX_PCIE_IO_ISSUE_DATA));
}

static inline unsigned int
__cxd9225gg_ioread32(const void __iomem *p)
{
	u32 addr = (unsigned long __force)p;

	pcie_raw_writel(0xF, CXD9225GG_PCIEX_READ_BYTE_ENABLE);
	pcie_raw_writel(addr | CXD9225GG_PCIEX_PCIE_IO_ISSUE_ADR_IDX_IO_EN,
			CXD9225GG_PCIEX_PCIE_IO_ISSUE_ADR);
	return(pcie_raw_readl(CXD9225GG_PCIEX_PCIE_IO_ISSUE_DATA));
}

static inline void
__cxd9225gg_ioread8_rep(const void __iomem *addr, void *dst, u32 count)
{
	while(count--) {
		*(u8 *)dst = __cxd9225gg_ioread8(addr);
		dst++;
	}
}

static inline void
__cxd9225gg_ioread16_rep(const void __iomem *addr, void *dst, u32 count)
{
	while(count--) {
		*(u16 *)dst = __cxd9225gg_ioread16(addr);
		dst += 2;
	}
}

static inline void
__cxd9225gg_ioread32_rep(const void __iomem *addr, void *dst, u32 count)
{
	while (count--) {
		*(u32 *)dst = __cxd9225gg_ioread32(addr);
		dst += 4;
	}
}

static inline void
__cxd9225gg_iowrite8(u8 v, const void __iomem *p)
{
	u8 byte;
	u32 addr = (unsigned long __force)p;

	byte = addr & 0x3;
	addr = addr & (~0x3);

	pcie_raw_writel(1 << byte, CXD9225GG_PCIEX_READ_BYTE_ENABLE);
	pcie_raw_writel(addr | CXD9225GG_PCIEX_PCIE_IO_ISSUE_ADR_IDX_IO_EN,
			CXD9225GG_PCIEX_PCIE_IO_ISSUE_ADR);
	pcie_raw_writeb((u8)(v & 0xFF),
			CXD9225GG_PCIEX_PCIE_IO_ISSUE_DATA + byte);
}

static inline void
__cxd9225gg_iowrite16(u16 v, const void __iomem *p)
{
	u32 addr = (unsigned long __force)p;

	pcie_raw_writel(0x3, CXD9225GG_PCIEX_READ_BYTE_ENABLE);
	pcie_raw_writel(addr | CXD9225GG_PCIEX_PCIE_IO_ISSUE_ADR_IDX_IO_EN,
			CXD9225GG_PCIEX_PCIE_IO_ISSUE_ADR);
	pcie_raw_writew(v, CXD9225GG_PCIEX_PCIE_IO_ISSUE_DATA);
}

static inline void
__cxd9225gg_iowrite32(u32 v, const void __iomem *p)
{
	u32 addr = (unsigned long __force)p;

	pcie_raw_writel(0xF, CXD9225GG_PCIEX_READ_BYTE_ENABLE);
	pcie_raw_writel(addr | CXD9225GG_PCIEX_PCIE_IO_ISSUE_ADR_IDX_IO_EN,
			CXD9225GG_PCIEX_PCIE_IO_ISSUE_ADR);
	pcie_raw_writel(v, CXD9225GG_PCIEX_PCIE_IO_ISSUE_DATA);
}

static inline void
__cxd9225gg_iowrite8_rep(void __iomem *addr, void *src, u32 count)
{
	while (count--) {
		__cxd9225gg_iowrite8(*(u8 *)src, addr);
		src++;
	}
}

static inline void
__cxd9225gg_iowrite16_rep(void __iomem *addr, void *src, u32 count)
{
	while (count--) {
		__cxd9225gg_iowrite16(*(u16 *)src, addr);
		src += 2;
	}
}

static inline void
__cxd9225gg_iowrite32_rep(void __iomem *addr, void *src, u32 count)
{
	while (count--) {
		__cxd9225gg_iowrite32(*(u32 *)src, addr);
		src += 4;
	}
}

#define ioread8(p)								\
	({									\
		unsigned int __v;						\
										\
		if (__is_pcie_io_address(p)) {					\
			__v = __cxd9225gg_ioread8(p);				\
		} else {							\
			__v = __raw_readb(p);					\
		}								\
										\
		__v;								\
	})									\

#define ioread16(p)								\
	({									\
		unsigned int __v;						\
										\
		if (__is_pcie_io_address(p)) {					\
			__v = __cxd9225gg_ioread16(p);				\
		} else {							\
			__v = le16_to_cpu((__force __le16)__raw_readw(p));	\
		}								\
										\
		__v;								\
	})

#define ioread32(p)								\
	({									\
		unsigned int __v;						\
										\
		if (__is_pcie_io_address(p)) {					\
			__v = __cxd9225gg_ioread32(p);				\
		} else {							\
			__v = le32_to_cpu((__force __le32)__raw_readl(p));	\
		}								\
										\
		 __v;								\
	})

#define iowrite8(v,p)								\
	({									\
		if (__is_pcie_io_address(p)) {					\
			__cxd9225gg_iowrite8((v), (p));				\
		} else {							\
			__raw_writeb((v), p);					\
		}								\
	})

#define iowrite16(v,p)								\
	({									\
		if (__is_pcie_io_address(p)) {					\
			__cxd9225gg_iowrite16((v), (p));			\
		} else {							\
			__raw_writew((__force __u16)cpu_to_le16(v), p);		\
		}								\
	})

#define iowrite32(v,p)								\
	({									\
		if (__is_pcie_io_address(p)) {					\
			__cxd9225gg_iowrite32((v), (p));			\
		} else {							\
			__raw_writel((__force __u32)cpu_to_le32(v), p);		\
		}								\
	})

#define ioread8_rep(p, d, c)							\
	({									\
		if (__is_pcie_io_address(p)) {					\
			__cxd9225gg_ioread8_rep((p), (d), (c));			\
		} else {							\
			__raw_readsb((p), (d), (c));				\
		}								\
	})									\

#define ioread16_rep(p, d, c)							\
	({									\
		if (__is_pcie_io_address(p)) {					\
			__cxd9225gg_ioread16_rep((p), (d), (c));		\
		} else {							\
			__raw_readsw((p), (d), (c));				\
		}								\
	})

#define ioread32_rep(p, d, c)							\
	({									\
		if (__is_pcie_io_address(p)) {					\
			__cxd9225gg_ioread32_rep((p), (d), (c));		\
		} else {							\
			__raw_readsl((p), (d), (c));				\
		}								\
	})

#define iowrite8_rep(p, s, c)							\
	({									\
		if (__is_pcie_io_address(p)) {					\
			__cxd9225gg_iowrite8_rep((p), (s), (c));		\
		} else {							\
			__raw_writesb((p), (s), (c));				\
		}								\
	})

#define iowrite16_rep(p, s, c)							\
	({									\
		if (__is_pcie_io_address(p)) {					\
			__cxd9225gg_iowrite16_rep((p), (s), (c));		\
		} else {							\
			__raw_writesb((p), (s) ,(c));				\
		}								\
	})

#define iowrite32_rep(p, s, c)							\
	({									\
		if (__is_pcie_io_address(p)) {					\
			__cxd9225gg_iowrite32_rep((p), (s), (c));		\
		} else {							\
			__raw_writesb((p), (s), (c));				\
		}								\
	})
extern void __iomem *ioport_map(unsigned long port, unsigned int nr);
extern void ioport_unmap(void __iomem *addr);
#endif /* __ASM_ARCH_CXD9225GG_IO_H */
