/*
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef __ASM_ARCH_CXD9225GG_PCI_H
#define __ASM_ARCH_CXD9225GG_PCI_H

#define PCI_DEVICE_ID_SONY_CXD9225GG_PCIC	(0x904B)
/* in AHB area */
#define CXD9225GG_PCI_BUSC_REG_OFFSET	0x00000000
#define CXD9225GG_PCI_BUSC_REG_PHYS_BASE	(CXD9225GG_AHB_REG_PHYS_BASE + \
					 CXD9225GG_PCI_BUSC_REG_OFFSET)
#define CXD9225GG_PCI_BUSC_REG_VIRT_BASE	0xFDEEA000
#define CXD9225GG_PCI_BUSC_REG_SIZE	SZ_4K

/* memory address space are mapped at vmalloc area on the fly */
#define CXD9225GG_PCI_WIN_2_PHYS_BASE	0xD0000000
#define CXD9225GG_PCI_WIN_2_SIZE		SZ_256M

/* IO and configuration address space */
#define CXD9225GG_PCI_WIN_1_PHYS_BASE	0xE0000000
#define CXD9225GG_PCI_WIN_1_VIRT_BASE	0xFDEF0000
#define CXD9225GG_PCI_WIN_1_SIZE		SZ_64K

/* The minimum PCI bus address where Linux will set to devices BARs */
#define PCIBIOS_MIN_IO			CXD9225GG_PCI_WIN_1_PHYS_BASE
#define PCIBIOS_MIN_MEM			CXD9225GG_PCI_WIN_2_PHYS_BASE
#define pcibios_assign_all_busses()     1

/*
 * PCI contoroller registers (on AHB bus)
 */
#define CXD9225GG_PCI_PCIAHB_WIN1_CTR_OFFSET	0x0000
#define CXD9225GG_PCI_PCIAHB_WIN2_CTR_OFFSET	0x0004
#define CXD9225GG_PCI_PCIAHB_DCT_OFFSET		0x0008
#define CXD9225GG_PCI_AHBPCI_WIN1_CTR_OFFSET	0x0010
#define CXD9225GG_PCI_AHBPCI_WIN2_CTR_OFFSET	0x0014
/* #define CXD9225GG_PCI_AHBPCI_WIN3_CTR_OFFSET	0x0018 not implemented */
#define CXD9225GG_PCI_AHBPCI_DCT_CTR_OFFSET	0x001C
#define CXD9225GG_PCI_PCI_INT_ENABLE_OFFSET	0x0020
#define CXD9225GG_PCI_PCI_INT_STATUS_OFFSET	0x0024
#define CXD9225GG_PCI_PCIERR_ADDR_OFFSET		0x0028
#define CXD9225GG_PCI_AHBERR_ADDR_OFFSET		0x002C
#define CXD9225GG_PCI_AHB_BUS_CTR_OFFSET		0x0030
#define CXD9225GG_PCI_PCI_ARBITER_CTR_OFFSET	0x0040
#define CXD9225GG_PCI_PCI_EXT_CTR_OFFSET		0x0044
#define CXD9225GG_PCI_PCI_UNIT_REV_OFFSET		0x0048
#define CXD9225GG_PCI_PCI_UNIT_DEBUG_OFFSET	0x004C
#define CXD9225GG_PCI_PCI_PME_CTR_OFFSET		0x0050

/*
 * Register field definitions
 */

/*
 * PCI->AHB Window control register
 * PCIAHB_WIN{1,2}_CTR
 */
#define CXD9225GG_PCI_PCIAHB_WIN_CTR_PREFETCH_MASK		(3 << 0)
#define CXD9225GG_PCI_PCIAHB_WIN_CTR_PREFETCH_DISABLE	(0 << 0)
#define CXD9225GG_PCI_PCIAHB_WIN_CTR_PREFETCH_4		(1 << 0)
#define CXD9225GG_PCI_PCIAHB_WIN_CTR_PREFETCH_8		(2 << 0)
#define CXD9225GG_PCI_PCIAHB_WIN_CTR_PREFETCH_16		(3 << 0)

#define CXD9225GG_PCI_PCIAHB_WIN_CTR_ENDIAN_MASK		(7 << 6)
#define CXD9225GG_PCI_PCIAHB_WIN_CTR_ENDIAN_THRU		(0 << 6)
#define CXD9225GG_PCI_PCIAHB_WIN_CTR_ENDIAN_ASWAP		(1 << 6)
#define CXD9225GG_PCI_PCIAHB_WIN_CTR_ENDIAN_BSWAP		(2 << 6)
#define CXD9225GG_PCI_PCIAHB_WIN_CTR_ENDIAN_HWSWAP		(3 << 6)
#define CXD9225GG_PCI_PCIAHB_WIN_CTR_ENDIAN_ADRSWAP	(4 << 6)

#define CXD9225GG_PCI_PCIAHB_WIN_CTR_BASE_ADDR_MASK	(0xFFFFF000)

/*
 * PCI->AHB discard timer register
 * PCIAHB_DCT_CTR
 */
#define CXD9225GG_PCI_PCIAHB_DCT_CTR_DISCARD_ENABLE	(1 << 0)
/*
 * discard timer time=(val + 1) * 256 * PCI clock.
 * poweron reset default 0xf (61.4 us)
 */
#define CXD9225GG_PCI_PCIAHB_DCT_CTR_DISCARD_TIMER_MASK	(0x0000FFF0)

/*
 * AHB->PCI window control register
 * AHBPCI_WIN{1,2}_CTR
 */
/* WIN2 only */
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_PREFETCH_MASK		(1 << 0)
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_PREFETCH_ENABLE	(1 << 0)
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_PREFETCH_DISABLE	(0 << 0)

/* WIN1,2 */
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_PCICMD_MASK		(3 << 1)
/* WIN1 only */
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_PCICMD_SPECIAL	(0 << 1)
/* WIN1 only */
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_PCICMD_IO		(1 << 1)
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_PCICMD_MEM		(3 << 1)
/* WIN1 only */
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_PCICMD_CONFIG		(5 << 1)
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_PCICMD_MEM_MULT	(6 << 1)
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_PCICMD_MEM_LINE	(7 << 1)

/* WIN1 only */
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_CFGTYPE_MASK		(1 << 4)
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_CFGTYPE_0		(0 << 4)
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_CFGTYPE_1		(1 << 4)

/* WIN2 only */
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_BURST_MASK		(1 << 5)
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_BURST_ENABLE		(1 << 5)
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_BURST_DISABLE		(0 << 5)

/* WIN1,2 */
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_ENDIAN_MASK		(7 << 6)
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_ENDIAN_THRU		(0 << 6)
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_ENDIAN_ASWAP		(1 << 6)
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_ENDIAN_BSWAP		(2 << 6)
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_ENDIAN_HWSWAP		(3 << 6)
#define CXD9225GG_PCI_AHBPCI_WIN_CTR_ENDIAN_ADRSWAP	(4 << 6)

/*
 * AHB->PCI discard timer register
 * AHBPCI_DCT_CTR
 */
#define CXD9225GG_PCI_AHBPCI_DTC_CTR_DISCARD_ENABLE	(1 << 0)
/*
 * discard timer retry=(val + 1) * 16
 * poweron reset default 0xf (256 retries)
 */
#define CXD9225GG_PCI_AHBPCI_DCT_CTR_DISCARD_TIMER_MASK	(0x0000FFF0)

/*
 * PCI interrupt enable/status register
 */
#define CXD9225GG_PCI_PCI_INT_INT3				(1 << 19)
#define CXD9225GG_PCI_PCI_INT_INT2				(1 << 18)
#define CXD9225GG_PCI_PCI_INT_INT1				(1 << 17)
#define CXD9225GG_PCI_PCI_INT_INT0				(1 << 16)
#define CXD9225GG_PCI_PCI_INT_AHBPCI_WIN			(1 << 14)
#define CXD9225GG_PCI_PCI_INT_PCIAHB_WIN2			(1 << 13)
#define CXD9225GG_PCI_PCI_INT_PCIAHB_WIN1			(1 << 12)
#define CXD9225GG_PCI_PCI_INT_DMA_AHBPCI			(1 << 9)
#define CXD9225GG_PCI_PCI_INT_DMA_PCIAHB			(1 << 8)
#define CXD9225GG_PCI_PCI_INT_CLKRUN			(1 << 7)
#define CXD9225GG_PCI_PCI_INT_PM				(1 << 6)
#define CXD9225GG_PCI_PCI_INT_RESERR			(1 << 5)
#define CXD9225GG_PCI_PCI_INT_SIGSERR			(1 << 4)
#define CXD9225GG_PCI_PCI_INT_PERR				(1 << 3)
#define CXD9225GG_PCI_PCI_INT_REMABOART			(1 << 2)
#define CXD9225GG_PCI_PCI_INT_RETABORT			(1 << 1)
#define CXD9225GG_PCI_PCI_INT_SIGTABORT			(1 << 0)

/*
 * AHB bus controll register
 * AHB_BUS_CTR
 * controlls the access to AHB bus
 */
#define CXD9225GG_PCI_AHB_BUS_CTR_SMODE_READY_CTR		(1 << 17)
#define CXD9225GG_PCI_AHB_BUS_CTR_SMODE_READ_BURST		(1 << 16)
#define CXD9225GG_PCI_AHB_BUS_CTR_MMODE_HBUSREQ		(1 << 7)

#define CXD9225GG_PCI_AHB_BUS_CTR_MMODE_BOUNDARY_MASK	(3 << 5)
#define CXD9225GG_PCI_AHB_BUS_CTR_MMODE_BOUNDARY_NONE	(0 << 5)
#define CXD9225GG_PCI_AHB_BUS_CTR_MMODE_BOUNDARY_3F	(1 << 5)
#define CXD9225GG_PCI_AHB_BUS_CTR_MMODE_BOUNDARY_7F	(2 << 5)

#define CXD9225GG_PCI_AHB_BUS_CTR_MMODE_BURST_WIDTH_MASK	(3 << 3)
#define CXD9225GG_PCI_AHB_BUS_CTR_MMODE_BURST_WIDTH_16	(0 << 3)
#define CXD9225GG_PCI_AHB_BUS_CTR_MMODE_BURST_WIDTH_8	(1 << 3)
#define CXD9225GG_PCI_AHB_BUS_CTR_MMODE_BURST_WIDTH_4	(2 << 3)
#define CXD9225GG_PCI_AHB_BUS_CTR_MMODE_BURST_WIDTH_SINGLE	(3 << 3)

#define CXD9225GG_PCI_AHB_BUS_CTR_MMODE_WR_INCR		(1 << 2)

#define CXD9225GG_PCI_AHB_BUS_CTR_MMODE_HTRANS		(1 << 0)

/*
 * PCI arbiter controll register
 * PCI_ARBITER_CTR
 */
/*
 * enter park mode after PCI clock count when FRAME#=1 and IRDY#=1 detected
 * power on default = 0x07
 */
#define CXD9225GG_PCI_PCI_ARBITER_CTR_TIMER_SHIFT		(16)
#define CXD9225GG_PCI_PCI_ARBITER_CTR_TIMER_MASK		(0xff << 16)
/*
 * select bus master when entered park mode
 * 0: master = self, 1: master = last bus master
 */
#define CXD9225GG_PCI_PCI_ARBITER_CTR_BP_MODE		(1 << 12)

#define CXD9225GG_PCI_PCI_ARBITER_REQ7			(1 << 7)
#define CXD9225GG_PCI_PCI_ARBITER_REQ6			(1 << 6)
#define CXD9225GG_PCI_PCI_ARBITER_REQ5			(1 << 5)
#define CXD9225GG_PCI_PCI_ARBITER_REQ4			(1 << 4)
#define CXD9225GG_PCI_PCI_ARBITER_REQ3			(1 << 3)
#define CXD9225GG_PCI_PCI_ARBITER_REQ2			(1 << 2)
#define CXD9225GG_PCI_PCI_ARBITER_REQ1			(1 << 1)
#define CXD9225GG_PCI_PCI_ARBITER_REQ0			(1 << 0)

/*
 * PCI extension controll register
 * PCI_EXT_CTR
 */
/*
 * disable access to PCI communication registers from PCI bus (BAR0 access)
 */
#define CXD9225GG_PCI_PCI_EXT_CTR_COMREG_MASK		(1 << 1)
/*
 * write enable to SUBSYS ID register
 */
#define CXD9225GG_PCI_PCI_EXT_CTR_SUBSYS_WEN		(1 << 0)

/*
 * PCI power management
 */
#define CXD9225GG_PCI_PCI_PM_CTR_CLKRUN_CTR		(1 << 5)
/* read only */
#define CXD9225GG_PCI_PCI_PM_CTR_CLKRUN_STATE		(1 << 4)
#define CXD9225GG_PCI_PCI_PM_CTR_PME_CTR			(1 << 2)
/* read only */
#define CXD9225GG_PCI_PCI_PM_CTR_PM_STATE_MASK		(3 << 0)


/*
 * virtual addresses for PCI contoroller registers
 */
#define CXD9225GG_PCI_REG_VIRT(off)	(CXD9225GG_PCI_BUSC_REG_VIRT_BASE + off)

#define CXD9225GG_PCI_PCIAHB_WIN1_CTR_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_PCIAHB_WIN1_CTR_OFFSET)
#define CXD9225GG_PCI_PCIAHB_WIN2_CTR_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_PCIAHB_WIN2_CTR_OFFSET)
#define CXD9225GG_PCI_PCIAHB_DCT_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_PCIAHB_DCT_OFFSET)
#define CXD9225GG_PCI_AHBPCI_WIN1_CTR_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_AHBPCI_WIN1_CTR_OFFSET)
#define CXD9225GG_PCI_AHBPCI_WIN2_CTR_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_AHBPCI_WIN2_CTR_OFFSET)
#define CXD9225GG_PCI_AHBPCI_DCT_CTR_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_AHBPCI_DCT_CTR_OFFSET)
#define CXD9225GG_PCI_PCI_INT_ENABLE_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_PCI_INT_ENABLE_OFFSET)
#define CXD9225GG_PCI_PCI_INT_STATUS_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_PCI_INT_STATUS_OFFSET)
#define CXD9225GG_PCI_PCIERR_ADDR_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_PCIERR_ADDR_OFFSET)
#define CXD9225GG_PCI_AHBERR_ADDR_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_AHBERR_ADDR_OFFSET)
#define CXD9225GG_PCI_AHB_BUS_CTR_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_AHB_BUS_CTR_OFFSET)
#define CXD9225GG_PCI_PCI_ARBITER_CTR_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_PCI_ARBITER_CTR_OFFSET)
#define CXD9225GG_PCI_PCI_EXT_CTR_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_PCI_EXT_CTR_OFFSET)
#define CXD9225GG_PCI_PCI_UNIT_REV_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_PCI_UNIT_REV_OFFSET)
#define CXD9225GG_PCI_PCI_UNIT_DEBUG_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_PCI_UNIT_DEBUG_OFFSET)
#define CXD9225GG_PCI_PCI_PME_CTR_VIRT		\
	CXD9225GG_PCI_REG_VIRT(CXD9225GG_PCI_PCI_PME_CTR_OFFSET)

#endif
