/*
 *  TIMER definitions
 *
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef __ASM_ARCH_CXD9225GG_TIMER_H
#define __ASM_ARCH_CXD9225GG_TIMER_H

#define CXD9225GG_TIMER_TMD_OFFSET		0x00001000
#define CXD9225GG_TIMER_CMD_OFFSET		0x00001004
#define CXD9225GG_TIMER_TMCD_OFFSET	0x00001008
/* TMCD bits */
#define CXD9225GG_TIMER_TMCD_CAE		(1 << 0)
#define CXD9225GG_TIMER_TMCD_CAE_CLK_DIS	(0 << 0)
#define CXD9225GG_TIMER_TMCD_CAE_CLK_ENA	(1 << 0)

#define CXD9225GG_TIMER_TMCD_CE		(1 << 1)
#define CXD9225GG_TIMER_TMCD_CE_DIS	(0 << 1)
#define CXD9225GG_TIMER_TMCD_CE_ENA	(1 << 1)

#define CXD9225GG_TIMER_TMCD_CS		(7 << 4)
#define CXD9225GG_TIMER_TMCD_DIV_2		(0 << 4)
#define CXD9225GG_TIMER_TMCD_DIV_4		(1 << 4)
#define CXD9225GG_TIMER_TMCD_DIV_8		(2 << 4)
#define CXD9225GG_TIMER_TMCD_DIV_16	(3 << 4)
#define CXD9225GG_TIMER_TMCD_DIV_32	(4 << 4)
#define CXD9225GG_TIMER_TMCD_DIV_64	(5 << 4)
#define CXD9225GG_TIMER_TMCD_DIV_128	(6 << 4)
#define CXD9225GG_TIMER_TMCD_DIV_256	(7 << 4)

#define CXD9225GG_TIMER_TMCD_TEST		(1 << 7)
#define CXD9225GG_TIMER_TMCD_TEST_NORM	(0 << 7)
#define CXD9225GG_TIMER_TMCD_TEST_ENA	(1 << 7)

/* Offset address from APB area base */
#define CXD9225GG_TIMER_REG_OFFSET(n, r)	(r + 0x0010 * n)
/* Physical address */
#define CXD9225GG_TIMER_REG_PHYS(n, r)	(CXD9225GG_APB_REG_PHYS_BASE + \
					 CXD9225GG_TIMER_REG_OFFSET(n, r))
/* virtual address */
#define CXD9225GG_TIMER_REG_VIRT(n, r)	(CXD9225GG_APB_REG_VIRT_BASE + \
					 CXD9225GG_TIMER_REG_OFFSET(n, r))

/* CH0 */
#define CXD9225GG_TIMER_TMD0_PHYS		CXD9225GG_TIMER_REG_PHYS(0, \
					CXD9225GG_TIMER_TMD_OFFSET)
#define CXD9225GG_TIMER_CMD0_PHYS		CXD9225GG_TIMER_REG_PHYS(0, \
					CXD9225GG_TIMER_CMD_OFFSET)
#define CXD9225GG_TIMER_TMCD0_PHYS		CXD9225GG_TIMER_REG_PHYS(0, \
					CXD9225GG_TIMER_TMCD_OFFSET)

#define CXD9225GG_TIMER_TMD0_VIRT		CXD9225GG_TIMER_REG_VIRT(0, \
					CXD9225GG_TIMER_TMD_OFFSET)
#define CXD9225GG_TIMER_CMD0_VIRT		CXD9225GG_TIMER_REG_VIRT(0, \
					CXD9225GG_TIMER_CMD_OFFSET)
#define CXD9225GG_TIMER_TMCD0_VIRT		CXD9225GG_TIMER_REG_VIRT(0, \
					CXD9225GG_TIMER_TMCD_OFFSET)
/* CH1 */
#define CXD9225GG_TIMER_TMD1_PHYS		CXD9225GG_TIMER_REG_PHYS(1, \
					CXD9225GG_TIMER_TMD_OFFSET)
#define CXD9225GG_TIMER_CMD1_PHYS		CXD9225GG_TIMER_REG_PHYS(1, \
					CXD9225GG_TIMER_CMD_OFFSET)
#define CXD9225GG_TIMER_TMCD1_PHYS		CXD9225GG_TIMER_REG_PHYS(1, \
					CXD9225GG_TIMER_TMCD_OFFSET)

#define CXD9225GG_TIMER_TMD1_VIRT		CXD9225GG_TIMER_REG_VIRT(1, \
					CXD9225GG_TIMER_TMD_OFFSET)
#define CXD9225GG_TIMER_CMD1_VIRT		CXD9225GG_TIMER_REG_VIRT(1, \
					CXD9225GG_TIMER_CMD_OFFSET)
#define CXD9225GG_TIMER_TMCD1_VIRT		CXD9225GG_TIMER_REG_VIRT(1, \
					CXD9225GG_TIMER_TMCD_OFFSET)
/* CH2 */
#define CXD9225GG_TIMER_TMD2_PHYS		CXD9225GG_TIMER_REG_PHYS(2, \
					CXD9225GG_TIMER_TMD_OFFSET)
#define CXD9225GG_TIMER_CMD2_PHYS		CXD9225GG_TIMER_REG_PHYS(2, \
					CXD9225GG_TIMER_CMD_OFFSET)
#define CXD9225GG_TIMER_TMCD2_PHYS		CXD9225GG_TIMER_REG_PHYS(2, \
					CXD9225GG_TIMER_TMCD_OFFSET)

#define CXD9225GG_TIMER_TMD2_VIRT		CXD9225GG_TIMER_REG_VIRT(2, \
					CXD9225GG_TIMER_TMD_OFFSET)
#define CXD9225GG_TIMER_CMD2_VIRT		CXD9225GG_TIMER_REG_VIRT(2, \
					CXD9225GG_TIMER_CMD_OFFSET)
#define CXD9225GG_TIMER_TMCD2_VIRT		CXD9225GG_TIMER_REG_VIRT(2, \
					CXD9225GG_TIMER_TMCD_OFFSET)
/* CH3 */
#define CXD9225GG_TIMER_TMD3_PHYS		CXD9225GG_TIMER_REG_PHYS(3, \
					CXD9225GG_TIMER_TMD_OFFSET)
#define CXD9225GG_TIMER_CMD3_PHYS		CXD9225GG_TIMER_REG_PHYS(3, \
					CXD9225GG_TIMER_CMD_OFFSET)
#define CXD9225GG_TIMER_TMCD3_PHYS		CXD9225GG_TIMER_REG_PHYS(3, \
					CXD9225GG_TIMER_TMCD_OFFSET)

#define CXD9225GG_TIMER_TMD3_VIRT		CXD9225GG_TIMER_REG_VIRT(3, \
					CXD9225GG_TIMER_TMD_OFFSET)
#define CXD9225GG_TIMER_CMD3_VIRT		CXD9225GG_TIMER_REG_VIRT(3, \
					CXD9225GG_TIMER_CMD_OFFSET)
#define CXD9225GG_TIMER_TMCD3_VIRT		CXD9225GG_TIMER_REG_VIRT(3, \
					CXD9225GG_TIMER_TMCD_OFFSET)
/* CH4 */
#define CXD9225GG_TIMER_TMD4_PHYS		CXD9225GG_TIMER_REG_PHYS(4, \
					CXD9225GG_TIMER_TMD_OFFSET)
#define CXD9225GG_TIMER_CMD4_PHYS		CXD9225GG_TIMER_REG_PHYS(4, \
					CXD9225GG_TIMER_CMD_OFFSET)
#define CXD9225GG_TIMER_TMCD4_PHYS		CXD9225GG_TIMER_REG_PHYS(4, \
					CXD9225GG_TIMER_TMCD_OFFSET)

#define CXD9225GG_TIMER_TMD4_VIRT		CXD9225GG_TIMER_REG_VIRT(4, \
					CXD9225GG_TIMER_TMD_OFFSET)
#define CXD9225GG_TIMER_CMD4_VIRT		CXD9225GG_TIMER_REG_VIRT(4, \
					CXD9225GG_TIMER_CMD_OFFSET)
#define CXD9225GG_TIMER_TMCD4_VIRT		CXD9225GG_TIMER_REG_VIRT(4, \
					CXD9225GG_TIMER_TMCD_OFFSET)

/* CH5 .. CH8 are not used by Linux */

#endif /* __ASM_ARCH_CXD9225GG_TIMER_H */
