/*
 *  UART definitions
 *
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef __ASM_ARCH_CXD9225GG_UART_H
#define __ASM_ARCH_CXD9225GG_UART_H

/* UART0 Offset from AHB area top */
#define CXD9225GG_UART_REG_OFFSET		0x00004800
/* Area size of each UART */
#define CXD9225GG_UART_REG_SIZE		0x00000200

/* Nth UART offset address from APB area */
#define CXD9225GG_UART_REG_PHYS_BASE(n)	(CXD9225GG_APB_REG_PHYS_BASE + \
					 CXD9225GG_UART_REG_OFFSET + \
					 CXD9225GG_UART_REG_SIZE * n)
#define CXD9225GG_UART_REG_VIRT_BASE(n)	(CXD9225GG_APB_REG_VIRT_BASE + \
					 CXD9225GG_UART_REG_OFFSET + \
					 CXD9225GG_UART_REG_SIZE * n)

/* 8250/16x50 compatible registers */
#define CXD9225GG_UART_RBR_OFFSET		0x00000000
#define CXD9225GG_UART_THR_OFFSET		0x00000000
#define CXD9225GG_UART_DLL_OFFSET		0x00000000

#define CXD9225GG_UART_DLH_OFFSET		0x00000004
#define CXD9225GG_UART_IER_OFFSET		0x00000004

#define CXD9225GG_UART_IRR_OFFSET		0x00000008
#define CXD9225GG_UART_FCR_OFFSET		0x00000008

#define CXD9225GG_UART_LCR_OFFSET		0x0000000C

#define CXD9225GG_UART_MCR_OFFSET		0x00000010

#define CXD9225GG_UART_LSR_OFFSET		0x00000014

#define CXD9225GG_UART_MSR_OFFSET		0x00000018

#define CXD9225GG_UART_SCR_OFFSET		0x0000001C

/* Registers */
#define CXD9225GG_UART_USR_OFFSET		0x0000007C
#define CXD9225GG_UART_USR_BUSY		(1 << 0)

#define CXD9225GG_UART_FDR_OFFSET		0x00000100
#define CXD9225GG_UART_FDR_FDR4		(1 << 4)
#define CXD9225GG_UART_FDR_FDR6		(1 << 6)


/* register physical address */
#define CXD9225GG_UART_REG_PHYS(n, r)	(CXD9225GG_UART_REG_PHYS_BASE(n) + r)
/* register virtual address */
#define CXD9225GG_UART_REG_VIRT(n, r)	(CXD9225GG_UART_REG_VIRT_BASE(n) + r)

/* CH0 */
#define CXD9225GG_UART_RBR0_PHYS		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_RBR_OFFSET)
#define CXD9225GG_UART_THR0_PHYS		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_THR_OFFSET)
#define CXD9225GG_UART_DLL0_PHYS		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_DLL_OFFSET)
#define CXD9225GG_UART_DLH0_PHYS		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_DLH_OFFSET)
#define CXD9225GG_UART_IER0_PHYS		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_IER_OFFSET)
#define CXD9225GG_UART_IRR0_PHYS		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_IRR_OFFSET)
#define CXD9225GG_UART_FCR0_PHYS		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_FCC_OFFSET)
#define CXD9225GG_UART_LCR0_PHYS		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_LCR_OFFSET)
#define CXD9225GG_UART_MCR0_PHYS		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_MCR_OFFSET)
#define CXD9225GG_UART_LSR0_PHYS		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_LSR_OFFSET)
#define CXD9225GG_UART_MSR0_PHYS		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_MSR_OFFSET)
#define CXD9225GG_UART_SCR0_PHYS		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_SCR_OFFSET)
#define CXD9225GG_UART_USR0_PHYS		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_USR_OFFSET)
#define CXD9225GG_UART_FDR0_PHYS		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_FDR_OFFSET)

#define CXD9225GG_UART_RBR0_VIRT		CXD9225GG_UART_REG_VIRT(0, \
							CXD9225GG_UART_RBR_OFFSET)
#define CXD9225GG_UART_THR0_VIRT		CXD9225GG_UART_REG_VIRT(0, \
							CXD9225GG_UART_THR_OFFSET)
#define CXD9225GG_UART_DLL0_VIRT		CXD9225GG_UART_REG_VIRT(0, \
							CXD9225GG_UART_DLL_OFFSET)
#define CXD9225GG_UART_DLH0_VIRT		CXD9225GG_UART_REG_VIRT(0, \
							CXD9225GG_UART_DLH_OFFSET)
#define CXD9225GG_UART_IER0_VIRT		CXD9225GG_UART_REG_VIRT(0, \
							CXD9225GG_UART_IER_OFFSET)
#define CXD9225GG_UART_IRR0_VIRT		CXD9225GG_UART_REG_VIRT(0, \
							CXD9225GG_UART_IRR_OFFSET)
#define CXD9225GG_UART_FCR0_VIRT		CXD9225GG_UART_REG_VIRT(0, \
							CXD9225GG_UART_FCC_OFFSET)
#define CXD9225GG_UART_LCR0_VIRT		CXD9225GG_UART_REG_VIRT(0, \
							CXD9225GG_UART_LCR_OFFSET)
#define CXD9225GG_UART_MCR0_VIRT		CXD9225GG_UART_REG_VIRT(0, \
							CXD9225GG_UART_MCR_OFFSET)
#define CXD9225GG_UART_LSR0_VIRT		CXD9225GG_UART_REG_VIRT(0, \
							CXD9225GG_UART_LSR_OFFSET)
#define CXD9225GG_UART_MSR0_VIRT		CXD9225GG_UART_REG_VIRT(0, \
							CXD9225GG_UART_MSR_OFFSET)
#define CXD9225GG_UART_SCR0_VIRT		CXD9225GG_UART_REG_VIRT(0, \
							CXD9225GG_UART_SCR_OFFSET)
#define CXD9225GG_UART_USR0_VIRT		CXD9225GG_UART_REG_VIRT(0, \
							CXD9225GG_UART_USR_OFFSET)
#define CXD9225GG_UART_FDR0_VIRT		CXD9225GG_UART_REG_PHYS(0, \
							CXD9225GG_UART_FDR_OFFSET)

#endif /* __ASM_ARCH_CXD9225GG_UART_H */
