/*
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef __ASM_ARCH_CXD9225GG_VIC_H
#define __ASM_ARCH_CXD9225GG_VIC_H

#define CXD9225GG_INTC_CPU0_PHYS_BASE	0xEFFD0000
#define CXD9225GG_INTC_CPU0_VIRT_BASE	(CXD9225GG_INTC_CPU0_PHYS_BASE - \
					 CXD9225GG_1M_MAP_0_PHYS_BASE + \
					 CXD9225GG_1M_MAP_0_VIRT_BASE)

/* 4 registers */
#define CXD9225GG_INTC_ISL_OFFSET		0x00000060
#define CXD9225GG_INTC_IEN_OFFSET		0x00000080
#define CXD9225GG_INTC_IEC_OFFSET		0x000000A0
#define CXD9225GG_INTC_SWI_OFFSET		0x000000C0
#define CXD9225GG_INTC_SWC_OFFSET		0x000000E0
#define CXD9225GG_INTC_PLS_OFFSET		0x00000100
#define CXD9225GG_INTC_PIC_OFFSET		0x00000120
/* 8 registers */
#define CXD9225GG_INTC_EDGC_OFFSET		0x00000140
#define CXD9225GG_INTC_LVLC_OFFSET		0x00000180
/* 1 register */
#define CXD9225GG_INTC_PRLC_OFFSET		0x000001C4
#define CXD9225GG_INTC_UEN_OFFSET		0x000001C8
#define CXD9225GG_INTC_HVA_OFFSET		0x00000200
/* 128 registers */
#define CXD9225GG_INTC_VADX_OFFSET		0x00000400
#define CXD9225GG_INTC_PRLX_OFFSET		0x00000800

/*
 * Offset address from APB area
 */
#define CXD9225GG_INTC_REG_OFFSET(n, r)	((r) + ((n) << 2))

#define CXD9225GG_INTC_ISL_PHYS(n) (CXD9225GG_INTC_CPU0_PHYS_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_ISL_OFFSET))
#define CXD9225GG_INTC_ISL_VIRT(n) (CXD9225GG_INTC_CPU0_VIRT_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_ISL_OFFSET))

#define CXD9225GG_INTC_IEN_PHYS(n) (CXD9225GG_INTC_CPU0_PHYS_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_IEN_OFFSET))
#define CXD9225GG_INTC_IEN_VIRT(n) (CXD9225GG_INTC_CPU0_VIRT_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_IEN_OFFSET))

#define CXD9225GG_INTC_IEC_PHYS(n) (CXD9225GG_INTC_CPU0_PHYS_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_IEC_OFFSET))
#define CXD9225GG_INTC_IEC_VIRT(n) (CXD9225GG_INTC_CPU0_VIRT_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_IEC_OFFSET))

#define CXD9225GG_INTC_SWI_PHYS(n) (CXD9225GG_INTC_CPU0_PHYS_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_SWI_OFFSET))
#define CXD9225GG_INTC_SWI_VIRT(n) (CXD9225GG_INTC_CPU0_VIRT_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_SWI_OFFSET))

#define CXD9225GG_INTC_SWC_PHYS(n) (CXD9225GG_INTC_CPU0_PHYS_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_SWC_OFFSET))
#define CXD9225GG_INTC_SWC_VIRT(n) (CXD9225GG_INTC_CPU0_VIRT_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_SWC_OFFSET))

#define CXD9225GG_INTC_PLS_PHYS(n) (CXD9225GG_INTC_CPU0_PHYS_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_PLS_OFFSET))
#define CXD9225GG_INTC_PLS_VIRT(n) (CXD9225GG_INTC_CPU0_VIRT_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_PLS_OFFSET))

#define CXD9225GG_INTC_PIC_PHYS(n) (CXD9225GG_INTC_CPU0_PHYS_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_PIC_OFFSET))
#define CXD9225GG_INTC_PIC_VIRT(n) (CXD9225GG_INTC_CPU0_VIRT_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_PIC_OFFSET))

#define CXD9225GG_INTC_EDGC_PHYS(n) (CXD9225GG_INTC_CPU0_PHYS_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_EDGC_OFFSET))
#define CXD9225GG_INTC_EDGC_VIRT(n) (CXD9225GG_INTC_CPU0_VIRT_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_EDGC_OFFSET))

#define CXD9225GG_INTC_LVLC_PHYS(n) (CXD9225GG_INTC_CPU0_PHYS_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_LVLC_OFFSET))
#define CXD9225GG_INTC_LVLC_VIRT(n) (CXD9225GG_INTC_CPU0_VIRT_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_LVLC_OFFSET))

#define CXD9225GG_INTC_PRLC_PHYS (CXD9225GG_INTC_CPU0_PHYS_BASE + \
			CXD9225GG_INTC_PRLC_OFFSET)
#define CXD9225GG_INTC_PRLC_VIRT (CXD9225GG_INTC_CPU0_VIRT_BASE + \
			CXD9225GG_INTC_PRLC_OFFSET)

#define CXD9225GG_INTC_UEN_PHYS (CXD9225GG_INTC_CPU0_PHYS_BASE + \
			CXD9225GG_INTC_UEN_OFFSET)
#define CXD9225GG_INTC_UEN_VIRT (CXD9225GG_INTC_CPU0_VIRT_BASE + \
			CXD9225GG_INTC_UEN_OFFSET)

#define CXD9225GG_INTC_HVA_PHYS (CXD9225GG_INTC_CPU0_PHYS_BASE + \
			CXD9225GG_INTC_HVA_OFFSET)
#define CXD9225GG_INTC_HVA_VIRT (CXD9225GG_INTC_CPU0_VIRT_BASE + \
			CXD9225GG_INTC_HVA_OFFSET)

#define CXD9225GG_INTC_VADX_PHYS(n) (CXD9225GG_INTC_CPU0_PHYS_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_VADX_OFFSET))
#define CXD9225GG_INTC_VADX_VIRT(n) (CXD9225GG_INTC_CPU0_VIRT_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_VADX_OFFSET))

#define CXD9225GG_INTC_PRLX_PHYS(n) (CXD9225GG_INTC_CPU0_PHYS_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_PRLX_OFFSET))
#define CXD9225GG_INTC_PRLX_VIRT(n) (CXD9225GG_INTC_CPU0_VIRT_BASE + \
			CXD9225GG_INTC_REG_OFFSET(n, CXD9225GG_INTC_PRLX_OFFSET))


#define CXD9225GG_INT_CH_BLK               32
#define CXD9225GG_INT_SOURCES_NUM          120

#endif /* __ASM_ARCH_CXD9225GG_VIC_H */
