/*
 * Copyright (C) 2012 Sony Corporation
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/jiffies.h>
#include <linux/slab.h>
#include <linux/hardirq.h>
#include <asm/system.h>

#include <mach/kmsg_que.h>

#define circ_empty(circ)	((circ)->head == (circ)->tail)
#define circ_clear(circ)	((circ)->head = (circ)->tail = 0)


/* Return count in buffer.  */
#define CIRC_CNT_02(head,tail,size)     \
        (head >= tail) ? (head - tail) : (head + size - tail)

/* Return space available, 0..size-1.  We always leave one free char
   as a completely full buffer has head == tail, which is the same as
   empty.  */
#define CIRC_SPACE_02(head,tail,size)   \
        (head >= tail) ? (tail + size - head - 1) : (tail - head)


#define __SHOW_PENDING__
#ifdef  __SHOW_PENDING__
static char * __signo_to_str(int sig)
{
	char * str = NULL;
	switch (sig) {
		#define case_sig(s) case s: str=#s; break;
		case_sig(SIGHUP);
		case_sig(SIGINT);
		case_sig(SIGQUIT);
		case_sig(SIGILL);
		case_sig(SIGTRAP);
		case_sig(SIGABRT);
		//case_sig(SIGIOT);
		case_sig(SIGBUS);
		case_sig(SIGFPE);
		case_sig(SIGKILL);
		case_sig(SIGUSR1);
		case_sig(SIGSEGV);
		case_sig(SIGUSR2);
		case_sig(SIGPIPE);
		case_sig(SIGALRM);
		case_sig(SIGTERM);
		case_sig(SIGSTKFLT);
		case_sig(SIGCHLD);
		case_sig(SIGCONT);
		case_sig(SIGSTOP);
		case_sig(SIGTSTP);
		case_sig(SIGTTIN);
		case_sig(SIGTTOU);
		case_sig(SIGURG);
		case_sig(SIGXCPU);
		case_sig(SIGXFSZ);
		case_sig(SIGVTALRM);
		case_sig(SIGPROF);
		case_sig(SIGWINCH);
		case_sig(SIGIO);
		//case_sig(SIGPOLL);
		case_sig(SIGPWR);
		case_sig(SIGSYS);
		//case_sig(SIGUNUSED);
		case_sig(SIGSWI);
		case 40: str="SIG_EXPCD_HP"; break;
		case 41: str="SIG_EXPCD_OC"; break;
		case 42: str="SIG_EXPCD_SC"; break;
		default: str="Unknown Signal"; break;
	}
	return str;
}

static inline int __get_next_signo(unsigned long *s, unsigned long *m)
{
	unsigned long i, x;
	int sig = 0;
	
	switch (_NSIG_WORDS) {
	default:
		for (i = 0; i < _NSIG_WORDS; ++i, ++s, ++m)
			if ((x = *s &~ *m) != 0) {
				sig = ffz(~x) + i*_NSIG_BPW + 1;
				break;
			}
		break;

	case 2: if ((x = s[0] &~ m[0]) != 0)
			sig = 1;
		else if ((x = s[1] &~ m[1]) != 0)
			sig = _NSIG_BPW + 1;
		else
			break;
		sig += ffz(~x);
		break;

	case 1: if ((x = *s &~ *m) != 0)
			sig = ffz(~x) + 1;
		break;
	}
	
	return sig;
}

static inline void __del_signo(unsigned long *set, int sig)
{
	unsigned long idx = sig - 1;
	if (_NSIG_WORDS == 1)
		set[0] &= ~(1UL << idx);
	else
		set[idx / _NSIG_BPW] &= ~(1UL << (idx % _NSIG_BPW));
}

static inline void __show_pending_signals(sigset_t *sig_set, sigset_t *blocked)
{
	sigset_t tmp_set;
	int signo = -1;

	// prepare the temp set.
	memcpy(&tmp_set, sig_set, sizeof(tmp_set));


	// get next signal number.
	while ((signo = __get_next_signo(tmp_set.sig, blocked->sig)) != 0) {
		// del the signo from temp set.
		__del_signo(tmp_set.sig, signo);
		// show the signo out
		printk("\t%s(%d)\n", __signo_to_str(signo), signo);
	}
}

static inline void __show_pending_signals_tsk(struct task_struct *t)
{
	printk("[pid=%3d, tgid=%3d] pending=<%08lx %08lx>, shared_pending=<%08lx %08lx>\n",
		t->pid, t->tgid, t->pending.signal.sig[1], t->pending.signal.sig[0],
		t->signal->shared_pending.signal.sig[1], t->signal->shared_pending.signal.sig[0]);
	__show_pending_signals(&t->pending.signal, &t->blocked);
	__show_pending_signals(&t->signal->shared_pending.signal, &t->blocked);
}
#endif //__SHOW_PENDING__

/*
 * The core wakeup function.  Non-exclusive wakeups (nr_exclusive == 0) just
 * wake everything up.  If it's an exclusive wakeup (nr_exclusive == small +ve
 * number) then we wake all the non-exclusive tasks and one exclusive task.
 *
 * There are circumstances in which we can try to wake a task which has already
 * started to run but is not in state TASK_RUNNING.  try_to_wake_up() returns
 * zero in this (rare) case, and we handle it by continuing to scan the queue.
 */
static void __wake_up_common_02(wait_queue_head_t *q, unsigned int mode,
			     int nr_exclusive, int sync, void *key)
{
    struct list_head *tmp, *next;

    list_for_each_safe(tmp, next, &q->task_list)
    {
        wait_queue_t *curr = list_entry(tmp, wait_queue_t, task_list);
	if(NULL == curr){
		return;
	}
	
        unsigned flags = curr->flags;

        /* attention... */
        if (curr && curr->func)
        {
            if (curr->func(curr, mode, sync, key) &&
                          (flags & WQ_FLAG_EXCLUSIVE) && !--nr_exclusive)
            break;
        }
        else
        {
            kmsg_error("curr:0x%0x or curr->func is invalid. \n", (unsigned int)curr);
            return;
        }
    }
}



/* create a kmsg queue. */
int fastcall kmsg_que_create (kmsg_que_id_t *msg_que_id,
                     unsigned long msg_num,
                     unsigned long msg_len)
{
    struct kmsg_que_t *p_kmsg_que = NULL;

    if (NULL == msg_que_id)
    {
        kmsg_error("msg_id is invalid. \n");
        return KMSG_QUE_ERROR;
    }

    /* the number of msg must be 2^n. */
    if ( (KMSG_QUE_MIN_MSG > msg_num)
        || (KMSG_QUE_MAX_MSG < msg_num)
        || (KMSG_QUE_MSG_MAX_LEN < msg_len)
        )
    {
        kmsg_error(" msg_num or msg_len is invalid, please check them...\n");
        return KMSG_QUE_LEN_ERR;
    }

    if (msg_num & (msg_num - 1))
    {
        kmsg_error(" msg_num must be aligned by 2^n now.\n");
        return KMSG_QUE_LEN_ERR;
    }

    /* is it in isr context?? */
    if (in_interrupt())
    {
        kmsg_error(" you should NOT use it when in_interrupt()==true.\n");
        return KMSG_QUE_IRQ_ERR;
    }

    p_kmsg_que = kzalloc(sizeof(struct kmsg_que_t) + (msg_num * msg_len), GFP_KERNEL);
    if (NULL == p_kmsg_que)
    {
        kmsg_error("kmalloc for kmsg_que failed. \n");
        return KMSG_QUE_ERROR;
    }

    p_kmsg_que->msg_size = msg_len;
    p_kmsg_que->total_msg = msg_num;
    p_kmsg_que->msg_data = (p_kmsg_que + 1);
    p_kmsg_que->reserved = KMSG_QUE_MAGIC;

    init_waitqueue_head(&p_kmsg_que->mq_full_wait);

    init_waitqueue_head(&p_kmsg_que->mq_empty_wait);

    spin_lock_init(&p_kmsg_que->mq_lock);


    /* attention: */
    *msg_que_id = (kmsg_que_id_t)&(p_kmsg_que->reserved);
    return KMSG_QUE_OK;
}
EXPORT_SYMBOL(kmsg_que_create);


/* destroy a specific kmsg queue. */
int fastcall __sched kmsg_que_destroy (kmsg_que_id_t *msg_que_id /* , int option */)
{
    struct kmsg_que_t *p_kmsg_que = NULL;
    int ret = KMSG_QUE_ERROR;
    unsigned long flags;
    int count;

    if (NULL == msg_que_id)
    {
        kmsg_error(":msg_id is invalid. \n");
        return KMSG_QUE_ERROR;
    }

    if (KMSG_QUE_MAGIC != *(unsigned long *)(*msg_que_id))
    {
        kmsg_error(":magic number is modified. it means there are some errors in your code. \n");
    }

    p_kmsg_que = container_of((unsigned long *)(*msg_que_id), struct kmsg_que_t, reserved);

    spin_lock_irqsave(&p_kmsg_que->mq_lock, flags);

/* clean up. */

    /* get the msg-count in queue. */
    count = CIRC_CNT(p_kmsg_que->head,
                     p_kmsg_que->tail,
                     p_kmsg_que->total_msg);

    /* get the msg-count in queue. */
    if (p_kmsg_que->head >= p_kmsg_que->tail)
    {
        count = p_kmsg_que->head - p_kmsg_que->tail;
    }
    else
    {
        count = p_kmsg_que->head + p_kmsg_que->total_msg - p_kmsg_que->tail;
    }

    while (count > 0)
    {
        /* update the head position. */
        p_kmsg_que->tail = (p_kmsg_que->tail + 1) & (p_kmsg_que->total_msg - 1);

        /* wake up one of the threads in the wait-queue. */
        __wake_up_common_02(&p_kmsg_que->mq_empty_wait,
                            TASK_UNINTERRUPTIBLE | TASK_INTERRUPTIBLE,
                            1,
                            1,
                            NULL);
        count--;
    }



    *msg_que_id = (kmsg_que_id_t)0;

    ret = KMSG_QUE_OK;

    spin_unlock_irqrestore(&p_kmsg_que->mq_lock, flags);

    kfree(p_kmsg_que);

    return ret;
}
EXPORT_SYMBOL(kmsg_que_destroy);


/* send a msg to the specific queue.
 * option: no-wait.
 *         wait_forever
 *         some milliseconds.
 * return: =0 --> OK --attention....
 *         <0 --> error
 *         >0 --> timeout
 */
int fastcall __sched kmsg_que_send (const kmsg_que_id_t *msg_que_id,
                   void *msg_data,
                   int msg_len,
                   int option)
{
    struct kmsg_que_t *p_kmsg_que = NULL;
    int count;
    int ret;
    unsigned long flags;

    if ((NULL == msg_que_id) || (NULL == msg_data))
    {
        kmsg_error(":msg_id is invalid. \n");
        return KMSG_QUE_ERROR;
    }

    if (KMSG_QUE_MAGIC != *(unsigned long *)(*msg_que_id))
    {
        kmsg_error(":magic number is modified exceptionally. it means there are some errors in your code. \n");
        return KMSG_QUE_ERROR;
    }

    /* is it in isr context?? */
    if (in_interrupt())
    {
        /* attention: you should NOT block in isr(). */
        if (NO_WAIT != option)
        {
            kmsg_error(": you should NOT use 'WAIT_FOREVER' in ISR().\n");
            return KMSG_QUE_IRQ_ERR;
        }
    }

    p_kmsg_que = container_of((unsigned long *)(*msg_que_id), struct kmsg_que_t, reserved);

    /* check the msg-length. */
    if (msg_len > p_kmsg_que->msg_size)
    {
        kmsg_warning(": msg_len > p_kmsg_que->msg_size.\n");
        msg_len = p_kmsg_que->msg_size;
    }

    spin_lock_irqsave(&p_kmsg_que->mq_empty_wait.lock, flags);

    /* get the space-count in queue. */
    count = CIRC_SPACE(p_kmsg_que->head,
                       p_kmsg_que->tail,
                       p_kmsg_que->total_msg);


    if (!count)
    {
        DECLARE_WAITQUEUE(wait, current);

        wait.flags |= WQ_FLAG_EXCLUSIVE;
        __add_wait_queue_tail(&p_kmsg_que->mq_empty_wait, &wait);

        do
        {
            if (signal_pending(current))
            {
                ret = -ERESTARTSYS;
                __remove_wait_queue(&p_kmsg_que->mq_empty_wait, &wait);
                goto send_out;
            }
            __set_current_state(TASK_INTERRUPTIBLE);

            /* unlock... */
            spin_unlock_irqrestore(&p_kmsg_que->mq_empty_wait.lock, flags);

            if (NO_WAIT == option)
            {
                return KMSG_QUE_ERROR;
            }
            else if ((int)WAIT_FOREVER == option)
            {
                /* schedule-yield. */
                schedule();

                /* lock again... */
                spin_lock_irqsave(&p_kmsg_que->mq_empty_wait.lock, flags);
            }
            else
            {
                /* schedule-yield timeout. */
                option = schedule_timeout(option);

                /* lock again... */
                spin_lock_irqsave(&p_kmsg_que->mq_empty_wait.lock, flags);

                if (!option)
                {
                    ret = KMSG_QUE_TIMEOUT_ERR;
                    __remove_wait_queue(&p_kmsg_que->mq_empty_wait, &wait);
                    goto send_out;
                }
            }

        } while (!(count = CIRC_SPACE(p_kmsg_que->head,
                       p_kmsg_que->tail,
                       p_kmsg_que->total_msg)));

        /* get a space in the queue... */
        __remove_wait_queue(&p_kmsg_que->mq_empty_wait, &wait);
    }

    /* save the msg data. */
    memcpy(((char *)(p_kmsg_que->msg_data) + p_kmsg_que->head * p_kmsg_que->msg_size),
           msg_data, msg_len);

    /* update the head position. */
    p_kmsg_que->head = (p_kmsg_que->head + 1) % p_kmsg_que->total_msg;


    /* wake up one of the threads in the wait-queue. */
    __wake_up_common_02(&p_kmsg_que->mq_full_wait,
                        TASK_UNINTERRUPTIBLE | TASK_INTERRUPTIBLE,
                        1,
                        1,
                        NULL);
    ret = KMSG_QUE_OK;

send_out:
    spin_unlock_irqrestore(&p_kmsg_que->mq_empty_wait.lock, flags);

    /* preempt_check_resched_delayed() is no longer used. */
    preempt_check_resched();

    /* attention to the return-value, 0-->KMSG_QUE_OK. */
    return ret;
}
EXPORT_SYMBOL(kmsg_que_send);


int fastcall __sched kmsg_que_send_urgent (kmsg_que_id_t *msg_que_id,
                                           void *msg_data,
                                           int msg_len,
                                           int option)
{
    struct kmsg_que_t *p_kmsg_que = NULL;
    int count;
    int ret;
    unsigned long flags;
    unsigned long tail_tmp;

    if ((NULL == msg_que_id) || (NULL == msg_data))
    {
        kmsg_error(":msg_id is invalid. \n");
        return KMSG_QUE_ERROR;
    }

    if (KMSG_QUE_MAGIC != *(unsigned long *)(*msg_que_id))
    {
        kmsg_error(":magic number is modified exceptionally. it means there are some errors in your code. \n");
        return KMSG_QUE_ERROR;
    }

    /* is it in isr context?? */
    if (in_interrupt())
    {
        /* attention: you should NOT block in isr(). */
        if (NO_WAIT != option)
        {
            kmsg_error(": you should NOT use 'WAIT_FOREVER' in ISR().\n");
            return KMSG_QUE_IRQ_ERR;
        }
    }

    p_kmsg_que = container_of((unsigned long *)(*msg_que_id), struct kmsg_que_t, reserved);

    /* check the msg-length. */
    if (msg_len > p_kmsg_que->msg_size)
    {
        kmsg_warning(": msg_len > p_kmsg_que->msg_size.\n");
        msg_len = p_kmsg_que->msg_size;
    }

    spin_lock_irqsave(&p_kmsg_que->mq_empty_wait.lock, flags);

    /* get the space-count in queue. */
    count = CIRC_SPACE(p_kmsg_que->head,
                       p_kmsg_que->tail,
                       p_kmsg_que->total_msg);
    if (!count)
    {
        DECLARE_WAITQUEUE(wait, current);

        wait.flags |= WQ_FLAG_EXCLUSIVE;
        __add_wait_queue_tail(&p_kmsg_que->mq_empty_wait, &wait);

        do
        {
            if (signal_pending(current))
            {
                ret = -ERESTARTSYS;
                __remove_wait_queue(&p_kmsg_que->mq_empty_wait, &wait);
                goto send_urgent_out;
            }
            __set_current_state(TASK_INTERRUPTIBLE);

            /* unlock... */
            spin_unlock_irqrestore(&p_kmsg_que->mq_empty_wait.lock, flags);

            if (NO_WAIT == option)
            {
                return KMSG_QUE_ERROR;
            }
            else if ((int)WAIT_FOREVER == option)
            {
                /* schedule-yield. */
                schedule();

                /* lock again... */
                spin_lock_irqsave(&p_kmsg_que->mq_empty_wait.lock, flags);
            }
            else
            {
                /* schedule-yield timeout. */
                option = schedule_timeout(option);

                /* lock again... */
                spin_lock_irqsave(&p_kmsg_que->mq_empty_wait.lock, flags);

                if (!option)
                {
                    ret = KMSG_QUE_TIMEOUT_ERR;
                    __remove_wait_queue(&p_kmsg_que->mq_empty_wait, &wait);
                    goto send_urgent_out;
                }
            }

        } while (!(count = CIRC_SPACE(p_kmsg_que->head,
                       p_kmsg_que->tail,
                       p_kmsg_que->total_msg)));

        /* get a space in the queue... */
        __remove_wait_queue(&p_kmsg_que->mq_empty_wait, &wait);
    }

    /* calculate the head-position.. */
    if(0 == p_kmsg_que->tail)
    {
        tail_tmp = p_kmsg_que->total_msg - 1;
    }
    else
    {
        tail_tmp = p_kmsg_que->tail - 1;
    }

    /* save the msg data. */
    memcpy(((char *)(p_kmsg_que->msg_data) + tail_tmp * p_kmsg_que->msg_size),
          msg_data, msg_len);

    /* update the tail-pointer. */
    p_kmsg_que->tail = tail_tmp;

    /* wake up one of the threads in the wait-queue. */
    __wake_up_common_02(&p_kmsg_que->mq_full_wait,
                        TASK_UNINTERRUPTIBLE | TASK_INTERRUPTIBLE,
                        1,
                        1,
                        NULL);
    ret = KMSG_QUE_OK;

send_urgent_out:
    spin_unlock_irqrestore(&p_kmsg_que->mq_empty_wait.lock, flags);

    /* preempt_check_resched_delayed() is no longer used. */
    preempt_check_resched();

    /* attention to the return-value, 0-->KMSG_QUE_OK. */
    return ret;
}
EXPORT_SYMBOL(kmsg_que_send_urgent);


/* receive a msg from the specific queue.
 * option: no-wait.
 *         wait_forever
 *         some milliseconds.
 */
int fastcall __sched kmsg_que_receive (const kmsg_que_id_t *msg_que_id,
                      void *msg_data,
                      int msg_len,
                      int option)
{
    struct kmsg_que_t *p_kmsg_que = NULL;
    int count;
    int ret;
    unsigned long flags;

    if ((NULL == msg_que_id) || (NULL == msg_data))
    {
        kmsg_error(":msg_id is invalid. \n");
        return KMSG_QUE_ERROR;
    }

    if (KMSG_QUE_MAGIC != *(unsigned long *)(*msg_que_id))
    {
        kmsg_error(":magic number is modified exceptionally. it means there are some errors in your code. \n");
        return KMSG_QUE_ERROR;
    }

    /* is it in isr context?? */
    if (in_interrupt())
    {
        /* attention: you should NOT block in isr(). */
        if (NO_WAIT != option)
        {
            kmsg_error(": you should NOT use 'WAIT_FOREVER' in ISR().\n");
            return KMSG_QUE_IRQ_ERR;
        }
    }

    p_kmsg_que = container_of((unsigned long *)(*msg_que_id), struct kmsg_que_t, reserved);

    /* check the msg-length. */
    if (msg_len < p_kmsg_que->msg_size)
    {
        kmsg_warning(":msg_len < p_kmsg_que->msg_size.\n");
    }

    /* attention: no-matter is it empty or full lock. */
    spin_lock_irqsave(&p_kmsg_que->mq_full_wait.lock, flags);



    /* get the msg-count in queue. */
    count = CIRC_CNT(p_kmsg_que->head,
                     p_kmsg_que->tail,
                     p_kmsg_que->total_msg);

    if (!count)
    {
        DECLARE_WAITQUEUE(wait, current);

        wait.flags |= WQ_FLAG_EXCLUSIVE;
        __add_wait_queue_tail(&p_kmsg_que->mq_full_wait, &wait);

        do
        {
            if (signal_pending(current))
            {
				#ifdef __SHOW_PENDING__ //adds to see SIGNO
				__show_pending_signals_tsk(current);
				#endif
                ret = -ERESTARTSYS;
                __remove_wait_queue(&p_kmsg_que->mq_full_wait, &wait);
                goto recv_out;
            }
            __set_current_state(TASK_INTERRUPTIBLE);

            /* unlock... */
            spin_unlock_irqrestore(&p_kmsg_que->mq_full_wait.lock, flags);

            if (NO_WAIT == option)
            {
                return KMSG_QUE_ERROR;
            }
            else if ((int)WAIT_FOREVER == option)
            {
                /* schedule-yield. */
                schedule();

                /* lock again... */
                spin_lock_irqsave(&p_kmsg_que->mq_full_wait.lock, flags);
            }
            else
            {
                /* schedule-yield timeout. */
                option = schedule_timeout(option);

                /* lock again... */
                spin_lock_irqsave(&p_kmsg_que->mq_full_wait.lock, flags);

                if (!option)
                {
                    ret = KMSG_QUE_TIMEOUT_ERR;
                    __remove_wait_queue(&p_kmsg_que->mq_full_wait, &wait);
                    goto recv_out;
                }
            }

        } while (!(count = CIRC_CNT(p_kmsg_que->head,
                     p_kmsg_que->tail,
                     p_kmsg_que->total_msg)));

        /* get a space in the queue... */
        __remove_wait_queue(&p_kmsg_que->mq_full_wait, &wait);
    }

    /* get the msg-data. */
    memcpy(msg_data,
           ((char *)(p_kmsg_que->msg_data) + p_kmsg_que->tail * p_kmsg_que->msg_size),
           msg_len);

    /* update the head position. */
    p_kmsg_que->tail = (p_kmsg_que->tail + 1) % p_kmsg_que->total_msg;

    /* wake up one of the threads in the wait-queue. */
    __wake_up_common_02(&p_kmsg_que->mq_empty_wait,
                        TASK_UNINTERRUPTIBLE | TASK_INTERRUPTIBLE,
                        1,
                        1,
                        NULL);

    ret = KMSG_QUE_OK;

recv_out:
    spin_unlock_irqrestore(&p_kmsg_que->mq_full_wait.lock, flags);

    /* preempt_check_resched_delayed() is no longer used. */
    preempt_check_resched();

    /* attention to the return-value, 0-->KMSG_QUE_OK. */
    return ret;
}
EXPORT_SYMBOL(kmsg_que_receive);


/* get the number of messages in the specific queue now.
 * msg_que_id: .
 */
int fastcall kmsg_que_num_msgs (const kmsg_que_id_t *msg_que_id)
{
    struct kmsg_que_t *p_kmsg_que = NULL;
    int count;
    unsigned long flags;

    if (NULL == msg_que_id)
    {
        kmsg_error(":msg_id is invalid. \n");
        return KMSG_QUE_ERROR;
    }

    if (KMSG_QUE_MAGIC != *(unsigned long *)(*msg_que_id))
    {
        kmsg_error(":magic number is modified exceptionally. it means there are some errors in your code. \n");
        return KMSG_QUE_ERROR;
    }

    p_kmsg_que = container_of((unsigned long *)(*msg_que_id), struct kmsg_que_t, reserved);

    spin_lock_irqsave(&p_kmsg_que->mq_lock, flags);

    /* get the msg-count in queue. */
    count = CIRC_CNT(p_kmsg_que->head,
                     p_kmsg_que->tail,
                     p_kmsg_que->total_msg);

    spin_unlock_irqrestore(&p_kmsg_que->mq_lock, flags);

    return count;
}
EXPORT_SYMBOL(kmsg_que_num_msgs);

int fastcall __sched kmsg_que_cleanup (const kmsg_que_id_t *msg_que_id)
{
    struct kmsg_que_t *p_kmsg_que = NULL;
    int count;
    unsigned long flags;

    if (NULL == msg_que_id)
    {
        kmsg_error(":msg_id is invalid. \n");
        return KMSG_QUE_ERROR;
    }

    if (KMSG_QUE_MAGIC != *(unsigned long *)(*msg_que_id))
    {
        kmsg_error(":magic number is modified exceptionally. it means there are some errors in your code. \n");
        return KMSG_QUE_ERROR;
    }

    p_kmsg_que = container_of((unsigned long *)(*msg_que_id), struct kmsg_que_t, reserved);

    spin_lock_irqsave(&p_kmsg_que->mq_lock, flags);

    /* get the msg-count in queue. */
    count = CIRC_CNT(p_kmsg_que->head,
                     p_kmsg_que->tail,
                     p_kmsg_que->total_msg);

    while (count > 0)
    {
        /* update the head position. */
        p_kmsg_que->tail = (p_kmsg_que->tail + 1) & (p_kmsg_que->total_msg - 1);

        /* wake up one of the threads in the wait-queue. */
        __wake_up_common_02(&p_kmsg_que->mq_empty_wait,
                            TASK_UNINTERRUPTIBLE | TASK_INTERRUPTIBLE,
                            1,
                            1,
                            NULL);
        count--;
    }

    spin_unlock_irqrestore(&p_kmsg_que->mq_lock, flags);

    return KMSG_QUE_OK;
}
EXPORT_SYMBOL(kmsg_que_cleanup);
