/* 2014-08-28: File changed by Sony Corporation */
/*
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/pci.h>
#include <linux/pci_regs.h>
#include <linux/io.h>
#include <linux/delay.h>

#include <mach/hardware.h>
#include <asm/mach/pci.h>
#include <asm/system.h>
#include <asm/mach/map.h>
#include <mach/hardware.h>
#include <mach/irqs.h>

#include "pci.h"

#define IDSEL_SHIFT	(11)
#define IDSEL_MASK	(0xfffff800)

#define SLOT_SHIFT	(11)
#define SLOT_MASK	(0x0000f800)
#define ADDR_SHIFT	(16)
#define ADDR_MASK	(0x00ff0000)

static int cxd9225gg_pci_setup_config(struct pci_bus *bus, unsigned int devfn,
				   int where, u32 *confaddr, u32 *regsave)
{
	int ret = PCIBIOS_SUCCESSFUL;
	u32 val;
	u32 slot;

	*regsave = readl(CXD9225GG_PCI_AHBPCI_WIN1_CTR_VIRT);

	slot = PCI_SLOT(devfn);

	if (!bus->number) {
		/* type 0 */
		val = CXD9225GG_PCI_AHBPCI_WIN_CTR_CFGTYPE_0 |
			CXD9225GG_PCI_AHBPCI_WIN_CTR_PCICMD_CONFIG |
			(1 << (slot + IDSEL_SHIFT));
	} else {
		/* type 1 */
		pr_debug("%s: config type1 %d %x %x\n", __func__,
			 bus->number, devfn, where);
		val = CXD9225GG_PCI_AHBPCI_WIN_CTR_CFGTYPE_1 |
			CXD9225GG_PCI_AHBPCI_WIN_CTR_PCICMD_CONFIG |
			(bus->number << ADDR_SHIFT) |
			(slot << SLOT_SHIFT);
	}

	writel(val, CXD9225GG_PCI_AHBPCI_WIN1_CTR_VIRT);
	*confaddr = CXD9225GG_PCI_WIN_1_VIRT_BASE | (PCI_FUNC(devfn)  << 8) |
		    where;

	return ret;
}

static int cxd9225gg_pci_read_config(struct pci_bus *bus, unsigned int devfn,
				  int where, int size, u32 *val)
{
	unsigned long flags;
	u32 confaddr, regsave;
	int ret;

	local_irq_save(flags);

	ret = cxd9225gg_pci_setup_config(bus, devfn, where, &confaddr, &regsave);

	if (ret != PCIBIOS_SUCCESSFUL)
		goto done;

	switch (size) {
	case 4:
		*val = readl(confaddr);
		break;
	case 2:
		*val = readw(confaddr);
		break;
	case 1:
		*val = readb(confaddr);
		break;
	default:
		ret = PCIBIOS_FUNC_NOT_SUPPORTED;
		break;
	};
done:
	writel(regsave, CXD9225GG_PCI_AHBPCI_WIN1_CTR_VIRT);
	local_irq_restore(flags);

	return ret;
}

static int cxd9225gg_pci_write_config(struct pci_bus *bus, unsigned int devfn,
				   int where, int size, u32 val)
{
	unsigned long flags;
	u32 confaddr, regsave;
	int ret;

	local_irq_save(flags);

	ret = cxd9225gg_pci_setup_config(bus, devfn, where, &confaddr, &regsave);

	if (ret != PCIBIOS_SUCCESSFUL)
		goto done;

	switch (size) {
	case 4:
		writel(val, confaddr);
		break;
	case 2:
		writew(val, confaddr);
		break;
	case 1:
		writeb(val, confaddr);
		break;
	default:
		ret = PCIBIOS_BAD_REGISTER_NUMBER;
		break;
	};
done:
	writel(regsave, CXD9225GG_PCI_AHBPCI_WIN1_CTR_VIRT);
	local_irq_restore(flags);

	return ret;
}

static struct pci_ops cxd9225gg_pci_ops = {
	.read = cxd9225gg_pci_read_config,
	.write = cxd9225gg_pci_write_config,
};

static struct resource cxd9225gg_pci_memory_resource = {
	.name = "PCI memory space",
	.flags = IORESOURCE_MEM,
	.start = CXD9225GG_PCI_WIN_2_PHYS_BASE,
	.end = CXD9225GG_PCI_WIN_2_PHYS_BASE + CXD9225GG_PCI_WIN_2_SIZE - 1,
};

static struct resource cxd9225gg_pci_io_resource = {
	.name = "PCI IO space",
	.flags = IORESOURCE_IO,
	.start = CXD9225GG_PCI_WIN_1_PHYS_BASE,
	.end = CXD9225GG_PCI_WIN_1_PHYS_BASE + CXD9225GG_PCI_WIN_1_SIZE - 1,
};

static int __init cxd9225gg_pci_setup_resources(struct resource *resource[])
{
	int ret;

	ret = request_resource(&ioport_resource, &cxd9225gg_pci_io_resource);
	if (ret) {
		pr_err("%s: PCI IO space allocation failed\n", __func__);
		goto done;
	}

	ret = request_resource(&iomem_resource, &cxd9225gg_pci_memory_resource);
	if (ret) {
		pr_err("%s: PCI memory space allocation failed\n", __func__);
		goto fail_memory;
	}
	/*
	 * replace IO resources with my own in order to make sure
	 * that the address is within the window size.
	 * See pcibios_init_hw()
	 */
	resource[0] = &cxd9225gg_pci_io_resource;
	resource[1] = &cxd9225gg_pci_memory_resource;

	return 0;

fail_memory:
	release_resource(&cxd9225gg_pci_io_resource);
done:
	return ret;
}

struct pci_bus *cxd9225gg_pci_scan_bus(int nr, struct pci_sys_data *sys)
{
	if (nr) {
		pr_err("%s: nr (%d) != 0\n", __func__, nr);
		return NULL;
	}

	return pci_scan_bus(sys->busnr, &cxd9225gg_pci_ops, sys);
};

int __init cxd9225gg_pci_setup(int nr, struct pci_sys_data *sys)
{
	int ret;

	/* we only support one PCI root bus */
	if (nr) {
		pr_err("%s: no more than one PCI bus supported %d\n", __func__,
		       nr);
		ret = -EIO;
		goto done;
	}

	/* request resource */
	sys->mem_offset = 0;
	sys->io_offset = 0;
	ret = cxd9225gg_pci_setup_resources(sys->resource);
	return 1; /* zero means fail */
done:
	return ret;
};

static unsigned int ahbpci_win1_ctl;

extern void cxd9225gg_pci_suspend(void)
{
	ahbpci_win1_ctl	= readl(CXD9225GG_PCI_AHBPCI_WIN1_CTR_VIRT);
}

extern void cxd9225gg_pci_resume(void)
{
	u32 val;

	/*
	 * reset PCI controller
	 * Ideally REST# of PCI bus should be asserted
	 * while resetting PCI controller to avoid PCI deviceses'
	 * access.
	 * As we can assume the bootloader stops
	 * all PCI devices before invoking Linux, it's OK without
	 * resetting the PCI bus.
	 */

	val = readl(CXD9225GG_SYSC_HWRST1_VIRT);
	val |= CXD9225GG_SYSC_HWRST1_PCI;
	writel(val, CXD9225GG_SYSC_HWRST1_VIRT);
	udelay(1);
	val &= ~CXD9225GG_SYSC_HWRST1_PCI;
	writel(val, CXD9225GG_SYSC_HWRST1_VIRT);

	/* AHB_BUS_CTR */
	val = CXD9225GG_PCI_AHB_BUS_CTR_SMODE_READY_CTR |
		CXD9225GG_PCI_AHB_BUS_CTR_SMODE_READ_BURST |
		CXD9225GG_PCI_AHB_BUS_CTR_MMODE_BURST_WIDTH_16 |
		CXD9225GG_PCI_AHB_BUS_CTR_MMODE_HTRANS;
	writel(val, CXD9225GG_PCI_AHB_BUS_CTR_VIRT);

	/* interrupts */
	val = CXD9225GG_PCI_PCI_INT_INT0 | CXD9225GG_PCI_PCI_INT_INT1 |
		CXD9225GG_PCI_PCI_INT_INT2 | CXD9225GG_PCI_PCI_INT_INT3;
	writel(val, CXD9225GG_PCI_PCI_INT_ENABLE_VIRT);

	/* PCI Arbiter */
	val = (7 << CXD9225GG_PCI_PCI_ARBITER_CTR_TIMER_SHIFT) |
		CXD9225GG_PCI_PCI_ARBITER_CTR_BP_MODE |
		CXD9225GG_PCI_PCI_ARBITER_REQ7 |
		CXD9225GG_PCI_PCI_ARBITER_REQ6 |
		CXD9225GG_PCI_PCI_ARBITER_REQ5 |
		CXD9225GG_PCI_PCI_ARBITER_REQ4 |
		CXD9225GG_PCI_PCI_ARBITER_REQ3 |
		CXD9225GG_PCI_PCI_ARBITER_REQ2 |
		CXD9225GG_PCI_PCI_ARBITER_REQ1 |
		CXD9225GG_PCI_PCI_ARBITER_REQ0;
	writel(val, CXD9225GG_PCI_PCI_ARBITER_CTR_VIRT);

	/*
	 * AHB -> PCI window #1
	 * defaulted to IO access
	 * we use PCI IO addr == AHB physical mapping
	 */
	val = CXD9225GG_PCI_WIN_1_PHYS_BASE |
		CXD9225GG_PCI_AHBPCI_WIN_CTR_ENDIAN_THRU |
		CXD9225GG_PCI_AHBPCI_WIN_CTR_PCICMD_IO;
	writel(ahbpci_win1_ctl, CXD9225GG_PCI_AHBPCI_WIN1_CTR_VIRT);
	/*
	 * AHB -> PCI windows #2
	 * to PCI memory spece
	 */
	val = CXD9225GG_PCI_WIN_2_PHYS_BASE |
		CXD9225GG_PCI_AHBPCI_WIN_CTR_BURST_ENABLE |
		CXD9225GG_PCI_AHBPCI_WIN_CTR_PCICMD_MEM;
	writel(val, CXD9225GG_PCI_AHBPCI_WIN2_CTR_VIRT);

	/*
	 * PCI -> AHB windows #1, 512MB window
	 */
	val = CXD9225GG_DDR2_PHYS_BASE | /* system memory base */
		CXD9225GG_PCI_PCIAHB_WIN_CTR_PREFETCH_16;

	writel(val, CXD9225GG_PCI_PCIAHB_WIN1_CTR_VIRT);
	/*
	 * PCI -> AHB windows #2, 256 window
	 * to prevent crashing system, set the dummy AHB address
	 */
	val = CXD9225GG_PCI_AHB_DUMMY_ADDR_PHYS;
	writel(val, CXD9225GG_PCI_PCIAHB_WIN2_CTR_VIRT);
}

int __init cxd9225gg_pci_init(struct hw_pci *hwpci)
{
	u32 val;

	/*
	 * reset PCI controller
	 * Ideally REST# of PCI bus should be asserted
	 * while resetting PCI controller to avoid PCI deviceses'
	 * access.
	 * As we can assume the bootloader stops
	 * all PCI devices before invoking Linux, it's OK without
	 * resetting the PCI bus.
	 */

	val = readl(CXD9225GG_SYSC_HWRST1_VIRT);
	val |= CXD9225GG_SYSC_HWRST1_PCI;
	writel(val, CXD9225GG_SYSC_HWRST1_VIRT);
	udelay(1);
	val &= ~CXD9225GG_SYSC_HWRST1_PCI;
	writel(val, CXD9225GG_SYSC_HWRST1_VIRT);

	/* AHB_BUS_CTR */
	val = CXD9225GG_PCI_AHB_BUS_CTR_SMODE_READY_CTR |
		CXD9225GG_PCI_AHB_BUS_CTR_SMODE_READ_BURST |
		CXD9225GG_PCI_AHB_BUS_CTR_MMODE_BURST_WIDTH_16 |
		CXD9225GG_PCI_AHB_BUS_CTR_MMODE_HTRANS;
	writel(val, CXD9225GG_PCI_AHB_BUS_CTR_VIRT);

	/* interrupts */
	val = CXD9225GG_PCI_PCI_INT_INT0 | CXD9225GG_PCI_PCI_INT_INT1 |
		CXD9225GG_PCI_PCI_INT_INT2 | CXD9225GG_PCI_PCI_INT_INT3;
	writel(val, CXD9225GG_PCI_PCI_INT_ENABLE_VIRT);

	/* PCI Arbiter */
	val = (7 << CXD9225GG_PCI_PCI_ARBITER_CTR_TIMER_SHIFT) |
		CXD9225GG_PCI_PCI_ARBITER_CTR_BP_MODE |
		CXD9225GG_PCI_PCI_ARBITER_REQ7 |
		CXD9225GG_PCI_PCI_ARBITER_REQ6 |
		CXD9225GG_PCI_PCI_ARBITER_REQ5 |
		CXD9225GG_PCI_PCI_ARBITER_REQ4 |
		CXD9225GG_PCI_PCI_ARBITER_REQ3 |
		CXD9225GG_PCI_PCI_ARBITER_REQ2 |
		CXD9225GG_PCI_PCI_ARBITER_REQ1 |
		CXD9225GG_PCI_PCI_ARBITER_REQ0;
	writel(val, CXD9225GG_PCI_PCI_ARBITER_CTR_VIRT);

	/*
	 * AHB -> PCI window #1
	 * defaulted to IO access
	 * we use PCI IO addr == AHB physical mapping
	 */
	val = CXD9225GG_PCI_WIN_1_PHYS_BASE |
		CXD9225GG_PCI_AHBPCI_WIN_CTR_ENDIAN_THRU |
		CXD9225GG_PCI_AHBPCI_WIN_CTR_PCICMD_IO;
	writel(val, CXD9225GG_PCI_AHBPCI_WIN1_CTR_VIRT);
	/*
	 * AHB -> PCI windows #2
	 * to PCI memory spece
	 */
	val = CXD9225GG_PCI_WIN_2_PHYS_BASE |
		CXD9225GG_PCI_AHBPCI_WIN_CTR_BURST_ENABLE |
		CXD9225GG_PCI_AHBPCI_WIN_CTR_PCICMD_MEM;
	writel(val, CXD9225GG_PCI_AHBPCI_WIN2_CTR_VIRT);

	/*
	 * PCI -> AHB windows #1, 512MB window
	 */
	val = CXD9225GG_DDR2_PHYS_BASE | /* system memory base */
		CXD9225GG_PCI_PCIAHB_WIN_CTR_PREFETCH_16;

	writel(val, CXD9225GG_PCI_PCIAHB_WIN1_CTR_VIRT);
	/*
	 * PCI -> AHB windows #2, 256 window
	 * to prevent crashing system, set the dummy AHB address
	 */
	val = CXD9225GG_PCI_AHB_DUMMY_ADDR_PHYS;
	writel(val, CXD9225GG_PCI_PCIAHB_WIN2_CTR_VIRT);

	pci_common_init(hwpci);

	return 0;
}

/*
 * PCI driver for this controller to setup PCI configuration area
 */

static struct pci_device_id cxd9225gg_pci_tbl[] = {
	{PCI_DEVICE(PCI_VENDOR_ID_SONY, PCI_DEVICE_ID_SONY_CXD9225GG_PCIC)},
	{},
};

static int __devinit cxd9225gg_pci_driver_probe(struct pci_dev *pdev,
					     const struct pci_device_id *ent)
{
	u32 val;
	u16 word;

	/* configure BAR0 (PCI comunication area), BAR2(Memory 256M) */
	val = CXD9225GG_PCI_AHB_DUMMY_ADDR_PHYS;
	pci_write_config_dword(pdev, val, PCI_BASE_ADDRESS_0);
	pci_write_config_dword(pdev, val, PCI_BASE_ADDRESS_2);

	/* configure BAR1 (map system memory 512M) */
	val = CXD9225GG_DDR2_PHYS_BASE;
	pci_write_config_dword(pdev, val, PCI_BASE_ADDRESS_1);

	/* command register */
	word = PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER;
	pci_write_config_word(pdev, word, PCI_COMMAND);

	return 0;
};

static void __devexit cxd9225gg_pci_driver_remove(struct pci_dev *pdev)
{
	pci_write_config_word(pdev, 0, PCI_COMMAND);
}

static int cxd9225gg_pci_driver_suspend(struct pci_dev *pdev, pm_message_t mesg)
{
	return 0;
}

static int cxd9225gg_pci_driver_resume(struct pci_dev *pdev)
{
	u32 val;
	u16 word;

	/* configure BAR0 (PCI comunication area), BAR2(Memory 256M) */
	val = CXD9225GG_PCI_AHB_DUMMY_ADDR_PHYS;
	pci_write_config_dword(pdev, val, PCI_BASE_ADDRESS_0);
	pci_write_config_dword(pdev, val, PCI_BASE_ADDRESS_2);

	/* configure BAR1 (map system memory 512M) */
	val = CXD9225GG_DDR2_PHYS_BASE;
	pci_write_config_dword(pdev, val, PCI_BASE_ADDRESS_1);

	/* command register */
	word = PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER;
	pci_write_config_word(pdev, word, PCI_COMMAND);

	return 0;
}

static struct pci_driver cxd9225gg_pci_driver = {
	.name		= "CXD9225GG PCI controller",
	.id_table	= cxd9225gg_pci_tbl,
	.probe		= cxd9225gg_pci_driver_probe,
	.remove		= cxd9225gg_pci_driver_remove,
	.suspend        = cxd9225gg_pci_driver_suspend,
	.resume         = cxd9225gg_pci_driver_resume,
};

static int __init cxd9225gg_pci_driver_init(void)
{
	return pci_register_driver(&cxd9225gg_pci_driver);
}

static void __exit cxd9225gg_pci_driver_exit(void)
{
	pci_unregister_driver(&cxd9225gg_pci_driver);
}

device_initcall(cxd9225gg_pci_driver_init);
module_exit(cxd9225gg_pci_driver_exit);



/* slot (idsel) on primary PCI bus */
#define AVP13_PCI_IDSEL_IEEE1394	5
#define AVP13_PCI_IDSEL_UPD720102	6
#define AVP13_PCI_IDSEL_COMPACTPCI	7
#define AVP13_PCI_IDSEL_I82541PI	8
#define AVP13_PCI_IDSEL_TMS320DM6437	9
#define AVP13_PCI_IDSEL_CXD9225GG		10

#define IDSEL_SHIFT	(11)
#define IDSEL_MASK	(0xfffff800)

#define SLOT_SHIFT	(11)
#define SLOT_MASK	(0x0000f800)
#define ADDR_SHIFT	(16)
#define ADDR_MASK	(0x00ff0000)

static struct pci_sys_data *avp13_pci_sys_data;
/*
 * return slot (idsel) number correspoinding with the pci_dev
 * If the device conneced via PCI bridge, return the slot for
 * the PCI bridge
 */
static u8 avp13_pci_swizzle(struct pci_dev *dev, u8 *pin)
{
	if (dev->bus->number) {
		if (avp13_pci_sys_data->bus != dev->bus)
			pr_err("%s: dev->bus != sysdata->bus\n", __func__);
		/*
		 * secondary or later bus.
		 * For AVP13, only CompactPCI connector can be connected to
		 * PCI bridge
		 */
		*pin = 1;
		return AVP13_PCI_IDSEL_COMPACTPCI;
	} else {
		/* primary bus */
		return PCI_SLOT(dev->devfn);
	}
};

static int avp13_pci_map_irq(struct pci_dev *dev, u8 slot, u8 pin)
{
	int irq;
	/*
	 * slot	   pin	 irq	  device
	 *    5	     1	 68(INTB) IEEE1394
	 *    6	     1	 69(INTC) EHCI
	 *    7	     1	 67(INTA) CompactPCI
	 *    8	     1	 70(INTD) Ethernet
	 *   10	     1?	 71	  PCI controller
	 */

	/*
	 * slot number is assumed to be swizzled already
	 */
	switch (slot) {
	case AVP13_PCI_IDSEL_IEEE1394:
		irq = INT_PCI_INTB;
		break;
	case AVP13_PCI_IDSEL_UPD720102:
		irq = INT_PCI_INTC;
		break;
	case AVP13_PCI_IDSEL_I82541PI:
		irq = INT_PCI_INTD;
		break;
	case AVP13_PCI_IDSEL_CXD9225GG:
		irq = INT_PCI_UNIT;
		break;
	case AVP13_PCI_IDSEL_TMS320DM6437:
		pr_info("%s: DSP used?\n", __func__);
	case AVP13_PCI_IDSEL_COMPACTPCI:
	default:
		irq = INT_PCI_INTA;
		break;
	}

	return irq;
};

static int __init avp13_pci_setup(int nr, struct pci_sys_data *sys)
{
	avp13_pci_sys_data = sys;
	return cxd9225gg_pci_setup(nr, sys);
};

static struct hw_pci avp13_pci __initdata = {
	/*
	 * only INT_A of each slotis recognized so bridges can not be supported
	 */
	.swizzle	= avp13_pci_swizzle,
	.map_irq	= avp13_pci_map_irq,
	.nr_controllers	= 1,
	.setup		= avp13_pci_setup,
	.scan		= cxd9225gg_pci_scan_bus,
};

static int __init avp13_pci_init(void)
{
	return cxd9225gg_pci_init(&avp13_pci);
}

#ifdef CONFIG_SNSC_DEFERRED_INITCALLS_CXD9225GG_PCI
deferred_initcall(avp13_pci_init,
		  CONFIG_SNSC_DEFERRED_INITCALLS_GROUP_CXD9225GG_PCI);
#else
subsys_initcall(avp13_pci_init);
#endif
