/* 2014-08-28: File changed by Sony Corporation */
/*
 * Copyright (C) 2012 Sony Corporation
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/pci.h>
#include <linux/pci_regs.h>
#include <linux/io.h>
#include <linux/delay.h>

#include <asm/mach/pci.h>
#include <mach/gpio.h>

#include <mach/hardware.h>
#include <mach/cxd9225gg_Chipset.h>
#include <mach/cxd9225gg_IO.h>


/*==============================================================================
 * PORT 0 - PCIEx1L0
 * PORT 1 - PCIEx1L1
 * PORT 2 - PCIEx4L
*/

#define NUM_PCIEx1L_PORTS	2
#define NUM_PCIEx4L_PORTS	1

#define NUM_PCIE_PORTS 		(NUM_PCIEx1L_PORTS + NUM_PCIEx4L_PORTS)

#ifdef CONFIG_SNSC_DEFERRED_INITCALLS_CXD9225GG_PCIE
int defd_init_pcie_num = NUM_PCIE_PORTS;
#endif

/*==============================================================================
 * The following elements are some macro functions only used locally.
 */
#define pcie_printk(level, format, arg...) \
		printk(level "[PCIE] %s() @%d :> " format , \
		__FUNCTION__ , __LINE__ , ## arg)

#define CONFIG_PCIE_DEBUG
#ifdef  CONFIG_PCIE_DEBUG

#define pcie_critl(format, arg...) \
		pcie_printk(KERN_CRIT, format "\n" , ## arg)

#define pcie_errorl(format, arg...) \
		pcie_printk(KERN_ERR, format "\n" , ## arg)

#define pcie_warningl(format, arg...) \
		pcie_printk(KERN_WARNING, format "\n" , ## arg)

#define pcie_infol(format, arg...) \
		pcie_printk(KERN_NOTICE, format "\n" , ## arg)/*KERN_INFO*/

#define pcie_debugl(format, arg...) \
		pcie_printk(KERN_DEBUG, format "\n" , ## arg)/*KERN_DEBUG*/

#else

#define pcie_critl(format, arg...) \
	pcie_printk(KERN_CRIT, format "\n" , ## arg)

#define pcie_errorl(format, arg...) \
	pcie_printk(KERN_ERR, format "\n" , ## arg)

#define pcie_warningl(format, arg...)

#define pcie_infol(format, arg...)

#define pcie_debugl(format, arg...)

#endif


/*==============================================================================
 * Local Macro Functions
 */

/* bus/func number of PCIe Device */
#define PCIE_BUS_NO_FIXED 	0
#define PCIE_FNC_NO_FIXED 	0
#define PCIE_DEV_NO_AXIBR 	0

#define CXD9225GG_SYSC_GCTS1_VIRT		(CXD9225GG_SYSC_REG_VIRT_BASE \
					+ CXD9225GG_SYSC_GCTS1_OFFSET)
#define CXD9225GG_SYSC_GCTS2_VIRT		(CXD9225GG_SYSC_REG_VIRT_BASE \
					+ CXD9225GG_SYSC_GCTS2_OFFSET)

#define CXD9225GG_SYSC_CGE_PCIEX1L0 	(1<<16)	/* pciex1L0 clock gate enable */
#define CXD9225GG_SYSC_CGE_PCIEX1L1 	(1<<24)	/* pciex1L1 clock gate enable */
#define CXD9225GG_SYSC_CGE_PCIEX4L 	(1<<0)	/* pciex4L  clock gate enable */

#define CXD9225GG_SYSC_HWRST1_PCIEX1L0 	(1<<16)	/* pciex1L0 hardware reset */
#define CXD9225GG_SYSC_HWRST1_PCIEX1L1 	(1<<24)	/* pciex1L1 hardware reset */
#define CXD9225GG_SYSC_HWRST2_PCIEX4L 	(1<<0)	/* pciex4L  hardware reset */


/*------------------------------------------------------------------------------
 * Macro to read/write controller's registers
 */
#define _PCIE_VB_                       CXD9225GG_PCIE_BUSC_REG_VIRT_BASE

#define PCIE_REG_READ(port, offset, result) \
    do { result = readl(_PCIE_VB_ + ((port)*0x1000)+(offset)); } while(0)

#define PCIE_REG_WRITE(port, offset, value) \
    do { writel((value), _PCIE_VB_ + ((port) * 0x1000)+(offset)); } while(0)

#define PCIE_REG_READ8(port, offset, result) \
    do { result = readb(_PCIE_VB_ + ((port) * 0x1000)+(offset)); } while(0)

#define PCIE_REG_WRITE8(port, offset, value) \
    do { writeb((value), _PCIE_VB_ + ((port) * 0x1000)+(offset)); } while(0)

#define PCIE_REG_READ16(port, offset, result) \
    do { result = readw(_PCIE_VB_ + ((port) * 0x1000)+(offset)); } while(0)

#define PCIE_REG_WRITE16(port, offset, value) \
    do { writew((value), _PCIE_VB_ + ((port) * 0x1000)+(offset)); } while(0)

/*------------------------------------------------------------------------------
 * Device On PCIE PORT 0 <--> SLOT 11
 * Device On PCIE PORT 1 <--> SLOT 12
 * Device On PCIE PORT 2 <--> SLOT 13
 * Invalid PORT or SLOT --> (-1)
 * The above rule is controlled by config space read/write functions.
 */
inline static u32 port_to_pcislot(u32 port)
{
	switch (port) {
	case 0:
		return 11;
	case 1:
		return 12;
	case 2:
		return 13;
	default:
		return -1;
	}
	/* Never come here */
}

inline static u32 pcislot_to_port(u32 slot)
{
	switch (slot) {
	case 11:
		return 0;
	case 12:
		return 1;
	case 13:
		return 2;
	default:
		return -1;
	}
	/* Never come here */
}

#define PORT_TO_PCISLOT(port) (port_to_pcislot(port))
#define PCISLOT_TO_PORT(slot) (pcislot_to_port(slot))
#define DEVFN_TO_PORT(dev)      (PCISLOT_TO_PORT(PCI_SLOT(dev->devfn)))

//#define PORT_TO_EXPCD_SLOT(port) (s_exchg ? ((port+1)&1) : (port))
#define PORT_TO_EXPCD_SLOT(port) (s_exchg?((port==0)?1:((port==1)?0:port)):(port))

#define SLOT_X(port)    ((port)+'A')

/*==============================================================================
 * The following elements are some static variables only used locally.
 */
struct cxd9225gg_pcie_port {
	u32                     port;
	u32                     root_bus_nr;
	struct pci_bus          *pbus; /* One HC extends one bus */
	struct pci_dev          *pdev; /* One PCIE port has only one slot */
	spinlock_t              conf_lock;
	char                    memsp_name[16]; /* mem space name */
	char                    iosp_name[16]; /* io space name */
	struct resource         res[2];
	int 			linkup;
};

/*------------------------------------------------------------------------------
 * root_bus_nr must be init to be -1 !!
 * see bsp_pcie_bus_to_port() to find the reason.
 */
static struct cxd9225gg_pcie_port s_pcie_ports[NUM_PCIE_PORTS];

/*------------------------------------------------------------------------------
 * record the start number of PCIE bus. see bsp_pcie_cal_busnr_start().
 */
static int s_busnr_start = 0;

/*------------------------------------------------------------------------------
 * record the start number of PCIE bus. see bsp_pcie_cal_busnr_start().
 */
static int s_exchg = 0;

/*------------------------------------------------------------------------------
 * This func is used to find the private data of given bus number.
 */
static struct cxd9225gg_pcie_port * bsp_pcie_busnr_to_port(int busnr)
{
	int i;
	for (i = NUM_PCIE_PORTS-1; i >= 0; i--) {
		int rootbus = s_pcie_ports[i].root_bus_nr;
		/* NOTE: root_bus_nr must be inited as -1 !! */
		if (rootbus != -1 && rootbus <= busnr) {
			return &s_pcie_ports[i];
		}
	}

	return NULL;
}

/*------------------------------------------------------------------------------
 * This function is used to calculate how many pci buses there are already in
 * the system. Doing this is because pci controller might be inited before
 * pcie, but both pci and pcie use the same pci-subsys-framework, bus number
 * must not be conflict. So the first pcie bus number should be after the last
 * pci bus number.
 */
static int bsp_pcie_cal_busnr_start(void)
{
	int bnr = 0;
	struct list_head *n;
	/* pci_root_buses is defined in linux/pci.h */
	for (n = &pci_root_buses; n->next != &pci_root_buses; n=n->next, bnr++){}
	return bnr;
}

struct pci_dev * bsp_pcie_get_first_dev(struct pci_bus *bus)
{
	struct pci_dev * dev = NULL;
	list_for_each_entry(dev, &bus->devices, bus_list) {
		/* return the first newly found device*/
		return dev;
	}
	return NULL;
}

/*------------------------------------------------------------------------------
 * Return 1 if ExpressCard is pluged in the slot, otherwise return 0.
 */
static int bsp_pcie_slot_check(int port)
{
	int plugedin = 0;
	int pin = (PORT_TO_EXPCD_SLOT(port)==0) ? 1 : 5;
	if (port == 2)
		plugedin = 1; 	/* In PCIEx4 port, ExpressCard is always pluged in. */
	else
		plugedin = (!cxd9225gg_get_gpio_datain(CXD9225GG_GPIO_ID(CXD9225GG_GPIO_PORT_13,pin)));

	return plugedin;
}

/*------------------------------------------------------------------------------
 * Return 1 in case the power is on, 0 in case the power is off.
 */
static int bsp_pcie_slot_get_power(int port)
{
	int gpio_port = (PORT_TO_EXPCD_SLOT(port)==0)
			? CXD9225GG_GPIO_PORT_10
			: CXD9225GG_GPIO_PORT_11;
	if (port == 2)
		return 1;
	return cxd9225gg_get_gpio_datain(CXD9225GG_GPIO_ID(gpio_port,2));
}

/*------------------------------------------------------------------------------
 * pwr_up : 1 means power up, 0 means power down.
 */
static void bsp_pcie_slot_set_power(int port, int pwr_up)
{
	int gpio_port = (PORT_TO_EXPCD_SLOT(port)==0)
			? CXD9225GG_GPIO_PORT_10
			: CXD9225GG_GPIO_PORT_11;
	if (port == 2)
		return ;
	cxd9225gg_set_gpio_direction(CXD9225GG_GPIO_ID(gpio_port, 2), false);
	cxd9225gg_set_gpio_dataout(CXD9225GG_GPIO_ID(gpio_port,2),pwr_up);
}

/*------------------------------------------------------------------------------
 * Low level function to read device config space on pcie bus.
 * Params:
 * 	port    : host controller index [0,1]
 * 	type    : internal(1) or external(0) [0,1]
 * 		in internal case, busnr, slot and func are ignored.
 * 	busnr   : bus number that the device is on [0,255]
 *	slot    : slot number of the device [0,31]
 *	func    : func number of the device [0,7]
 *	offset  : register offset we want to read [0,4096)
 *	size    : size to read at a time {1,2,4}
 *		size will not be checked in this function.
 * Return the read result.
 * Ratinale:
 *	A pcie bus has at most 32 devices, a pcie device has at most 8 funcs.
 */
static u32 bsp_pcie_cfg_read(u32 port, u32 type, u32 busnr,
                                u32 slot, u32 func, u32 size, u32 offset)
{
	u32 reg_val;

	/* internal type is special */
	if (type==1) {
		type = CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_INTERNAL;
		func = slot = busnr = 0;
	} else {
		type = CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_EXTERNAL;
	}

	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_READ_BYTE_ENABLE,
					CXD9225GG_PCIEX_READ_BYTE_ENABLE_ALL_EN);

	reg_val = CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_CFGTYPE0
		| CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_ISSUE_EN
		| type
		| (func <<CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_FNCNUM_SHIFT)
		| (slot <<CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_DEVNUM_SHIFT)
		| (busnr <<CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_BUSNUM_SHIFT)
		| (offset &CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_EXREG_MASK);

	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR, reg_val);

	switch(size){
	case sizeof(u8):
		PCIE_REG_READ8 (port,
			CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_DATA+(offset&0x3),
						reg_val);
		break;
	case sizeof(u16):
		PCIE_REG_READ16(port,
			CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_DATA+(offset&0x2),
						reg_val);
		break;
	case sizeof(u32):
		PCIE_REG_READ  (port,
			CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_DATA, reg_val);
		break;
	default:
		/* should not happen */
		pcie_errorl("Invalid size %d", size);
		reg_val = 0xffffffff;
		break;
	}

	return reg_val;
}

/*------------------------------------------------------------------------------
 * Low level function to write device config space on pcie bus.
 * Params:
 *	port	: host controller index [0,1]
 *	type	: internal(1) or external(0) [0,1]
 *		in internal case, busnr, slot and func are ignored.
 *	busnr	: bus number that the device is on [0,255]
 *	slot	: slot number of the device [0,31]
 *	func	: func number of the device [0,7]
 *	offset	: register offset we want to write [0,4096)
 *	size	: size to write at a time {1,2,4}
 *		size will not be checked in this function.
 *	value	: value to write [0,0xffffffff]
 * Return the read result.
 * Ratinale 	:
 *	A pcie bus has at most 32 devices, a pcie device has at most 8 funcs.
 */
static void bsp_pcie_cfg_write(u32 port, u32 type, u32 busnr,
                        u32 slot, u32 func, u32 size, u32 offset, u32 value)
{
	u32 reg_val;

	/* internal type is special */
	if (type==1) {
		type = CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_INTERNAL;
		func = slot = busnr = 0;
	} else {
		type = CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_EXTERNAL;
	}

	reg_val = CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_CFGTYPE0
		| type
		| CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_ISSUE_EN
		| (func << CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_FNCNUM_SHIFT)
		| (slot << CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_DEVNUM_SHIFT)
		| (busnr << CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_BUSNUM_SHIFT)
		| (offset & CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_EXREG_MASK);

	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR, reg_val);

	switch (size) {
	case 1:
		PCIE_REG_WRITE8 (port,
			CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_DATA+(offset&0x03),
			value);
		break;
	case 2:
		PCIE_REG_WRITE16(port,
			CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_DATA+(offset&0x02),
			value);
		break;
	case 4:
		PCIE_REG_WRITE  (port,
			CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_DATA, value);
        break;
	default:
		/* should not happen */
		pcie_errorl("Invalid size %d", size);
		break;
	}
}

static void bsp_pcie_cfg_write_hwinit(u32 port, u32 size, u32 offset, u32 value)
{
	u32 reg_val;

	/* hwinit writing is always internal type */

	reg_val = CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_CFGTYPE0
		| CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_INTERNAL
		| CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_ISSUE_EN
		| CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_CFGHWINIT_EN
		| (0 << CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_FNCNUM_SHIFT)
		| (0 << CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_DEVNUM_SHIFT)
		| (0 << CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_BUSNUM_SHIFT)
		| (offset & CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_EXREG_MASK);

	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR, reg_val);

	switch (size) {
	case 1:
		PCIE_REG_WRITE8 (port,
			CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_DATA+(offset&0x03),
			value);
		break;
	case 2:
		PCIE_REG_WRITE16(port,
			CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_DATA+(offset&0x02),
			value);
		break;
	case 4:
		PCIE_REG_WRITE  (port,
			CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_DATA, value);
        break;
	default:
		/* should not happen */
		pcie_errorl("Invalid size %d", size);
		break;
	}
}

/*------------------------------------------------------------------------------
 * Wrapper for internal read/write/modify.
 */
inline static u32 bsp_pcie_cfg_read_in(u32 port, u32 size, u32 offset)
{
	return bsp_pcie_cfg_read(port,1,0,0,0,size,offset);
}

inline static void bsp_pcie_cfg_write_in(u32 port, u32 size, u32 offset, u32 value)
{
	bsp_pcie_cfg_write(port,1,0,0,0,size,offset,value);
}

inline static void bsp_pcie_cfg_modify_in(u32 port, u32 size, u32 offset, u32 value, u32 mask)
{
	u32 reg_val = bsp_pcie_cfg_read_in(port, size, offset);
	reg_val = (reg_val & ~mask) | (value & mask);
	bsp_pcie_cfg_write_in(port, size, offset, reg_val);
}

/*------------------------------------------------------------------------------
 * Wrapper for external read/write/modify.
 */
inline static u32 bsp_pcie_cfg_read_ex(u32 port, u32 size, u32 offset)
{
	return bsp_pcie_cfg_read(port,0,0,PORT_TO_PCISLOT(port),0,size,offset);
}

inline static void bsp_pcie_cfg_write_ex(u32 port, u32 size, u32 offset, u32 value)
{
	bsp_pcie_cfg_write(port,0,0,PORT_TO_PCISLOT(port),0,size,offset,value);
}

inline static void bsp_pcie_cfg_modify_ex(u32 port, u32 size, u32 offset, u32 value, u32 mask)
{
	u32 reg_val = bsp_pcie_cfg_read_ex(port, size, offset);
	reg_val = (reg_val & ~mask) | (value & mask);
	bsp_pcie_cfg_write_ex(port, size, offset, reg_val);
}

/*------------------------------------------------------------------------------
 * Read device config space on pcie bus.
 * Return 0 on success, otherwise on failure.
 */
static int cxd9225gg_pcie_read_config(struct pci_bus *bus, unsigned int devfn,
                                        int offset, int size, u32 *val)
{
	struct cxd9225gg_pcie_port * pp = NULL;
	unsigned long flags;
	u32 tmp;

	/* do param check */
	if (!bus || (devfn > 255) || (offset>=4096) || (offset<0)
		|| (size != 1 && size != 2 && size != 4) || (!val)) {
		pcie_errorl("Bad param !");
		return -1;
	}

	/* we do not support sub pci bus */
	if (bus->parent) {return -1;}

	pp = bsp_pcie_busnr_to_port(bus->number);
	if (!pp) {
		pcie_errorl("No PCIE port on bus number %d", bus->number);
		return -1;
	}

	if (PORT_TO_PCISLOT(pp->port) != PCI_SLOT(devfn)) {
		return -1;
	}

	if (!pp->linkup) {return -1;}

	spin_lock_irqsave(&pp->conf_lock, flags);
	tmp = bsp_pcie_cfg_read_ex(pp->port, size, offset);

	spin_unlock_irqrestore(&pp->conf_lock, flags);


	/* we do not change value in val, unless we read successfully. */
	*val = tmp;

	return 0;
}

/*------------------------------------------------------------------------------
 * Write device config space on pcie bus.
 * Return 0 on success, otherwise on failure.
 */
static int cxd9225gg_pcie_write_config(struct pci_bus *bus, unsigned int devfn,
                                        int offset, int size, u32 val)
{
	struct cxd9225gg_pcie_port * pp = NULL;
	unsigned long flags;

	/* do param check */
	if (!bus || (devfn > 255) || (offset>=4096) || (offset<0)
		|| (size != 1 && size != 2 && size != 4)) {
		pcie_errorl("Bad param !");
		return -1;
	}

	/* we do not support sub pci bus */
	if (bus->parent) {return -1;}

	pp = bsp_pcie_busnr_to_port(bus->number);
	if (!pp) {
		pcie_errorl("No PCIE port on bus number %d", bus->number);
		return -1;
	}

	if (PORT_TO_PCISLOT(pp->port) != PCI_SLOT(devfn)) {
		return -1;
	}

	if (!pp->linkup) {return -1;}

	spin_lock_irqsave(&pp->conf_lock, flags);
	bsp_pcie_cfg_write_ex(pp->port, size, offset, val);

	spin_unlock_irqrestore(&pp->conf_lock, flags);

	return 0;
}

/*------------------------------------------------------------------------------
 * Modify device config space on pcie bus, with a mask.
 * Return 0 on success, otherwise on failure.
 */
static __used int cxd9225gg_pcie_modify_config(struct pci_bus *bus, unsigned int devfn,
									int offset, int size, u32 val, u32 mask)
{
	u32 reg_val;
	if (cxd9225gg_pcie_read_config(bus, devfn, offset, size, &reg_val)) {
		/* return failed value means no device here */
		return -1;
	}
	reg_val = (reg_val & ~mask) | (val | mask);
	return cxd9225gg_pcie_write_config(bus, devfn, offset, size, reg_val);
}

/*------------------------------------------------------------------------------
 * Set reset state, reset=1 means to reset the hardware.
 */
static void bsp_pcie_port_set_reset_state(int port, int reset)
{
	u32 val;

	if(port == 2)
		val = readl(CXD9225GG_SYSC_HWRST2_VIRT);
	else
		val = readl(CXD9225GG_SYSC_HWRST1_VIRT);

	if(reset)
		switch(port) {
		case 0:
			/* PCIEx1L0 reset */
			val |= CXD9225GG_SYSC_HWRST1_PCIEX1L0;
			writel(val, CXD9225GG_SYSC_HWRST1_VIRT);
			break;
		case 1:
			/* PCIEx1L1 reset */
			val |= CXD9225GG_SYSC_HWRST1_PCIEX1L1;
			writel(val, CXD9225GG_SYSC_HWRST1_VIRT);
			break;
		case 2:
			/* PCIEx4L reset */
			val |= CXD9225GG_SYSC_HWRST2_PCIEX4L;
			writel(val, CXD9225GG_SYSC_HWRST2_VIRT);
		}
	else
		switch(port) {
		case 0:
			/* PCIEx1L0 normal */
			val &= ~CXD9225GG_SYSC_HWRST1_PCIEX1L0;
			writel(val, CXD9225GG_SYSC_HWRST1_VIRT);
			break;
		case 1:
			/* PCIEx1L1 normal */
			val &= ~CXD9225GG_SYSC_HWRST1_PCIEX1L1;
			writel(val, CXD9225GG_SYSC_HWRST1_VIRT);
			break;
		case 2:
			/* PCIEx4L normal */
			val &= ~CXD9225GG_SYSC_HWRST2_PCIEX4L;
			writel(val, CXD9225GG_SYSC_HWRST2_VIRT);
		}
}

/*------------------------------------------------------------------------------
 * Set clock gate, gate_on=1 means to set clock gate on.
 * Clock Gate On means do not assign clock resource to the PCIE HC.
 */
static void bsp_pcie_port_set_clock_gate(int port, int gate_on)
{
	u32 val;

	if(port == 2)
		val = readl(CXD9225GG_SYSC_GCTS2_VIRT);
	else
		val = readl(CXD9225GG_SYSC_GCTS1_VIRT);

	if(gate_on)
		switch(port) {
		case 0:
			/* PCIEx1L0 clock gate on */
			val |= CXD9225GG_SYSC_CGE_PCIEX1L0;
			writel(val, CXD9225GG_SYSC_GCTS1_VIRT);
			break;
		case 1:
			/* PCIEx1L1 clock gate on */
			val |= CXD9225GG_SYSC_CGE_PCIEX1L1;
			writel(val, CXD9225GG_SYSC_GCTS1_VIRT);
			break;
		case 2:
			/* PCIEx4L clock gate on */
			val |= CXD9225GG_SYSC_CGE_PCIEX4L;
			writel(val, CXD9225GG_SYSC_GCTS2_VIRT);
		}
	else
		switch(port) {
		case 0:
			/* PCIEx1L0 clock gate off */
			val &= ~CXD9225GG_SYSC_CGE_PCIEX1L0;
			writel(val, CXD9225GG_SYSC_GCTS1_VIRT);
			break;
		case 1:
			/* PCIEx1L1 clock gate off */
			val &= ~CXD9225GG_SYSC_CGE_PCIEX1L1;
			writel(val, CXD9225GG_SYSC_GCTS1_VIRT);
			break;
		case 2:
			/* PCIEx4L clock gate off */
			val &= ~CXD9225GG_SYSC_CGE_PCIEX4L;
			writel(val, CXD9225GG_SYSC_GCTS2_VIRT);
		}
}

/*------------------------------------------------------------------------------
 * Shutdown the clock for PCIE bus by setting the clock gate on.
 * The main purpose of doing this is to save power.
 */
void bsp_pcie_port_gated_clock_on(int port)
{
	bsp_pcie_port_set_reset_state(port, 1);
	bsp_pcie_port_set_clock_gate(port, 1);
}
EXPORT_SYMBOL(bsp_pcie_port_gated_clock_on);

/*------------------------------------------------------------------------------
 * Stop gating the bus clock so that PCIE HC is enabled.
 */
void bsp_pcie_port_gated_clock_off(int port)
{
	bsp_pcie_port_set_clock_gate(port, 0);
	bsp_pcie_port_set_reset_state(port, 0);
}
EXPORT_SYMBOL(bsp_pcie_port_gated_clock_off);

static int g_port_breath_ms = 10;
static int g_max_lnkup_loop = 10;
static unsigned int g_probe_cnt = 0;
static unsigned int g_failure_cnt = 0;

/*------------------------------------------------------------------------------
 * This function reset the hardware by writing SYSC registers.
 */
static void bsp_pcie_port_reset_hw(int port)
{
	bsp_pcie_port_gated_clock_on(port);

	bsp_pcie_port_gated_clock_off(port);

	/* Give a breath to ensure the power state and clock resource is
	 * stable before we really accessing the registers of the PCIE HC.
	 * It is especially important for saturn board. */
	if (g_port_breath_ms > 0) mdelay(g_port_breath_ms);

	/* PCIe core reset assert */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_RESET, CXD9225GG_PCIEX_RESET_RST_DEFAULT);

	/* Mode Root Complex */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_MODE_INTERFACE,
					CXD9225GG_PCIEX_MODE_INTERFACE_MODE_PORT_RC);

	/* PCIe core reset deassert */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_RESET,
				(CXD9225GG_PCIEX_RESET_RST_B_UNRESET
				| CXD9225GG_PCIEX_RESET_RST_GP_B_UNRESET
				| CXD9225GG_PCIEX_RESET_RST_RSM_B_UNRESET
				| CXD9225GG_PCIEX_RESET_RST_CFG_B_UNRESET
				| CXD9225GG_PCIEX_RESET_RST_LOAD_B_UNRESET
				| CXD9225GG_PCIEX_RESET_RST_PS_B_UNRESET
				| CXD9225GG_PCIEX_RESET_RST_OUT_B_UNRESET));
}

static void bsp_pcie_port_reset_core(int port)
{
	unsigned int pciexRegVal;

	/* Disable PCIE INTA */
	PCIE_REG_READ (port, CXD9225GG_PCIEX_PCI_INTX_RCV_INT_ENABLE, pciexRegVal);
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_PCI_INTX_RCV_INT_ENABLE,
		(pciexRegVal & ~CXD9225GG_PCIEX_PCI_INTX_RCV_INT_ENABLE_INTA_EN));

	/* BAR0 (Mem space, Bus master Disable) */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR,
		CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_CFGTYPE0
		| CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_INTERNAL
		| CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_ISSUE_EN
		| (PCIE_FNC_NO_FIXED << CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_FNCNUM_SHIFT)
		| (PCIE_DEV_NO_AXIBR << CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_DEVNUM_SHIFT)
		| (PCIE_BUS_NO_FIXED << CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_ADR_BUSNUM_SHIFT)
		| CXD9225GG_PCIECFG_RC_CMD_AND_STS
		);
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_PCIE_CONFIG_ISSUE_DATA,
		CXD9225GG_PCIECFG_RC_COMMAND_MEM_SPACE_DIS
		| CXD9225GG_PCIECFG_RC_COMMAND_BUS_MASTER_DIS
		| CXD9225GG_PCIECFG_RC_COMMAND_PARITY_ERR_RES_DIS
		| CXD9225GG_PCIECFG_RC_COMMAND_SERR_DIS
		| CXD9225GG_PCIECFG_RC_COMMAND_INT_DIS
		);

	/* PCIe core reset assert */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_RESET, (CXD9225GG_PCIEX_RESET_RST_B_RESET
		| CXD9225GG_PCIEX_RESET_RST_GP_B_RESET
		| CXD9225GG_PCIEX_RESET_RST_RSM_B_RESET
		| CXD9225GG_PCIEX_RESET_RST_CFG_B_RESET
		| CXD9225GG_PCIEX_RESET_RST_LOAD_B_RESET
		| CXD9225GG_PCIEX_RESET_RST_PS_B_RESET
		| CXD9225GG_PCIEX_RESET_RST_OUT_B_RESET)
		);
}

/*------------------------------------------------------------------------------
 * Check if current state is link-down. Return 0 if it is.
 */
static int bsp_pcie_port_lnkdn_check(int port)
{
	u32 reg_val;

	/* LinkDown status clear */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_PCIE_ERRINT_STATUS,
					CXD9225GG_PCIEX_PCIE_ERRINT_STS_LNKDN);

	/* LinkDown check */
	PCIE_REG_READ (port, CXD9225GG_PCIEX_PCIE_CORE_CONTROL_STATUS, reg_val);

	if ((reg_val & CXD9225GG_PCIEX_PCIE_CORE_CTL_STS_LINKDN) == 0) {
		pcie_critl("REG[0x400]=0x%08x", reg_val);
		return -1;
	}

	return 0;
}

/*------------------------------------------------------------------------------
 * Check if current state is link-up. Return 0 if it is.
 */
static int bsp_pcie_port_lnkup_check(int port)
{
	u32 loop, reg_val, core_status;

	/* LinkDown interrupt enable */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_PCIE_ERRINT_ENABLE,
				CXD9225GG_PCIEX_PCIE_ERRINT_ENA_LNKDN_EN);

	/* wait for linking up */
	loop = 0;
	do {
		PCIE_REG_READ( port, CXD9225GG_PCIEX_INT_TABLE, reg_val);
		if (reg_val & CXD9225GG_PCIEX_INT_TABLE_PCIE_ERR_INT) {
			break;
		}
		mdelay(10);
		loop++;
	}
	while (loop < g_max_lnkup_loop);

	PCIE_REG_READ (port, CXD9225GG_PCIEX_PCIE_CORE_CONTROL_STATUS, core_status);
	pcie_critl("REG[0x%03x]=0x%08x, REG[0x%03x]=0x%08x, loop=%d",
			CXD9225GG_PCIEX_INT_TABLE, reg_val,
			CXD9225GG_PCIEX_PCIE_CORE_CONTROL_STATUS, core_status, loop);

	/* LinkDown interrupt disable */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_PCIE_ERRINT_ENABLE, 0);

	if (loop >= g_max_lnkup_loop) {
		return -1;
	}

	return 0;
}

/*------------------------------------------------------------------------------
 * Set the AXI-PCIE Bridge Windows.
 * If size is set to 0, then the bridge window is disabled.
 */
static int bsp_pcie_set_pcie_bridge_window(int port, int win, u32 base, u32 dest, u32 size)
{
	u32 addr_base=0, addr_mask=0, addr_dest_low=0, addr_dest_high=0;

	switch (win) {
	case 0:
		addr_base      = CXD9225GG_PCIEX_PCIE_WINDOW0_BASE;
		addr_mask      = CXD9225GG_PCIEX_PCIE_WINDOW0_MASK;
		addr_dest_low  = CXD9225GG_PCIEX_PCIE_DESTINATION_LOW0;
		addr_dest_high = CXD9225GG_PCIEX_PCIE_DESTINATION_UP0;
		break;
	case 1:
		addr_base      = CXD9225GG_PCIEX_PCIE_WINDOW1_BASE;
		addr_mask      = CXD9225GG_PCIEX_PCIE_WINDOW1_MASK;
		addr_dest_low  = CXD9225GG_PCIEX_PCIE_DESTINATION_LOW1;
		addr_dest_high = CXD9225GG_PCIEX_PCIE_DESTINATION_UP1;
		break;
	case 2:
		addr_base      = CXD9225GG_PCIEX_PCIE_WINDOW2_BASE;
		addr_mask      = CXD9225GG_PCIEX_PCIE_WINDOW2_MASK;
		addr_dest_low  = CXD9225GG_PCIEX_PCIE_DESTINATION_LOW2;
		addr_dest_high = CXD9225GG_PCIEX_PCIE_DESTINATION_UP2;
		break;
	default:
		pcie_errorl("Invalid PCIE Bridge Window Number! (%d)", win);
		panic("PCIE HC on CXD9225GG only has 3 Bridge Windows!");
		return -1;
	}

	if (size != 0) {
		base = ((base & CXD9225GG_PCIEX_PCIE_WINDOW_BASE_PWBASE_MASK)
			| CXD9225GG_PCIEX_PCIE_WINDOW_BASE_NORMAL_READ
			| CXD9225GG_PCIEX_PCIE_WINDOW_BASE_PCIE_WND_EN);
		size = ((size-1) & CXD9225GG_PCIEX_PCIE_WINDOW_MASK_PWMASK_MASK);
		dest = (dest & CXD9225GG_PCIEX_PCIE_DESTINATION_LOW_PDEST_MASK);

		PCIE_REG_WRITE (port, addr_base, base);
		PCIE_REG_WRITE (port, addr_mask, size);
		PCIE_REG_WRITE (port, addr_dest_low, dest);
		PCIE_REG_WRITE (port, addr_dest_high, 0);

		base &= CXD9225GG_PCIEX_PCIE_WINDOW_BASE_PWBASE_MASK;
		size += 1+(~CXD9225GG_PCIEX_PCIE_WINDOW_MASK_PWMASK_MASK);
		pcie_debugl("AXI->PCIE: Port%d, Base=0x%08x, Dest=0x%08x, Size=0x%08x",
				port, base, dest, size);

	} else {
		// just disable the bridge window.
		PCIE_REG_WRITE (port, addr_base,
				CXD9225GG_PCIEX_PCIE_WINDOW_BASE_PCIE_WND_DIS);
	}
	return 0;
}


/*------------------------------------------------------------------------------
 * Set the PCIE->AXI Bridge Windows.
 * If size is set to 0, then the bridge window is disabled.
 */
static int bsp_pcie_set_axi_bridge_window(int port, int win, u32 base, u32 dest, u32 size)
{

	u32 addr_base=0, addr_mask=0, addr_dest=0;

	switch (win) {
	case 0:
		addr_base = CXD9225GG_PCIEX_AXI_WINDOW0_BASE;
		addr_mask = CXD9225GG_PCIEX_AXI_WINDOW0_MASK;
		addr_dest = CXD9225GG_PCIEX_AXI_DESTINATION0;
		break;
	case 1:
		addr_base = CXD9225GG_PCIEX_AXI_WINDOW1_BASE;
		addr_mask = CXD9225GG_PCIEX_AXI_WINDOW1_MASK;
		addr_dest = CXD9225GG_PCIEX_AXI_DESTINATION1;
		break;
	case 2:
		addr_base = CXD9225GG_PCIEX_AXI_WINDOW2_BASE;
		addr_mask = CXD9225GG_PCIEX_AXI_WINDOW2_MASK;
		addr_dest = CXD9225GG_PCIEX_AXI_DESTINATION2;
		break;
	case 3:
		addr_base = CXD9225GG_PCIEX_AXI_WINDOW3_BASE;
		addr_mask = CXD9225GG_PCIEX_AXI_WINDOW3_MASK;
		addr_dest = CXD9225GG_PCIEX_AXI_DESTINATION3;
		break;
	default:
		pcie_errorl("Invalid AXI Bridge Window Number! (%d)", win);
		panic("PCIE HC on CXD9225GG only has 4 Bridge Windows!");
		return -1;
	}

	if (size != 0) {
		base = ((base & CXD9225GG_PCIEX_AXI_WINDOW_BASE_AWBASE_MASK)
			| CXD9225GG_PCIEX_AXI_WINDOW_BASE_AXI_MSTR_SZ64
			| CXD9225GG_PCIEX_AXI_WINDOW_BASE_AXI_WND_EN);
		size = ((size-1) & CXD9225GG_PCIEX_AXI_WINDOW_MASK_AWMASK_MASK);
		dest = (dest & CXD9225GG_PCIEX_AXI_DESTINATION_ADEST_MASK);

		PCIE_REG_WRITE (port, addr_base, base);
		PCIE_REG_WRITE (port, addr_mask, size);
		PCIE_REG_WRITE (port, addr_dest, dest);

		base &= CXD9225GG_PCIEX_AXI_WINDOW_BASE_AWBASE_MASK;
		size += 1+(~CXD9225GG_PCIEX_AXI_WINDOW_MASK_AWMASK_MASK);
		pcie_debugl("PCIE->AXI: Port%d, Base=0x%08x, Dest=0x%08x, Size=0x%08x",
				port, base, dest, size);

	} else {
		// just disable the bridge window.
		PCIE_REG_WRITE (port, addr_base,
				CXD9225GG_PCIEX_AXI_WINDOW_BASE_AXI_WND_DIS);
	}
	return 0;
}

/*------------------------------------------------------------------------------
 * Fixup the AXI Bridge Registers so that the PCIE Host Controller works.
 */
static void bsp_pcie_port_fixup(int port)
{
	/* LinkDown interrupt disable */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_PCIE_ERRINT_ENABLE, 0);

	/* LinkDown status clear */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_PCIE_ERRINT_STATUS,
				CXD9225GG_PCIEX_PCIE_ERRINT_STS_LNKDN);

	/* Memory Base / Limit */
	bsp_pcie_cfg_write_in(port, 4, CXD9225GG_PCIECFG_RC_MEMBASE_LIMIT,
			((CXD9225GG_PCIE_AXI_MEM_LIMIT
			>> CXD9225GG_PCIECFG_RC_MEMLIMIT_SHIFT)
			| (CXD9225GG_PCIE_AXI_MEM_BASE
			>> CXD9225GG_PCIECFG_RC_MEMBASE_SHIFT)));

	/* BAR0 (Base Address) */
	bsp_pcie_cfg_write_in(port, 4, CXD9225GG_PCIECFG_RC_BASE_ADDRESS_0,
			CXD9225GG_PCIE_AXI_BAR0_BASE);

	/* BAR0 (Base Address Mask)  H/W type reg */
	bsp_pcie_cfg_write_hwinit(port, 4,
			CXD9225GG_PCIECFG_RC_BASE_ADR_MASK0_L,
			CXD9225GG_PCIE_AXI_BAR0_MASK);

	/* BAR0 (Mem space, Bus master Enable) */
	bsp_pcie_cfg_write_in(port, 4, CXD9225GG_PCIECFG_RC_CMD_AND_STS,
			(CXD9225GG_PCIECFG_RC_COMMAND_MEM_SPACE_EN
			| CXD9225GG_PCIECFG_RC_COMMAND_BUS_MASTER_EN
			| CXD9225GG_PCIECFG_RC_COMMAND_PARITY_ERR_RES_EN
			| CXD9225GG_PCIECFG_RC_COMMAND_SERR_EN
			| CXD9225GG_PCIECFG_RC_COMMAND_INT_EN));

	if (port == 2) {
		/* BAR0 (Mem space, IO space, Bus master Enable) */
		bsp_pcie_cfg_write_ex(port, 4, CXD9225GG_PCIECFG_RC_CMD_AND_STS,
				      (CXD9225GG_PCIECFG_RC_COMMAND_IO_SPACE_EN
				       | CXD9225GG_PCIECFG_RC_COMMAND_MEM_SPACE_EN
				       | CXD9225GG_PCIECFG_RC_COMMAND_BUS_MASTER_EN));
	}


	/* Disable all AXI bridge windows at first */
	bsp_pcie_set_axi_bridge_window(port, 0, 0, 0, 0);
	bsp_pcie_set_axi_bridge_window(port, 1, 0, 0, 0);
	bsp_pcie_set_axi_bridge_window(port, 2, 0, 0, 0);
	bsp_pcie_set_axi_bridge_window(port, 3, 0, 0, 0);

	/* Disable all PCIE bridge windows at first */
	bsp_pcie_set_pcie_bridge_window(port, 0, 0, 0, 0);
	bsp_pcie_set_pcie_bridge_window(port, 1, 0, 0, 0);
	bsp_pcie_set_pcie_bridge_window(port, 2, 0, 0, 0);

	switch(port) {
	case 0:
		bsp_pcie_set_pcie_bridge_window(port, 0,
				CXD9225GG_AXI_PCIE_PWBASE0,
				CXD9225GG_AXI_PCIE_PWDEST0,
				CXD9225GG_PCIEX1L0_MEM_SPACE_SIZE);
		bsp_pcie_set_axi_bridge_window(port, 0,
				CXD9225GG_PCIE_AXI_BAR0_AWBASE0,
				CXD9225GG_SDRAM_REAL_BASE,
				CXD9225GG_SDRAM_SIZE);
		break;
	case 1:
		bsp_pcie_set_pcie_bridge_window(port, 0,
				CXD9225GG_AXI_PCIE_PWBASE1,
				CXD9225GG_AXI_PCIE_PWDEST1,
				CXD9225GG_PCIEX1L1_MEM_SPACE_SIZE);
		bsp_pcie_set_axi_bridge_window(port, 0,
				CXD9225GG_PCIE_AXI_BAR0_AWBASE0,
				CXD9225GG_SDRAM_REAL_BASE,
				CXD9225GG_SDRAM_SIZE);
		break;
	case 2:
		bsp_pcie_set_pcie_bridge_window(port, 0,
				CXD9225GG_AXI_PCIE_PWBASE2,
				CXD9225GG_AXI_PCIE_PWDEST2,
				CXD9225GG_PCIEX4L_MEM_SPACE_SIZE);
		bsp_pcie_set_axi_bridge_window(port, 0,
				CXD9225GG_PCIE_AXI_BAR0_AWBASE0,
				CXD9225GG_SDRAM_REAL_BASE,
				CXD9225GG_SDRAM_SIZE);
	}

	/* Device capabilities 2 reg */
	bsp_pcie_cfg_write_hwinit(port, 4,
			CXD9225GG_PCIECFG_RC_DEVICE_CAPABILITY2,
			CXD9225GG_PCIECFG_RC_DEVICE_CAPA2_CMPTORANG_A);

	/* Device control 2 reg */
	bsp_pcie_cfg_write_in(port, 4,
			CXD9225GG_PCIECFG_RC_DEVICE_CTL_STS_2,
			CXD9225GG_PCIECFG_RC_DEVICE_CTL2_CMPTORANG_B);

	if (port == 2) {
		/* Set the registers as ASMEDIA requrired */
		/* Need to set before spin-up */
		bsp_pcie_cfg_modify_ex(port, 1, 0x0CA4, 0xAB, 0xFF);
		bsp_pcie_cfg_modify_ex(port, 1, 0x0CA5, 0x26, 0xFF);
		bsp_pcie_cfg_modify_ex(port, 1, 0x0CAE, 0x12, 0xFF);
		bsp_pcie_cfg_modify_ex(port, 1, 0x0DA4, 0xAB, 0xFF);
		bsp_pcie_cfg_modify_ex(port, 1, 0x0DA5, 0x26, 0xFF);
		bsp_pcie_cfg_modify_ex(port, 1, 0x0DAE, 0x12, 0xFF);
		bsp_pcie_cfg_modify_ex(port, 1, 0x0A01, 0x0, 0x3);
		bsp_pcie_cfg_modify_ex(port, 1, 0x0A12, 0x18, 0x18);

		/* Set spin-up device control bit to enable SSD */
		bsp_pcie_cfg_modify_ex(port, 1, CXD9225GG_PCIECFG_SPIN_UP_DEVICE, 0x3, 0x3);
	}

	/* PCIe GPIO GPE default setting */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_GENERAL_PURPOSE_ENABLE0,
			RB_CXD9225GG_PCIE_GPE0_DEFAULT);
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_GENERAL_PURPOSE_ENABLE1,
			RB_CXD9225GG_PCIE_GPE1_DEFAULT);
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_GENERAL_PURPOSE_ENABLE2,
			RB_CXD9225GG_PCIE_GPE2_DEFAULT);
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_GENERAL_PURPOSE_ENABLE3,
			RB_CXD9225GG_PCIE_GPE3_DEFAULT);

	/*
	 * Further init :
	 *   The following codes are in pciehConfigAttach().
	 *   We move them here because they will not mix the init procedure.
	 */

	/* Interrupt Pending reg clear */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_PCI_INTX_RCV_INT_STATUS, 0xf);

	/* Interrupt LinkDown Enable */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_PCIE_ERRINT_ENABLE,
			CXD9225GG_PCIEX_PCIE_ERRINT_ENA_LNKDN_EN
			|CXD9225GG_PCIEX_PCIE_ERRINT_ENA_CA_EN);

	/* Error register clear */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_AXI_MASTER_ERRINT_ENABLE, 0);

	/*
	 * Always Open the Receiving of INTA.
	 * Use CPU intr controller mask to enable/disable the irq.
	 * Namely, use enable_irq() and diable_irq().
	 */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_PCI_INTX_RCV_INT_ENABLE, 0x00000001);

	/* Open the err intr */
	PCIE_REG_WRITE (port, CXD9225GG_PCIEX_AXI_MASTER_ERRINT_ENABLE, 0x101);
}

/*------------------------------------------------------------------------------
 * Initialize the PCIE port.
 * Return 0 or 1 in success case.
 * 1 - linkup done.
 * 0 - linkup ignored.
 * less than 0 - error.
 */
#if 0
int bsp_pcie_port_init(int port, int is_pow_on)
{
	int result = 0;

	if (port < 0 || port > NUM_PCIE_PORTS) {
		pcie_errorl("Invalid port %d", port);
		return -1;
	}

	(void)bsp_pcie_port_reset_hw(port);

	if (is_pow_on) {
		g_probe_cnt++;

		if (bsp_pcie_port_lnkdn_check(port)) {
			result = -2;
		} else
		/* If link-up failed, (10 x g_max_lnkup_loop)ms would be taken. */
		if (bsp_pcie_port_lnkup_check(port)) {
			result = -3;
		} else {
			result = 1;
		}

		if (result > 0) {
			pcie_critl("Port %d Link-Up successfully !! [Failure=%u/%u]",
				port, g_failure_cnt, g_probe_cnt);
		} else {
			g_failure_cnt++;
			pcie_critl("Port %d Link-Up failed !! [Failure=%u/%u]",
					port, g_failure_cnt, g_probe_cnt);
		}
	}

	if (is_pow_on && (result>0))
		bsp_pcie_port_fixup(port);

	/* if no card is plugged in, then gate the clock to save power. */
	if (!is_pow_on) {
		bsp_pcie_port_gated_clock_on(port);
	}

	return result;
}
#else
int bsp_pcie_port_init(int port, int is_pow_on)
{
	int result = 0;

	if (port < 0 || port > NUM_PCIE_PORTS) {
		pcie_errorl("Invalid Port %d", port);
		return -1;
	}

	(void)bsp_pcie_port_reset_hw(port);

	if (is_pow_on) {
		int port_breath_ms = g_port_breath_ms;
		int max_lnkup_loop = g_max_lnkup_loop;
		int i_try = 3;

		g_probe_cnt++;

		do {
			if (bsp_pcie_port_lnkdn_check(port)) {
				result = -2; /* LinkDown failed */
			} else
			/* If link-up failed, (10 x g_max_lnkup_loop)ms would be taken. */
			if (bsp_pcie_port_lnkup_check(port)) {
				result = -3; /* LinkUp failed */
			} else {
				result = 1; /* OK and jump out */
				break;
			}

			/*
			 * If link-up failed, we enlarge port_breath_ms and max_lnkup_loop
			 * a little bit, then set power off->on, clock-gate on->off,
			 * and do link-down and link-up again for 2 more times.
			*/
			g_port_breath_ms += 10; /* more 10ms */
			g_max_lnkup_loop += 4;  /* more 40ms */

			if (!bsp_pcie_slot_check(port)) {
				pcie_errorl("No Card is found in Slot%c.", SLOT_X(port));
				result = -1; /* Device Connection Error */
				break;
			}
			bsp_pcie_slot_set_power(port, 0); /* Power Off */
			mdelay(10);
			bsp_pcie_slot_set_power(port, 1); /* Power On */
			if (!bsp_pcie_slot_get_power(port)) {
				pcie_errorl("Failed to power up the PCIE Slot%c !!", SLOT_X(port));
				result = -1; /* Device Connection Error */
				break;
			}

			(void)bsp_pcie_port_reset_hw(port);
		} while (--i_try);

		/* restore the default value. */
		g_port_breath_ms = port_breath_ms;
		g_max_lnkup_loop = max_lnkup_loop;

		if (result > 0) {
			pcie_critl("Port %d Link-Up successfully !! [Failure=%u/%u]",
				port, g_failure_cnt, g_probe_cnt);
		} else {
			g_failure_cnt++;
			pcie_critl("Port %d Link-Up failed !! [Failure=%u/%u]",
					port, g_failure_cnt, g_probe_cnt);
		}
	}

	if (is_pow_on && result>0) {
		bsp_pcie_port_fixup(port);
	}

	/* if no card is plugged in, then gate the clock to save power. */
	if (!is_pow_on) {
		bsp_pcie_port_gated_clock_on(port);
	}

	return result;
}
#endif

/*------------------------------------------------------------------------------
 * Function entries to read and write device config space on pcie bus.
 * According to the PCI subsys framework for ARM, this structure is needed by
 * func pci_scan_bus() which is called in func cxd9225gg_pcie_scan().
 */
static struct pci_ops cxd9225gg_pciehc_ops = {
	.read = cxd9225gg_pcie_read_config,
	.write = cxd9225gg_pcie_write_config,
};

/*------------------------------------------------------------------------------
 * Params:
 *	nr			: index of this host controller
 *	sys			: system specific data
 *
 * Return:
 *	According to PCI subsys framework,
 *	return less than 0 means to stop the whole bus subsys setup procedure.
 *	return 0 means to stop setup procedure of this HC, if there are
 *		more than 1 controllers, other ones can still get setup.
 *	return greater than 0 means this controller setup is successfully done.
 *	Typically, the return values may be 1, 0, or -1.
 */
static int cxd9225gg_pciehc_setup(int port, struct pci_sys_data * sys)
{
	struct cxd9225gg_pcie_port * pp;
	static int s_first = 1;
	int is_on = 0;
	int is_linkup = 0;

	pcie_debugl("start.");

	/* do param check */
	if (!sys || (port < 0 || port >= NUM_PCIE_PORTS)) {
		pcie_errorl("Bad param! port=%d, sys=0x%08x", port, (u32)sys);
		return 0;
	}
#ifdef CONFIG_SNSC_DEFERRED_INITCALLS_CXD9225GG_PCIE
	if ((port < defd_init_pcie_num) &&
	    (port == 0 || port == 1 || port == 2)) {
#else
	if (port == 0 || port == 1 || port == 2) {
#endif
		/* if ExpressCard is in the slot, power it up */
		if (bsp_pcie_slot_check(port)) {
			bsp_pcie_slot_set_power(port, 1);

			is_on = bsp_pcie_slot_get_power(port);
			if (!is_on) {
				pcie_errorl("Failed to power up the PCIE port %d !!", port);
				return 0;
			}

			pcie_warningl("Found ExpressCard in port %d", port);
		} else {
			pcie_warningl("No ExpressCard in port %d", port);
		}
	} else
	{
		pcie_errorl("PCIE port %d is not supported now!!", port);
		return 0;
	}

	/* init the pcie port */
	is_linkup = bsp_pcie_port_init(port, is_on);
	if (is_linkup < 0) {
		pcie_errorl("Failed to init the PCIE port %d !!", port);
		is_linkup=0;
		//return 0;
	}

	/* add s_busnr_start because there're already some pci buses.
	 * but only the first port needs this, the following ports count on
	 * the first one. */
	if (s_first) {
		s_first = 0;
		sys->busnr += s_busnr_start;
	}

	/* init private data structure [begin] */
	pp = &s_pcie_ports[port];
	pp->port = port;
	pp->root_bus_nr = sys->busnr;
	pp->linkup = is_linkup;

	/* setup IORESOURCE_MEM */
	snprintf(pp->memsp_name, sizeof(pp->memsp_name), "PCIe %d MEM", pp->port);
	pp->memsp_name[sizeof(pp->memsp_name) - 1] = 0;

	/* offset: start of the bus resource to the start of the cpu mem space */
	switch(port) {
	case 0:
		pp->res[0].name = pp->memsp_name;
		pp->res[0].start = CXD9225GG_PCIE_1X_0_PHYS_BASE;
		pp->res[0].end = pp->res[0].start + SZ_4K - 1;
		pp->res[0].flags = IORESOURCE_MEM;
		break;
	case 1:
		pp->res[0].name = pp->memsp_name;
		pp->res[0].start = CXD9225GG_PCIE_1X_1_PHYS_BASE;
		pp->res[0].end = pp->res[0].start + SZ_4K - 1;
		pp->res[0].flags = IORESOURCE_MEM;
		break;
	case 2:
		pp->res[0].name = pp->memsp_name;
		pp->res[0].start = CXD9225GG_PCIE_4X_PHYS_BASE +
			CXD9225GG_PCIE_4X_SIZE;
		pp->res[0].end = pp->res[0].start + SZ_4K - 1;
		pp->res[0].flags = IORESOURCE_MEM;
	}

	if (request_resource(&iomem_resource, &pp->res[0])) {
		panic("[PCIE] Request PCIe Memory resource failed !!\n");
	}
	/* init private data structure [end] */

	/* init sys data */
	sys->mem_offset = pp->res[0].start;
	sys->resource[0] = &pp->res[0];
	if (port == 2) {
		/* Setup dummy IORESOURCE_IO as sys->resource[1] */
		snprintf(pp->iosp_name, sizeof(pp->iosp_name), "PCIe %d IO", pp->port);
		pp->iosp_name[sizeof(pp->iosp_name) - 1] = 0;
		pp->res[1].name = pp->iosp_name;
		pp->res[1].start = CXD9225GG_PCIE_4X_PHYS_BASE;
		pp->res[1].end = pp->res[1].start + CXD9225GG_PCIE_4X_SIZE - 1;
		pp->res[1].flags = IORESOURCE_IO;
		if (request_resource(&ioport_resource, &pp->res[1])) {
			panic("[PCIE] Request PCIe IO resource failed !!\n");
		}
		sys->io_offset = 0;
		sys->resource[1] = &pp->res[1];
	} else {
		sys->resource[1] = NULL;
	}
	sys->resource[2] = NULL;

	pcie_debugl("done.");

	if(is_linkup<0)
		return 0;
	else
		return 1;
}

/*------------------------------------------------------------------------------
 * Begin to scan the pcie host controller. Doing this means to create and init
 * a root bus, then recursively create and init its sub buses if it has.
 * Return Null means failed to create the root bus, so stop doing anything.
 * Otherwise, return the root bus.
 */
static struct pci_bus * cxd9225gg_pciehc_scan(int port, struct pci_sys_data * sys)
{
	/*
	port		: index number of the controller
	sys->busnr 	: index number of the bus
	If controller_0 has 5 buses, controller_1's bus number is bus_5, not bus_1.
	*/
	struct pci_bus *root_bus;
	struct pci_dev *dev;
	int irq;

	pcie_debugl("start.");

	/* do param check */
	if (!sys || (port < 0 || port >= NUM_PCIE_PORTS)) {
		pcie_errorl("Bad param! port=%d, sys=0x%08x", port, (u32)sys);
		return 0;
	}

	/*
	 * Use pci_scan_bus() is another choice, which will immediately
	 * adds the newly found devices to the PCI subsys.
	 * Since in pci_common_init() function, the pci_bus_add_devices()
	 * will be called anyway, then we have no need to the adding job
	 * right here.
	 * Therefore, we only do pci_scan_bus_parented(), not pci_scan_bus().
	 */
	root_bus = pci_scan_bus_parented(NULL, sys->busnr, &cxd9225gg_pciehc_ops, sys);

	/*
	 * We MUST fix irq here, because pci_fixup_irqs() which is called in
	 * pci_common_init() will only fixup those pci_dev already in
	 * global_list, while our pci_dev is not there right now.
	 * It won't be added into global_list until pci_bus_add_devices(),
	 * which runs at end of pci_common_init. So, if it is not done here,
	 * it will be done never.
	 */
	dev = bsp_pcie_get_first_dev(root_bus);
	if (dev) {
		pcie_infol("Device [%s] has been found !!", pci_name(dev));

		switch(port) {
		case 0: irq = INT_PCIEX0_INTA; break;
		case 1: irq = INT_PCIEX1_INTA; break;
		case 2: irq = INT_PCIEX2_INTA; break;
		default: irq = -1; break;
		}
		dev->irq = irq;
		/*
		 * Write the irq back so that it might be used by others.
		 * It's not forced, but high recommended.
		 */
		pci_write_config_byte(dev, PCI_INTERRUPT_LINE, irq);
	}

	s_pcie_ports[port].pbus = root_bus;
	s_pcie_ports[port].pdev = dev;

	pcie_debugl("done.");

	return root_bus;
}

/*------------------------------------------------------------------------------
 * Map the device to the right irq. This return value will be written back
 * to the config space of the device, and it's driver might request that irq.
 * If the return value is invalid, such as 0xffffffff, this irq will not work.
 */
static int cxd9225gg_pciehc_map_irq(struct pci_dev *dev, u8 slot, u8 pin)
{
	/* find bus nr of the device, then find the controller nr of the bus. */
	struct cxd9225gg_pcie_port * pp = NULL;
	int irq = -1;

	/* do param check */
	if (!dev) {
		pcie_errorl("Bad param! dev=0x%08x", (u32)dev);
		return 0;
	}

	pp = bsp_pcie_busnr_to_port(dev->bus->number);
	if (!pp) {
		pcie_errorl("No PCIE port on bus number %d", dev->bus->number);
		return -1;
	}

	switch(pp->port) {
	case 0: irq = INT_PCIEX0_INTA; break;
	case 1: irq = INT_PCIEX1_INTA; break;
	case 2: irq = INT_PCIEX2_INTA;
	}

	return irq;
}

/*------------------------------------------------------------------------------
 * All the function entries are called during initialization of the pcie host.
 */
static __initdata struct hw_pci cxd9225gg_pciehc_hw = {
	.nr_controllers = NUM_PCIE_PORTS,
	.setup = cxd9225gg_pciehc_setup,
	.scan = cxd9225gg_pciehc_scan,
	.map_irq = cxd9225gg_pciehc_map_irq,
};

/*------------------------------------------------------------------------------
 * This is a Data-Abort Exception handler for cxd9225gg PCIE bus.
 * > If a PCIE device is just removed when we are accessing it, a Data-Abort
 *   Exception is raised.
 * > The default Data-Abort Exception handler for this situation prints some
 *   informations and halts the task (so called, die).
 * > But we only want to know the event but not to halt the task. So here we
 *   implement our own handler, and hook it onto that exception.
 * > Our handler finally returns a ZERO to tell the do_DataAbort() function
 *   the exception is solved, do not get into the default die-way.
 */
#include <linux/exception_monitor.h>
void __weak fake_die(const char *str, struct pt_regs *regs, int err) {}
static int bsp_pcie_data_abort(unsigned long addr, unsigned int fsr,
				struct pt_regs *regs)
{
	/*
	 * we print the error logs as we are about to die,
	 * but actually we are not to die.
	*/
	extern void fake_die(const char * str,struct pt_regs * regs,int err);

	printk(KERN_ALERT "=========================="
		"========================================\n");
	printk(KERN_ALERT "Unhandled data abort: imprecise external abort "
		"(0x%03x) at 0x%08lx\n", fsr, addr);

	fake_die("", regs, fsr);

	printk(KERN_ALERT "=========================="
		"========================================\n");

	/* return 0 to tell kernel that we are good. */
	return 0;
}


#define CXD9225GG_PCIE_DEBUG
#ifdef  CXD9225GG_PCIE_DEBUG
#include <linux/proc_fs.h>
#include <linux/uaccess.h>
#define __PCIE_DBG_NODE "pcie_debug"

static int pcie_proc_read_dbg(char *page, char **start, off_t off,
			 int count, int *eof, void *data)
{
	char *p = page;
	int len;

	p += sprintf(p, "port_breath_ms: %d\n"
					"max_lnkup_loop: %d\n",
					g_port_breath_ms, g_max_lnkup_loop);

	len = (p - page) - off;
	if (len < 0)
		len = 0;

	*eof = (len <= count) ? 1 : 0;
	*start = page + off;

	return len;
}

static int pcie_proc_write_dbg(struct file *file, const char __user *user_buf,
			  unsigned long count, void *data)
{
	int ret = 0;
	char * str = NULL;
	char * p = NULL;
	int port_breath_ms = -1;
	int max_lnkup_loop = -1;

	str = (char*)__get_free_page(GFP_KERNEL);
	if (!str) { ret = -ENOMEM; goto _out; }

	if (copy_from_user(str, user_buf, count)) { ret = -EFAULT; goto _out; }
	str[count] = '\0';
	if (str[count-1] == '\n') str[count-1]='\0';

	/* see if it is a command that we can recognize. */
	if (str[0] == 'c') {
		g_probe_cnt = 0;
		g_failure_cnt = 0;
	}
	else
	{	/* parse the options into integers. */
		p = str;
		ret = get_option(&p, &port_breath_ms);
		if (ret == 2) {
			ret = get_option (&p, &max_lnkup_loop);
		}
		if (ret == 0) { ret = -EINVAL; goto _out; }

		/* assign the result */
		if (port_breath_ms >= 0) g_port_breath_ms = port_breath_ms;
		if (max_lnkup_loop >= 0) g_max_lnkup_loop = max_lnkup_loop;
	}

	ret = count;

_out:
	if (str) free_page((unsigned long)str);
	return ret;
}

int __init pcie_proc_init_dbg(void)
{
	struct proc_dir_entry *res = create_proc_entry(__PCIE_DBG_NODE,0,NULL);

	if (res) {
		res->read_proc  = pcie_proc_read_dbg;
		res->write_proc = pcie_proc_write_dbg;
		res->data       = NULL;
	}
	return 0;
}

void __exit pcie_proc_exit_dbg(void)
{
	remove_proc_entry(__PCIE_DBG_NODE, NULL);
}
#endif /* CXD9225GG_PCIE_DEBUG */


/*------------------------------------------------------------------------------
 * This is the module entry of the pcie host controllers driver.
 */
static int __init cxd9225gg_pciehc_init(void)
{
	pcie_debugl("Start of PCIE Host Controller Initialization");
	/* hook imprecise external abort for PCIE bus */
	hook_fault_code(22, bsp_pcie_data_abort, SIGBUS, "imprecise external abort");

	{//init global variables.
		int i=0;
		memset(s_pcie_ports, 0, sizeof(s_pcie_ports));
		for (i=0; i<NUM_PCIE_PORTS; i++) {
			s_pcie_ports[i].root_bus_nr = -1;
		}
	}

	//skip pci bus. Note, pci bus must be setup before pcie bus.
	s_busnr_start = bsp_pcie_cal_busnr_start();
	pcie_infol("PCIE bus number start from %d", s_busnr_start);

	//check pcie slot exchange state.
	s_exchg = 0;
	pcie_infol("Slot Exchange = %s", s_exchg ?"ON" :"OFF");
	if (s_exchg) {
		pcie_infol("So,  port0 --> slot1,  port1 --> slot0");
	}

	//really do pcie bus setup.
	pci_common_init(&cxd9225gg_pciehc_hw);

#ifdef CXD9225GG_PCIE_DEBUG
	pcie_proc_init_dbg();
#endif

	pcie_debugl("End of PCIE Host Controller Initialization");

	return 0;
}
#ifdef CONFIG_SNSC_DEFERRED_INITCALLS_CXD9225GG_PCIE
deferred_initcall(cxd9225gg_pciehc_init,
		  CONFIG_SNSC_DEFERRED_INITCALLS_GROUP_CXD9225GG_PCIE);
#else
subsys_initcall(cxd9225gg_pciehc_init);
#endif

/*==============================================================================
 * The following functions are exported to other modules.
 */

/*------------------------------------------------------------------------------
 * Attach the device which is pluged into the PCIE port slot.
 * Return 0 in success case, otherwise return non-zero value.
 * This is an external interface.
 */
int cxd9225gg_pcie_port_attach(int port)
{
	struct pci_bus *bus = s_pcie_ports[port].pbus;
	struct pci_dev *dev = NULL;
	int irq = -1;

	if (port != 0 && port != 1) {
		pcie_errorl("Bad param : port = %d", port);
		return -1;
	}

	/* Tolerate detach mismatch :
	 * This is just what we do in detach function, in case double attach
	 * without any detach, which is so-called detach mismatch.
	 */
	if (s_pcie_ports[port].pdev) {
		pci_remove_bus_device(s_pcie_ports[port].pdev);
		s_pcie_ports[port].pdev = NULL;
		pcie_errorl("Detach mismatch !! It is tolerated but not good.");
	}

	/* Re-initialize the PCIE port, so that we can find external device */
	s_pcie_ports[port].linkup = bsp_pcie_port_init(port, 1);
	if (s_pcie_ports[port].linkup < 0) {
		pcie_errorl("Failed to init the PCIE port %d !!", port);
		return -1;
	}

	/* scan child bus */
	pci_scan_child_bus(bus);

	/* find the newly pluged device */
	dev = bsp_pcie_get_first_dev(bus);
	if (!dev) {
		pcie_errorl("Can not find new device in port %d !!", port);
		return -1;
	}
	pcie_debugl("Device [%s] has been found !!", pci_name(dev));

	/* record the pci_dev pointer */
	s_pcie_ports[port].pdev = dev;

	/* fixup device resources with the bus resources */
	pci_bus_assign_resources(bus);

	/* fixup device irq */
	switch(port) {
	case 0: irq = INT_PCIEX0_INTA; break;
	case 1: irq = INT_PCIEX1_INTA; break;
	case 2: irq = INT_PCIEX2_INTA;
	}
	dev->irq = irq;

	pci_write_config_byte(dev, PCI_INTERRUPT_LINE, irq);

	/* Now the pci_dev is well prepared */
	pcie_infol("Port %d is attached! New device [%s]", port, pci_name(dev));

	/* add device to subsystem (notify the driver) */
	pci_bus_add_devices(bus);

	return 0;
}
EXPORT_SYMBOL(cxd9225gg_pcie_port_attach);


/*------------------------------------------------------------------------------
 * Detach the device which is pulled out of the PCIE port slot.
 * This is an external interface.
 */
void cxd9225gg_pcie_port_detach(int port)
{
	if (port != 0 && port != 1) {
		pcie_errorl("Bad param : port = %d", port);
		return;
	}

	if (s_pcie_ports[port].pdev) {
		/* Actually, we know there is only one device to be remove */
		pci_remove_bus_device(s_pcie_ports[port].pdev);

		/* Clear the pointer field, it is important */
		s_pcie_ports[port].pdev = NULL;

		pcie_infol("Port %d is detached !", port);

		/* reset pcie core */
		bsp_pcie_port_reset_core(port);

		/* shutdown clock to save power */
		bsp_pcie_port_gated_clock_on(port);

	} else {
		pcie_errorl("Nothing to be detached from port %d !", port);
	}

	/*
	 * XXX might be some more disable things to do.
	 * E.g. the PM things of the HC it self.
	 */
}
EXPORT_SYMBOL(cxd9225gg_pcie_port_detach);


/*------------------------------------------------------------------------------
 * Set PCIE Host Controller's Bridge Window
 * This is an external interface.
 */
void pciehc_set_bridge_window(int port, int win, u32 base, u32 dest, u32 size)
{
	bsp_pcie_set_pcie_bridge_window(port, win, base, dest, size);
}
EXPORT_SYMBOL(pciehc_set_bridge_window);

EXPORT_SYMBOL(bsp_pcie_port_init);
