/* 2014-08-28: File changed by Sony Corporation */
/*
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/suspend.h>
#include <linux/device.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/interrupt.h>
#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif
#include <linux/io.h>
#include <linux/uaccess.h>
#include <linux/irq.h>

#include <asm/irq.h>
#include <asm/mach/time.h>

#include <mach/system.h>
#include <linux/em_export.h>

#ifdef CONFIG_SNSC_SSBOOT
#include <linux/ssboot.h>
#include <asm/pgalloc.h>
#include <asm/hardware/cache-l2x0.h>
#include "power.h"
#include "irq.h"
#include "pci.h"
#ifdef CONFIG_CXD9225GG_SSBOOT_ADJUST_SCHED_CLOCK
#include "timer.h"
#endif
#endif

int emxx_sleep_while_idle;
EXPORT_SYMBOL(emxx_sleep_while_idle);

#ifdef CONFIG_PM
#include <linux/cpu.h>
#include <linux/suspend.h>
#include <linux/reboot.h>

#ifdef CONFIG_SNSC_SSBOOT
extern int  cxd9225gg_suspend(void *func);
extern void cxd9225gg_resume(void);

/* buffer to save cpu context */
u_int32_t suspend_context_area[CONTEXT_SIZE_BYTES / sizeof(u_int32_t)];

/*
 * list of virtual addresses which must be mapped as VA=PA while
 * booting by Snapshot Boot.
 */

static void *lpmap_addr[] = {
	suspend_context_area,
	cxd9225gg_resume,
};

static struct saved_pmd {
	pmd_t *pmd;
#ifndef CONFIG_CPU_AFE
	pmd_t value[2];
#else
	pmd_t value[4];
#endif
} saved_pmd[ARRAY_SIZE(lpmap_addr)];

static int
cxd9225gg_ssboot_prepare_pgd(void)
{
	pmd_t *pmd_p, *pmd_v, *pmd_sv;
	unsigned long virt, phys;
	int i;

	for (i = 0; i < ARRAY_SIZE(lpmap_addr); i++) {
		virt = (unsigned long)lpmap_addr[i];
		phys = virt_to_phys((void *)virt);
		pmd_v = pmd_offset(pgd_offset(current->mm, virt), virt);
		pmd_p = pmd_offset(pgd_offset(current->mm, phys), phys);
		pmd_sv = (pmd_t *)&saved_pmd[i].value;
		saved_pmd[i].pmd = pmd_p;
		copy_pmd(pmd_sv, pmd_p);
		copy_pmd(pmd_p, pmd_v);
	}

	return 0;
}

static int
cxd9225gg_ssboot_restore_pgd(void)
{
	pmd_t *pmd_p, *pmd_sv;
	int i;

	for (i = ARRAY_SIZE(lpmap_addr) - 1; i >= 0; i--) {
		pmd_p = saved_pmd[i].pmd;
		pmd_sv = (pmd_t *)&saved_pmd[i].value;
		copy_pmd(pmd_p, pmd_sv);
	}

	return 0;
}

static void cxd9225gg_save_state(void)
{
	cxd9225gg_pci_suspend();
	cxd9225gg_suspend_irq();
	cxd9225gg_ssboot_prepare_pgd();
#ifdef CONFIG_CXD9225GG_SSBOOT_ADJUST_SCHED_CLOCK
	cxd9225gg_sched_clock_suspend();
#endif
}

static void cxd9225gg_restore_state(void)
{
	cxd9225gg_ssboot_restore_pgd();
	cxd9225gg_resume_irq();
	cxd9225gg_pci_resume();
#ifdef CONFIG_CXD9225GG_SSBOOT_ADJUST_SCHED_CLOCK
	cxd9225gg_sched_clock_resume();
#endif
}
#endif /* CONFIG_SNSC_SSBOOT */

static int cxd9225gg_pm_enter(suspend_state_t state)
{
#ifdef CONFIG_SNSC_SSBOOT
	int ret = 0;

	if (PM_SUSPEND_SNAPSHOT == state) {
		cxd9225gg_save_state();
		ret = cxd9225gg_suspend(ssboot_snapshot);
		cxd9225gg_restore_state();
		return ret;
	}
#endif
	return 0;
}

static int cxd9225gg_pm_valid(suspend_state_t state)
{
#ifdef CONFIG_SNSC_SSBOOT
	if (PM_SUSPEND_SNAPSHOT == state)
		return 1;
#endif
	return 0;
}

static void cxd9225gg_pm_finish(void)
{
}

static struct platform_suspend_ops cxd9225gg_pm_ops = {
	.valid  = cxd9225gg_pm_valid,
	.enter  = cxd9225gg_pm_enter,
	.finish = cxd9225gg_pm_finish,
};

#ifdef CONFIG_SNSC_SSBOOT
/*
 * Snapshot Boot support
 */

#define CXD9225GG_NUI_TARGET_ID_MAJOR 0xc
#define CXD9225GG_NUI_TARGET_ID_MINOR 0x0

#define SSBOOT_CXD9225GG_TARGET_ID                        \
	( ((CXD9225GG_NUI_TARGET_ID_MAJOR & 0xff) << 8)   \
	| ((CXD9225GG_NUI_TARGET_ID_MINOR & 0xff) << 0))

static void __init cxd9225gg_ssboot_init(void)
{
#ifdef CONFIG_MACH_AVP13
	ssboot_set_target_id(SSBOOT_CXD9225GG_TARGET_ID);
#else
#error Please assign target ID for ssboot.
#endif /* CONFIG_MACH_AVP13 */
}
#endif /* CONFIG_SNSC_SSBOOT */

int __init cxd9225gg_pm_init(void)
{
	suspend_set_ops(&cxd9225gg_pm_ops);

#ifdef CONFIG_SNSC_SSBOOT
	cxd9225gg_ssboot_init();
#endif /* CONFIG_SNSC_SSBOOT */
	return 0;
}
device_initcall(cxd9225gg_pm_init);

void em_reboot(int force)
{
	machine_restart(NULL);
}
#endif
