/*
 * CPU state save & restore routines for Snapshot Boot
 *
 * Copyright 2013 Sony Corporation
 *
 * This code is implemented based on following code:
 */
/*
 * arch/arm/mach-tegra/cortex-a9.S
 *
 * CPU state save & restore routines for CPU hotplug
 *
 * Copyright (c) 2010, NVIDIA Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <linux/linkage.h>
#include <asm/memory.h>
#include <asm/domain.h>
#include <asm/cache.h>
#include <asm/vfpmacros.h>
#include <asm/system.h>

/*
 * spooled CPU context is 1KB / CPU
 */
#define CTX_SP		0
#define CTX_CPSR	4
#define CTX_SPSR	8
#define CTX_CPACR	12
#define CTX_CSSELR	16
#define CTX_SCTLR	20
#define CTX_ACTLR	24
#define CTX_PCTLR	28

#define CTX_FPEXC	32
#define CTX_FPSCR	36

#define CTX_TTBR0	48
#define CTX_TTBR1	52
#define CTX_TTBCR	56
#define CTX_DACR	60
#define CTX_PAR		64
#define CTX_PRRR	68
#define CTX_NMRR	72
#define CTX_VBAR	76
#define CTX_CONTEXTIDR	80
#define CTX_TPIDRURW	84
#define CTX_TPIDRURO	88
#define CTX_TPIDRPRW	92

#define CTX_SVC_SP	0
#define CTX_SVC_LR	-1	@ stored on stack
#define CTX_SVC_SPSR	8

#define CTX_SYS_SP	96
#define CTX_SYS_LR	100

#define CTX_ABT_SPSR	112
#define CTX_ABT_SP	116
#define CTX_ABT_LR	120

#define CTX_UND_SPSR	128
#define CTX_UND_SP	132
#define CTX_UND_LR	136

#define CTX_IRQ_SPSR	144
#define CTX_IRQ_SP	148
#define CTX_IRQ_LR	152

#define CTX_FIQ_SPSR	160
#define CTX_FIQ_R8	164
#define CTX_FIQ_R9	168
#define CTX_FIQ_R10	172
#define CTX_FIQ_R11	178
#define CTX_FIQ_R12	180
#define CTX_FIQ_SP	184
#define CTX_FIQ_LR	188

#define CTX_VFP_REGS	256
#define CTX_VFP_SIZE	(32 * 8)

#define CTX_DIAG_CTR	512

#include "power.h"

/*
 * Suspend/resume routine
 */
	.text

        .align L1_CACHE_SHIFT
ENTRY(cxd9225gg_suspend)

	push	{r3-r12, lr}
	mov	lr, r0			@ keep function address to call in 'lr'

	ldr     r8,=suspend_context_area

	mov	r0, sp
	mrs	r1, cpsr
	mrs	r2, spsr

	mrc	p15, 0, r3, c1, c0, 2	@ cpacr
	stmia	r8, {r0-r3}

	add	r9, r8, #CTX_SCTLR
	mrc	p15, 0, r0, c1, c0, 0	@ sctlr
	mrc	p15, 0, r1, c1, c0, 1	@ actlr
	stmia	r9, {r0-r1}

	add	r9, r8, #CTX_TTBR0
	mrc	p15, 0, r0, c2, c0, 0	@ TTBR0
	mrc	p15, 0, r1, c2, c0, 1	@ TTBR1
	mrc	p15, 0, r2, c2, c0, 2	@ TTBCR
	mrc	p15, 0, r3, c3, c0, 0	@ DACR
	mrc	p15, 0, r4, c7, c4, 0	@ PAR
	mrc	p15, 0, r5, c10, c2, 0	@ PRRR
	mrc	p15, 0, r6, c10, c2, 1	@ NMRR
	mrc	p15, 0, r7, c12, c0, 0	@ VBAR
	stmia	r9!, {r0-r7}

	mrc	p15, 0, r0, c13, c0, 1	@ CONTEXTIDR
	mrc	p15, 0, r1, c13, c0, 2	@ TPIDRURW
	mrc	p15, 0, r2, c13, c0, 3	@ TPIDRURO
	mrc	p15, 0, r3, c13, c0, 4	@ TPIDRPRW
	stmia	r9, {r0-r3}

	cps	0x1f			@ SYS mode
	add	r9, r8, #CTX_SYS_SP
	stmia	r9, {sp,lr}

	cps	0x17			@ Abort mode
	mrs	r12, spsr
	add	r9, r8, #CTX_ABT_SPSR
	stmia	r9, {r12,sp,lr}

	cps	0x12			@ IRQ mode
	mrs	r12, spsr
	add	r9, r8, #CTX_IRQ_SPSR
	stmia	r9, {r12,sp,lr}

	cps	0x1b			@ Undefined mode
	mrs	r12, spsr
	add	r9, r8, #CTX_UND_SPSR
	stmia	r9, {r12,sp,lr}

	mov	r0, r8
	add	r1, r8, #CTX_FIQ_SPSR
	cps	0x11			@ FIQ mode
	mrs	r7, spsr
	stmia	r1, {r7-r12,sp,lr}

	cps     0x13                    @ back to SVC

	/*
	 * call the function kept in 'lr', with the resume address.
	 */
	ldr	r0, =cxd9225gg_resume
	blx	lr

	pop	{r3-r12, lr}
	mov	pc, lr

ENDPROC(cxd9225gg_suspend)

	.align L1_CACHE_SHIFT
ENTRY(cxd9225gg_resume)
	/*
	 * Assume MMU is diabled.
	 */

	ldr     r0,=suspend_context_area
	sub	r0, r0, #PAGE_OFFSET
	add	r0, r0, #PHYS_OFFSET

	cps	0x11			@ FIQ mode
	add	r1, r0, #CTX_FIQ_SPSR
	ldmia	r1, {r7-r12,sp,lr}
	msr	spsr_fsxc, r7

	cps	0x12			@ IRQ mode
	add	r1, r0, #CTX_IRQ_SPSR
	ldmia	r1, {r12, sp, lr}
	msr	spsr_fsxc, r12

	cps	0x17			@ abort mode
	add	r1, r0, #CTX_ABT_SPSR
	ldmia	r1, {r12, sp, lr}
	msr	spsr_fsxc, r12

	cps	0x1f			@ SYS mode
	add	r1, r0, #CTX_SYS_SP
	ldmia	r1, {sp, lr}

	cps	0x1b			@ Undefined mode
	add	r1, r0, #CTX_UND_SPSR
	ldmia	r1, {r12, sp, lr}
	msr	spsr_fsxc, r12

	cps	0x13			@ back to SVC
	mov	r8, r0

	add	r9, r8, #CTX_SCTLR
	ldmia	r9, {r0-r1}

	mcr	p15, 0, r0, c1, c0, 0	@ sctlr
	mcr	p15, 0, r1, c1, c0, 1	@ actlr

	/*
	 * now MMU is enabled/restored, but still some other registers
	 * must be restored to touch kernel virtual addresses.
	 */

	add	r9, r8, #CTX_TTBR0
	ldmia	r9!, {r0-r7}

	mcr	p15, 0, r4, c7, c4, 0	@ PAR
	mcr	p15, 0, r7, c12, c0, 0	@ VBAR
	mcr	p15, 0, r3, c3, c0, 0	@ DACR

	mov     r10, #0
	mcr     p15, 0, r10, c7, c5, 4   @ ISB

	mcr	p15, 0, r2, c2, c0, 2	@ TTBCR
	mcr     p15, 0, r10, c7, c5, 4   @ ISB

	mcr	p15, 0, r5, c10, c2, 0	@ PRRR
	mcr     p15, 0, r10, c7, c5, 4   @ ISB

	mcr	p15, 0, r6, c10, c2, 1	@ NMRR
	mcr     p15, 0, r10, c7, c5, 4   @ ISB

	ldmia	r9, {r4-r7}

	mcr	p15, 0, r5, c13, c0, 2	@ TPIDRURW
	mcr	p15, 0, r6, c13, c0, 3	@ TPIDRURO
	mcr	p15, 0, r7, c13, c0, 4	@ TPIDRPRW

	ldmia	r8, {r5-r7, lr}

	/* perform context switch to previous context */
	mov	r9, #0
	mcr	p15, 0, r9, c13, c0, 1	@ set reserved context
	mcr     p15, 0, r9, c7, c5, 4   @ ISB

	mcr	p15, 0, r0, c2, c0, 0	@ TTBR0
	mcr     p15, 0, r9, c7, c5, 4   @ ISB

	mcr	p15, 0, r4, c13, c0, 1	@ CONTEXTIDR
	mcr     p15, 0, r9, c7, c5, 4   @ ISB

	mcr	p15, 0, r1, c2, c0, 1	@ TTBR1
	mcr     p15, 0, r9, c7, c5, 4   @ ISB

	mov	r4, #0
	mcr     p15, 0, r4, c8, c5, 0   @ invalidate instruction TLB
	mcr     p15, 0, r4, c8, c6, 0   @ invalidate data TLB
	mcr	p15, 0, r4, c8, c7, 0	@ invalidate unified TLB
	mcr	p15, 0, r4, c7, c5, 6	@ flush BTAC
	mcr	p15, 0, r4, c7, c5, 0	@ flush instruction cache

	mcr     p15, 0, r4, c7, c10, 4  @ DSB
	mcr     p15, 0, r4, c7, c5, 4   @ ISB

	/*
	 * now MMU is completely restored.
	 * we can touch kernel virtual addresses (stack, etc.)
	 */

	mov	sp, r5
	msr	cpsr_cxsf, r6
	msr	spsr_cxsf, r7

	mcr	p15, 0, lr, c1, c0, 2	@ cpacr

	/* finally, restore the stack and return */
	pop	{r3-r12, lr}

	/* return value */
	mov	r0, #0

	mov	pc, lr

ENDPROC(cxd9225gg_resume)
