/* 2014-08-28: File changed by Sony Corporation */
/*
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * Timer #0 is used as clock_event_device and timer #1 is used as clocksource.
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/clockchips.h>
#include <linux/clocksource.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/cnt32_to_63.h>
#include <mach/hardware.h>
#include <mach/irqs.h>
#include <asm/mach/time.h>
#include "timer.h"
#ifdef CONFIG_SNSC_SSBOOT
#include <linux/ssboot.h>
#endif /* CONFIG_SNSC_SSBOOT */

#define CXD9225GG_TICK_TIMER_DIVISOR (2)
#define CXD9225GG_TICK_TIMER_INTVL_VALUE \
	(CXD9225GG_TIMER_CLK / HZ / CXD9225GG_TICK_TIMER_DIVISOR)

/*
 * clockevent (TIMER 0)
 */
static int cxd9225gg_clockevent_next_event(unsigned long delta,
					struct clock_event_device *dev)
{
	unsigned long flags;
	u8 val;

	if (delta == 0)
		return -ETIME;

	local_irq_save(flags);

	val = readb(CXD9225GG_TIMER_TMCD0_VIRT);
	val &= ~(CXD9225GG_TIMER_TMCD_CAE | CXD9225GG_TIMER_TMCD_CE);
	writeb(val, CXD9225GG_TIMER_TMCD0_VIRT);

	do{
		val = readb(CXD9225GG_TIMER_TMCD0_VIRT);
	}while( val & (CXD9225GG_TIMER_TMCD_CAE | CXD9225GG_TIMER_TMCD_CE));

	/* new timer value */
	writel(delta, CXD9225GG_TIMER_CMD0_VIRT);
	/* restart timer */
	val |= (CXD9225GG_TIMER_TMCD_CAE | CXD9225GG_TIMER_TMCD_CE);
	writeb(val, CXD9225GG_TIMER_TMCD0_VIRT);

	local_irq_restore(flags);

	return 0;
}

static void cxd9225gg_clockevent_mode(enum clock_event_mode newmode,
				   struct clock_event_device *dev)
{
	unsigned long flags;
	u8 val;

	local_irq_save(flags);

	switch (newmode) {
	case CLOCK_EVT_MODE_PERIODIC:
		val = readb(CXD9225GG_TIMER_TMCD0_VIRT);
		val &= ~(CXD9225GG_TIMER_TMCD_CAE | CXD9225GG_TIMER_TMCD_CE);
		writeb(val, CXD9225GG_TIMER_TMCD0_VIRT);
		do{
			val = readb(CXD9225GG_TIMER_TMCD0_VIRT);
		}while( val & (CXD9225GG_TIMER_TMCD_CAE | CXD9225GG_TIMER_TMCD_CE));
		/* reload counter */
		writel(CXD9225GG_TICK_TIMER_INTVL_VALUE, CXD9225GG_TIMER_CMD0_VIRT);
		/* restart */
		val |= (CXD9225GG_TIMER_TMCD_CAE | CXD9225GG_TIMER_TMCD_CE);
		writeb(val, CXD9225GG_TIMER_TMCD0_VIRT);
		break;
	case CLOCK_EVT_MODE_ONESHOT:
		val = readb(CXD9225GG_TIMER_TMCD0_VIRT);
		val &= ~(CXD9225GG_TIMER_TMCD_CAE | CXD9225GG_TIMER_TMCD_CE);
		writeb(val, CXD9225GG_TIMER_TMCD0_VIRT);
		break;
	case CLOCK_EVT_MODE_SHUTDOWN:
		val = readb(CXD9225GG_TIMER_TMCD0_VIRT);
		val &= ~CXD9225GG_TIMER_TMCD_CAE;
		writeb(val, CXD9225GG_TIMER_TMCD0_VIRT);
		break;
	case CLOCK_EVT_MODE_RESUME:
		val = readb(CXD9225GG_TIMER_TMCD0_VIRT);
		val |= CXD9225GG_TIMER_TMCD_CAE;
		writeb(val, CXD9225GG_TIMER_TMCD0_VIRT);
		break;
	default:
		break;
	};

	local_irq_restore(flags);
}

static struct clock_event_device cxd9225gg_timer0_event = {
	.name		= "timer0",
	.features	= CLOCK_EVT_FEAT_ONESHOT | CLOCK_EVT_FEAT_PERIODIC,
	.shift		= 32,
	.rating		= 300,
	.irq		= INT_TIMER0,
	.cpumask 	= cpu_all_mask,
	.set_next_event	= cxd9225gg_clockevent_next_event,
	.set_mode	= cxd9225gg_clockevent_mode,
};

static irqreturn_t cxd9225gg_timer0_interrupt(int irq, void *dev_id)
{
	if (cxd9225gg_timer0_event.mode == CLOCK_EVT_MODE_ONESHOT) {
		u8 val;
		/*
		 * Reset and stop timer.
		 * set_next_event will get called and enabled in
		 * event_handler()
		 */
		val = readb(CXD9225GG_TIMER_TMCD0_VIRT);
		writeb(val & ~(CXD9225GG_TIMER_TMCD_CAE | CXD9225GG_TIMER_TMCD_CE),
		       CXD9225GG_TIMER_TMCD0_VIRT);
	}

	cxd9225gg_timer0_event.event_handler(&cxd9225gg_timer0_event);

	return IRQ_HANDLED;
}

static struct irqaction cxd9225gg_timer0_irq = {
	.name		= "timer0",
	.flags		= IRQF_DISABLED | IRQF_TIMER | IRQF_IRQPOLL,
	.handler	= cxd9225gg_timer0_interrupt,
	.dev_id		= &cxd9225gg_timer0_event,
};


/*
 * clock source (TIMER 1)
 */

/* Inherits bootloader's setting */
#define CXD9225GG_TIMER_CLOCKSOURCE_DIVISOR	4
#define CXD9225GG_TIMER_CLOCKSOURCE_FREQ \
	(CXD9225GG_TIMER_CLK / CXD9225GG_TIMER_CLOCKSOURCE_DIVISOR)

static cycle_t notrace cxd9225gg_clocksource_read(void)
{
	u64 val;
	unsigned long irqflag;

	local_irq_save(irqflag);

	/*
	 * As our clocksource's multiplier is not even number,
	 * make sure that the topmost garbage bit is clear
	 */
	val = cnt32_to_63(readl(CXD9225GG_TIMER_TMD1_VIRT)) &
		~0x8000000000000000ULL;

	local_irq_restore(irqflag);

	return val;
}


static struct clocksource cxd9225gg_clock_source = {
	.name		= "timer1",
	.shift		= 20,
	.rating		= 300,
	.read		= cxd9225gg_clocksource_read,
	.mask		= CLOCKSOURCE_MASK(32),
	.flags		= CLOCK_SOURCE_IS_CONTINUOUS,
};

static int cxd9225gg_timer_init = 0;

unsigned long long notrace cxd9225gg_timer_nsec(void)
{
	u64 val;
	unsigned long irqflag;
	if (!cxd9225gg_timer_init)
		return 0;

	local_irq_save(irqflag);

	/*
	 * As our clocksource's multiplier is not even number,
	 * make sure that the topmost garbage bit is clear
	 */
	val = cnt32_to_63(readl(CXD9225GG_TIMER_TMD1_VIRT)) &
		~0x8000000000000000ULL;

	/* for timer1 takes 60ns per cycle, so ns = cycle * 60 */
	val = (val<<6)-(val<<2);

	local_irq_restore(irqflag);

	return val;
}

#ifdef CONFIG_CXD9225GG_SSBOOT_ADJUST_SCHED_CLOCK
static unsigned long long sched_clock_saved;
static unsigned long long sched_clock_adjust = 0;
#endif /* CONFIG_CXD9225GG_SSBOOT_ADJUST_SCHED_CLOCK */

/*
 * sched_clock support
 */
unsigned long long notrace sched_clock(void)
{
#ifdef CONFIG_CXD9225GG_SSBOOT_ADJUST_SCHED_CLOCK
	return cxd9225gg_timer_nsec() + sched_clock_adjust;
#else
	return cxd9225gg_timer_nsec();
#endif
}

#ifdef CONFIG_CXD9225GG_SSBOOT_ADJUST_SCHED_CLOCK
void cxd9225gg_sched_clock_suspend(void)
{
	/* save current sched_clock */
	sched_clock_saved = sched_clock();
}

void cxd9225gg_sched_clock_resume(void)
{
	/* restoration is required only at resume */
	if (!ssboot_is_resumed())
		return;

	/* calculate adjust value */
	sched_clock_adjust = sched_clock_saved - cxd9225gg_timer_nsec();
}
#endif /* CONFIG_CXD9225GG_SSBOOT_ADJUST_SCHED_CLOCK */

void __init cxd9225gg_init_timers(void)
{
	/* make sure sched_clock HW Timer CMD is 0xffffffff. */
	u32 val = 0;
	val = readl(CXD9225GG_TIMER_CMD1_VIRT);
	if (val != 0xffffffff) {
		printk("%s() %d: CXD9225GG_TIMER_CMD1_VIRT=0x%08x\n",
				__func__, __LINE__, val);
		val = 0xffffffff;
		writel(val, CXD9225GG_TIMER_CMD1_VIRT);
	}

	/*
	 * clockevent (tick timer)
	 */
	setup_irq(cxd9225gg_timer0_event.irq, &cxd9225gg_timer0_irq);

	cxd9225gg_timer0_event.mult = div_sc(CXD9225GG_TIMER_CLK/CXD9225GG_TICK_TIMER_DIVISOR
		, NSEC_PER_SEC, cxd9225gg_timer0_event.shift);
	cxd9225gg_timer0_event.max_delta_ns =
	  clockevent_delta2ns(0xffffffff, &cxd9225gg_timer0_event);
	cxd9225gg_timer0_event.min_delta_ns =
	  clockevent_delta2ns(0x400, &cxd9225gg_timer0_event);

	clockevents_register_device(&cxd9225gg_timer0_event);

	/*
	 * clock source
	 */
	cxd9225gg_clock_source.mult =
		clocksource_hz2mult(CXD9225GG_TIMER_CLOCKSOURCE_FREQ,
				    cxd9225gg_clock_source.shift);
	clocksource_register(&cxd9225gg_clock_source);
	cxd9225gg_timer_init = 1;
}

struct sys_timer cxd9225gg_timer  = {
	.init = cxd9225gg_init_timers,
};
