/* 2014-08-28: File changed by Sony Corporation */
/*
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**********************************
     Include Files
**********************************/
#include<linux/init.h>
#include <linux/module.h>
#include<linux/types.h>
#include<linux/kdev_t.h>
#include<linux/fs.h>
#include<linux/cdev.h>
//#include<linux/kernel.h>    //printk()
#include<linux/slab.h>
#include<linux/ioctl.h>
#include<asm/uaccess.h>
#include<linux/delay.h>    //mdelay()

/**********************************
    Define Macros
**********************************/
/*
 * Device Major ID
 */
//for char devices
#define CXD9225GG_MAJOR_ID_COMPLIANCE_TEST_MODE (263)
#define COMPLIANCE_TEST_NR_DEVS (1)

    typedef struct port_mode_t
    {
	    unsigned int port;
	    unsigned int mode;
    }PORT_MODE;

/*
 * usb_compliance test magic number
 */
#define COMPLIANCE_TEST_IOC_MAGIC '['
#define USB_HOST_TEST_IOC _IOW(COMPLIANCE_TEST_IOC_MAGIC, 0, PORT_MODE*)
#define COMPLIANCE_TEST_IOC_MAXNR 1

/**********************************
     Module Variables
**********************************/
struct usb_compliance_test_dev
{
	struct cdev cdev;
};

int usb_compliance_test_major =  CXD9225GG_MAJOR_ID_COMPLIANCE_TEST_MODE;
int usb_compliance_test_minor =  0;
int usb_compliance_test_devs_nr = COMPLIANCE_TEST_NR_DEVS;	/* number of bare pcie host devices */
struct usb_compliance_test_dev *usb_compliance_test_devices;  	/* allocated in pcie_host_init_module */

/**********************************
     External Functions
**********************************/
//extern int pciehTestModeEnable(void);
extern int usbHostTestMode(u32 port, u32 mode);
/**********************************
     Module Functions
**********************************/
/*
 * Open and close
 */
int usb_compliance_test_open(struct inode *inode, struct file *filp)
{
	return 0;          /* success */
}

int usb_compliance_test_release(struct inode *inode, struct file *filp)
{
	return 0;
}

/*
 * Data management: read and write
 */
ssize_t usb_compliance_test_read(struct file *filp, char __user *buf, size_t count,
                loff_t *f_pos)
{
	return 0;
}

ssize_t usb_compliance_test_write(struct file *filp, const char __user *buf, size_t count,
                loff_t *f_pos)
{
	return count;
}

/*
 * The ioctl() implementation
 */
int usb_compliance_test_ioctl(struct inode *inode, struct file *filp,
                 unsigned int cmd, unsigned long arg)
{
	int err = 0;
	int retval = 0;
    PORT_MODE port_mode;

	if (_IOC_TYPE(cmd) != COMPLIANCE_TEST_IOC_MAGIC)
	{
		printk(KERN_ALERT "invalid cmd magic number\n");
		return -ENOTTY;
    }
	if (_IOC_NR(cmd) > COMPLIANCE_TEST_IOC_MAXNR)
	{
		printk(KERN_ALERT "invalid cmd number\n");
		return -ENOTTY;
    }

	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
	else if (_IOC_DIR(cmd) & _IOC_WRITE)
		err =  !access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
	if (err)
	{
		printk(KERN_ALERT "invalid cmd direction\n");
	    return -EFAULT;
	}

	switch(cmd)
	{
      case USB_HOST_TEST_IOC:
        if(copy_from_user(&port_mode, (void __user *)arg, sizeof(PORT_MODE)))
			{
				return -EFAULT;
			}
        retval = usbHostTestMode(port_mode.port, port_mode.mode);
		break;
	  default:  /* redundant, as cmd was checked against MAXNR */
	    printk(KERN_ALERT "invalid cmd\n");
      	return -ENOTTY;
	}
	return retval;
}

struct file_operations usb_compliance_test_fops = {
	.owner =    THIS_MODULE,
	.read =     usb_compliance_test_read,
	.write =    usb_compliance_test_write,
	.ioctl =    usb_compliance_test_ioctl,
	.open =     usb_compliance_test_open,
	.release =  usb_compliance_test_release
};

/*
 * Set up the char_dev structure for this device.
 */
static void usb_host_setup_cdev(struct usb_compliance_test_dev *dev, int index)
{
	int err, devno = MKDEV(usb_compliance_test_major, usb_compliance_test_minor + index);

	cdev_init(&dev->cdev, &usb_compliance_test_fops);
	dev->cdev.owner = THIS_MODULE;
	err = cdev_add (&dev->cdev, devno, 1);

	if (err)
		printk(KERN_NOTICE "Error %d adding pcie%d", err, index);
}

/*
 * The cleanup function is used to handle initialization failures as well.
 * Therefore, it must be careful to work correctly even if some of the items
 * have not been initialized
 */
static void usb_compliance_test_cleanup_module(void)
{
	int i;
	dev_t devno = MKDEV(usb_compliance_test_major, usb_compliance_test_minor);

	/* Release char dev entries */
	if (usb_compliance_test_devices)
	{
		for (i = 0; i < usb_compliance_test_devs_nr; i++)
        {
			cdev_del(&usb_compliance_test_devices[i].cdev);
		}
		kfree(usb_compliance_test_devices);
	}

	/* Release device numbers */
	unregister_chrdev_region(devno, usb_compliance_test_devs_nr);

}

static int __init usb_compliance_test_init_module(void)
{
	int result, i;
	dev_t dev = 0;

    /* Get device numbers */
	if (usb_compliance_test_major)
    {
		dev = MKDEV(usb_compliance_test_major, usb_compliance_test_minor);
		result = register_chrdev_region(dev, usb_compliance_test_devs_nr, "usb_compliance_test");
	}
	else
	{
		result = alloc_chrdev_region(&dev, usb_compliance_test_minor, usb_compliance_test_devs_nr,"usb_compliance_test");
		usb_compliance_test_major = MAJOR(dev);
	}
	if (result < 0)
	{
		printk(KERN_WARNING "usb_compliance_test: can't get major %d\n", usb_compliance_test_major);
		return result;
	}

    /* allocate the devices as the number */
	usb_compliance_test_devices = kmalloc(usb_compliance_test_devs_nr * sizeof(struct usb_compliance_test_dev), GFP_KERNEL);
	if (!usb_compliance_test_devices)
	{
		result = -ENOMEM;
		goto fail;
	}

	memset(usb_compliance_test_devices, 0, usb_compliance_test_devs_nr * sizeof(struct usb_compliance_test_dev));

    /* Initialize each device */
	for (i = 0; i < usb_compliance_test_devs_nr; i++)
	{
		usb_host_setup_cdev(&usb_compliance_test_devices[i], i);
	}

    /* succeed */
	return 0;

  fail:
	usb_compliance_test_cleanup_module();
	return result;
}

module_init(usb_compliance_test_init_module);
module_exit(usb_compliance_test_cleanup_module);

MODULE_LICENSE("Dual BSD/GPL");
