/* 2014-08-28: File changed by Sony Corporation */
/*
 *Copyright (C) 2012 Sony Corporation
 *This program is free software; you can redistribute it and/or modify
 *it under the terms of the GNU General Public License, version 2, as
 *published by the Free Software Foundation.
 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *GNU General Public License for more details.
 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/****************************************************************************
 * Filename:     cxd9225gg_udc.c
 * Description:  the implemention of udc driver for cxd9225gg cpu. 
 *****************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/mm.h>
#include <linux/moduleparam.h>
#include <linux/platform_device.h>
#include <linux/proc_fs.h>
#include <linux/debugfs.h>
#include <linux/seq_file.h>
#include <linux/kthread.h>
#include <linux/dma-mapping.h>

#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/usb/gadget.h>
#include <linux/usb/otg.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/unaligned.h>

#include "cxd9225gg_udc.h"
#include "cxd9225gg_usbExt.h"
#include "cxd9225gg_usbEvalLib.h"

#include <mach/udc_mass_priority.h>

#ifdef UDC_PROC_DEBUG
#include <linux/proc_fs.h>
#include <linux/semaphore.h>

struct _udc_stat udc_stat;
static spinlock_t udc_stat_lock;

static int udc_proc_read_stat(char *page, char **start, off_t off,
			 int count, int *eof, void *data)
{
	struct _udc_stat * pstat = (struct _udc_stat *)data;
	char *p = page;
	int len;
	unsigned long irq_flags;
	struct _udc_stat st;

	spin_lock_irqsave(&udc_stat_lock, irq_flags);
	memcpy(&st, pstat, sizeof(struct _udc_stat));
	spin_unlock_irqrestore(&udc_stat_lock, irq_flags);

	p += sprintf(p, "%s"
		"isr_step=%lld, que_step=%lld, vbus=%lld, "
		"isr_wakeup=%lld, isr_loop=%lld\n"
		"epc: intr=%lld, cmpl=%lld, wait=%lld # "
		"bak: intr=%lld, wakeup=%lld, loop=%lld\n"
		"brg: intr=%lld, cmpl=%lld, wait=%lld # "
		"bak: intr=%lld, wakeup=%lld, loop=%lld\n"
		"in_dma=%lld, out_dma=%lld, watch_dog=%lld\n"
		, (st.epc_cmpl_cnt==st.epc_wait_cnt)?"":"ERROR may occur!\n"
		, st.isr_step,       st.que_step,       st.vbus_cnt
		, st.isr_thd_wakeup, st.isr_thd_loop
		, st.epc_intr_cnt,   st.epc_cmpl_cnt,   st.epc_wait_cnt
		, st.epc_intr_bak,   st.epc_wakeup_bak, st.epc_loop_bak
		, st.brg_intr_cnt,   st.brg_cmpl_cnt,   st.brg_wait_cnt
		, st.brg_intr_bak,   st.brg_wakeup_bak, st.brg_loop_bak
		, st.in_dma_cnt,     st.out_dma_cnt,    st.watch_dog_cnt
		);

	len = (p - page) - off;
	if (len < 0)
		len = 0;

	*eof = (len <= count) ? 1 : 0;
	*start = page + off;

	return len;
}

static int udc_proc_write_stat(struct file *file, const char __user *buffer,
			  unsigned long count, void *data)
{
	return count;
}

int __init udc_proc_init_stat(void)
{
	struct proc_dir_entry *res = create_proc_entry("udc_stat",0,NULL);

	if (res) {
		memset(&udc_stat, 0, sizeof(udc_stat));
		spin_lock_init(&udc_stat_lock);

		res->read_proc  = udc_proc_read_stat;
		res->write_proc = udc_proc_write_stat;
		res->data       = &udc_stat;
	}
	return 0;
}
#endif /* UDC_PROC_DEBUG */

/*  */
#define DRIVER_DESC     "CXD9225GG USB Device Controller Gadget"
#define DRIVER_VERSION  "11 March 2011"
#define DRIVER_AUTHOR   "Sony Corporation."

/* default invalid value for dma-address. */
#ifdef DMA_ADDR_INVALID
#undef DMA_ADDR_INVALID
#endif
#define DMA_ADDR_INVALID    (~(dma_addr_t)0)


/* for bulk-dma support. */
#ifdef CONFIG_CXD9225GG_UDC_USE_DMA
static const unsigned int cxd9225gg_udc_use_dma = 1;
#else /* !USE_DMA */
static const unsigned int cxd9225gg_udc_use_dma = 0;
#endif

static const char driver_name [] = "cxd9225gg_udc";
static const char driver_desc [] = DRIVER_DESC;

static struct cxd9225gg_udc    *cxd9225gg_udc_controller = NULL;

/* debug fs root for cxd9225gg udc. */
static struct dentry    *cxd9225gg_udc_debugfs_root;

/* read usb-function-config
 * only for cxd9225gg board.
 */
static int board_usb_disable = 0;

/* endpoint setup for cxd9225gg udc. */
extern void usb_mass_enable(void);

static int usb_mass_flag = 0;

static inline struct cxd9225gg_ep *to_cxd9225gg_ep(struct usb_ep *_ep_)
{
    return container_of(_ep_, struct cxd9225gg_ep, ep);
}

static inline struct cxd9225gg_udc *to_cxd9225gg_udc(struct usb_gadget *_gadget_)
{
    return container_of(_gadget_, struct cxd9225gg_udc, gadget);
}

static inline struct cxd9225gg_req *to_cxd9225gg_req(struct usb_request *_req_)
{
    return container_of(_req_, struct cxd9225gg_req, req);
}


epcd_fasync_sigcall p_expcdsigcallcb;

int usbConnectionNotifyRegister(void* func)
{
    if(func)
    {
        p_expcdsigcallcb=(epcd_fasync_sigcall)func ;
    }
    else
    {
        udc_error(" func is NULL !!");
        return -1;
    }
    return 0;
}
EXPORT_SYMBOL(usbConnectionNotifyRegister);

int usbConnectionNotifyUnRegister(void)
{
    if(p_expcdsigcallcb)
    {
        p_expcdsigcallcb=NULL ;
    }
    return 0;
}
EXPORT_SYMBOL(usbConnectionNotifyUnRegister);

static struct completion do_notify;

/*
 * usbIsrExec() and usbVbusWaitTask() call this func to wake up udc_q thread.
*/
void notify_to_udc_daemon(void)
{
    complete(&do_notify);
}

void wait_for_notify(void)
{
    wait_for_completion_interruptible(&do_notify);
}

/* the daemon thread for cxd9225gg udc's ep request. */
static int cxd9225gg_udc_queue_thread(void *data)
{
    struct cxd9225gg_udc *udc_dev = (struct cxd9225gg_udc *)data;
    struct cxd9225gg_ep  *cur_ep;
    int    ep_num = 0;
    int    ret_val;

    current->flags |= PF_MEMALLOC;

    udc_debug("entry test... \n");

    /* wait for a start-up semaphore. */
    down(&udc_dev->thread_sem);

    /* main loop. */
    do
    {
        struct cxd9225gg_req *udc_req = NULL;

        /* triggered by other thread */
        UDC_STAT_QUE_STEP(__LINE__);
        wait_for_notify();
        UDC_STAT_QUE_STEP(__LINE__);

        do {
            /* try to get a udc_req */
            spin_lock_irq(&udc_dev->lock);
            if (!list_empty(&udc_dev->ep[1].queue)) {
                cur_ep = &udc_dev->ep[1]; 
                ep_num = 1; 
            }
            else
            if (!list_empty(&udc_dev->ep[2].queue)) {
                cur_ep = &udc_dev->ep[2];  
                ep_num = 2; 
            }
            else
            {
                udc_debug("there is no request for any endpoints. \n");
                cur_ep = NULL;
                ep_num = -1;  
            }
            if (cur_ep) {
                udc_req = list_entry(cur_ep->queue.next, struct cxd9225gg_req, queue);
            }
            udc_dev->cur_req = udc_req;
            spin_unlock_irq(&udc_dev->lock);
            
	    /* continue the do-while loop, until any udc_req is gotten. */
            if (!udc_req) {
                udc_debug("begin to relinquish... \n");
                if (kthread_should_stop()) {
                    udc_error(" is stopped, and exit. \n"); 
                    break;
                }
                
                UDC_STAT_QUE_STEP(__LINE__);
                set_current_state(TASK_INTERRUPTIBLE);
                up(&udc_dev->thread_sem);
                schedule();
                down(&udc_dev->thread_sem);
                set_current_state(TASK_RUNNING);
                UDC_STAT_QUE_STEP(__LINE__);
            }
        } while (!udc_req);

        /* dispatch the request... */
        if (1 == ep_num)
        {
            if (udc_dev->ep_bulk_recv)
            {
                /* lock, and modify... */
                spin_lock_irqsave(&udc_dev->lock, udc_dev->m_flags);
                udc_dev->ep_bulk_recv = 0;
                spin_unlock_irqrestore(&udc_dev->lock, udc_dev->m_flags);

                ret_val = usbFncCopyDataFromEpBuf(cur_ep, udc_req);
                if (ret_val)
                {
                    udc_error("CopyDataFromEpBuf() failed. \n");
                }
            }
            else
            {
                udc_debug("ep1-recv buffer is NOT ready, so switch to ep2-send. \n"); 

                /* whether are there some data in ep2-queue... */
                if (!list_empty(&udc_dev->ep[2].queue) /* && udc_dev->ep_bulk_send */)
                {
                    udc_debug("find a request in its ep2-send-queue. \n");
                    cur_ep = &udc_dev->ep[2];
                    udc_req = list_entry(cur_ep->queue.next, struct cxd9225gg_req, queue);

                    goto ep2_entry;
                }
            }
        }
        else if (2 == ep_num)
        {
ep2_entry: 
            ret_val = usbFncCopyDataToEpBuf(cur_ep, udc_req); 
            if (ret_val)
            {
                udc_error("CopyDataToEpBuf() failed. \n");
            }
        }
        else if (0 == ep_num)
        {
            udc_debug("---- Exception ---- for ep0. \n");
        }
        else
        {
            /* should never reach here. */
            udc_error("---- Exception ---- for ep%d. \n", ep_num);
        }

    } while (1);

    /* release the semaphore to notify others 'exit'. */
    up(&udc_dev->thread_sem);

    udc_error("exit test. \n");

    return 0;
}

/*
 * fsg call this func to wake up udc_q thread.
*/
static void cxd9225gg_udc_request(struct cxd9225gg_udc *udc_dev)
{
    /*  */
    if (1 != udc_dev->run_flag)
    {
        udc_error(" thread for cxd9225gg udc is killed or suspended... \n");

        /* clean up the queue for all endpoints. */
        return;
    }

    /* wake up it... */
    if (!udc_dev->cur_req)
    {
        wake_up_process(udc_dev->thread);
    }
}

/**
 * cxd9225gg_udc_init_queue - initialise a queue structure.
 * @udc_dev: cxd9225gg_udc device
 *
 * Initialise cxd9225gg udc request queue.
 */
static int cxd9225gg_udc_init_queue(struct cxd9225gg_udc *udc_dev)
{
    udc_debug("entry test. \n");

    udc_dev->run_flag = 1;
    udc_dev->cur_req = NULL;

    /*  attention: we should give the sema somewhere else. */
    init_MUTEX(&udc_dev->thread_sem);

    init_completion(&do_notify);

    /* create the kthread for cxd9225gg-udc daemon. */
    udc_dev->thread = kthread_run(cxd9225gg_udc_queue_thread, udc_dev, "cxd9225gg_udc_qd");
    if (IS_ERR(udc_dev->thread))
    {
        udc_error("kthread_run for cxd9225gg-udc-thread failed. \n");
        return -1;
    }

    /* attention. set preemption and its priority. */
    #ifdef CXD9225GG_TASK_PRIORITY_UDC_REQ_Q
    {
        struct sched_param tmp_param;
        tmp_param.sched_priority = CXD9225GG_TASK_PRIORITY_UDC_REQ_Q; 
        sched_setscheduler(udc_dev->thread, SCHED_FIFO, &tmp_param);
    }
    #endif /* CXD9225GG_TASK_PRIORITY_UDC_REQ_Q */

    udc_debug("exit--OK. \n");

    return 0;
}

/* clean up the udc thread queue. */
static void cxd9225gg_udc_cleanup_queue(struct cxd9225gg_udc *udc_dev)
{
    unsigned long flags;

    udc_debug("entry test. \n");

    /* Make sure the queue isn't suspended, as that will deadlock */

    /* Mark that we should start throwing out stragglers */
    spin_lock_irqsave(&udc_dev->lock, flags);
    udc_dev->run_flag = 0;
    udc_dev->cur_req = NULL;

    spin_unlock_irqrestore(&udc_dev->lock, flags);

    /* Then terminate our worker thread */

    udc_debug("exit test. \n");

    udc_dev->run_flag = 1;
    udc_dev->cur_req = NULL;
    return;
}

/**
 * cxd9225gg_udc_queue_suspend - suspend a cxd9225gg udc request queue
 * @udc_dev: cxd9225gg udc to suspend
 *
 * Stop the cxd9225gg request queue, and wait for our thread to
 * complete any outstanding requests.  This ensures that we
 * won't suspend while a request is being processed.
 */
int cxd9225gg_udc_queue_suspend(struct cxd9225gg_udc *udc_dev)
{
    unsigned long flags;

    udc_debug("entry test. \n");

    /* whether is it running. */
    if (1 == udc_dev->run_flag)
    {
        /* to stop gadget-driver's request-queue. */
        spin_lock_irqsave(&udc_dev->lock, flags);
        udc_dev->run_flag = 2;
        spin_unlock_irqrestore(&udc_dev->lock, flags);

        /* get the common mutex. */
        down(&udc_dev->thread_sem);
    }

    udc_debug("exit test. \n");

    return 0;
}

/**
 * cxd9225gg_udc_queue_resume - resume a previously suspended dialbo udc request queue
 * @udc_dev: cxd9225gg udc to suspend
 */
void cxd9225gg_udc_queue_resume(struct cxd9225gg_udc *udc_dev)
{
    unsigned long flags;

    udc_debug("entry test. \n");

    /* whether is it suspended. */
    if (2 == udc_dev->run_flag)
    {
        /* to restart the gadget-driver's request-queue.  */
        spin_lock_irqsave(&udc_dev->lock, flags);
        udc_dev->run_flag = 1;
        spin_unlock_irqrestore(&udc_dev->lock, flags);

        /* release the common mutex. */
        up(&udc_dev->thread_sem);
    }

    udc_debug("exit test. \n");

    return;
}

static int cxd9225gg_ep_enable(struct usb_ep *_ep,
                            const struct usb_endpoint_descriptor *desc
                            )
{
    struct cxd9225gg_ep  *ep = container_of(_ep, struct cxd9225gg_ep, ep);
    struct cxd9225gg_udc *udc;
    unsigned long flags;
    int   ep_num;
    u16   maxp;

    udc_debug("ep-%d.\n", ep->num);

    /* catch various bogus parameters */
    if (!_ep || !desc || ep->desc
        || desc->bDescriptorType != USB_DT_ENDPOINT
        )
    {
        udc_error(" bad ep or descriptor. \n");
        return -EINVAL;
    }

    maxp = le16_to_cpu (desc->wMaxPacketSize);
    if ((desc->bmAttributes == USB_ENDPOINT_XFER_BULK
          && maxp != ep->maxpacket)
        || !desc->wMaxPacketSize
        )
    {
    	if (maxp == USB_MAXPACKET_HIGH || maxp == USB_MAXPACKET_FULL) {
    	    /* we accept dual-speed */
            ep->maxpacket    = maxp;
            ep->ep.maxpacket = maxp;
    	} else {
            udc_error("%s: bad maxpacket. ep_max=%d, desc_max=%d\n", 
                _ep->name, ep->maxpacket, maxp);
            return -ERANGE;
    	}
    }

#ifdef  CONFIG_CXD9225GG_UDC_USE_ISO
    if ((desc->bmAttributes == USB_ENDPOINT_XFER_ISOC
        && desc->bInterval != 1))
    {
        /* hardware wants period = 1; USB allows 2^(Interval-1) */
        udc_error("%s, unsupported ISO period %dms. \n",
                   _ep->name, 1 << (desc->bInterval - 1));
        return -EDOM;
    }
#else
    if (desc->bmAttributes == USB_ENDPOINT_XFER_ISOC)
    {
        udc_error("%s, ISO NOT yet. \n", _ep->name);
        return -EDOM;
    }
#endif

    /* xfer types must match, except that interrupt ~= bulk */
    if (ep->bmAttributes != desc->bmAttributes
        && ep->bmAttributes != USB_ENDPOINT_XFER_BULK
        && desc->bmAttributes != USB_ENDPOINT_XFER_INT
        )
    {
        udc_error("%s type mismatch. \n", _ep->name);
        return -EINVAL;
    }

    udc = ep->udc;
    if (!udc->driver || udc->gadget.speed == USB_SPEED_UNKNOWN)
    {
        udc_error(" bogus device state. \n");
        return -ESHUTDOWN;
    }

    /* lock. */
    spin_lock_irqsave(&udc->lock, flags);

    ep->desc = desc;
    ep->stopped = 0;

    /* check the validation.  */
    if (maxp != ep->ep.maxpacket)
    {
        udc_error("desc->maxp != ep->ep.maxpacket. \n");
    }


    /* set endpoint to initial state */
    ep->has_dma = 0;

    /* to stall this endpoint first... */

/* for iso supported. */
#ifdef CONFIG_CXD9225GG_UDC_USE_ISO
    if (ep->bmAttributes == USB_ENDPOINT_XFER_ISOC)
    {
        list_add(&ep->iso, &udc->iso);
    }
#endif

    /* unlock... */
    spin_unlock_irqrestore(&udc->lock, flags);

    /* iterate the endpoint list for the matched one. */
    for (ep_num = 0; ep_num < 3; ep_num++)
    {
        if (ep == &udc->ep[ep_num])
        {
            break;
        }
    }

    if (ep_num >= 3)
    {
        udc_error("fatal -- can NOT find a matched one for ep:0x%0x. \n", (unsigned int)ep);
        return -1;
    }

    if (ep->num != ep_num)
    {
        udc_error("fatal -- ep->num:%d != ep_num:%d. \n", ep->num, ep_num);
        ep->num = ep_num;
    }

    if (ep->num)
    {
        udc_debug("we use dma for ep:%d. \n", ep->num);
        ep->has_dma = 1;
    }
    else
    {
        ep->has_dma = 0;
    }

	usbFncEndpointAssign(udc, ep_num); //ZZZ

    udc_debug("ep-%d(%s) enabled.\n", ep->num, _ep->name);

    return 0;
}

static int cxd9225gg_ep_disable(struct usb_ep *_ep)
{
    struct cxd9225gg_ep  *ep = container_of(_ep, struct cxd9225gg_ep, ep);

    udc_debug("ep-%d.\n", ep->num);

    if (!_ep || !ep->desc)
    {
        udc_error("input arg invalid, %s not enabled. \n", _ep ? ep->ep.name : NULL);
        return -EINVAL;
    }

    /* lock. */
    spin_lock_irqsave(&ep->udc->lock, ep->udc->m_flags);

    ep->desc = NULL;

    /* completion-ack... */
    cxd9225gg_udc_nuke (ep, -ESHUTDOWN);

    ep->ep.maxpacket = ep->maxpacket;
    ep->has_dma = 0;

    /* to stall the specific endpoint. */

    /* attention: NOT supported now. */
#ifdef CONFIG_CXD9225GG_UDC_USE_ISO
    list_del_init(&ep->iso);
#endif

    /* unlock. */
    spin_unlock_irqrestore(&ep->udc->lock, ep->udc->m_flags);

    udc_debug("exit, %s disabled.\n", _ep->name);

    return 0;
}

/* allocate a request for the specific endpoint. */
static struct usb_request *cxd9225gg_alloc_request(struct usb_ep *_ep, gfp_t gfp_flags)
{
    struct cxd9225gg_req *_req;

    udc_debug("entry test, ep:%s, gfp:0x%02x. \n", _ep->name, (unsigned int)gfp_flags);

    if (!_ep)
    {
        udc_error("ep is NULL. \n");
        return NULL;
    }

    /* allocate and cleanup. */
    _req = kzalloc(sizeof(*_req), gfp_flags);
    if (!_req)
    {
        udc_error("kzalloc for req failed. \n");
        return NULL;
    }

    _req->req.dma = DMA_ADDR_INVALID;
    INIT_LIST_HEAD (&_req->queue);

    return &_req->req;
}

/* release the request for the specific endpoint. */
static void cxd9225gg_free_request(struct usb_ep *_ep, struct usb_request *_req)
{
    struct cxd9225gg_req *req = container_of(_req, struct cxd9225gg_req, req);

    udc_debug("entry test, (%p, %p). \n", to_cxd9225gg_ep(_ep), _req);

    if (!_ep || !_req /* || (!ep->desc && strcmp(ep->name, "ep0")) */ )
    {
        udc_error("input arg is invalid. \n");
        return;
    }

    WARN_ON (!list_empty (&req->queue));

    kfree (req);

    return;
}

/*-------------------------------------------------------------------------*/
/* complete the request for the specific endpoint.
 * attention: the caller must hold the lock.
 */
void cxd9225gg_udc_done(struct cxd9225gg_ep *ep, struct cxd9225gg_req *req, int status)
{
    unsigned    stopped = ep->stopped;

    udc_debug("entry test, udc-ep:%p, udc-req:%p, sts:0x%0x. \n",
               ep, req, (unsigned int)status);

    /* delete it. */
    list_del_init(&req->queue);

    /* attention to unlikely... */
    if (unlikely(req->req.status == -EINPROGRESS))
    {
        req->req.status = status;
    }
    else
    {
        status = req->req.status;
    }

    udc_debug("complete %s usb-req %p stat %d len %u/%u\n",
              ep->ep.name, &req->req, status,
              req->req.actual, req->req.length);

    /* wether should we do unmap for dma. */
    if (cxd9225gg_udc_use_dma && ep->has_dma)
    {
        if (req->mapped && (DMA_ADDR_INVALID != req->req.dma))
        {
            udc_debug(" dma address should be un-mapped. \n");

            dma_unmap_single(ep->udc->gadget.dev.parent,
                             req->req.dma,
                             req->req.length,
                             (ep->bEndpointAddress & 0x01)
                               ? DMA_TO_DEVICE : DMA_FROM_DEVICE);

            req->req.dma = DMA_ADDR_INVALID;
            req->mapped = 0;
        }
        else if (DMA_ADDR_INVALID != req->req.dma)
        {
            udc_debug(" dma address need NOT to be un-mapped, but do sync. \n");

            if (!(ep->bEndpointAddress & 0x01))
            {
                dma_sync_single_for_cpu(ep->udc->gadget.dev.parent,
                                        req->req.dma,
                                        req->req.length,
                                        DMA_FROM_DEVICE);
            }
        }
        else
        {
            udc_error("fatal error--dma:0x%0x, mapped:%lu. \n",
                      (unsigned int)req->req.dma, req->mapped);
        }
    }

    /* don't modify queue heads during completion callback */
    ep->stopped = 1;

    /* attention: the caller hold the lock.
     * unlock first, then lock again.
     */
    spin_unlock_irqrestore(&ep->udc->lock, ep->udc->m_flags);
    req->req.complete(&ep->ep, &req->req);
    spin_lock_irqsave(&ep->udc->lock, ep->udc->m_flags);

    ep->stopped = stopped;

    return;
}


/* enqueue a request to the specific ep-queue. */
static int cxd9225gg_ep_queue(struct usb_ep *_ep,
                           struct usb_request *_req,
                           gfp_t gfp_flags
                           )
{
    struct cxd9225gg_ep  *ep = container_of(_ep, struct cxd9225gg_ep, ep);
    struct cxd9225gg_req *req = container_of(_req, struct cxd9225gg_req, req);
    struct cxd9225gg_udc *udc;
    unsigned long flags;
    int   ret_val;

#ifdef CONFIG_CXD9225GG_UDC_USE_ISO
    int   is_iso = 0;
#endif

    /* catch various bogus parameters */
    if (!_req || !req->req.complete || !req->req.buf
        || !list_empty(&req->queue))
    {
        udc_error("input bad params\n");
        return -EINVAL;
    }

    if (!_ep || (!ep->desc && ep->bEndpointAddress))
    {
        udc_error(" bad ep. \n");
        return -EINVAL;
    }

    udc_debug("%s  queue, ep:%p  req:%p, len:%d,  buf:%p, gfp:0x%0x. \n",
              ep->ep.name, ep, _req, _req->length, _req->buf, (unsigned int)gfp_flags);



/* if iso endpoint is supported. */
#ifdef CONFIG_CXD9225GG_UDC_USE_ISO
    /*  */
    if (ep->bmAttributes == USB_ENDPOINT_XFER_ISOC)
    {
        udc_error("we can NOT support ISOC now. \n");

        if (req->req.length > ep->ep.maxpacket)
        {
            udc_error("req->req.length > ep->ep.maxpacket. \n");
            return -EMSGSIZE;
        }

        is_iso = 1;
    }
#endif

/* we can handle partial packet. */

    udc = ep->udc;
    if (!udc->driver || udc->gadget.speed == USB_SPEED_UNKNOWN)
    {
        udc_error("is shutting down now. \n");
        return -ESHUTDOWN;
    }

    /* whether should we do map for dma. */
    if (cxd9225gg_udc_use_dma && ep->has_dma)
    {
        if (req->req.dma == DMA_ADDR_INVALID)
        {
            udc_debug(" dma address should be mapped. \n");

            /* in current cxd9225gg-platform--it's the same as __pa or virt_to_bus. */
            req->req.dma = dma_map_single(ep->udc->gadget.dev.parent,
                                          req->req.buf,
                                          req->req.length,
                                          (ep->bEndpointAddress & 0x01)
                                            ? DMA_TO_DEVICE : DMA_FROM_DEVICE);

            req->mapped = 1;
        }
        else
        {
            udc_debug(" dma address has been already here, sync is . \n");

            if (ep->bEndpointAddress & 0x01)
            {
                dma_sync_single_for_device(ep->udc->gadget.dev.parent,
                                           req->req.dma,
                                           req->req.length,
                                           DMA_TO_DEVICE);
            }
            req->mapped = 0;
        }
    }

    /* lock... */
    spin_lock_irqsave(&udc->lock, flags);

    req->req.status = -EINPROGRESS;
    req->req.actual = 0;

    /* whether is the ep-queue empty... */
    if (list_empty(&ep->queue) /* && !ep->stopped */)
    {
        if (0 == ep->num)
        {
            if ( USB_CTRL_TRANS_WRITE == udc->usb_target->ulCtrlTransType )
            {
                udc_debug("begin to call PioDataRecieve. \n");
                ret_val = usbPioDataRecieve(&udc->ep[0], req);
                if (ret_val)
                {
                    udc_error("usbPioDataRecieve() failed. \n");
                }
                udc_debug("after of PioDataRecieve. \n");

                req = NULL;
            }
            else if ( USB_CTRL_TRANS_READ == udc->usb_target->ulCtrlTransType )
            {
                udc_debug("begin to call PioDataSend. \n");
                ret_val = usbPioDataSend(&udc->ep[0], req);
                if (ret_val)
                {
                    udc_error("usbPioDataSend() failed. \n");
                }
                udc_debug("after of PioDataSend. \n");

                req = NULL;
            }
            else
            {
                udc_error("-------status error for ep0--------. \n");
            }
        }
        else if (1 == ep->num)
        {
            udc_debug("enqueue a request for ep1-recv. \n");

            if (ep->desc->bEndpointAddress & USB_DIR_IN)
            {
                udc_error("bEndpointAddress is NOT OUT for ep1-recv. \n");
            }
        }
        else if (2 == ep->num)
        {
            udc_debug("enqueue a request for ep2-send. \n");

            if (0 == (ep->desc->bEndpointAddress & USB_DIR_IN))
            {
                udc_error("bEndpointAddress is NOT IN for ep2-send. \n");
            }
        }
        else
        {
            udc_error("fatal error--should NOT reach here, ep-num=%d. \n", ep->num);
        }
    }

    /* irq handler advances the queue */
    if (req != NULL)
    {
        list_add_tail(&req->queue, &ep->queue);

        /* if it suspend, then we should wake it up. */
        cxd9225gg_udc_request(ep->udc);
    }

    /* unlock. */
    spin_unlock_irqrestore(&udc->lock, flags);

    udc_debug("exit test. \n");

    return 0;
}

/* dequeue the request from the specific ep's queue. */
static int cxd9225gg_ep_dequeue(struct usb_ep *_ep, struct usb_request *_req)
{
    struct cxd9225gg_ep  *ep = container_of(_ep, struct cxd9225gg_ep, ep);
    struct cxd9225gg_req *req;

    udc_debug("entry test. \n");

    if (!_ep || !_req)
    {
        udc_error("input bad params\n");
        return -EINVAL;
    }

    /* lock. */
    spin_lock_irqsave(&ep->udc->lock, ep->udc->m_flags);

    /* make sure it's actually queued on this endpoint */
    list_for_each_entry (req, &ep->queue, queue)
    {
        if (&req->req == _req)
        {
            udc_debug("find a matched req. \n");
            break;
        }
    }

    /* whether is it matched. */
    if (&req->req != _req)
    {
        udc_error("error: the request is NOT matched in ep%d queue. \n", ep->num);

        spin_unlock_irqrestore(&ep->udc->lock, ep->udc->m_flags);
        return -EINVAL;
    }

    udc_debug("begin to call cxd9225gg_udc_done for ep%d. \n", ep->num);

    cxd9225gg_udc_done(ep, req, -ECONNRESET);

    /* unlock. */
    spin_unlock_irqrestore(&ep->udc->lock, ep->udc->m_flags);

    return 0;
}

static int cxd9225gg_ep_set_halt(struct usb_ep *_ep, int value)
{
    struct cxd9225gg_ep  *ep = container_of(_ep, struct cxd9225gg_ep, ep);
    unsigned long flags;
    int   status = -EOPNOTSUPP;
    //extern int usbPioDataSend_ack(struct cxd9225gg_ep  *udc_ep);

    udc_debug("ep:%p(%s.%d), value:%d. \n", ep, ep->ep.name, ep->num, value);

    /* lock... */
    spin_lock_irqsave(&ep->udc->lock, flags);

    /* just use protocol stalls for ep0; real halts are annoying */
    if (ep->bEndpointAddress == 0)
    {
        if (!ep->udc->ep0_pending)
        {
            udc_error("ep0_pending is 0, so can NOT to halt. \n");

            status = -EINVAL;
        }
        else if (value)
        {
            udc_debug("begin to halt ep0... \n");

            if (ep->udc->ep0_set_config)
            {
                udc_error("error changing config?\n");

                /* clear the configuration. */
            }

            /* attention: to halt ep0. */
            (void)usbFncEndpointStateSet(USB_EP0_RX, 1);
            //(void)usbPioDataSend_ack(ep);

            ep->udc->ep0_pending = 0;
            status = 0;
        }
        else /* NOP */
        {
            udc_debug("begin to resume ep0... \n");

            /* attention: to halt ep0. */
            (void)usbFncEndpointStateSet(USB_EP0_RX, 0);
            //(void)usbPioDataSend_ack(ep);

            status = 0;
        }

        /* otherwise, all active non-ISO endpoints can halt */
    }
    else if (ep->bmAttributes != USB_ENDPOINT_XFER_ISOC && ep->desc)
    {
        /* halt or clear. */
        if (value)
        {
            /* IN endpoints must already be idle */
            if ((ep->bEndpointAddress & USB_DIR_IN)
                && !list_empty(&ep->queue))
            {
                udc_error("there are data in queue to send, so can NOT halt ep%d. \n", ep->num);

                status = -EAGAIN;
                goto done;
            }

            /* to halt the specific ep. */
        }
        else
        {
            /* to clear the halt for the specific ep. */
        }

        /* attention: to halt/resume ep0. */
        status = usbFncEndpointStateSet(ep->bEndpointAddress, value);
        //(void)usbPioDataSend_ack(ep);
    }

done:
    udc_debug("exit, %s %s halt stat %d\n",
              ep->ep.name, value ? "set" : "clear", status);

    /* unlock. */
    spin_unlock_irqrestore(&ep->udc->lock, flags);

    return status;
}

/* file operations for cxd9225gg-udc'e endpoints. */
static struct usb_ep_ops cxd9225gg_ep_ops =
{
  .enable   = cxd9225gg_ep_enable,
  .disable  = cxd9225gg_ep_disable,

  .alloc_request  = cxd9225gg_alloc_request,
  .free_request = cxd9225gg_free_request,

  .queue    = cxd9225gg_ep_queue,
  .dequeue  = cxd9225gg_ep_dequeue,

  .set_halt = cxd9225gg_ep_set_halt,
};

static int cxd9225gg_get_frame(struct usb_gadget *gadget)
{
    udc_error("this function is NOT supported now.. \n");

    return -EOPNOTSUPP;
}

static int cxd9225gg_wakeup(struct usb_gadget *gadget)
{
    udc_error("entry test, for otg....\n");

    return 0;
}

/* self power callback. */
static int cxd9225gg_set_selfpowered(struct usb_gadget *gadget, int is_selfpowered)
{
    udc_error("we are self-powered now... \n");

    return 0;
}


/*
 * Called by whatever detects VBUS sessions:  external transceiver
 * driver, or maybe GPIO0 VBUS IRQ.  May request 48 MHz clock.
 */
static int cxd9225gg_vbus_session(struct usb_gadget *gadget, int is_active)
{
    udc_error("we have done it in irq-handler directly. \n");

    return 0;
}

static int cxd9225gg_vbus_draw(struct usb_gadget *gadget, unsigned mA)
{
    udc_error("we have done it in irq-handler directly. \n");

    return 0;
}


/* operations for cxd9225gg-udc gadget. */
static struct usb_gadget_ops cxd9225gg_gadget_ops =
{
  .get_frame    = cxd9225gg_get_frame,
  .wakeup     = cxd9225gg_wakeup,
  .set_selfpowered  = cxd9225gg_set_selfpowered,
  .vbus_session   = cxd9225gg_vbus_session,
  .vbus_draw    = cxd9225gg_vbus_draw,
};


static struct cxd9225gg_udc cxd9225gg_udc_memory =
{
    /* gadget object. */
    .gadget = {
        .ops        = &cxd9225gg_gadget_ops,
        .ep0        = &cxd9225gg_udc_memory.ep[0].ep,
        .name       = driver_name,
        .dev = {
            .init_name = "gadget",
        },
    },

    /* control endpoint */
    .ep[0] = {
        .num    = 0,
        .ep = {
            .name       = "ep0",
            .ops        = &cxd9225gg_ep_ops,
            .maxpacket  = USB_MAXPSIZE_64,
        },
        .udc    = &cxd9225gg_udc_memory,
        .maxpacket  = USB_MAXPSIZE_64,
        .bEndpointAddress = USB_EP0_RX, //attention...
        .bmAttributes   = USB_ENDPOINT_XFER_CONTROL,
    },

    /* first group of endpoints */
    .ep[1] = {
        .num    = 1,
        .ep = {
            .name       = "ep1-bulk",
            .ops        = &cxd9225gg_ep_ops,
            .maxpacket  = USB_MAXPACKET_HIGH,
        },
        .udc        = &cxd9225gg_udc_memory,
        .maxpacket  = USB_MAXPACKET_HIGH,
        .bEndpointAddress = USB_EP1_RX,
        .bmAttributes   = USB_ENDPOINT_XFER_BULK,
    },

    .ep[2] = {
        .num        = 2,
        .ep = {
            .name       = "ep2-bulk",
            .ops        = &cxd9225gg_ep_ops,
            .maxpacket  = USB_MAXPACKET_HIGH,
        },
        .udc        = &cxd9225gg_udc_memory,
        .maxpacket  = USB_MAXPACKET_HIGH,
        .bEndpointAddress = USB_EP2_TX,
        .bmAttributes   = USB_ENDPOINT_XFER_BULK,
    },


};


/* dequeue ALL requests for a specific endpoint; caller holds udc->lock */
void cxd9225gg_udc_nuke(struct cxd9225gg_ep *ep, int status)
{
    struct cxd9225gg_req *req;

    udc_debug("entry test. \n");

    ep->stopped = 1;

    /* clean up the queue.  */
    while (!list_empty(&ep->queue))
    {
        req = list_entry(ep->queue.next, struct cxd9225gg_req, queue);

        cxd9225gg_udc_done(ep, req, status);
    }

    return;
}

/* caller holds udc->lock */
void cxd9225gg_udc_quiesce(struct cxd9225gg_udc *udc)
{
    struct cxd9225gg_ep  *ep;

    udc_debug("entry test. \n");

    udc->gadget.speed = USB_SPEED_UNKNOWN;

    cxd9225gg_udc_nuke(&udc->ep[0], -ESHUTDOWN);

    list_for_each_entry (ep, &udc->gadget.ep_list, ep.ep_list)
    {
        cxd9225gg_udc_nuke(ep, -ESHUTDOWN);
    }

    return;
}


static void cxd9225gg_udc_disable(struct cxd9225gg_udc *dev)
{
    int ret;

    udc_debug(" entry test. \n");

    /* disable the udc. */
    ret = usbFncDisable(dev->usb_target);
    if (ret)
    {
        udc_error("usbFncDisable failed. \n");
        return;
    }

    /* Set speed to unknown */
    dev->gadget.speed = USB_SPEED_UNKNOWN;
    cxd9225gg_udc_cleanup_queue(dev);

    udc_debug("exit OK... \n");

    return;
}

static void cxd9225gg_udc_enable(struct cxd9225gg_udc *dev)
{
    int ret_val;

    udc_debug("entry test. \n");

    /* enable the udc. */
    ret_val = usbFncEnable(dev->usb_target);
    if (ret_val)
    {
        udc_error("usbFncEnable failed. \n");
        return;
    }

    udc_debug("exit OK... \n");

    return;
}

int cxd9225gg_udc_get_mass_flag(void)
{
    return usb_mass_flag;
}
void cxd9225gg_udc_set_mass_flag(int flag)
{
    usb_mass_flag = flag;
}

int cxd9225gg_udc_find_media(int findno, int ispcie)
{
    if (!cxd9225gg_udc_controller->driver) {
        udc_error("driver is NULL, so we do nothing.\n");
        return 0;
    }
    return cxd9225gg_udc_controller->driver->find_media(findno, ispcie);
}
int cxd9225gg_udc_remove_media(int findno)
{
    if (!cxd9225gg_udc_controller->driver) {
        udc_error("driver is NULL, so we do nothing.\n");
        return 0;
    }
    return cxd9225gg_udc_controller->driver->remove_media(findno);
}

void usb_mass_enable(void)
{
    int   status = -ENODEV;
    if (!cxd9225gg_udc_controller->driver) {
        udc_error("driver is NULL, so we do nothing.\n");
        return;
    }
    status = cxd9225gg_udc_controller->driver->bind(&cxd9225gg_udc_controller->gadget); //fsg_bind(struct usb_configuration * c, struct usb_function * f)
    if (status)
    {
        udc_error("bind to %s --> %d\n", cxd9225gg_udc_controller->driver->driver.name, status);
        cxd9225gg_udc_controller->gadget.dev.driver = NULL;
        cxd9225gg_udc_controller->driver = NULL;
        return;
    }

    /* attach device, and enable related irqs. */
    status = usbFncAttach(cxd9225gg_udc_controller);
    if (status)
    {
        udc_error("usbFncAttach() failed. \n");

    }
}

int usb_gadget_register_driver (struct usb_gadget_driver *driver)
{
    int   status = -ENODEV;
    struct cxd9225gg_ep  *ep;
    unsigned long flags;

    udc_debug("entry test. \n");

    /* basic sanity tests */
    if (!cxd9225gg_udc_controller)
    {
        udc_error("cxd9225gg-udc-device has been released already. \n");
        return -ENODEV;
    }

    if (!driver
        // FIXME if otg, check:  driver->is_otg
        || driver->speed < USB_SPEED_FULL
        || !driver->bind
        || !driver->setup)
    {
        udc_error("some input args are invalid. \n");
        return -EINVAL;
    }

    /* lock. */
    spin_lock_irqsave(&cxd9225gg_udc_controller->lock, flags);

    if (cxd9225gg_udc_controller->driver)
    {
      spin_unlock_irqrestore(&cxd9225gg_udc_controller->lock, flags);

      udc_error("udc->driver has been here already, can NOT attach again. \n");

      return -EBUSY;
    }

    /* reset state */
    list_for_each_entry (ep, &cxd9225gg_udc_controller->gadget.ep_list, ep.ep_list)
    {
        ep->irqs = 0;
        udc_debug("loop for ep, ep:%p. \n", ep);
    }

    cxd9225gg_udc_controller->ep0_pending = 0;
    cxd9225gg_udc_controller->ep[0].irqs = 0;
    cxd9225gg_udc_controller->softconnect = 1;

    /* hook up the driver */
    driver->driver.bus = NULL;
    cxd9225gg_udc_controller->driver = driver;
    cxd9225gg_udc_controller->gadget.dev.driver = &driver->driver;

    /* unlock... */
    spin_unlock_irqrestore(&cxd9225gg_udc_controller->lock, flags);

    /* bind the gadget... */
    status = driver->bind(&cxd9225gg_udc_controller->gadget); //fsg_bind(struct usb_configuration * c, struct usb_function * f)
    if (status)
    {
        udc_error("bind to %s --> %d\n", driver->driver.name, status);
        cxd9225gg_udc_controller->gadget.dev.driver = NULL;
        cxd9225gg_udc_controller->driver = NULL;
        goto done;
    }

    udc_debug("bound to driver %s\n", driver->driver.name);

    /* do we need to do it here. */
    cxd9225gg_udc_enable(cxd9225gg_udc_controller);

    /* we should enable the interrupt here.... */

 /* attention: we do it when usb_gadget_register_driver. */
    /* attach device, and enable related irqs. */
    status = usbFncAttach(cxd9225gg_udc_controller);
    if (status)
    {
        udc_error("usbFncAttach() failed. \n");
        goto done;
    }

    /* connect to bus through transceiver,
     * in fact, we should connect to usb-bus when cxd9225gg-mass-on.
     */
    udc_debug(" need NOT to pull up for cxd9225gg. \n");

    udc_debug("exit--OK.\n");

done:
    return status;
}

int usb_gadget_unregister_driver (struct usb_gadget_driver *driver)
{
   int   status = -ENODEV;

    udc_debug("entry test. \n");

    /* basic sanity tests */
    if (!cxd9225gg_udc_controller)
    {
        udc_error("cxd9225gg-udc-device has been released already. \n");
        return -ENODEV;
    }

    if (!driver || driver != cxd9225gg_udc_controller->driver || !driver->unbind)
    {
        udc_error("some input args are invalid. \n");
        return -EINVAL;
    }

    /* disconnect/detach usb device from usb-bus. */
    status = usbFncDetach(cxd9225gg_udc_controller);
    if (status)
    {
        udc_error("usbFncDetach() failed. \n");
    }

    spin_lock_irqsave(&cxd9225gg_udc_controller->lock, cxd9225gg_udc_controller->m_flags);
    /* cleanup all queues of cxd9225gg-udc. */
    cxd9225gg_udc_quiesce(cxd9225gg_udc_controller);
    spin_unlock_irqrestore(&cxd9225gg_udc_controller->lock, cxd9225gg_udc_controller->m_flags);

    /* unbind the gadget... */
    driver->unbind(&cxd9225gg_udc_controller->gadget);
    cxd9225gg_udc_controller->gadget.dev.driver = NULL;
    cxd9225gg_udc_controller->driver = NULL;

    /* Disable udc */
    cxd9225gg_udc_disable(cxd9225gg_udc_controller);
    notify_to_udc_daemon();
    udc_debug("unregistered driver '%s'\n", driver->driver.name);

    return 0;
}



#ifdef CONFIG_USB_GADGET_DEBUG_FILES

#include <linux/seq_file.h>

/*  */
static const char cxd9225gg_proc_filename[] = "driver/cxd9225gg_udc";

/*  */
static void proc_ep_show(struct seq_file *s, struct cxd9225gg_ep *ep)
{
    udc_error("we've NOT supported it yet. \n");
}

/*  */
static int cxd9225gg_proc_udc_show(struct seq_file *s, void *_)
{
    udc_error("we've NOT supported it yet. \n");

    return 0;
}

static int cxd9225gg_proc_udc_open(struct inode *inode, struct file *file)
{
    return single_open(file, cxd9225gg_proc_udc_show, NULL);
}

/*  */
static const struct file_operations cxd9225gg_proc_ops =
{
  .owner    = THIS_MODULE,
  .open     = cxd9225gg_proc_udc_open,
  .read     = seq_read,
  .llseek   = seq_lseek,
  .release  = single_release,
};

/*  */
static void create_proc_file(void)
{
    proc_create(cxd9225gg_proc_filename, 0, NULL, &cxd9225gg_proc_ops);
}

/*  */
static void remove_proc_file(void)
{
    remove_proc_entry(cxd9225gg_proc_filename, NULL);
}

#else

static inline void create_proc_file(void) {}
static inline void remove_proc_file(void) {}

#endif


int cxd9225gg_ep_setup(struct cxd9225gg_udc *udc_dev, int ep_num)
{
    struct cxd9225gg_ep  *ep;

    udc_debug("entry, ep:%d. \n", ep_num);

    /* check. */
    if (CXD9225GG_UDC_ENDPOINTS <= ep_num)
    {
        udc_error("ep_num:%d is invalid. \n", ep_num);
        return -1;
    }

    /* OUT endpoints first, then IN */
    ep = &udc_dev->ep[ep_num & 0x7f];

    INIT_LIST_HEAD(&ep->queue);

#ifdef CONFIG_CXD9225GG_UDC_USE_ISO
    INIT_LIST_HEAD(&ep->iso);
#endif

    /* attach the udc device with ep. */
    ep->udc = udc_dev;

    //attention, already done by the caller in fact.
    if (ep->num)
    {
        udc_debug("we use dma for ep:%d. \n", ep->num);
        ep->has_dma = 1;
    }
    else
    {
        ep->has_dma = 0;
    }

    /* already done, for dynamically.  */
    ep->ep.ops = &cxd9225gg_ep_ops;

    /* attention, except for ep0. */
    if (ep_num)
    {
        list_add_tail (&ep->ep.ep_list, &udc_dev->gadget.ep_list);
    }

/* do it when the bus is attached. */

    return 0;
}

/* release the device for cxd9225gg-udc. */
static void cxd9225gg_udc_release(struct device *dev)
{
    udc_debug("entry test. \n");

    if (!cxd9225gg_udc_controller)
    {
        udc_error("already released, please check... \n");
        return;
    }

    /* to notify the end of deletion. */
    complete(cxd9225gg_udc_controller->done);

    /* clean it. */
    cxd9225gg_udc_controller = NULL;

    return;
}

static struct cxd9225gg_udc *cxd9225gg_udc_setup(USBTARGET *target)
{
    struct cxd9225gg_udc *udc;
    int ret_val;

    udc_debug("entry test. \n");

    udc = &cxd9225gg_udc_memory;

    udc->usb_target = target;


    /* init spinlock. */
    spin_lock_init (&udc->lock);

    //already done.
    udc->gadget.ops = &cxd9225gg_gadget_ops;
    udc->gadget.ep0 = &udc->ep[0].ep;

    INIT_LIST_HEAD(&udc->gadget.ep_list);

//attention... no iso now.
#ifdef CONFIG_CXD9225GG_UDC_USE_ISO
    INIT_LIST_HEAD(&udc->iso);
#endif

    udc->gadget.speed = USB_SPEED_UNKNOWN;
    udc->gadget.name = driver_name;

    /* initialize the cxd9225gg_udc. */
    device_initialize(&udc->gadget.dev);
    strcpy (udc->gadget.dev.bus_id, "gadget");
    udc->gadget.dev.release = cxd9225gg_udc_release;
    udc->gadget.dev.parent = NULL; //&odev->dev;
    if (cxd9225gg_udc_use_dma)
    {
        udc->gadget.dev.dma_mask = 0;
    }

//do it out of this function, since there would be some config-otg...

    /* ep0 is special; put it right after the SETUP buffer */
    /* for ep0 rx/tx. */
    ret_val = cxd9225gg_ep_setup(udc, 0);
    if (ret_val)
    {
        udc_error("cxd9225gg_ep_setup() failed for ep0. \n");
        return NULL;
    }

/* in fact, initially we should disable all non-ep0 endpoints */
    /* for ep1 rx. */
    ret_val = cxd9225gg_ep_setup(udc, 1);
    if (ret_val)
    {
        udc_error("cxd9225gg_ep_setup() failed for ep1-rx. \n");
        return NULL;
    }

    /* for ep2 tx. */
    ret_val = cxd9225gg_ep_setup(udc, 2);
    if (ret_val)
    {
        udc_error("cxd9225gg_ep_setup() failed for ep2-tx. \n");
        return NULL;
    }

    udc_debug("exit OK. \n");

    return udc;
}


/* probe the cxd9225gg udc. */
static int cxd9225gg_udc_probe(USBTARGET *target)
{
    int status;

    udc_debug("entry test. \n");

    /* a "gadget" abstracts/virtualizes the controller */
    cxd9225gg_udc_controller = cxd9225gg_udc_setup(target);
    if (NULL == cxd9225gg_udc_controller)
    {
        udc_error("cxd9225gg_udc_setup() failed. \n");
        return -1;
    }

/* not support otg now. */
#if defined(CONFIG_USB_OHCI_HCD) || defined(CONFIG_USB_OHCI_HCD_MODULE)
    cxd9225gg_udc_controller->gadget.is_otg = 0; //(config->otg != 0);
#endif

    /* add the gadget-device to system. */
    status = device_add(&cxd9225gg_udc_controller->gadget.dev);
    if (status)
    {
        udc_error("device_add() for cxd9225gg-udc-device failed. \n");
        return -1;
    }

    /* initialize the daenon thread and queue for cxd9225gg udc. */
    status = cxd9225gg_udc_init_queue(cxd9225gg_udc_controller);
    if (status)
    {
        udc_error("cxd9225gg_udc_init_queue() for cxd9225gg-udc failed. \n");
        return -1;
    }

/* for debugging. */
#ifdef CXD9225GG_UDC_DEBUG_PROC
    /*  */
    create_proc_file();
#else
    /*  */
    if (cxd9225gg_udc_debugfs_root)
    {
        udc->regs_info = debugfs_create_file("registers", S_IRUGO,
                                            cxd9225gg_udc_debugfs_root,
                                            udc,
                                            &cxd9225gg_proc_ops);
        if (IS_ERR(udc->regs_info))
        {
            udc_error("debugfs file creation failed %ld\n", PTR_ERR(udc->regs_info));

            udc->regs_info = NULL;
        }
    }
#endif

    udc_debug("exit----OK. \n");

    return 0;
}

/* remove the cxd9225gg udc. */
static int cxd9225gg_udc_remove(void)
{
    DECLARE_COMPLETION_ONSTACK(done);

    if (!cxd9225gg_udc_controller)
    {
        udc_error("already removed. \n");
        return -ENODEV;
    }

    /* whether the gadget-driver has been removed already... */
    if (cxd9225gg_udc_controller->driver)
    {
        udc_error("udc->driver is still used now... \n");
        return -EBUSY;
    }

    /* for sync-release. */
    cxd9225gg_udc_controller->done = &done;

    /* attention...clean up the kthread and queue of cxd9225gg udc. */
    cxd9225gg_udc_cleanup_queue(cxd9225gg_udc_controller);

/* for debugging. */
#ifdef CXD9225GG_UDC_DEBUG_PROC
    /*  */
    remove_proc_file();
#else
    /*  */
    debugfs_remove(cxd9225gg_udc_controller->regs_info);
#endif

/* we would do it in other place. */

    /* delete the device and wait for release-ok. */
    device_unregister(&cxd9225gg_udc_controller->gadget.dev);
    wait_for_completion(&done);

    /* Disable cxd9225gg-udc */
    cxd9225gg_udc_disable(cxd9225gg_udc_controller);

    return 0;
}

/* init entry for this module. */
static int __init udc_init(void)
{
    int retval;
    USBTARGET *cxd9225gg_target;

    udc_debug("%s, version: " DRIVER_VERSION
          #ifdef  CONFIG_CXD9225GG_UDC_USE_ISO
              " (iso)"
          #endif
              "%s\n", driver_desc,
              cxd9225gg_udc_use_dma ?  " (dma)" : "");

    cxd9225gg_udc_debugfs_root = debugfs_create_dir(driver_name, NULL);
    if (IS_ERR(cxd9225gg_udc_debugfs_root))
    {
        udc_error("%s: debugfs dir creation failed %ld. \n",
                  driver_name, PTR_ERR(cxd9225gg_udc_debugfs_root));
        cxd9225gg_udc_debugfs_root = NULL;
    }

    /* initialize the hardware of cxd9225gg-udc. */
    retval = usbHwInitialize();
    if (retval)
    {
        udc_error("usbHwInitialize() failed. \n");

        if (0 < retval)
        {
            board_usb_disable = 1;
        }

        goto err_exit1;
    }

    udc_debug("usbHwInitialize--succeed. \n");

    /* initialize the software resources. */
    cxd9225gg_target = usbInitialize();
    if (NULL == cxd9225gg_target)
    {
        udc_error("usbInitialize() failed. \n");
        goto err_exit2;
    }

    udc_debug("usbInitialize--succeed. \n");

    /* to probe actively. */
    retval = cxd9225gg_udc_probe(cxd9225gg_target);
    if (retval)
    {
        udc_error("cxd9225gg_udc_probe() failed. \n");
        goto err_exit3;
    }

    udc_debug("cxd9225gg_udc_probe--succeed. \n");

    /* attention...attach the udc-device for cxd9225gg udc target. */
    cxd9225gg_target->udc_dev = cxd9225gg_udc_controller;

/* attention: we do it when usb_gadget_register_driver. */

    retval = request_irq(INT_USBDEV, usbBridgeIsr, IRQF_DISABLED, "usbBridgeIsr", cxd9225gg_udc_controller);
    if (retval)
    {
        udc_error(" request irq --usbBridgeIsr failed. \n");
        goto err_exit3;
    }

    p_expcdsigcallcb = NULL;

#ifdef UDC_PROC_DEBUG
    udc_proc_init_stat();
#endif /* UDC_PROC_DEBUG */

    udc_debug("exit--OK. \n");

    return 0;

err_exit3:
    usbTerminate();
err_exit2:
    usbHwTerminate();
err_exit1:

    debugfs_remove(cxd9225gg_udc_debugfs_root);

    udc_error(" exit -- failed. \n");

    return retval;
}


/* exit entry for this module. */
static void __exit udc_exit(void)
{
    int ret;

    if (board_usb_disable)
    {
        udc_error("USB UDC is disabled on this cxd9225gg board. \n");

        board_usb_disable = 0;

        return;
    }

    free_irq(INT_USBDEV, cxd9225gg_udc_controller);
    ret = usbFncDetach(cxd9225gg_udc_controller);
    if (ret)
    {
        udc_error("usbFncDetach() failed. \n");
    }

    cxd9225gg_udc_controller->usb_target->udc_dev = NULL;

    ret = cxd9225gg_udc_remove();
    if (ret)
    {
        udc_error("cxd9225gg_udc_remove() failed. \n");
    }

    /* release the resources of cxd9225gg udc driver module. */
    usbTerminate();

    /* terminate the hardware of cxd9225gg udc. */
    usbHwTerminate();

    debugfs_remove(cxd9225gg_udc_debugfs_root);
    cxd9225gg_udc_debugfs_root = NULL;
    p_expcdsigcallcb = NULL;

    return;
}

/*  */
EXPORT_SYMBOL(usb_gadget_unregister_driver);
EXPORT_SYMBOL(usb_gadget_register_driver);

/*  */
#ifdef CONFIG_SNSC_DEFERRED_INITCALLS_USB
deferred_initcall(udc_init,
                  CONFIG_SNSC_DEFERRED_INITCALLS_GROUP_USB_CXD9225GG_UDC);
#else
module_init(udc_init);
#endif
module_exit(udc_exit);

EXPORT_SYMBOL(cxd9225gg_udc_remove_media);
EXPORT_SYMBOL(cxd9225gg_udc_find_media);
EXPORT_SYMBOL(cxd9225gg_udc_get_mass_flag);
EXPORT_SYMBOL(cxd9225gg_udc_set_mass_flag);	
