/*
 *Copyright (C) 2012 Sony Corporation
 *This program is free software; you can redistribute it and/or modify
 *it under the terms of the GNU General Public License, version 2, as
 *published by the Free Software Foundation.
 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *GNU General Public License for more details.
 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/****************************************************************************
 * Filename:     cxd9225gg_udc.h
 * Description:  the definitions of udc driver for cxd9225gg cpu.
 *****************************************************************************/

#ifndef _CXD9225GG_UDC_H_
#define _CXD9225GG_UDC_H_

#include <linux/list.h>
#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/usb/gadget.h>

#include "cxd9225gg_usbEvalLib.h"

#define CXD9225GG_UDC_DEBUG_PROC

#define CXD9225GG_UDC_ENDPOINTS       USB_NUM_ENDPOINTS

/* bulk DMA seems to be behaving for both IN and OUT
 * we can do it by kconfig...
 */
#define CONFIG_CXD9225GG_UDC_USE_DMA



/* forward declaration. */
struct cxd9225gg_udc;
struct usbEndpoint;
struct usbTarget;


enum ep0_state
{
        EP0_IDLE,
        EP0_IN_DATA_PHASE,
        EP0_OUT_DATA_PHASE,
        EP0_END_XFER,
        EP0_STALL,
};

/* the definition of request for cxd9225gg udc. */
struct cxd9225gg_req
{
    struct usb_request  req;
    struct list_head    queue;
    unsigned long       dma_bytes;
    unsigned long       mapped;
};

/* the definition of endpoint for cxd9225gg udc. */
struct cxd9225gg_ep
{
    struct usb_ep       ep;
    struct list_head    queue;
    int                 num;
    unsigned long       irqs;

/* for iso support. */
#ifdef CONFIG_CXD9225GG_UDC_USE_ISO
    struct list_head    iso;
#endif

    const struct usb_endpoint_descriptor  *desc;

    char        name[16];

    u16       maxpacket;
    u8        bEndpointAddress;
    u8        bmAttributes;

    unsigned      double_buf:1;
    unsigned      stopped:1;
    unsigned      fnf:1;
    unsigned      has_dma:1;

    struct cxd9225gg_udc     *udc;
};


struct cxd9225gg_udc
{
    struct usb_gadget   gadget;
    struct usb_gadget_driver  *driver;

    spinlock_t      lock;
    unsigned long   m_flags;

    /* in fact, there are 16 endpoints for cxd9225gg-cpu's udc... */
    struct cxd9225gg_ep      ep[CXD9225GG_UDC_ENDPOINTS];

#ifdef CONFIG_CXD9225GG_UDC_USE_ISO
    struct list_head    iso;
#endif

    unsigned      softconnect:1;
    unsigned      vbus_active:1;
    unsigned      ep0_pending:1;
    unsigned      ep0_in:1;
    unsigned      ep0_set_config:1;
    unsigned      ep0_reset_config:1;
    unsigned      ep0_setup:1;
    unsigned      clk_requested:1;
    unsigned      ep_bulk_recv:1;
    unsigned      ep_bulk_send:1;

    struct completion   *done;

    int                 address;

    struct cxd9225gg_req   *cur_req;
    struct dentry       *regs_info;
    struct usbTarget *usb_target;

    /* the mutex for current queue-thread. */
    struct semaphore    thread_sem;
    struct task_struct  *thread;
    int volatile        run_flag;

};


/* for EPn-in/out notification. */
void notify_to_udc_daemon(void);
extern int cxd9225gg_udc_get_mass_flag(void);
extern void cxd9225gg_udc_set_mass_flag(int flag);
extern int cxd9225gg_udc_find_media(int findno, int ispcie);
extern int cxd9225gg_udc_remove_media(int findno);

int cxd9225gg_ep_setup(struct cxd9225gg_udc *udc_dev, int ep_num);

/* resume the queue of cxd9225gg udc. */
void cxd9225gg_udc_queue_resume(struct cxd9225gg_udc *udc_dev);

/* susend the queue of cxd9225gg udc. */
int cxd9225gg_udc_queue_suspend(struct cxd9225gg_udc *udc_dev);

/* complete the request for the specific endpoint. */
void cxd9225gg_udc_done(struct cxd9225gg_ep *ep, struct cxd9225gg_req *req, int status);

/* dequeue ALL requests for a specific endpoint; caller holds udc->lock */
void cxd9225gg_udc_nuke(struct cxd9225gg_ep *ep, int status);


extern int volatile s_measure_started;
extern int volatile s_bridge_irq_count;
extern int volatile s_epc_irq_count;
extern int volatile s_isr_sw_count;

typedef  int (*epcd_fasync_sigcall)(int sig, void *param, int size);


#define UDC_PROC_DEBUG /* for /proc/udc_stat */
#ifdef UDC_PROC_DEBUG

struct _udc_stat {
	/* view of step in threads */
	unsigned long long isr_step;
	unsigned long long que_step;

	/* view of how many processing times */
	unsigned long long isr_thd_wakeup;
	unsigned long long isr_thd_loop;

	/* view of status related to bridge interrupt */
	unsigned long long brg_intr_cnt;
	unsigned long long brg_cmpl_cnt;
	unsigned long long brg_wait_cnt;
	unsigned long long brg_wakeup_bak;
	unsigned long long brg_loop_bak;
	unsigned long long brg_intr_bak;

	/* view of status related to epc interrupt */
	unsigned long long epc_intr_cnt;
	unsigned long long epc_cmpl_cnt;
	unsigned long long epc_wait_cnt;
	unsigned long long epc_wakeup_bak;
	unsigned long long epc_loop_bak;
	unsigned long long epc_intr_bak;

	/* view of number of times doing in/out dma */
	unsigned long long out_dma_cnt;
	unsigned long long in_dma_cnt;

	/* view of other things */
	unsigned long long vbus_cnt;
	unsigned long long watch_dog_cnt;
};
extern struct _udc_stat udc_stat;

#define UDC_STAT_ISR_STEP(x)            do{udc_stat.isr_step = (x);}while(0)
#define UDC_STAT_QUE_STEP(x)            do{udc_stat.que_step = (x);}while(0)

#define UDC_STAT_OUT_DMA_INC()          do{udc_stat.out_dma_cnt++;}while(0)
#define UDC_STAT_IN_DMA_INC()           do{udc_stat.in_dma_cnt++;}while(0)

#define UDC_STAT_EPC_INTR_INC()         do{udc_stat.epc_intr_cnt++;}while(0)
#define UDC_STAT_BRG_INTR_INC()         do{udc_stat.brg_intr_cnt++;}while(0)
#define UDC_STAT_VBUS_INTR_INC()        do{udc_stat.vbus_cnt++;}while(0)

#define UDC_STAT_ISR_WAKEUP_INC()       do{udc_stat.isr_thd_wakeup++;}while(0)
#define UDC_STAT_ISR_LOOP_INC()         do{udc_stat.isr_thd_loop++;}while(0)

#define UDC_STAT_EPC_CMPL_INC()         do{udc_stat.epc_cmpl_cnt++;}while(0)
#define UDC_STAT_EPC_WAIT_INC()         do{udc_stat.epc_wait_cnt++;}while(0)
#define UDC_STAT_BRG_CMPL_INC()         do{udc_stat.brg_cmpl_cnt++;}while(0)
#define UDC_STAT_BRG_WAIT_INC()         do{udc_stat.brg_wait_cnt++;}while(0)

#define UDC_STAT_EPC_BACKUP() \
	do {\
		udc_stat.epc_intr_bak   = udc_stat.epc_intr_cnt;\
		udc_stat.epc_wakeup_bak = udc_stat.isr_thd_wakeup;\
		udc_stat.epc_loop_bak   = udc_stat.isr_thd_loop;\
	} while (0)
#define UDC_STAT_BRG_BACKUP() \
	do {\
		udc_stat.brg_intr_bak   = udc_stat.brg_intr_cnt;\
		udc_stat.brg_wakeup_bak = udc_stat.isr_thd_wakeup;\
		udc_stat.brg_loop_bak   = udc_stat.isr_thd_loop;\
	} while (0)

#define UDC_STAT_WATCH_DOG_INC()        do{udc_stat.watch_dog_cnt++;}while(0)

#else /* !UDC_PROC_DEBUG */

#define UDC_STAT_ISR_STEP(x)
#define UDC_STAT_QUE_STEP(x)
#define UDC_STAT_OUT_DMA_INC()
#define UDC_STAT_IN_DMA_INC()
#define UDC_STAT_EPC_INTR_INC()
#define UDC_STAT_BRG_INTR_INC()
#define UDC_STAT_VBUS_INTR_INC()
#define UDC_STAT_ISR_WAKEUP_INC()
#define UDC_STAT_ISR_LOOP_INC()
#define UDC_STAT_EPC_CMPL_INC()
#define UDC_STAT_EPC_WAIT_INC()
#define UDC_STAT_BRG_CMPL_INC()
#define UDC_STAT_BRG_WAIT_INC()
#define UDC_STAT_EPC_BACKUP()
#define UDC_STAT_BRG_BACKUP()
#define UDC_STAT_WATCH_DOG_INC()

#endif /* UDC_PROC_DEBUG */

#endif /* end of _CXD9225GG_UDC_H_. */
