/*
 * Copyright (C) 2012 Sony Corporation 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License, version 2, as 
 * published by the Free Software Foundation. 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef __INCusbh
#define __INCusbh


#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/completion.h>
#include <linux/semaphore.h>
#include <linux/mutex.h>

#include <mach/cxd9225gg_types.h>

#include  <mach/cxd9225gg_Chipset.h>

#include "cxd9225gg_udc_common.h"


/* trace function of udc module. */
#ifdef CONFIG_UDC_DEBUG
    //temporarily for debugging output.....
    #define udc_debug(fmt, args...)      \
          printk(KERN_ERR "%s() %d: " fmt, __func__, __LINE__, ## args)

    #define udc_warning(fmt, args...)      \
          printk(KERN_ERR "%s() %d: " fmt, __func__, __LINE__, ## args)
#else
  #define udc_debug(fmt, args...)
  #define udc_warning(fmt, args...)
#endif

  #define udc_error(fmt, args...)      \
          printk(KERN_ERR "%s() %d: " fmt, __func__, __LINE__, ## args)


/**
 *  USBFUNC Driver Version
 */
#define USB_VERSION    (0x00000300)

#define USBFUNC_U2F_INT_EDGE
#define USBFUNC_U2F_EPC_INT_EDGE



/**
 * Registers define
 */
#define USB_TC_OFFSET              CXD9225GG_USBFUNCC_BASE
#define USB_GATED_CTS_USBFUNC      CXD9225GG_GATED_CTS_USBFUNC

/**
 * EPC Registers define
 */
#define     USB_BIT_00                             (0x00000001)
#define     USB_BIT_01                             (0x00000002)
#define     USB_BIT_02                             (0x00000004)
#define     USB_BIT_03                             (0x00000008)
#define     USB_BIT_04                             (0x00000010)
#define     USB_BIT_05                             (0x00000020)
#define     USB_BIT_06                             (0x00000040)
#define     USB_BIT_07                             (0x00000080)
#define     USB_BIT_08                             (0x00000100)
#define     USB_BIT_09                             (0x00000200)
#define     USB_BIT_10                             (0x00000400)
#define     USB_BIT_11                             (0x00000800)
#define     USB_BIT_12                             (0x00001000)
#define     USB_BIT_13                             (0x00002000)
#define     USB_BIT_14                             (0x00004000)
#define     USB_BIT_15                             (0x00008000)
#define     USB_BIT_16                             (0x00010000)
#define     USB_BIT_17                             (0x00020000)
#define     USB_BIT_18                             (0x00040000)
#define     USB_BIT_19                             (0x00080000)
#define     USB_BIT_20                             (0x00100000)
#define     USB_BIT_21                             (0x00200000)
#define     USB_BIT_22                             (0x00400000)
#define     USB_BIT_23                             (0x00800000)
#define     USB_BIT_24                             (0x01000000)
#define     USB_BIT_25                             (0x02000000)
#define     USB_BIT_26                             (0x04000000)
#define     USB_BIT_27                             (0x08000000)
#define     USB_BIT_28                             (0x10000000)
#define     USB_BIT_29                             (0x20000000)
#define     USB_BIT_30                             (0x40000000)
#define     USB_BIT_31                             (0x80000000)

/**
 * USB Control Register
 */
#define    USB_USBCTRL_REG             (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0000))

#define    USB_USBCTRL_TESTMODE2_BIT               USB_BIT_18
#define    USB_USBCTRL_TESTMODE1_BIT               USB_BIT_17
#define    USB_USBCTRL_TESTMODE0_BIT               USB_BIT_16
#define    USB_USBCTRL_INT_SEL_BIT                 USB_BIT_10
#define    USB_USBCTRL_CONSTFS_BIT                 USB_BIT_09
#define    USB_USBCTRL_SOF_RCV_BIT                 USB_BIT_08
#define    USB_USBCTRL_RSUM_IN_BIT                 USB_BIT_07
#define    USB_USBCTRL_SUSPEND_BIT                 USB_BIT_06
#define    USB_USBCTRL_CONF_BIT                    USB_BIT_05
#define    USB_USBCTRL_DEFAULT_BIT                 USB_BIT_04
#define    USB_USBCTRL_CONNECTB_BIT                USB_BIT_03
#define    USB_USBCTRL_PUE2_BIT                    USB_BIT_02
#define    USB_USBCTRL_PHY_RST_BIT                 USB_BIT_01
#define    USB_USBCTRL_F_RST_BIT                   USB_BIT_00

#define    USB_USBCTRL_INIT                  ( USB_USBCTRL_INT_SEL_BIT \
                                                    | USB_USBCTRL_SOF_RCV_BIT )
/**
 * USB Status Register
 */
#define    USB_USBSTS_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0004))

#define    USB_USBSTS_SPEED_MODE_BIT               USB_BIT_06
#define    USB_USBSTS_CONF_BIT                     USB_BIT_05
#define    USB_USBSTS_DEFAULT_BIT                  USB_BIT_04
#define    USB_USBSTS_USB_RST_BIT                  USB_BIT_03
#define    USB_USBSTS_SPND_OUT_BIT                 USB_BIT_02
#define    USB_USBSTS_RSUM_OUT_BIT                 USB_BIT_01

/**
 * USB Address Register
 */
#define    USB_USBADDR_REG             (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0008))

#define    USB_USBADDR_ADDR_6_BIT                  USB_BIT_22
#define    USB_USBADDR_ADDR_5_BIT                  USB_BIT_21
#define    USB_USBADDR_ADDR_4_BIT                  USB_BIT_20
#define    USB_USBADDR_ADDR_3_BIT                  USB_BIT_19
#define    USB_USBADDR_ADDR_2_BIT                  USB_BIT_18
#define    USB_USBADDR_ADDR_1_BIT                  USB_BIT_17
#define    USB_USBADDR_ADDR_0_BIT                  USB_BIT_16
#define    USB_USBADDR_SOF_STATUS_BIT              USB_BIT_15

#define    USB_USB_ADDRESS_MASK                    ( ~(USB_USBADDR_ADDR_6_BIT \
                                                        |   USB_USBADDR_ADDR_5_BIT \
                                                        |   USB_USBADDR_ADDR_4_BIT \
                                                        |   USB_USBADDR_ADDR_3_BIT \
                                                        |   USB_USBADDR_ADDR_2_BIT \
                                                        |   USB_USBADDR_ADDR_1_BIT \
                                                        |   USB_USBADDR_ADDR_0_BIT) )

/**
 * Scratch pad Register
 */
#define    USB_SCRATCHPAD_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x000C))

/**
 * Setup Data0 Register
 */
#define    USB_SETUPDATA0_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0018))

/**
 * Setup Data1_Register
 */
#define    USB_SETUPDATA1_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x001C))

/**
 * USB Interrupt Status Register
 */
#define    USB_USBINTSTS_REG               (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0020))

#define    USB_USBINTSTS_EP15_INT_BIT              USB_BIT_23
#define    USB_USBINTSTS_EP14_INT_BIT              USB_BIT_22
#define    USB_USBINTSTS_EP13_INT_BIT              USB_BIT_21
#define    USB_USBINTSTS_EP12_INT_BIT              USB_BIT_20
#define    USB_USBINTSTS_EP11_INT_BIT              USB_BIT_19
#define    USB_USBINTSTS_EP10_INT_BIT              USB_BIT_18
#define    USB_USBINTSTS_EP9_INT_BIT               USB_BIT_17
#define    USB_USBINTSTS_EP8_INT_BIT               USB_BIT_16
#define    USB_USBINTSTS_EP7_INT_BIT               USB_BIT_15
#define    USB_USBINTSTS_EP6_INT_BIT               USB_BIT_14
#define    USB_USBINTSTS_EP5_INT_BIT               USB_BIT_13
#define    USB_USBINTSTS_EP4_INT_BIT               USB_BIT_12
#define    USB_USBINTSTS_EP3_INT_BIT               USB_BIT_11
#define    USB_USBINTSTS_EP2_INT_BIT               USB_BIT_10
#define    USB_USBINTSTS_EP1_INT_BIT               USB_BIT_09
#define    USB_USBINTSTS_EP0_INT_BIT               USB_BIT_08
#define    USB_USBINTSTS_SPEED_MODE_INT_BIT        USB_BIT_06
#define    USB_USBINTSTS_SOF_ERROR_INT_BIT         USB_BIT_05
#define    USB_USBINTSTS_SOF_INT_BIT               USB_BIT_04
#define    USB_USBINTSTS_USB_RST_INT_BIT           USB_BIT_03
#define    USB_USBINTSTS_SPND_INT_BIT              USB_BIT_02
#define    USB_USBINTSTS_RSUM_INT_BIT              USB_BIT_01

/**
 * USB Interrupt Enable Register
 */
#define    USB_USBINTEN_REG                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0024))

#define    USB_USBINTEN_EP15_EN_BIT                USB_BIT_23
#define    USB_USBINTEN_EP14_EN_BIT                USB_BIT_22
#define    USB_USBINTEN_EP13_EN_BIT                USB_BIT_21
#define    USB_USBINTEN_EP12_EN_BIT                USB_BIT_20
#define    USB_USBINTEN_EP11_EN_BIT                USB_BIT_19
#define    USB_USBINTEN_EP10_EN_BIT                USB_BIT_18
#define    USB_USBINTEN_EP9_EN_BIT                 USB_BIT_17
#define    USB_USBINTEN_EP8_EN_BIT                 USB_BIT_16
#define    USB_USBINTEN_EP7_EN_BIT                 USB_BIT_15
#define    USB_USBINTEN_EP6_EN_BIT                 USB_BIT_14
#define    USB_USBINTEN_EP5_EN_BIT                 USB_BIT_13
#define    USB_USBINTEN_EP4_EN_BIT                 USB_BIT_12
#define    USB_USBINTEN_EP3_EN_BIT                 USB_BIT_11
#define    USB_USBINTEN_EP2_EN_BIT                 USB_BIT_10
#define    USB_USBINTEN_EP1_EN_BIT                 USB_BIT_09
#define    USB_USBINTEN_EP0_EN_BIT                 USB_BIT_08
#define    USB_USBINTEN_SPEED_MODE_EN_BIT          USB_BIT_06
#define    USB_USBINTEN_SOF_ERROR_EN_BIT           USB_BIT_05
#define    USB_USBINTEN_SOF_EN_BIT                 USB_BIT_04
#define    USB_USBINTEN_USB_RST_EN_BIT             USB_BIT_03
#define    USB_USBINTEN_SPND_EN_BIT                USB_BIT_02
#define    USB_USBINTEN_RSUM_EN_BIT                USB_BIT_01

#define     USB_USBINTEN_INIT                      ( USB_USBINTEN_EP2_EN_BIT \
                                                        | USB_USBINTEN_EP1_EN_BIT \
                                                        | USB_USBINTEN_EP0_EN_BIT \
                                                        | USB_USBINTEN_USB_RST_EN_BIT \
                                                        | USB_USBINTEN_SPND_EN_BIT \
                                                        | USB_USBINTEN_RSUM_EN_BIT )

#define     USB_USBINTEN_ALL_DISABLE       (0)

/**
 * EP0 Control Register
 */
#define    USB_EP0CTRL_REG             (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0028))

#define     USB_EPCTRL_REG(n)          (*(volatile unsigned long *)(USB_TC_OFFSET + ((n==0)?0x0028:0x0020) + (n<<5)))
#define     USB_EPCTRL_REG_ADR(n)          (USB_TC_OFFSET + ((n==0)?0x0028:0x0020) + (n<<5))

#define    USB_EP0CTRL_STGSEL_BIT                  USB_BIT_18
#define    USB_EP0CTRL_OVERSEL_BIT                 USB_BIT_17
#define    USB_EP0CTRL_AUTO_BIT                    USB_BIT_16
#define    USB_EP0CTRL_PIDCLR_BIT                  USB_BIT_09
#define    USB_EP0CTRL_BCLR_BIT                    USB_BIT_08
#define    USB_EP0CTRL_DEND_BIT                    USB_BIT_07
#define    USB_EP0CTRL_DW1_BIT                     USB_BIT_06
#define    USB_EP0CTRL_DW0_BIT                     USB_BIT_05
#define    USB_EP0CTRL_INAK_EN_BIT                 USB_BIT_04
#define    USB_EP0CTRL_PERR_NAK_CLR_BIT            USB_BIT_03
#define    USB_EP0CTRL_STL_BIT                     USB_BIT_02
#define    USB_EP0CTRL_INAK_BIT                    USB_BIT_01
#define    USB_EP0CTRL_ONAK_BIT                    USB_BIT_00

#define    USB_EPCTRL_DW_MASK      (USB_EP0CTRL_DW1_BIT | USB_EP0CTRL_DW0_BIT)

/**
 * EP0 Status Register
 */
#define    USB_EP0STS_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x002C))

#define    USB_EPSTS_REG(n)            (n==0)?(*(volatile unsigned long *)(USB_TC_OFFSET + 0x002C)): \
                                                   (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0024 + (n << 5)))

#define    USB_EP0STS_PID_BIT                      USB_BIT_18
#define    USB_EP0STS_PERR_NAK_BIT                 USB_BIT_17
#define    USB_EP0STS_PERR_NAK_INT_BIT             USB_BIT_16
#define    USB_EP0STS_OUT_NAK_INT_BIT              USB_BIT_15
#define    USB_EP0STS_OUT_NULL_BIT                 USB_BIT_14
#define    USB_EP0STS_OUT_FULL_BIT                 USB_BIT_13
#define    USB_EP0STS_OUT_EMPTY_BIT                USB_BIT_12
#define    USB_EP0STS_IN_NAK_INT_BIT               USB_BIT_11
#define    USB_EP0STS_IN_DATA_BIT                  USB_BIT_10
#define    USB_EP0STS_IN_FULL_BIT                  USB_BIT_09
#define    USB_EP0STS_IN_EMPTY_BIT                 USB_BIT_08
#define    USB_EP0STS_OUT_NULL_INT_BIT             USB_BIT_07
#define    USB_EP0STS_OUT_OR_INT_BIT               USB_BIT_06
#define    USB_EP0STS_OUT_INT_BIT                  USB_BIT_05
#define    USB_EP0STS_IN_INT_BIT                   USB_BIT_04
#define    USB_EP0STS_STALL_INT_BIT                USB_BIT_03
#define    USB_EP0STS_STG_END_INT_BIT              USB_BIT_02
#define    USB_EP0STS_STG_START_INT_BIT            USB_BIT_01
#define    USB_EP0STS_SETUP_INT_BIT                USB_BIT_00


/**
 * EP0 Interrupt Enable Register
 */
#define    USB_EP0INTEN_REG                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0030))

#define    USB_EP0INTEN_PERR_NAK_EN_BIT            USB_BIT_16
#define    USB_EP0INTEN_OUT_NAK_EN_BIT             USB_BIT_15
#define    USB_EP0INTEN_IN_NAK_EN_BIT              USB_BIT_11
#define    USB_EP0INTEN_OUT_NULL_EN_BIT            USB_BIT_07
#define    USB_EP0INTEN_OUT_OR_EN_BIT              USB_BIT_06
#define    USB_EP0INTEN_OUT_EN_BIT                 USB_BIT_05
#define    USB_EP0INTEN_IN_EN_BIT                  USB_BIT_04
#define    USB_EP0INTEN_STALL_EN_BIT               USB_BIT_03
#define    USB_EP0INTEN_STG_END_EN_BIT             USB_BIT_02
#define    USB_EP0INTEN_STG_START_EN_BIT           USB_BIT_01
#define    USB_EP0INTEN_SETUP_EN_BIT               USB_BIT_00

#define    USB_EP0INTEN_INIT                       ( USB_EP0INTEN_OUT_EN_BIT \
                                                        | USB_EP0INTEN_IN_EN_BIT \
                                                        | USB_EP0INTEN_SETUP_EN_BIT )

/**
 * EP0 Length Register
 */
#define    USB_EP0LEN_REG                  (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0034))
#define    USB_EPLEN_REG(n)                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0034 + (n << 5)))

#define    USB_EP0LEN_LDATA_6_BIT                  USB_BIT_06
#define    USB_EP0LEN_LDATA_5_BIT                  USB_BIT_05
#define    USB_EP0LEN_LDATA_4_BIT                  USB_BIT_04
#define    USB_EP0LEN_LDATA_3_BIT                  USB_BIT_03
#define    USB_EP0LEN_LDATA_2_BIT                  USB_BIT_02
#define    USB_EP0LEN_LDATA_1_BIT                  USB_BIT_01
#define    USB_EP0LEN_LDATA_0_BIT                  USB_BIT_00

#define    USB_EPLEN_LENGTH_MASK                   0x07ff

/**
 * EP Read Register
 */
#define    USB_EP0READ_REG                 (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0038))
#define    USB_EP1READ_REG                 (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0058))
#define    USB_EP2READ_REG                 (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0078))
#define    USB_EPREAD_REG(n)               (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0038 + (n << 5)))

/**
 * EP Write Register
 */
#define    USB_EP0WRITE_REG                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x003C))
#define    USB_EP1WRITE_REG                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x005C))
#define    USB_EP2WRITE_REG                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x007C))
#define     USB_EPWRITE_REG(n)             (*(volatile unsigned long *)(USB_TC_OFFSET + 0x003C + (n << 5)))

/**
 * EP Control Register
 */
#define    USB_EP1CTRL_REG                 (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0040))
#define    USB_EP2CTRL_REG                 (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0060))

#define    USB_EPCTRL_EN_BIT                       USB_BIT_31
#define    USB_EPCTRL_BUF_TYPE_BIT                 USB_BIT_30
#define    USB_EPCTRL_DIR0_BIT                     USB_BIT_26
#define    USB_EPCTRL_MODE_1_BIT                   USB_BIT_25
#define    USB_EPCTRL_MODE_0_BIT                   USB_BIT_24
#define    USB_EPCTRL_OVERSEL_BIT                  USB_BIT_17
#define    USB_EPCTRL_AUTO_BIT                     USB_BIT_16
#define    USB_EPCTRL_IPIDCLR_BIT                  USB_BIT_11
#define    USB_EPCTRL_OPIDCLR_BIT                  USB_BIT_10
#define    USB_EPCTRL_BCLR_BIT                     USB_BIT_09
#define    USB_EPCTRL_CBCLR_BIT                    USB_BIT_08
#define    USB_EPCTRL_DEND_BIT                     USB_BIT_07
#define    USB_EPCTRL_DW_1_BIT                     USB_BIT_06
#define    USB_EPCTRL_DW_0_BIT                     USB_BIT_05
#define    USB_EPCTRL_OSTL_EN_BIT                  USB_BIT_04
#define    USB_EPCTRL_ISTL_BIT                     USB_BIT_03
#define    USB_EPCTRL_OSTL_BIT                     USB_BIT_02
#define    USB_EPCTRL_ONAK_BIT                     USB_BIT_00

#define    USB_EPCTRL_DIR0_OUT                     USB_EPCTRL_DIR0_BIT

/**
 * EP Status Register
 */
#define    USB_EP1STS_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0044))
#define    USB_EP2STS_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0064))

#define    USB_EPSTS_OPID_BIT                      USB_BIT_28
#define    USB_EPSTS_OUT_END_INT_BIT               USB_BIT_23
#define    USB_EPSTS_OUT_OR_INT_BIT                USB_BIT_22
#define    USB_EPSTS_OUT_NAK_ERR_INT_BIT           USB_BIT_21
#define    USB_EPSTS_OUT_STALL_INT_BIT             USB_BIT_20
#define    USB_EPSTS_OUT_INT_BIT                   USB_BIT_19
#define    USB_EPSTS_OUT_NULL_INT_BIT              USB_BIT_18
#define    USB_EPSTS_OUT_FULL_BIT                  USB_BIT_17
#define    USB_EPSTS_OUT_EMPTY_BIT                 USB_BIT_16
#define    USB_EPSTS_IPID_BIT                      USB_BIT_10
#define    USB_EPSTS_IN_END_INT_BIT                USB_BIT_07
#define    USB_EPSTS_IN_NAK_ERR_INT_BIT            USB_BIT_05
#define    USB_EPSTS_IN_STALL_INT_BIT              USB_BIT_04
#define    USB_EPSTS_IN_INT_BIT                    USB_BIT_03
#define    USB_EPSTS_IN_DATA_BIT                   USB_BIT_02
#define    USB_EPSTS_IN_FULL_BIT                   USB_BIT_01
#define    USB_EPSTS_IN_EMPTY_BIT                  USB_BIT_00

#define     USB_EPSTS_INT_MASK                     (   USB_EPSTS_OUT_END_INT_BIT      |   \
                                                            USB_EPSTS_OUT_OR_INT_BIT       |   \
                                                            USB_EPSTS_OUT_NAK_ERR_INT_BIT  |   \
                                                            USB_EPSTS_OUT_STALL_INT_BIT    |   \
                                                            USB_EPSTS_OUT_INT_BIT          |   \
                                                            USB_EPSTS_OUT_NULL_INT_BIT     |   \
                                                            USB_EPSTS_IN_END_INT_BIT       |   \
                                                            USB_EPSTS_IN_NAK_ERR_INT_BIT   |   \
                                                            USB_EPSTS_IN_STALL_INT_BIT     |   \
                                                            USB_EPSTS_IN_INT_BIT   )

/**
 * EP Interrupt Enable Register
 */
#define    USB_EP1INTEN_REG                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0048))
#define    USB_EP2INTEN_REG                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0068))

#define    USB_EPINTEN_OUT_END_EN_BIT              USB_BIT_23
#define    USB_EPINTEN_OUT_OR_EN_BIT               USB_BIT_22
#define    USB_EPINTEN_OUT_NAK_ERR_EN_BIT          USB_BIT_21
#define    USB_EPINTEN_OUT_STALL_EN_BIT            USB_BIT_20
#define    USB_EPINTEN_OUT_EN_BIT                  USB_BIT_19
#define    USB_EPINTEN_OUT_NULL_EN_BIT             USB_BIT_18
#define    USB_EPINTEN_IN_END_EN_BIT               USB_BIT_07
#define    USB_EPINTEN_IN_NAK_ERR_EN_BIT           USB_BIT_05
#define    USB_EPINTEN_IN_STALL_EN_BIT             USB_BIT_04
#define    USB_EPINTEN_IN_EN_BIT                   USB_BIT_03

#define    USB_EP1INTEN_INIT                       (USB_EPINTEN_OUT_EN_BIT)

#define    USB_EP2INTEN_INIT                       ( USB_EPINTEN_IN_END_EN_BIT \
                                                        | USB_EPINTEN_IN_NAK_ERR_EN_BIT )
/**
 * EP DMA Control Register
 */
#define    USB_EP1DMACTRL_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x004C))
#define    USB_EP2DMACTRL_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x006C))

#define    USB_EPDMACTRL_BUS_SEL_1_BIT             USB_BIT_13
#define    USB_EPDMACTRL_BUS_SEL_0_BIT             USB_BIT_12
#define    USB_EPDMACTRL_STOP_MODE_BIT             USB_BIT_11
#define    USB_EPDMACTRL_DEND_SET_BIT              USB_BIT_10
#define    USB_EPDMACTRL_BURST_SET_BIT             USB_BIT_09
#define    USB_EPDMACTRL_STOP_SET_BIT              USB_BIT_08
#define    USB_EPDMACTRL_DMA_EN_BIT                USB_BIT_04
#define    USB_EPDMACTRL_DMAMODE2_BIT              USB_BIT_02
#define    USB_EPDMACTRL_DMAMODE0_BIT              USB_BIT_00


/**
 * EP Maxpacket & BaseAddres
 */
#define    USB_EP1PCKTADDR_REG             (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0050))
#define    USB_EP2PCKTADDR_REG             (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0070))

#define    USB_EPPCKTADDR_BASEAD_12_BIT            USB_BIT_28
#define    USB_EPPCKTADDR_BASEAD_11_BIT            USB_BIT_27
#define    USB_EPPCKTADDR_BASEAD_10_BIT            USB_BIT_26
#define    USB_EPPCKTADDR_BASEAD_9_BIT             USB_BIT_25
#define    USB_EPPCKTADDR_BASEAD_8_BIT             USB_BIT_24
#define    USB_EPPCKTADDR_BASEAD_7_BIT             USB_BIT_23
#define    USB_EPPCKTADDR_BASEAD_6_BIT             USB_BIT_22
#define    USB_EPPCKTADDR_BASEAD_5_BIT             USB_BIT_21
#define    USB_EPPCKTADDR_BASEAD_4_BIT             USB_BIT_20
#define    USB_EPPCKTADDR_BASEAD_3_BIT             USB_BIT_19
#define    USB_EPPCKTADDR_BASEAD_2_BIT             USB_BIT_18
#define    USB_EPPCKTADDR_BASEAD_1_BIT             USB_BIT_17
#define    USB_EPPCKTADDR_BASEAD_0_BIT             USB_BIT_16
#define    USB_EPPCKTADDR_MPKT_10_BIT              USB_BIT_10
#define    USB_EPPCKTADDR_MPKT_9_BIT               USB_BIT_09
#define    USB_EPPCKTADDR_MPKT_8_BIT               USB_BIT_08
#define    USB_EPPCKTADDR_MPKT_7_BIT               USB_BIT_07
#define    USB_EPPCKTADDR_MPKT_6_BIT               USB_BIT_06
#define    USB_EPPCKTADDR_MPKT_5_BIT               USB_BIT_05
#define    USB_EPPCKTADDR_MPKT_4_BIT               USB_BIT_04
#define    USB_EPPCKTADDR_MPKT_3_BIT               USB_BIT_03
#define    USB_EPPCKTADDR_MPKT_2_BIT               USB_BIT_02
#define    USB_EPPCKTADDR_MPKT_1_BIT               USB_BIT_01
#define    USB_EPPCKTADDR_MPKT_0_BIT               USB_BIT_00

#define     USB_MAXPACKET_MASK             0x7FF

#define     USB_EP1_BASEAD                 (0x00200000)    /* EP1 base address is 0x20h */
#define     USB_EP2_BASEAD                 (0x01200000)    /* EP2 base address is 0x120h (EP1 size is 256Byte) */
#define     USB_MPKT_HS                        (0x0200)    /* HS Mode MaxPacketSize (512byte) */
#define     USB_MPKT_FS                        (0x0040)    /* FS Mode MaxPacketSize  (64byte) */

/**
 * EP Length & DMA count
 */
#define    USB_EP1LENDCNT_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0054))
#define    USB_EP2LENDCNT_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0074))

#define    USB_EPLENDCNT_EPDMACNT_8_BIT                USB_BIT_24
#define    USB_EPLENDCNT_EPDMACNT_7_BIT                USB_BIT_23
#define    USB_EPLENDCNT_EPDMACNT_6_BIT                USB_BIT_22
#define    USB_EPLENDCNT_EPDMACNT_5_BIT                USB_BIT_21
#define    USB_EPLENDCNT_EPDMACNT_4_BIT                USB_BIT_20
#define    USB_EPLENDCNT_EPDMACNT_3_BIT                USB_BIT_19
#define    USB_EPLENDCNT_EPDMACNT_2_BIT                USB_BIT_18
#define    USB_EPLENDCNT_EPDMACNT_1_BIT                USB_BIT_17
#define    USB_EPLENDCNT_EPDMACNT_0_BIT                USB_BIT_16
#define    USB_EPLENDCNT_EPLDATA_10_BIT                USB_BIT_10
#define    USB_EPLENDCNT_EPLDATA_9_BIT                 USB_BIT_09
#define    USB_EPLENDCNT_EPLDATA_8_BIT                 USB_BIT_08
#define    USB_EPLENDCNT_EPLDATA_7_BIT                 USB_BIT_07
#define    USB_EPLENDCNT_EPLDATA_6_BIT                 USB_BIT_06
#define    USB_EPLENDCNT_EPLDATA_5_BIT                 USB_BIT_05
#define    USB_EPLENDCNT_EPLDATA_4_BIT                 USB_BIT_04
#define    USB_EPLENDCNT_EPLDATA_3_BIT                 USB_BIT_03
#define    USB_EPLENDCNT_EPLDATA_2_BIT                 USB_BIT_02
#define    USB_EPLENDCNT_EPLDATA_1_BIT                 USB_BIT_01
#define    USB_EPLENDCNT_EPLDATA_0_BIT                 USB_BIT_00

#define     USB_EPLENDCNT_EPDMACNT_MASK                (~ (USB_EPLENDCNT_EPDMACNT_8_BIT \
                                                            | USB_EPLENDCNT_EPDMACNT_7_BIT \
                                                            | USB_EPLENDCNT_EPDMACNT_6_BIT \
                                                            | USB_EPLENDCNT_EPDMACNT_5_BIT \
                                                            | USB_EPLENDCNT_EPDMACNT_4_BIT \
                                                            | USB_EPLENDCNT_EPDMACNT_3_BIT \
                                                            | USB_EPLENDCNT_EPDMACNT_2_BIT \
                                                            | USB_EPLENDCNT_EPDMACNT_1_BIT \
                                                            | USB_EPLENDCNT_EPDMACNT_0_BIT) )

#define     USB_EPLENDCNT_EPLDATA_MASK                 (~ (USB_EPLENDCNT_EPLDATA_10_BIT \
                                                            | USB_EPLENDCNT_EPLDATA_9_BIT \
                                                            | USB_EPLENDCNT_EPLDATA_8_BIT \
                                                            | USB_EPLENDCNT_EPLDATA_7_BIT \
                                                            | USB_EPLENDCNT_EPLDATA_6_BIT \
                                                            | USB_EPLENDCNT_EPLDATA_5_BIT \
                                                            | USB_EPLENDCNT_EPLDATA_4_BIT \
                                                            | USB_EPLENDCNT_EPLDATA_3_BIT \
                                                            | USB_EPLENDCNT_EPLDATA_2_BIT \
                                                            | USB_EPLENDCNT_EPLDATA_1_BIT \
                                                            | USB_EPLENDCNT_EPLDATA_0_BIT ))



/**
 *EP3~EP6
 */

#define    USB_EP3_CONTROL_REG             (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0080))
#define    USB_EP3_STATUS_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0084))
#define    USB_EP3_INT_ENA_REG             (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0088))
#define    USB_EP3_DMA_CTRL_REG            (*(volatile unsigned long *)(USB_TC_OFFSET + 0x008C))
#define    USB_EP3_PCKT_ADRS_REG           (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0090))
#define    USB_EP3_LEN_DCNT_REG            (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0094))
#define    USB_EP3_READ_REG                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0098))
#define    USB_EP3_WRITE_REG               (*(volatile unsigned long *)(USB_TC_OFFSET + 0x009C))
#define    USB_EP4_CONTROL_REG             (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00A0))
#define    USB_EP4_STATUS_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00A4))
#define    USB_EP4_INT_ENA_REG             (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00A8))
#define    USB_EP4_DMA_CTRL_REG            (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00AC))
#define    USB_EP4_PCKT_ADRS_REG           (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00B0))
#define    USB_EP4_LEN_DCNT_REG            (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00B4))
#define    USB_EP4_READ_REG                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00B8))
#define    USB_EP4_WRITE_REG               (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00BC))
#define    USB_EP5_CONTROL_REG             (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00C0))
#define    USB_EP5_STATUS_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00C4))
#define    USB_EP5_INT_ENA_REG             (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00C8))
#define    USB_EP5_DMA_CTRL_REG            (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00CC))
#define    USB_EP5_PCKT_ADRS_REG           (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00D0))
#define    USB_EP5_LEN_DCNT_REG            (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00D4))
#define    USB_EP5_READ_REG                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00D8))
#define    USB_EP5_WRITE_REG               (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00DC))
#define    USB_EP6_CONTROL_REG             (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00E0))
#define    USB_EP6_STATUS_REG              (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00E4))
#define    USB_EP6_INT_ENA_REG             (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00E8))
#define    USB_EP6_DMA_CTRL_REG            (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00EC))
#define    USB_EP6_PCKT_ADRS_REG           (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00F0))
#define    USB_EP6_LEN_DCNT_REG            (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00F4))
#define    USB_EP6_READ_REG                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00F8))
#define    USB_EP6_WRITE_REG               (*(volatile unsigned long *)(USB_TC_OFFSET + 0x00FC))



/**
 * AHB Bridge Registers define
 */
/**
 * AHBSCTR Register
 */
#define    USB_AHBSCTR_REG                 (*(volatile unsigned long *)(USB_TC_OFFSET + 0x1000))

#define    USB_AHBSCTR_NOT_RETRY_MASTER_15_BIT         USB_BIT_31
#define    USB_AHBSCTR_NOT_RETRY_MASTER_14_BIT         USB_BIT_30
#define    USB_AHBSCTR_NOT_RETRY_MASTER_13_BIT         USB_BIT_29
#define    USB_AHBSCTR_NOT_RETRY_MASTER_12_BIT         USB_BIT_28
#define    USB_AHBSCTR_NOT_RETRY_MASTER_11_BIT         USB_BIT_27
#define    USB_AHBSCTR_NOT_RETRY_MASTER_10_BIT         USB_BIT_26
#define    USB_AHBSCTR_NOT_RETRY_MASTER_9_BIT          USB_BIT_25
#define    USB_AHBSCTR_NOT_RETRY_MASTER_8_BIT          USB_BIT_24
#define    USB_AHBSCTR_NOT_RETRY_MASTER_7_BIT          USB_BIT_23
#define    USB_AHBSCTR_NOT_RETRY_MASTER_6_BIT          USB_BIT_22
#define    USB_AHBSCTR_NOT_RETRY_MASTER_5_BIT          USB_BIT_21
#define    USB_AHBSCTR_NOT_RETRY_MASTER_4_BIT          USB_BIT_20
#define    USB_AHBSCTR_NOT_RETRY_MASTER_3_BIT          USB_BIT_19
#define    USB_AHBSCTR_NOT_RETRY_MASTER_2_BIT          USB_BIT_18
#define    USB_AHBSCTR_NOT_RETRY_MASTER_1_BIT          USB_BIT_17
#define    USB_AHBSCTR_NOT_RETRY_MASTER_0_BIT          USB_BIT_16
#define    USB_AHBSCTR_WAIT_MODE_BIT                   USB_BIT_00 /*  In Slave,  Wait Control HREADY = 0b */

#define     USB_AHBSCTR_INIT                   ( USB_AHBSCTR_NOT_RETRY_MASTER_ALL_OFF \
                                                        | USB_AHBSCTR_WAIT_MODE_BIT )

#define     USB_AHBSCTR_NOT_RETRY_MASTER_ALL_OFF       (0x00000000)    /*  default value */

/**
 * AHBMCTR Register
 */
#define    USB_AHBMCTR_REG                 (*(volatile unsigned long *)(USB_TC_OFFSET + 0x1004))


#define    USB_AHBMCTR_ARBITER_CTR_BIT                 USB_BIT_31
#define    USB_AHBMCTR_MCYCLE_RST_BIT                  USB_BIT_12
#define    USB_AHBMCTR_ENDIAN_CTR_1_BIT                USB_BIT_09
#define    USB_AHBMCTR_ENDIAN_CTR_0_BIT                USB_BIT_08
#define    USB_AHBMCTR_HBUSREQ_MODE_BIT                USB_BIT_05
#define    USB_AHBMCTR_HTRANS_MODE_BIT                 USB_BIT_04
#define    USB_AHBMCTR_WBURST_TYPE_BIT                 USB_BIT_02
#define    USB_AHBMCTR_BURST_TYPE_1_BIT                USB_BIT_01
#define    USB_AHBMCTR_BURST_TYPE_0_BIT                USB_BIT_00

#define     USB_AHBMCTR_INIT                   ( USB_AHBMCTR_HBUSREQ_MODE_BIT \
                                                        | USB_AHBMCTR_HTRANS_MODE_BIT \
                                                        | USB_AHBMCTR_WBURST_TYPE_BIT )            /* AHBMCTR default value */
/**
 * AHBBINT Register
 */
#define    USB_AHBBINT_REG                 (*(volatile unsigned long *)(USB_TC_OFFSET + 0x1008))

#define    USB_AHBBINT_DMA_ENDINT_15_BIT               USB_BIT_31
#define    USB_AHBBINT_DMA_ENDINT_14_BIT               USB_BIT_30
#define    USB_AHBBINT_DMA_ENDINT_13_BIT               USB_BIT_29
#define    USB_AHBBINT_DMA_ENDINT_12_BIT               USB_BIT_28
#define    USB_AHBBINT_DMA_ENDINT_11_BIT               USB_BIT_27
#define    USB_AHBBINT_DMA_ENDINT_10_BIT               USB_BIT_26
#define    USB_AHBBINT_DMA_ENDINT_9_BIT                USB_BIT_25
#define    USB_AHBBINT_DMA_ENDINT_8_BIT                USB_BIT_24
#define    USB_AHBBINT_DMA_ENDINT_7_BIT                USB_BIT_23
#define    USB_AHBBINT_DMA_ENDINT_6_BIT                USB_BIT_22
#define    USB_AHBBINT_DMA_ENDINT_5_BIT                USB_BIT_21
#define    USB_AHBBINT_DMA_ENDINT_4_BIT                USB_BIT_20
#define    USB_AHBBINT_DMA_ENDINT_3_BIT                USB_BIT_19
#define    USB_AHBBINT_DMA_ENDINT_2_BIT                USB_BIT_18
#define    USB_AHBBINT_DMA_ENDINT_1_BIT                USB_BIT_17
#define    USB_AHBBINT_VBUS_INT_BIT                    USB_BIT_13
#define    USB_AHBBINT_MBUS_ERRINT_BIT                 USB_BIT_06
#define    USB_AHBBINT_SBUS_ERRINT1_BIT                USB_BIT_05
#define    USB_AHBBINT_SBUS_ERRINT0_BIT                USB_BIT_04
#define    USB_AHBBINT_ERR_MASTER_3_BIT                USB_BIT_03
#define    USB_AHBBINT_ERR_MASTER_2_BIT                USB_BIT_02
#define    USB_AHBBINT_ERR_MASTER_1_BIT                USB_BIT_01
#define    USB_AHBBINT_ERR_MASTER_0_BIT                USB_BIT_00


#define     USB_AHBBINT_BUSERR_MASK                    (USB_AHBBINT_MBUS_ERRINT_BIT \
                                                            | USB_AHBBINT_SBUS_ERRINT1_BIT \
                                                            | USB_AHBBINT_SBUS_ERRINT0_BIT)


#define     USB_AHBBINT_CHECK_MASK                     ( USB_AHBBINT_DMA_ENDINT_2_BIT \
                                                            | USB_AHBBINT_DMA_ENDINT_1_BIT \
                                                            | USB_AHBBINT_BUSERR_MASK )

/**
 * AHBBINTEN Register
 */
#define    USB_AHBBINTEN_REG                   (*(volatile unsigned long *)(USB_TC_OFFSET + 0x100C))

#define    USB_AHBBINTEN_DMA_ENDINTEN_15_BIT           USB_BIT_31 /*  for EP15 DAM comp interrupt */
#define    USB_AHBBINTEN_DMA_ENDINTEN_14_BIT           USB_BIT_30 /*  for EP14 DAM comp interrupt */
#define    USB_AHBBINTEN_DMA_ENDINTEN_13_BIT           USB_BIT_29 /*  for EP13 DAM comp interrupt */
#define    USB_AHBBINTEN_DMA_ENDINTEN_12_BIT           USB_BIT_28 /*  for EP12 DAM comp interrupt */
#define    USB_AHBBINTEN_DMA_ENDINTEN_11_BIT           USB_BIT_27 /*  for EP11 DAM comp interrupt */
#define    USB_AHBBINTEN_DMA_ENDINTEN_10_BIT           USB_BIT_26 /*  for EP10 DAM comp interrupt */
#define    USB_AHBBINTEN_DMA_ENDINTEN_9_BIT            USB_BIT_25 /*  for EP9 DAM comp interrupt */
#define    USB_AHBBINTEN_DMA_ENDINTEN_8_BIT            USB_BIT_24 /*  for EP8 DAM comp interrupt*/
#define    USB_AHBBINTEN_DMA_ENDINTEN_7_BIT            USB_BIT_23 /*  for EP7 DAM comp interrupt*/
#define    USB_AHBBINTEN_DMA_ENDINTEN_6_BIT            USB_BIT_22 /*  for EP6 DAM comp interrupt*/
#define    USB_AHBBINTEN_DMA_ENDINTEN_5_BIT            USB_BIT_21 /*  for EP5 DAM comp interrupt*/
#define    USB_AHBBINTEN_DMA_ENDINTEN_4_BIT            USB_BIT_20 /*  for EP4 DAM comp interrupt*/
#define    USB_AHBBINTEN_DMA_ENDINTEN_3_BIT            USB_BIT_19 /*  for EP3 DAM comp interrupt*/
#define    USB_AHBBINTEN_DMA_ENDINTEN_2_BIT            USB_BIT_18 /*  for EP2 DAM comp interrupt*/
#define    USB_AHBBINTEN_DMA_ENDINTEN_1_BIT            USB_BIT_17 /*  for EP1 DAM comp interrupt*/

#define    USB_AHBBINTEN_VBUS_INTEN_BIT                USB_BIT_13 /*  for VBUS */
#define    USB_AHBBINTEN_MBUS_ERRINTEN_BIT             USB_BIT_06 /*  for AHB master */
#define    USB_AHBBINTEN_SBUS_ERRINT1EN_BIT            USB_BIT_05 /*  for AHBSCTR Regiter */
#define    USB_AHBBINTEN_SBUS_ERRINT0EN_BIT            USB_BIT_04 /*  for AHB slave bus error */

#define     USB_AHBBINTEN_INIT                         ( USB_AHBBINTEN_DMA_ENDINTEN_2_BIT \
                                                            | USB_AHBBINTEN_DMA_ENDINTEN_1_BIT \
                                                            | USB_AHBBINTEN_VBUS_INTEN_BIT \
                                                            | USB_AHBBINTEN_MBUS_ERRINTEN_BIT \
                                                            | USB_AHBBINTEN_SBUS_ERRINT1EN_BIT \
                                                            | USB_AHBBINTEN_SBUS_ERRINT0EN_BIT )

#define     USB_AHBBINTEN_ALL_DISABLE      (0)


/**
 * EPCTR Register
 */
#define    USB_EPCTR_REG                       (*(volatile unsigned long *)(USB_TC_OFFSET + 0x1010))

#define    USB_EPCTR_DIRPD_BIT                         USB_BIT_12 /*  direct power down */
#define    USB_EPCTR_VBUS_LEVEL_BIT                    USB_BIT_08 /*  VBUS = 1    */
#define    USB_EPCTR_PLL_RESUME_BIT                    USB_BIT_05 /*  resume pll  */
#define    USB_EPCTR_PLL_LOCK_BIT                      USB_BIT_04 /*  PLL locked  */
#define    USB_EPCTR_EPC_RST_BIT                       USB_BIT_00 /*  EPC reset   */

/**
 * EPTST Register
 */
#define    USB_EPTST_REG                       (*(volatile unsigned long *)(USB_TC_OFFSET + 0x1014))

#define    USB_EPTST_PHY_TST_BIT                       USB_BIT_01
#define    USB_EPTST_PHY_TSTCLK_BIT                    USB_BIT_00

/**
 * USBSSVER Register
 */
#define    USB_USBSSVER_REG                    (*(volatile unsigned long *)(USB_TC_OFFSET + 0x1020))

#define    USB_USBSSVER_AHBB_VER_7_BIT                 USB_BIT_23
#define    USB_USBSSVER_AHBB_VER_6_BIT                 USB_BIT_22
#define    USB_USBSSVER_AHBB_VER_5_BIT                 USB_BIT_21
#define    USB_USBSSVER_AHBB_VER_4_BIT                 USB_BIT_20
#define    USB_USBSSVER_AHBB_VER_3_BIT                 USB_BIT_19
#define    USB_USBSSVER_AHBB_VER_2_BIT                 USB_BIT_18
#define    USB_USBSSVER_AHBB_VER_1_BIT                 USB_BIT_17
#define    USB_USBSSVER_AHBB_VER_0_BIT                 USB_BIT_16
#define    USB_USBSSVER_EPC_VER_7_BIT                  USB_BIT_15
#define    USB_USBSSVER_EPC_VER_6_BIT                  USB_BIT_14
#define    USB_USBSSVER_EPC_VER_5_BIT                  USB_BIT_13
#define    USB_USBSSVER_EPC_VER_4_BIT                  USB_BIT_12
#define    USB_USBSSVER_EPC_VER_3_BIT                  USB_BIT_11
#define    USB_USBSSVER_EPC_VER_2_BIT                  USB_BIT_10
#define    USB_USBSSVER_EPC_VER_1_BIT                  USB_BIT_09
#define    USB_USBSSVER_EPC_VER_0_BIT                  USB_BIT_08
#define    USB_USBSSVER_SS_VER_7_BIT                   USB_BIT_07
#define    USB_USBSSVER_SS_VER_6_BIT                   USB_BIT_06
#define    USB_USBSSVER_SS_VER_5_BIT                   USB_BIT_05
#define    USB_USBSSVER_SS_VER_4_BIT                   USB_BIT_04
#define    USB_USBSSVER_SS_VER_3_BIT                   USB_BIT_03
#define    USB_USBSSVER_SS_VER_2_BIT                   USB_BIT_02
#define    USB_USBSSVER_SS_VER_1_BIT                   USB_BIT_01
#define    USB_USBSSVER_SS_VER_0_BIT                   USB_BIT_00

/**
 * USBSSCONF Register
 */
#define    USB_USBSSCONF_REG                   (*(volatile unsigned long *)(USB_TC_OFFSET + 0x1024))

#define    USB_USBSSCONF_EP_AVAILABLE_15_BIT           USB_BIT_31
#define    USB_USBSSCONF_EP_AVAILABLE_14_BIT           USB_BIT_30
#define    USB_USBSSCONF_EP_AVAILABLE_13_BIT           USB_BIT_29
#define    USB_USBSSCONF_EP_AVAILABLE_12_BIT           USB_BIT_28
#define    USB_USBSSCONF_EP_AVAILABLE_11_BIT           USB_BIT_27
#define    USB_USBSSCONF_EP_AVAILABLE_10_BIT           USB_BIT_26
#define    USB_USBSSCONF_EP_AVAILABLE_9_BIT            USB_BIT_25
#define    USB_USBSSCONF_EP_AVAILABLE_8_BIT            USB_BIT_24
#define    USB_USBSSCONF_EP_AVAILABLE_7_BIT            USB_BIT_23
#define    USB_USBSSCONF_EP_AVAILABLE_6_BIT            USB_BIT_22
#define    USB_USBSSCONF_EP_AVAILABLE_5_BIT            USB_BIT_21
#define    USB_USBSSCONF_EP_AVAILABLE_4_BIT            USB_BIT_20
#define    USB_USBSSCONF_EP_AVAILABLE_3_BIT            USB_BIT_19
#define    USB_USBSSCONF_EP_AVAILABLE_2_BIT            USB_BIT_18
#define    USB_USBSSCONF_EP_AVAILABLE_1_BIT            USB_BIT_17
#define    USB_USBSSCONF_EP_AVAILABLE_0_BIT            USB_BIT_16
#define    USB_USBSSCONF_DMA_AVAILABLE_15_BIT          USB_BIT_15
#define    USB_USBSSCONF_DMA_AVAILABLE_14_BIT          USB_BIT_14
#define    USB_USBSSCONF_DMA_AVAILABLE_13_BIT          USB_BIT_13
#define    USB_USBSSCONF_DMA_AVAILABLE_12_BIT          USB_BIT_12
#define    USB_USBSSCONF_DMA_AVAILABLE_11_BIT          USB_BIT_11
#define    USB_USBSSCONF_DMA_AVAILABLE_10_BIT          USB_BIT_10
#define    USB_USBSSCONF_DMA_AVAILABLE_9_BIT           USB_BIT_09
#define    USB_USBSSCONF_DMA_AVAILABLE_8_BIT           USB_BIT_08
#define    USB_USBSSCONF_DMA_AVAILABLE_7_BIT           USB_BIT_07
#define    USB_USBSSCONF_DMA_AVAILABLE_6_BIT           USB_BIT_06
#define    USB_USBSSCONF_DMA_AVAILABLE_5_BIT           USB_BIT_05
#define    USB_USBSSCONF_DMA_AVAILABLE_4_BIT           USB_BIT_04
#define    USB_USBSSCONF_DMA_AVAILABLE_3_BIT           USB_BIT_03
#define    USB_USBSSCONF_DMA_AVAILABLE_2_BIT           USB_BIT_02
#define    USB_USBSSCONF_DMA_AVAILABLE_1_BIT           USB_BIT_01
#define    USB_USBSSCONF_DMA_AVAILABLE_0_BIT           USB_BIT_00

/**
 * EP1DCR1 Register
 */
#define    USB_EP1DCR1_REG                 (*(volatile unsigned long *)(USB_TC_OFFSET + 0x1110))
#define    USB_EP2DCR1_REG                 (*(volatile unsigned long *)(USB_TC_OFFSET + 0x1120))

#define    USB_EPDCR1_EPDMACNT_7_BIT                   USB_BIT_23
#define    USB_EPDCR1_EPDMACNT_6_BIT                   USB_BIT_22
#define    USB_EPDCR1_EPDMACNT_5_BIT                   USB_BIT_21
#define    USB_EPDCR1_EPDMACNT_4_BIT                   USB_BIT_20
#define    USB_EPDCR1_EPDMACNT_3_BIT                   USB_BIT_19
#define    USB_EPDCR1_EPDMACNT_2_BIT                   USB_BIT_18
#define    USB_EPDCR1_EPDMACNT_1_BIT                   USB_BIT_17
#define    USB_EPDCR1_EPDMACNT_0_BIT                   USB_BIT_16
#define    USB_EPDCR1_EPDMADIR0_BIT                    USB_BIT_01
#define    USB_EPDCR1_EPDMAREQEN_BIT                   USB_BIT_00

#define    USB_EPDCR1_EPDMACNT_MASK                    0x00FF0000

/**
 * EPnDCR2 Register
 */
#define    USB_EP1DCR2_REG                 (*(volatile unsigned long *)(USB_TC_OFFSET + 0x1114))
#define    USB_EP2DCR2_REG                 (*(volatile unsigned long *)(USB_TC_OFFSET + 0x1124))

#define    USB_EPDCR2_EPLMPKT_10_BIT                   USB_BIT_26
#define    USB_EPDCR2_EPLMPKT_9_BIT                    USB_BIT_25
#define    USB_EPDCR2_EPLMPKT_8_BIT                    USB_BIT_24
#define    USB_EPDCR2_EPLMPKT_7_BIT                    USB_BIT_23
#define    USB_EPDCR2_EPLMPKT_6_BIT                    USB_BIT_22
#define    USB_EPDCR2_EPLMPKT_5_BIT                    USB_BIT_21
#define    USB_EPDCR2_EPLMPKT_4_BIT                    USB_BIT_20
#define    USB_EPDCR2_EPLMPKT_3_BIT                    USB_BIT_19
#define    USB_EPDCR2_EPLMPKT_2_BIT                    USB_BIT_18
#define    USB_EPDCR2_EPLMPKT_1_BIT                    USB_BIT_17
#define    USB_EPDCR2_EPLMPKT_0_BIT                    USB_BIT_16
#define    USB_EPDCR2_EPMPKT_10_BIT                    USB_BIT_10
#define    USB_EPDCR2_EPMPKT_9_BIT                     USB_BIT_09
#define    USB_EPDCR2_EPMPKT_8_BIT                     USB_BIT_08
#define    USB_EPDCR2_EPMPKT_7_BIT                     USB_BIT_07
#define    USB_EPDCR2_EPMPKT_6_BIT                     USB_BIT_06
#define    USB_EPDCR2_EPMPKT_5_BIT                     USB_BIT_05
#define    USB_EPDCR2_EPMPKT_4_BIT                     USB_BIT_04
#define    USB_EPDCR2_EPMPKT_3_BIT                     USB_BIT_03
#define    USB_EPDCR2_EPMPKT_2_BIT                     USB_BIT_02
#define    USB_EPDCR2_EPMPKT_1_BIT                     USB_BIT_01
#define    USB_EPDCR2_EPMPKT_0_BIT                     USB_BIT_00

#define    USB_EPDCR2_EPMPKT_MASK                  (~ (USB_EPDCR2_EPMPKT_10_BIT \
                                                            | USB_EPDCR2_EPMPKT_9_BIT \
                                                            | USB_EPDCR2_EPMPKT_8_BIT \
                                                            | USB_EPDCR2_EPMPKT_7_BIT \
                                                            | USB_EPDCR2_EPMPKT_6_BIT \
                                                            | USB_EPDCR2_EPMPKT_5_BIT \
                                                            | USB_EPDCR2_EPMPKT_4_BIT \
                                                            | USB_EPDCR2_EPMPKT_3_BIT \
                                                            | USB_EPDCR2_EPMPKT_2_BIT \
                                                            | USB_EPDCR2_EPMPKT_1_BIT \
                                                            | USB_EPDCR2_EPMPKT_0_BIT) )

/**
 * EP1TADR Register
 */
#define    USB_EP1TADR_REG                 (*(volatile unsigned long *)(USB_TC_OFFSET + 0x1118))
#define    USB_EP2TADR_REG                 (*(volatile unsigned long *)(USB_TC_OFFSET + 0x1128))



/**
 *EP3~EP6
 */


/**
 * Interrupt bitmap
 */
#define    USB_BRIDGE_INT      (0x0001)
#define    USB_EPC_INT         (0x0002)
#define    USB_VBUS_TASK_INT   (0x0010)
#define    USB_SUSPEND_REQ_INT (0x0020)
#define    USB_SEQUENCE_CTRL_INT   (0x0040)

/**
 * Number Endpoint
 */
#define    USB_NUM_ENDPOINTS   (0x04)
#define    USB_OUT_ENDPOINT    (0x00000003)    /* OUT EP bit mask EP0(bit0),EP1(bit1) */
#define    USB_IN_ENDPOINT     (0x00050000)    /* IN EP bit mask EP0(bit16),EP2(bit18) */


#define    USB_STACK_SIZE      (4096)

#define    USB_FUNC_TIMEOUT    1000  /* millisecond. */

/**
 * Endpoint Index
 */
#define    USB_EP0_RX              (0x00)  /* EndPoint Number bit4:0-RX,1-TX,bit3-0:EndPoint Number */
#define    USB_EP0_TX              (0x01)  /* bit3-0:serial number, bit7-4:EndPoint Number */
#define    USB_EP1_RX              (0x12)
#define    USB_EP2_TX              (0x23)
#define    USB_EP_INDEX_MASK       (0xf0)  /* from Endpoint index */
#define    USB_EP_SERIAL_NUM_MASK  (0x0f)

/**
 * Endpoint Pending
 */
#define    USB_PENDING_EP0RX          (0x0001)
#define    USB_PENDING_EP0TX          (0x0002)
#define    USB_PENDING_EP1RX          (0x0004)
#define    USB_PENDING_EP2TX          (0x0008)

/**
 * check SP_MODE bit max polling count
 */
#define    USB_SPMODE_POLLING_COUNT   (8000)

/**
 * Gatedclock Control
 */
#define    USB_GATEDCLOCK_ENABLE           1       /* GatedClockControl enable */
#define    USB_GATEDCLOCK_DISABLE          0       /* GatedClockControl disable */

/** 
*standard USB feature selectors
*/

#define USB_FSEL_DEV_REMOTE_WAKEUP	1
#define USB_FSEL_DEV_TEST_MODE      2
#define USB_FSEL_DEV_ENDPOINT_HALT	0


/**
 * USBTESTMODE
 */
#define    USB_TESTMODE_SIFT                       16
#define    USB_AHBBINT_DMA_ENDINT_MASKDE_NORMAL    (0x00)
#define    USB_USB_TESTMODE_TEST_J                 (0x01)
#define    USB_USB_TESTMODE_TEST_K                 (0x02)
#define    USB_USB_TESTMODE_TEST_SE0_NAK           (0x03)
#define    USB_USB_TESTMODE_TEST_PACKET            (0x04)
#define    USB_USB_TESTMODE_TEST_FORCE_EN          (0x05)

/**
 * Setup packet analysis
 */
#define     USB_SETCONFIGRATION        (0x09)      /* No Data Control */
#define     USB_SETINTERFACE           (0x0B)      /* No Data Control */
#define     USB_CLEAR_FEATURE          (0x01)      /* No Data Control */
#define     USB_GET_CONFIGURATION      (0x08)      /* control read */
#define     USB_GET_DESCRIPTOR         (0x06)      /* control read */
#define     USB_GET_INTERFACE          (0x0A)      /* control read */
#define     USB_GET_STATUS             (0x00)      /* control read */
#define     USB_SET_ADDRESS            (0x05)      /* No Data control */
#define     USB_SET_DESCRIPTOR         (0x07)      /* control write */
#define     USB_SET_FEATURE            (0x03)      /* No Data control */
#define     USB_SYNC_FRAME             (0x0C)      /* control read */
#define     USB_GET_MAX_LUN            (0xFE)      /* get max logical unit number */

/**
 * max paket size
 */
#define     USB_MAXPACKET_HIGH         (512)
#define     USB_MAXPACKET_FULL         (64)

/**
 * Standard Request Element
 */
#define     USB_BREQUEST               (1)
#define     USB_WVALUE                 (2)

/**
 * Standard Request Size
 */
#define     USB_STANDERD_REQUEST_SIZE  (8)

/**
 * Endpoint operation define
 */
#define     USB_MAXPSIZE_64            (0x40)      /* EP0 */
#define     USB_EP_PENDING_MASK        (0x000F)
#define     USB_EP0RX_MASK             (0xFFFFFFDF)/*EP0RX interrupt = 0*/
#define     USB_EP_BCLR_MAX_LOOP       (200)   /* EP buffer clear loop count */
#define     USB_EP_READ_REG_SIZE_BYTE  (4)     /* EP READ register size 4Byte */
#define     USB_EP_WRITE_REG_SIZE_BYTE (4)     /* EP WRITE register size 4Byte */
#define     USB_EP_DW_BIT              (5)     /* EP0 Control Reg. DW bit position*/

/**
 * error
 */
#define     USB_MBUSERR_STR            "V1TYPEF MBUSERR"
#define     USB_SBUSERR1_STR           "V1TYPEF SBUSERR1"
#define     USB_SBUSERR2_STR           "V1TYPEF SBUSERR2"

/**
 *  SelfDiagnose
 */
#define     USB_SCRATCH_PAD_RST    (0x00000000)
#define     USB_SCRATCH_PAD_ALL_F  (0xFFFFFFFF)


/* Interrupt status masks */
#define USB_DISCONNECT_INTERRUPT_MASK    0x01  /* disconnect interrupt mask */
#define USB_RESET_INTERRUPT_MASK         0x02  /* reset interrupt mask */
#define USB_SUSPEND_INTERRUPT_MASK       0x04  /* suspend interrupt mask */
#define USB_RESUME_INTERRUPT_MASK        0x08  /* resume interrupt mask */
#define USB_ENDPOINT_INTERRUPT_MASK      0x10  /* endpoint interrupt mask */

/* Endpoint interrupt status masks */
#define USB_ENDPOINT_INTERRUPT_PENDING_MASK  0x01  /* endpoint interrupt */

/* pending mask */
#define USB_ENDPOINT_SETUP_PID_MASK          0x02  /* setup interrupt */
#define USB_ENDPOINT_OUT_PID_MASK            0x04  /* out interrupt mask */
#define USB_ENDPOINT_IN_PID_MASK             0x08  /* in interrupt mask */
#define USB_ENDPOINT_TRANSFER_STATUS_MASK    0x0FF0  /* endpoint transfer */
              /* status mask */

/* Endpoint interrupt status error code values */
#define USB_ENDPOINT_TRANSFER_SUCCESS        0x0000  /* transfer success */
#define USB_ENDPOINT_DATA_TOGGLE_ERROR       0x00E0  /* data toggle error */
#define USB_ENDPOINT_PID_MISMATCH            0x00F0  /* PID mismatch */
#define USB_ENDPOINT_COMMN_FAULT             0x0100  /*Communication Fault*/
#define USB_ENDPOINT_STALL_ERROR             0x0110  /* stall error */
#define USB_ENDPOINT_DATA_OVERRUN            0x0120  /* data overrun */
#define USB_ENDPOINT_DATA_UNDERRUN           0x0130  /* data underrun */
#define USB_ENDPOINT_CRC_ERROR               0x0140  /* CRC error */
#define USB_ENDPOINT_TIMEOUT_ERROR           0x0150  /* Time Out error */
#define USB_ENDPOINT_BIT_STUFF_ERROR         0x0160  /* bit stuff error */


#endif    /* __INCusbV1TypeF2h */
