/*
 * Copyright (C) 2012 Sony Corporation 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License, version 2, as 
 * published by the Free Software Foundation. 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/kthread.h>
#include <linux/dma-mapping.h>
#include <linux/completion.h>
#include <linux/usb.h>
#include <linux/usb/gadget.h>
#include <mach/udc_mass_priority.h>
#include "cxd9225gg_udc.h"
#include "cxd9225gg_usbExt.h"
#include "cxd9225gg_usbEvalLib.h"
#include <mach/cxd9225gg_IO.h>
#include <linux/io.h>
#include <mach/cxd9225gg_types.h>

#define _SYSCON_VIRT(addr)  ((addr)-CXD9225GG_AXI_REG_PHYS_BASE+CXD9225GG_AXI_REG_VIRT_BASE)
#define RB_CXD9225GG_HWVER_CXD9225GG    (0)  /* HW version ID */

#define cust_printl(level, format, arg...) \
		printk(level "[CUSTOM] %s() %d : " format "\n",__func__,__LINE__,##arg)

#define BOOT_BCT_HWCONFIG_DISABLE_USBDEVICE        (0) /**< Disable USB Device */
#define BOOT_BCT_HWCONFIG_ENABLE_USBDEVICE         (1) /**< Enable USB Device */

typedef struct
    {
    UINT8  intNum;		/* Interrupt number */
    UINT8  EdgeOrLevel;		/* Edge or Level */
    UINT32 errCode;		/* Error code */
    } CXD9225GG_ERRINT;

//==============================================================================

UINT8 cxd9225ggBspBctConfigurationDataGet(UINT32 bctId)
{
	return 0;
}
EXPORT_SYMBOL(cxd9225ggBspBctConfigurationDataGet);

static int __init cxd9225gg_syscon_error_handlers(void)
{
	return 0;
}
arch_initcall(cxd9225gg_syscon_error_handlers);


//==============================================================================

static int usbGatedClockOn(void)
{
    /* the 3rd byte of CXD9225GG_SYSC_GCTS2, see datasheet of SYSC registers */
    writeb(1, CXD9225GG_SYSC_GCTS2+2);
    return 0;
}

static int usbGatedClockOff(void)
{
    /* the 3rd byte of CXD9225GG_SYSC_GCTS2, see datasheet of SYSC registers */
    writeb(0, CXD9225GG_SYSC_GCTS2+2);
    return 0;
}

static int  s_bInitialized = 0;

USBTARGET         g_UsbTargetInfo;

const PUSBTARGET  g_UsbpTarget = &g_UsbTargetInfo;

USBENDPOINT       g_UsbEndpoint0Recv;
USBENDPOINT       g_UsbEndpoint0Send;
USBENDPOINT       g_UsbEndpoint1Recv;
USBENDPOINT       g_UsbEndpoint2Send;

const PUSBENDPOINT    g_UsbpEndpoint[] =
{
    &g_UsbEndpoint0Recv ,
    &g_UsbEndpoint0Send ,
    &g_UsbEndpoint1Recv ,
    &g_UsbEndpoint2Send
};

/* externs */
extern unsigned short sysHwVer(int unit);
extern int usbEpcReset(void);
extern int usbBridgeInit(void);

/* forward declaration */
int usbHwInitialize(void);

USBTARGET *usbInitialize(void);

int usbConnectionNotify(int plug_flag);
unsigned long usbVersionGet(void);
unsigned int usbConnectStatusGet(void);
unsigned int usbEnableFlgGet(void);

int usbSelfDiagnose(int SDFlg);

void usbBulkOutRemainSizeGet(int *dmaResult,
                                        unsigned long *dmaRemainSize);

void usbBulkInRemainSizeGet(int *dmaResult,
                                       unsigned long *dmaRemainSize);


int usbHwInitialize(void)
{
    int    ret = 0;
    unsigned long    ulLoopCnt;
    unsigned char    usbDeviceValue = BOOT_BCT_HWCONFIG_DISABLE_USBDEVICE;

    udc_error("please use cxd9225ggBspBctConfigurationDataGet to get usb-config. \n");

    usbDeviceValue = BOOT_BCT_HWCONFIG_ENABLE_USBDEVICE;

    /* USBDevice */
    if(BOOT_BCT_HWCONFIG_DISABLE_USBDEVICE == usbDeviceValue)
    {
        udc_error("system can NOT support USB Client fucntion, so return directly. \n");
        return 1; //attention...
    }

    if ( usbGatedClockOff() != 0 )
    {
        udc_error(" [FATAL]usbGatedClockOff() disable Fail. \n");
    }


    /* EPC -- reset */
    ret = usbEpcReset();
    if (ret)
    {
        udc_error(" usbEpcReset() failed. \n");
        return -1;
    }

    /* wait for PLL for EPC to be stable. */
    for (ulLoopCnt = 0; ulLoopCnt < 100; ulLoopCnt++)
    {
        if (USB_EPCTR_PLL_LOCK_BIT == (USB_EPCTR_REG & USB_EPCTR_PLL_LOCK_BIT))
        {
            udc_debug("EPC PLL is locked. \n");
            break;
        }
    }

    /* whether is it unstable.  */
    if (ulLoopCnt >= 100)
    {
        udc_error(" FATAL ERROR -- PLL Lock Wait TIME OUT... \n");
        return -1;
    }

    /* disable all. */
    USB_USBINTEN_REG = 0;  /* all mask EPC  */
    USB_EP0INTEN_REG = 0;  /* all mask EPC  */
    USB_EP1INTEN_REG = 0;  /* all mask EPC  */
    USB_EP2INTEN_REG = 0;  /* all mask EPC  */
    USB_AHBBINTEN_REG = 0; /* all mask bridge  */

    /* Bridge -- initialize. */
    ret = usbBridgeInit();
    if (ret)
    {
        udc_error(" usbBridgeInit() failed. \n");
        return -1;
    }

    udc_debug("exit -- OK. \n");

    return 0;
}

int usbHwReInit(void)
{
    int    ret = 0;
    unsigned char    usbDeviceValue = BOOT_BCT_HWCONFIG_DISABLE_USBDEVICE;

    usbDeviceValue = BOOT_BCT_HWCONFIG_ENABLE_USBDEVICE;

    /* USBDevice */
    if(BOOT_BCT_HWCONFIG_DISABLE_USBDEVICE == usbDeviceValue)
    {
        udc_error("system can NOT support USB Client fucntion, so return directly. \n");
        return 1;
    }
    if ( usbGatedClockOff() != 0 )
    {
        udc_error(" [FATAL]usbGatedClockOff() disable Fail. \n");
    } 

    /* EPC -- reset */
    ret = usbEpcReset();
    if (ret)
    {
        udc_error(" usbEpcReset() failed. \n");
        return -1;
    }
    USB_AHBBINTEN_REG = 0; /* all mask bridge  */

    /* Bridge -- initialize. */
    ret = usbBridgeInit();
    if (ret)
    {
        udc_error(" usbBridgeInit() failed. \n");
        return -1;
    }

    udc_debug("exit -- OK. \n");
    return 0;
}

void usbHwTerminate(void)
{
    udc_debug("entry test. \n");

    udc_error("attention: you should do something here. \n");

    return;
}


struct usbTarget *usbInitialize(void)
{
    PUSBTARGET       pTarget      = NULL;
    PUSBENDPOINT     pEndpoint    = NULL;
    unsigned long               uloopCount;
    int                         ret_val;

    /* whether is it initialized. */
    if (s_bInitialized)
    {
        udc_error(" already initialized, and return directly. \n");
        return g_UsbpTarget;
    }
    s_bInitialized = 1;

    /* initialize the related interrupt sources. */
    ret_val = usbIntcSetting();
    if (ret_val)
    {
        udc_error("usbIntcSetting() failed. \n");
        return NULL;
    }

//attention... we need another type of notification-method to do it...

    pTarget = g_UsbpTarget;
    memset(pTarget, 0, sizeof( *pTarget ) );

    spin_lock_init(&pTarget->lock);

    atomic_clear_mask(0xff, (unsigned long *)&(pTarget->thread_exit));

    /* initialize all ep descriptor. */
    for (uloopCount = 0; uloopCount < USB_NUM_ENDPOINTS; uloopCount++)
    {
        pEndpoint = g_UsbpEndpoint[uloopCount];
        memset( pEndpoint, 0, sizeof( *pEndpoint ) );
    }

    /* get the head of ep-list. */
    pTarget->pEndPointInfoTop = (PUSBENDPOINT *)&g_UsbpEndpoint[0];

    init_completion(&(pTarget->sUsbIntSemId));

    /* create the task for usb isr handling. */
    pTarget->hThd_UsbIsr = kthread_run(usbIsrExec,
                                       pTarget,
                                       "usb-isr");
    if (IS_ERR(pTarget->hThd_UsbIsr))
    {
        udc_error("can NOT create kthread for USB's IsrExec(). \n");
        return NULL;
    }

    /* attention. set preemption and its priority. */
    #ifdef CXD9225GG_TASK_PRIORITY_UDC_ISR
    {
        struct sched_param tmp_param;
        tmp_param.sched_priority = CXD9225GG_TASK_PRIORITY_UDC_ISR;
        sched_setscheduler(pTarget->hThd_UsbIsr, SCHED_FIFO, &tmp_param);
    }
    #endif /* CXD9225GG_TASK_PRIORITY_UDC_ISR */

    /* semaphore for VBus-Wait-task. */
    init_completion(&(pTarget->hSem4VbusWaitTask));

    /* create the vbus-wait-task. */
    pTarget->hThd4VbusWait = kthread_run(usbVbusWaitTask,
                                         pTarget, //g_UsbpTarget,
                                         "usb-vbus");
    if (IS_ERR(pTarget->hThd4VbusWait))
    {
        udc_error("can NOT create kthread for VbusWaitTask(). \n");
        return NULL;
    }

    /* attention. set preemption and its priority. */
    #ifdef CXD9225GG_TASK_PRIORITY_UDC_VBUS
    {
        struct sched_param tmp_param;
        tmp_param.sched_priority = CXD9225GG_TASK_PRIORITY_UDC_VBUS;
        sched_setscheduler(pTarget->hThd4VbusWait, SCHED_FIFO, &tmp_param);
    }
    #endif /* CXD9225GG_TASK_PRIORITY_UDC_VBUS */

    /* API mutex. */
    mutex_init(&(pTarget->hSemApiExecuting));

    /* attention -- full. */
    init_completion(&(pTarget->sUsbPlugInOutSemId));
    complete(&(pTarget->sUsbPlugInOutSemId));

    pTarget->uTransitVersion = sysHwVer(RB_CXD9225GG_HWVER_CXD9225GG);

    udc_debug(" cxd9225gg--HW-version: 0x%0x. \n", pTarget->uTransitVersion);

    pTarget->usbPlugFlg = 0;
    pTarget->usbEnableFlg = 0;

    /* whether is it connected to vbus now... */
    if( USB_EPCTR_VBUS_LEVEL_BIT != (USB_EPCTR_REG & USB_EPCTR_VBUS_LEVEL_BIT) )
    {
        udc_debug(" vbus is NOT connected now, so init power-off. \n");

        ret_val = usbPowerStateControl(&pTarget->ulPowState,
                                                  USB_POW_EVENT_INIT_OFF,
                                                  &pTarget->bIsPllUnlocked);
    }
    else
    {
        udc_debug(" vbus is connected now, so init power-on. \n");

        ret_val = usbPowerStateControl(&pTarget->ulPowState,
                                                  USB_POW_EVENT_INIT_ON,
                                                  &pTarget->bIsPllUnlocked);
        pTarget->usbConnectStatus = USB_CONNECTED_STATUS;
    }

    if (ret_val != 0)
    {
        udc_error("[FATAL] Power state change error !!\n");
        return NULL;
    }

    udc_debug(" exit--SUCCESS. \n");

    return pTarget;
}


int usbReInit(PUSBTARGET       pTarget)
{
    PUSBENDPOINT     pEndpoint    = NULL;
    unsigned long               uloopCount;
    int                         ret_val;

    /* initialize all ep descriptor. */
    for (uloopCount = 0; uloopCount < USB_NUM_ENDPOINTS; uloopCount++)
    {
        pEndpoint = g_UsbpEndpoint[uloopCount];
        memset( pEndpoint, 0, sizeof( *pEndpoint ) );
    }

    /* get the head of ep-list. */
    pTarget->pEndPointInfoTop = (PUSBENDPOINT *)&g_UsbpEndpoint[0];

    pTarget->udc_dev->gadget.dev.dma_mask = 0;

//do it out of this function, since there would be some config-otg...

    /* ep0 is special; put it right after the SETUP buffer */
    /* for ep0 rx/tx. */
    ret_val = cxd9225gg_ep_setup(pTarget->udc_dev, 0);
    if (ret_val)
    {
        udc_error("cxd9225gg_ep_setup() failed for ep0. \n");
        return ret_val;
    }

/* in fact, initially we should disable all non-ep0 endpoints */

    /* for ep1 rx. */
    ret_val = cxd9225gg_ep_setup(pTarget->udc_dev, 1);
    if (ret_val)
    {
        udc_error("cxd9225gg_ep_setup() failed for ep1-rx. \n");
        return ret_val;
    }

    /* for ep2 tx. */
    ret_val = cxd9225gg_ep_setup(pTarget->udc_dev, 2);
    if (ret_val)
    {
        udc_error("cxd9225gg_ep_setup() failed for ep2-tx. \n");
        return ret_val;
    }

    udc_error(" usbReInitexit--SUCCESS. \n");
    pTarget->udc_dev->ep_bulk_recv=0;
    pTarget->udc_dev->ep_bulk_send=0;
    pTarget->udc_dev->cur_req = NULL;
    return ret_val;
}


void usbTerminate(void)
{
    udc_debug("entry test. \n");

    udc_debug("attention: you should do something here. \n");

    if (NULL != g_UsbpTarget->hThd4VbusWait)
    {
        atomic_inc(&g_UsbpTarget->thread_exit);

        complete(&(g_UsbpTarget->hSem4VbusWaitTask));

        kthread_stop(g_UsbpTarget->hThd4VbusWait);
        g_UsbpTarget->hThd4VbusWait = NULL;
    }

    s_bInitialized = 0;

    return;
}


extern epcd_fasync_sigcall p_expcdsigcallcb;

int usbConnectionNotify(int plug_flag)
{
    udc_debug("entry test. \n");

    /* send message MSDEV_USB_CONNECT to notify msdev application. */
    //MSDEV_USB_CONNECT
    udc_debug("send MSDEV_USB_CONNECT or disconnect to notify msdev app. \n");

    /* plug in or out... */
    if (plug_flag)
    {
        /* notify app and wait for response.  */
        if(p_expcdsigcallcb)
            p_expcdsigcallcb(48, &plug_flag, sizeof(plug_flag));
        else
            udc_error("p_expcdsigcallcb is NULL\n");
        if (plug_flag)
        {
            return 1;
        }
        else
        {
            return 0;
        }
    }
    else
    {
        /* notify app and return directly... */
        if(p_expcdsigcallcb)
            p_expcdsigcallcb(49, &plug_flag, sizeof(plug_flag));
        else
            udc_error("p_expcdsigcallcb is NULL\n");
        if (plug_flag)
        {
            return 1;
        }
        else
        {
            return 0;
        }

        return 0;
    }
}


unsigned int usbConnectStatusGet(void)
{
    udc_debug(" START ..... usbConnectStatus(0x%08x). \n", g_UsbpTarget->usbConnectStatus);

    return g_UsbpTarget->usbConnectStatus;
}
EXPORT_SYMBOL(usbConnectStatusGet);


unsigned int usbEnableFlgGet(void)
{
    udc_debug(" START ..... usbEnableFlgGet(0x%08x). \n", g_UsbpTarget->usbEnableFlg);
    udc_error(" START ..... usbEnableFlgGet(0x%08x). \n", g_UsbpTarget->usbEnableFlg);

    return g_UsbpTarget->usbEnableFlg;
}


unsigned long usbVersionGet(void)
{
    udc_debug(" START ..... END(0x%08x). \n", USB_VERSION);
    return(USB_VERSION);
}


int usbSelfDiagnose(int SDFlg)
{
    int    status;
    int    ret = 0;
    unsigned char   bctVal = 0;

    udc_debug(" entry test. \n");

    if (0 != g_UsbpTarget->usbEnableFlg)
    {
        udc_error(" [FATAL]Invalid Operation(Storage Mode). \n");
        return -1;
    }

    /* USB Func SUCCESS  */
    bctVal = 1;
    if (bctVal == BOOT_BCT_HWCONFIG_DISABLE_USBDEVICE)
    {
        udc_error("system can NOT support USB Client fucntion, so return directly. \n");
        return 0;
    }

    if ( usbGatedClockOff() != 0 )
    {
        udc_error(" [FATAL]usbGatedClockOff() disable Fail. \n");
    }

    ret = usbReturnfromDirectPowerDown();
    usbPowerResume();
    if (ret)
    {
        udc_error(" [FATAL]ReturnDirectPowerDown Fail. \n");
        return -2;
    }

    /* ID switch. */
    switch (SDFlg)
    {
        /* USB_DIAGNOSE_REGISTER_RW */
        case USB_DIAGNOSE_REGISTER_RW :
            status = usbDiagReg();
            break;

        /* USB_DIAGNOSE_POWERSW */
        case USB_DIAGNOSE_POWERSW:
            status = usbDiagPower();
            break;

        default:
            udc_error(" [WARNING]Invalid Diagnose Flag. \n");
            /*    SelfDiagnose-PROC-12:  */
            status = -3;
            break;
    }

    usbEpcReset();

    if( USB_DPD_RET_FAIL == usbDirectPowerDown() )
    {
        udc_error("[FATAL]DirectPowerDown Fail. \n");
        return -4;
    }

    if( usbGatedClockOn() != 0 )
    {
        udc_error("[FATAL]usbGatedClockOn() enable Fail. \n");
    }

    udc_debug(" exit -- status(%d). \n", status);

    return status;
}


void usbBulkOutRemainSizeGet(int *dmaResult,
                                        unsigned long *dmaRemainSize
                                        )
{
    udc_debug("entry test... \n");

    /* BulkOut DMA dmaResult */
    (*dmaResult)     = g_UsbpTarget->bBulkOutResult;

    /*  BulkOut  DMA dmaRemainSize  */
    (*dmaRemainSize) = g_UsbpTarget->uBulkOutRemainSize;

    /*  BulkOut DMA   0  */
    g_UsbpTarget->bBulkOutResult     = 0;

    /*   BulkOut  DMA  0  */
    g_UsbpTarget->uBulkOutRemainSize = 0;

    udc_debug(" END (dmaResult=%d, dmaRemainSize=%lu). \n", *dmaResult, *dmaRemainSize);

    return;
}


void usbBulkInRemainSizeGet(int *dmaResult,
                                       unsigned long *dmaRemainSize
                                       )
{
    udc_debug("entry test... \n");

    /* BulkIn DMA  dmaResult */
    (*dmaResult)     = g_UsbpTarget->bBulkInResult;

    /*  BulkIn  DMA dmaRemainSize  */
    (*dmaRemainSize) = g_UsbpTarget->uBulkInRemainSize;

    /*   BulkIn DMA  0  */
    g_UsbpTarget->bBulkInResult     = 0;

    /* BulkIn   DMA  0 */
    g_UsbpTarget->uBulkInRemainSize = 0;

    udc_debug("END (dmaResult=%d, dmaRemainSize=%lu). \n", *dmaResult, *dmaRemainSize);

    return;
}
