/*
 * Copyright (C) 2012 Sony Corporation 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License, version 2, as 
 * published by the Free Software Foundation. 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/mm.h>
#include <linux/moduleparam.h>

#include "cxd9225gg_udc.h"
#include "cxd9225gg_usbExt.h"
#include "cxd9225gg_usbEvalLib.h"


int usbFncAddressSet(int dev_addr);
int usbFncSignalResume(struct cxd9225gg_udc *udc_dev);
int usbFncDeviceFeatureSet(int feature, int index);
int usbFncDeviceFeatureClear(int feature);


int usbFncAddressSet(int dev_addr)
{
    unsigned long      usb_addr_data = 0;
    unsigned long      uTcReg = 0;

    /* -- USB_ADDRESS_REG -- deviceAddress -- */
    udc_debug(" old USBADDR_REG = 0x%08x, new deviceAddress = 0x%04x",
              (unsigned int)USB_USBADDR_REG,
              (unsigned int)dev_addr);

    if (0 == dev_addr)
    {
        udc_error("usually the newly-attached device use '0' as default address, except for 'reset'. \n");
    }

    /* 16bit -- USB_ADDR, 32bit cast 15bit  0  */
    usb_addr_data = ((unsigned long)(dev_addr)) & USB_USB_ADDRESS_MASK;

    /* USB ADDR  --- 16bit shift. */
    usb_addr_data = usb_addr_data << 16;

    /*  SOF_STATUS (bit15) 0 */
    uTcReg = USB_USBADDR_REG & USB_USBADDR_SOF_STATUS_BIT;

    /* or USB device address */
    uTcReg |= usb_addr_data;

    /* write back to register. */
    USB_USBADDR_REG = uTcReg;

    udc_debug(" configured USBADDR_REG = 0x%08x. \n", (unsigned int)USB_USBADDR_REG);

    return 0;
}


int usbFncSignalResume(struct cxd9225gg_udc *udc_dev)
{
    udc_error("According to the specification (NOT SUPPORT). \n");
    return -ENOTSUPP;
}


int usbFncDeviceFeatureSet(int feature, int index)
{
    unsigned long      uTestMode       = 0;
    unsigned long      lpcnt           = 0;
    unsigned long      ulTmp;

    /* Feature */
    if (feature != USB_FSEL_DEV_TEST_MODE)
    {
        udc_error(" Feature is not TEST MODE. \n");
        return -1;
    }

    /* index */
    switch (index >> 8)
    {
        case USB_USB_TESTMODE_TEST_J :
            uTestMode = (CXD9225GG_UINT32)USB_USB_TESTMODE_TEST_J;
            uTestMode = uTestMode << USB_TESTMODE_SIFT;
            break;

        case USB_USB_TESTMODE_TEST_K :
            uTestMode = (CXD9225GG_UINT32)USB_USB_TESTMODE_TEST_K;
            uTestMode = uTestMode << USB_TESTMODE_SIFT;
            break;

        case USB_USB_TESTMODE_TEST_SE0_NAK :
            uTestMode = (CXD9225GG_UINT32)USB_USB_TESTMODE_TEST_SE0_NAK;
            uTestMode = uTestMode << USB_TESTMODE_SIFT;
            break;

        case USB_USB_TESTMODE_TEST_PACKET :
            uTestMode = (CXD9225GG_UINT32)USB_USB_TESTMODE_TEST_PACKET;
            uTestMode = uTestMode << USB_TESTMODE_SIFT;
            break;

        case USB_USB_TESTMODE_TEST_FORCE_EN :
            uTestMode = (CXD9225GG_UINT32)USB_USB_TESTMODE_TEST_FORCE_EN;
            uTestMode = uTestMode << USB_TESTMODE_SIFT;
            break;

        default :
            udc_error(" BAD PARAM (uTestSelector=0x%0x). \n", index);
            return -1;
    }


    /* EP0 */
    USB_EP0CTRL_REG |= USB_EP0CTRL_BCLR_BIT;

    /* EP0 (IN_DATA or IN_EMPTY  ) */
    for (lpcnt = 0; lpcnt < USB_EP_BCLR_MAX_LOOP; lpcnt++)
    {
        if ( (0 == (USB_EP0STS_REG & USB_EP0STS_IN_DATA_BIT) )
            && (0 != (USB_EP0STS_REG & USB_EP0STS_IN_EMPTY_BIT) ) )
        {
            break;
        }
    }

    /* EP0 */
    if (lpcnt >= USB_EP_BCLR_MAX_LOOP)
    {
        udc_error(" EP0 BCLR TIMEOUT !!\n");
        return -1;
    }

    ulTmp = USB_EP0CTRL_REG;
    ulTmp |= USB_EP0CTRL_INAK_EN_BIT;
    ulTmp &= ~USB_EP0CTRL_INAK_BIT;
    USB_EP0CTRL_REG = ulTmp;

    /* NULL  */
    USB_EP0CTRL_REG |= USB_EP0CTRL_DEND_BIT;

    msleep(5);

    USB_USBCTRL_REG |= (uTestMode);

    udc_debug("USB_USBCTRL_REG=0x%08x\n", (unsigned int)USB_USBCTRL_REG);

    return 0;
}


int usbFncDeviceFeatureClear(int feature)
{
    udc_debug("entry test. \n");

    if (feature !=  USB_FSEL_DEV_TEST_MODE)
    {
        udc_error(" uFeatureSelector is not TEST MODE (feature = 0x%0x\n",feature);
        return -1;
    }

    udc_debug("exit test. \n");

    return 0;
}
