/* 2014-08-28: File changed by Sony Corporation */
/*
 * Copyright (C) 2012 Sony Corporation 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License, version 2, as 
 * published by the Free Software Foundation. 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/errno.h>

#include <linux/usb.h>
#include <linux/usb/gadget.h>
#include <linux/dma-mapping.h>

#include "cxd9225gg_udc.h"
#include "cxd9225gg_usbExt.h"
#include "cxd9225gg_usbEvalLib.h"


#define USB_SETCONFIG_VALUE_SUPPORT

/* Bulk-only class specific requests */
#define USB_REQ_BULK_RESET			0xff

extern    PUSBENDPOINT g_UsbpEndpoint[USB_NUM_ENDPOINTS];
extern PUSBTARGET    g_UsbpTarget;


int usbFncEndpointAssign(struct cxd9225gg_udc *udc_dev, int ep_num);
int usbFncEndpointRelease(struct cxd9225gg_udc *udc_dev, int ep_num);
int usbFncEndpointStateSet(int endpointIndex, int stall);  
int usbFncEndpointStatusGet(int endpointIndex, int *p_stall); 
int usbFncIsBufferEmpty(int endpointIndex, int *p_buf_empty);
int usbFncCopyDataFromEpBuf(struct cxd9225gg_ep  *udc_ep, 
                                       struct cxd9225gg_req *udc_req);
int usbFncCopyDataToEpBuf(struct cxd9225gg_ep  *udc_ep, 
                                     struct cxd9225gg_req *udc_req);
int usbSetupPioDataRecieve(struct cxd9225gg_ep  *udc_ep, 
                                      struct cxd9225gg_req *udc_req); 
int usbPioDataRecieve(struct cxd9225gg_ep  *udc_ep, 
                                 struct cxd9225gg_req *udc_req); 
int usbPioDataSend(struct cxd9225gg_ep  *udc_ep, 
                             struct cxd9225gg_req *udc_req);  
static int usbBulkOutDma(struct cxd9225gg_ep  *udc_ep, 
                                    struct cxd9225gg_req *udc_req);  
static int usbBulkOutDmaMultiPacket(struct cxd9225gg_ep  *udc_ep, 
                                               struct cxd9225gg_req *udc_req,
                                               unsigned long     ulBytes,
                                               unsigned long     *pulTransfered,
                                               unsigned long     *pulRemain);  
static int usbBulkInDma(struct cxd9225gg_ep  *udc_ep, 
                                   struct cxd9225gg_req *udc_req);  
int usbDmaStop(unsigned char  uEpNo);  
static int usbBulkOutDmaShortPacket2(unsigned long       ulBytes,
                                                unsigned char       *pucStartPtr,
                                                unsigned long       *pulTransfered,
                                                unsigned long       *pulRemain,
                                                PUSBENDPOINT pEndpointInfo);  
static int usbBulkInDmaMax(unsigned long     ulBytes,
                                      unsigned char     *pucStartPtr,
                                      unsigned long     *pulTransfered,
                                      unsigned long     *pulRemain,
                                      unsigned long     ulMaxPacketSize,
                                      PUSBENDPOINT pEndpointInfo);

int usbPioDataSend_ack(struct cxd9225gg_ep  *udc_ep);  

/* return endpoint index by windex */
static int getEpIndexByWindex(u16 wIndex)
{
	int num,dir;
	num =  le16_to_cpu(wIndex) & USB_ENDPOINT_NUMBER_MASK;
	if(num == 0)
	{
		dir = (le16_to_cpu(wIndex) & USB_ENDPOINT_DIR_MASK) ? 1 : 0;
		if(dir)
			return USB_EP0_TX;
		return USB_EP0_RX;
	}
	else if(num == 1)
		return USB_EP1_RX;
	else if(num == 2)
		return USB_EP2_TX;
	else
		return -1;
}


static void getStatusComplete(struct usb_ep *ep, struct usb_request *req)
{
	if (ep == NULL || req == NULL) {
		err("EINVAL");
		return;
	}
	req->buf = NULL;
	kfree(req);
}

static int usbFncStdGetstatus(struct cxd9225gg_ep *ep, 
									   struct usb_ctrlrequest *ctrl)
{
	u16 usb_status = 0;
	int ep_index;
	struct usb_request  *req = NULL;
	int status = 0;

	if ((ctrl->bRequestType & USB_RECIP_MASK) == USB_RECIP_DEVICE) {
		/* Get device status */
		usb_status = 1 << USB_DEVICE_SELF_POWERED;
	} else if ((ctrl->bRequestType & USB_RECIP_MASK) == USB_RECIP_INTERFACE) {
		/* Get interface status */
		/* We don't have interface information in udc driver */
		usb_status = 0;
	} else if ((ctrl->bRequestType & USB_RECIP_MASK) == USB_RECIP_ENDPOINT) {
		/* Get endpoint index */
		ep_index = getEpIndexByWindex(ctrl->wIndex);

		usbFncEndpointStatusGet(ep_index, &status);
			
		usb_status = status << USB_ENDPOINT_HALT;
	}

	req = usb_ep_alloc_request(&ep->ep, GFP_KERNEL);
	if (req == NULL)
		return -ENOMEM;
	
	req->complete = getStatusComplete;
	req->length   = 2;
	req->buf      = kzalloc(req->length, GFP_KERNEL);
	if (req->buf == NULL){
		status = -ENOMEM;
		goto err_free_req;
	}
	*((u16 *) req->buf) = cpu_to_le16(usb_status);

	/* data phase */
	status = usb_ep_queue(&ep->ep, req, GFP_KERNEL);
	if(status)
		goto err_free_buf;
	
	return 0;
	
err_free_buf:
	kfree(req->buf);
err_free_req:
	usb_ep_free_request(&ep->ep, req);
	return status;
}

int usbFncEndpointAssign(struct cxd9225gg_udc *udc_dev, int ep_num)
{
    PUSBTARGET           pTarget     = NULL;
    unsigned char                   transferType  = 0;      /* trasfer type of endpoint */
    unsigned char                   direction     = 0;      /* direction 1-IN, 0-OUT */
    unsigned short                  maxPacketSize = 0;      /* max packet size */
    unsigned long                   ulEpNum   = 0;          /* EndPoint */
    unsigned long                   ulTmp   = 0;
    unsigned char                   endpointIndex = 0;      /* endpoint index */
        
    udc_debug("entry test. \n");

    if( !(USB_EPCTR_REG & USB_EPCTR_PLL_LOCK_BIT) ){
        udc_error("PLL is not locked\n");
        goto exit;
    }
    
    if ((!udc_dev) || (!udc_dev->usb_target) || (ep_num >= 3))
    {
        udc_error("udc_dev or ep_num is invalid. \n");
        return -1; 
    }

    pTarget = udc_dev->usb_target;


    /* except for ep0. */
    if (ep_num)
    {
        transferType = USB_ENDPOINT_XFER_BULK;

        maxPacketSize = USB_MAXPACKET_HIGH;

        if (2 == ep_num)
        {
            direction = USB_DIR_IN; 
        }
        else 
        {
            direction = USB_DIR_OUT; 
        }
    }
    else
    {
        udc_debug("pEndptDescr is NULL for ep0. \n");
        
        transferType = USB_ENDPOINT_XFER_CONTROL;
        maxPacketSize = USB_MAXPSIZE_64; 

        direction = USB_DIR_OUT;
    }

    if (USB_ENDPOINT_XFER_CONTROL == transferType)    
    {
        udc_debug("to configure USB_EP0_RX/TX.\n");
        
        if (maxPacketSize > USB_MAXPSIZE_64)
        {
            udc_error(" maxPacketSize > 64byte for ep0-control. \n");
            return -1;
        }
        else
        {
            maxPacketSize = USB_MAXPSIZE_64;
        }
        endpointIndex = 0;
       
        /* attention...for ep0, we should configure both tx/rx.  */
        {
            /* Buffer?Release Suspend */
            /* Buffer Clear & PID Reset */ 
            USB_EP0CTRL_REG |= (USB_EP0CTRL_BCLR_BIT | USB_EP0CTRL_PIDCLR_BIT );

            USB_EP0CTRL_REG &= ~USB_EP0CTRL_STL_BIT;
            ulTmp = USB_EP0INTEN_SETUP_EN_BIT /*  | USB_EP0INTEN_OUT_EN_BIT */;
            USB_EP0INTEN_REG |= ulTmp;
            if ( ulTmp != ( USB_EP0INTEN_REG & ulTmp) )
            {
                udc_error(" EP0_RX int enable setting error. \n");
                return -1; 
            }

            /* Buffer Clear & PID Reset */
            USB_EP0CTRL_REG |= (USB_EP0CTRL_BCLR_BIT | USB_EP0CTRL_PIDCLR_BIT );

            USB_EP0CTRL_REG &= ~USB_EP0CTRL_STL_BIT;
            ulTmp = USB_EP0INTEN_STG_END_EN_BIT | USB_EP0INTEN_STG_START_EN_BIT;

            /* USB_EP0INTEN_IN_EN_BIT */
            USB_EP0INTEN_REG |= ulTmp;
            if ( ulTmp != ( USB_EP0INTEN_REG & ulTmp) )
            {
                udc_error(" EP0_TX int enable setting error. \n");
                return -1;
            }
        }
    }
    else
    {
        if (USB_USBSTS_SPEED_MODE_BIT == (USB_USBSTS_REG & USB_USBSTS_SPEED_MODE_BIT))
        {
            udc_debug("speed-bit is high. \n");
            
            maxPacketSize = USB_MAXPACKET_HIGH;
        }
        else
        {
            udc_debug("speed-bit is full. \n");
            
            maxPacketSize = USB_MAXPACKET_FULL;
        }

        /* in or out. */ 
        if (direction == USB_DIR_OUT)
        {
            udc_debug("to configure USB_EP1_RX. \n");
            
            /* OUT */
            endpointIndex = USB_EP1_RX;
            ulEpNum = endpointIndex & USB_EP_SERIAL_NUM_MASK; 

            /* MaxPacketSize */
            if( USB_MAXPACKET_HIGH == maxPacketSize )
            {
                USB_EP1PCKTADDR_REG = ((USB_EP1PCKTADDR_REG & ~USB_MAXPACKET_MASK) | USB_MPKT_HS);
            }
            else
            {
                USB_EP1PCKTADDR_REG = ((USB_EP1PCKTADDR_REG & ~USB_MAXPACKET_MASK) | USB_MPKT_FS);
            }
            
            USB_EP1DMACTRL_REG = 0;  /* DMA Control Reg */
            
            /* Buffer */
            USB_EP1CTRL_REG |= (USB_EPCTRL_OPIDCLR_BIT | USB_EPCTRL_BCLR_BIT | USB_EPCTRL_CBCLR_BIT);

            /* EP1 (EP1 Control Register) */ 
            USB_EP1CTRL_REG = (USB_EP1CTRL_REG \
                                    | USB_EPCTRL_DIR0_OUT) 
                                    | USB_EPCTRL_AUTO_BIT; 
            
            if ( (USB_EPCTRL_DIR0_OUT | USB_EPCTRL_AUTO_BIT) 
                 != (USB_EP1CTRL_REG & (USB_EPCTRL_DIR0_OUT |USB_EPCTRL_AUTO_BIT)) 
                )
            {
                udc_error(" EP1CTRL setting error!!\n");
                return -1; 
            }
            
            /* EP1?(EP1 Interrupt Enable Register) */
            USB_EP1INTEN_REG = USB_EP1INTEN_INIT;
            if (USB_EP1INTEN_INIT != USB_EP1INTEN_REG)
            {
                udc_error(" EP1INTEN setting error!!\n");
                return -1;
            }

            USB_AHBBINTEN_REG |= USB_AHBBINTEN_DMA_ENDINTEN_1_BIT;
            USB_AHBBINT_REG |= USB_AHBBINT_DMA_ENDINT_1_BIT;

            /* EP1?EP1 Control Register */
            USB_EP1CTRL_REG |= USB_EPCTRL_EN_BIT;
            if (USB_EPCTRL_EN_BIT != (USB_EP1CTRL_REG & USB_EPCTRL_EN_BIT))
            {
                udc_error(" EP1 Enable error. \n");
                return -1;
            }
            
            USB_EP1CTRL_REG &= ~USB_EPCTRL_ONAK_BIT;
            
            /* stall */
            USB_EP1CTRL_REG &= ~USB_EPCTRL_ISTL_BIT;
            ulTmp = USB_EP1CTRL_REG;
            ulTmp |= USB_EPCTRL_OSTL_EN_BIT;
            ulTmp &= ~USB_EPCTRL_OSTL_BIT;
            USB_EP1CTRL_REG = ulTmp; 

            init_completion(&(g_UsbpEndpoint[ulEpNum]->hSemEpcDmaEndInt)); 
            init_completion(&(g_UsbpEndpoint[ulEpNum]->hSemBridgeDmaEndInt)); 
        
        }
        else
        {
            udc_debug("to configure USB_EP2_TX. \n");
            
            /* OUT?EP2_TX */
            endpointIndex = USB_EP2_TX;
            USB_EP2DMACTRL_REG = 0;  /* DMA Control Reg? */
            USB_EP2CTRL_REG |= (USB_EPCTRL_IPIDCLR_BIT | USB_EPCTRL_BCLR_BIT | USB_EPCTRL_CBCLR_BIT);

            /* MaxPacketSize?*/
            if( USB_MAXPACKET_HIGH == maxPacketSize )
            {
                USB_EP2PCKTADDR_REG = ((USB_EP2PCKTADDR_REG & ~USB_MAXPACKET_MASK) | USB_MPKT_HS);
            }
            else
            {
                USB_EP2PCKTADDR_REG = ((USB_EP2PCKTADDR_REG & ~USB_MAXPACKET_MASK) | USB_MPKT_FS);
            }

            /* EP2?(EP2 Control Register)*/
            USB_EP2CTRL_REG = (USB_EP2CTRL_REG & ~USB_EPCTRL_DIR0_OUT) 
                         | USB_EPCTRL_AUTO_BIT;
            
            if( USB_EPCTRL_AUTO_BIT != (USB_EP2CTRL_REG 
                  & (USB_EPCTRL_DIR0_OUT | USB_EPCTRL_AUTO_BIT)) )
            {
                udc_error(" EP2CTRL setting error!!\n");
                return -1;
            }
           
            /* EP2?(EP2 Interrupt Enable Register)*/
            USB_EP2INTEN_REG = USB_EP2INTEN_INIT;
            if(USB_EP2INTEN_INIT != USB_EP2INTEN_REG)
            {
                udc_error(" [FATAL]EP2INTEN setting error!!\n");
                return -1;
            }

            USB_AHBBINTEN_REG |= USB_AHBBINTEN_DMA_ENDINTEN_2_BIT;
            USB_AHBBINT_REG |= USB_AHBBINT_DMA_ENDINT_2_BIT;

            /* EP2?(EP2 Control Register)*/
            USB_EP2CTRL_REG |= USB_EPCTRL_EN_BIT;
            if(USB_EPCTRL_EN_BIT != (USB_EP2CTRL_REG & USB_EPCTRL_EN_BIT))
            {
                udc_error(" EP2 Enable error!!\n"); 
                return -1;
            }
            
            /* stall */
            USB_EP2CTRL_REG &= ~USB_EPCTRL_ISTL_BIT;
            ulTmp = USB_EP2CTRL_REG;
            ulTmp |= USB_EPCTRL_OSTL_EN_BIT;
            ulTmp &= ~USB_EPCTRL_OSTL_BIT;
            USB_EP2CTRL_REG = ulTmp;
        }
    }

    udc_debug(" Assign ep index = 0x%04x. \n", endpointIndex);

    /* attention: should NOT use ep_num here... */
    ulEpNum = endpointIndex & USB_EP_SERIAL_NUM_MASK;

    /* BulEpNum */  
    g_UsbpEndpoint[ulEpNum]->transferType  = transferType;
    g_UsbpEndpoint[ulEpNum]->direction     = direction;
    g_UsbpEndpoint[ulEpNum]->maxPacketSize = maxPacketSize;
    g_UsbpEndpoint[ulEpNum]->endpointIndex = endpointIndex; 

    /* attention--initial full. */
    init_completion(&(g_UsbpEndpoint[ulEpNum]->hSem4PipeSerialize)); 
    complete(&(g_UsbpEndpoint[ulEpNum]->hSem4PipeSerialize)); 
    init_completion(&(g_UsbpEndpoint[ulEpNum]->hSem4WaitForInterrupt));   
    
    /* attention...for ep0-in.  */
    if (0 == ulEpNum)
    {
        /* BulEpNum  */
        g_UsbpEndpoint[1]->transferType  = transferType;
        g_UsbpEndpoint[1]->direction     = USB_DIR_IN;
        g_UsbpEndpoint[1]->maxPacketSize = maxPacketSize;
        g_UsbpEndpoint[1]->endpointIndex = USB_EP0_TX; 
    
        /* attention--initial full. */
        init_completion(&(g_UsbpEndpoint[1]->hSem4PipeSerialize));
        complete(&(g_UsbpEndpoint[1]->hSem4PipeSerialize)); 
        init_completion(&(g_UsbpEndpoint[1]->hSem4WaitForInterrupt)); 

        g_UsbpEndpoint[1]->bIsAssigned = 1; 
    }

    /* control */
    init_completion(&(pTarget->hSem4SyncCtrlTransfer));

    g_UsbpEndpoint[ulEpNum]->bIsAssigned = 1;

exit:
    return 0;
}


int usbFncEndpointRelease(struct cxd9225gg_udc *udc_dev, int ep_num) 
{
    unsigned long flags;
    PUSBENDPOINT    pEndpoint;
    PUSBENDPOINT    pEndpoint_2;  

    udc_debug("entry test, ep:%d. \n", ep_num); 

    if ((!udc_dev) || (!udc_dev->usb_target) || (ep_num >= 3))
    {
        udc_error("udc_dev or ep_num is invalid. \n");
        return -1; 
    }

    /* ep-num calculation. */
    if (ep_num)
    {
        pEndpoint = *(udc_dev->usb_target->pEndPointInfoTop + ep_num + 1);
        pEndpoint_2 = NULL; 
    }
    else
    {
        pEndpoint = *(udc_dev->usb_target->pEndPointInfoTop);
        pEndpoint_2 = *(udc_dev->usb_target->pEndPointInfoTop + 1); 
    }

    switch (pEndpoint->endpointIndex)
    {
        /* EPn Reg  */
        case USB_EP0_RX:
            udc_debug(" endpointIndex is EP0_RX. \n");
            break;
            
        case USB_EP0_TX:
            udc_debug(" endpointIndex is EP0_TX. \n");
            break;
            
        case USB_EP1_RX:
            udc_debug(" endpointIndex is EP1_RX. \n");
            break;
            
        case USB_EP2_TX:
            udc_debug(" endpointIndex is EP2_TX. \n");
            break;
            
        default:
            udc_error(" endpointIndex is Error. \n");
            break;
    }

    spin_lock_irqsave(&udc_dev->lock, flags); 

    /* cleanup the specific endpoint. */
        memset( pEndpoint, 0, sizeof( *pEndpoint ) );
    pEndpoint->bIsAssigned = 0; //already been cleanup. 

    /* attention: clean up ep0-in. */
    if (pEndpoint_2)
    {
        memset( pEndpoint_2, 0, sizeof( *pEndpoint_2 ) );
        pEndpoint_2->bIsAssigned = 0;  //already been cleanup.  
    }

    spin_unlock_irqrestore(&udc_dev->lock, flags); 
    
    return 0;
}



int usbFncEndpointStateSet(int endpointIndex, int stall)  
{
    unsigned long               ulTmp;
    unsigned long               lpcnt = 0;
    unsigned long               regIntDummy = 0;
    
    udc_debug("endpointIndex=0x%02x, stall=%d. \n", endpointIndex, stall); 

    switch ( endpointIndex )
    {
        case USB_EP0_RX:
            udc_debug(" endpointIndex is EP0_RX. \n");
            break;
            
        case USB_EP0_TX:
            udc_debug(" endpointIndex is EP0_TX. \n");
            break;
            
        case USB_EP1_RX:
            udc_debug(" endpointIndex is EP1_RX. \n");
            break;
            
        case USB_EP2_TX:
            udc_debug(" endpointIndex is EP2_TX. \n");
            break;
            
        default:
            udc_error(" endpointIndex is Error. \n"); 
            break;
    }
 

    /*  STALL state. */
    if (stall)
    {
        udc_debug(" state is STALL (0x%02x). \n", endpointIndex);

        switch (endpointIndex)
        {
            /* EP0_RX EP0_TX.  */
            case USB_EP0_RX:
            case USB_EP0_TX:
                /* EP0 STALL. */
                USB_EP0CTRL_REG |= USB_EP0CTRL_STL_BIT;
                break;

            /* EP1_RX. */
            case USB_EP1_RX:
                /* EP1 STALL. */
                USB_EP1CTRL_REG = (USB_EP1CTRL_REG |
                                        USB_EPCTRL_OSTL_EN_BIT |
                                        USB_EPCTRL_OSTL_BIT);
                break;

            /* EP2_TX  */
            case USB_EP2_TX:
                /* EP2 STALL  */
                USB_EP2CTRL_REG |= USB_EPCTRL_ISTL_BIT;
                break;

            default:
                udc_error(" endpointIndex is Error. \n");
                return -1; 
                //break;
        }
    }
    /*  UNSTALL  */
    else
    {
        udc_debug("state is UN-STALL (0x%02x). \n", endpointIndex);

        switch (endpointIndex)
        {
            /* EP0_RX  EP0_TX   */
            case USB_EP0_RX:
            case USB_EP0_TX:
                /* EP0  */
                USB_EP0CTRL_REG |= USB_EP0CTRL_BCLR_BIT;
                
                /* EP0  */
                /* IN_DATA   0b  IN_EMPTY  1b  */
                for(lpcnt = 0; lpcnt < USB_EP_BCLR_MAX_LOOP; lpcnt++)
                {
                    if( (USB_EP0STS_IN_DATA_BIT != (USB_EP0STS_REG & USB_EP0STS_IN_DATA_BIT) )
                        && (USB_EP0STS_IN_EMPTY_BIT == (USB_EP0STS_REG & USB_EP0STS_IN_EMPTY_BIT) ) )
                    {
                        break;
                    }
                }
                
                /* EP0  */
                if (lpcnt >= USB_EP_BCLR_MAX_LOOP)
                {
                    /* error */
                    udc_error(" EP0 Buffer Clear Time OUT. \n");
                    return -1;
                }
                
                /* EP0 UNSTALL */
                USB_EP0STS_REG = ~USB_EP0STS_STALL_INT_BIT;
                regIntDummy = USB_EP0STS_REG;  /* Dummy Read */
                break;

            /* EP1_RX  */
            case USB_EP1_RX:
                /* EP1  */
                USB_EP1CTRL_REG |= USB_EPCTRL_BCLR_BIT;

                /* EP1 STALL  */
                for (lpcnt = 0; lpcnt < USB_EP_BCLR_MAX_LOOP; lpcnt++)
                {
                    if( (USB_EPSTS_OUT_EMPTY_BIT == (USB_EP1STS_REG & USB_EPSTS_OUT_EMPTY_BIT) ) )
                    {
                        break;
                    }
                }
                
                /* EP1 */
                if (lpcnt >= USB_EP_BCLR_MAX_LOOP)
                {
                    udc_error(" EP1 Buffer Clear Time OUT. \n");
                    return -1; 
                }
                
                /* Cntrol Reg */
                ulTmp = USB_EP1CTRL_REG;
                ulTmp &= ~USB_EPCTRL_OSTL_BIT;
                ulTmp |= USB_EPCTRL_OSTL_EN_BIT;
                USB_EP1CTRL_REG = ulTmp;
                USB_EP1CTRL_REG |= USB_EPCTRL_OPIDCLR_BIT;
                break;

            /* EP2_TX  */
            case USB_EP2_TX:
                /* EP2   */
                USB_EP2CTRL_REG |= USB_EPCTRL_BCLR_BIT;
                
                /* EP2 IN  */
                /* IN_DATA  0b  IN_EMPTY  1b  */
                for (lpcnt = 0; lpcnt < USB_EP_BCLR_MAX_LOOP; lpcnt++)
                {
                    if( (USB_EPSTS_IN_DATA_BIT != (USB_EP2STS_REG & USB_EPSTS_IN_DATA_BIT) )
                        && (USB_EPSTS_IN_EMPTY_BIT == (USB_EP2STS_REG & USB_EPSTS_IN_EMPTY_BIT) ) )
                    {
                        break;
                    }
                }
                
                /* EP2  */
                if (lpcnt >= USB_EP_BCLR_MAX_LOOP)
                {
                    udc_error(" EP2 Buffer Clear Time OUT. \n");
                    return -1; 
                }
                
                /* EP2 STALL  */
                USB_EP2STS_REG = ~USB_EPSTS_IN_STALL_INT_BIT;
                regIntDummy = USB_EP2STS_REG;  /* Dummy Read */

                /* Cntrol Reg  */
                USB_EP2CTRL_REG &= ~USB_EPCTRL_ISTL_BIT;
                USB_EP2CTRL_REG |= USB_EPCTRL_IPIDCLR_BIT;
                break;
                
            default:
                udc_error(" endpointIndex Error. \n");
                return -1;
                //break;
        }
    }

    udc_debug("endpointIndex=0x%02x, stall=%d. \n", endpointIndex, stall);
    
    return 0;
}



int usbFncEndpointStatusGet(int endpointIndex, int *p_stall)
{
    udc_debug("entry test, endpointIndex:%d. \n", endpointIndex); 
 
    /* attention: endpointIndex is NOT ep_num.  */
    if (!p_stall) 
    {
        udc_error(" p_stall is NULL, endpointIndex:0x%0x. \n", endpointIndex);
        return -1; 
    }

    switch ( endpointIndex )
    {
        case USB_EP0_RX:
            udc_debug(" endpointIndex is EP0_RX");
            break;
            
        case USB_EP0_TX:
            udc_debug(" endpointIndex is EP0_TX");
            break;
            
        case USB_EP1_RX:
            udc_debug(" endpointIndex is EP1_RX");
            break;
            
        case USB_EP2_TX:
            udc_debug(" endpointIndex is EP2_TX");
            break;
            
        default:
            udc_error("  is Error"); 
            break;
    }

    switch (endpointIndex)
    {
        /* EP0_RX, EP0_TX   */
        case USB_EP0_RX:
        case USB_EP0_TX:
            /* EP0 STALL */
            /*  CDR :EP0CTRL STL_BIT. */
            if (USB_EP0CTRL_STL_BIT == (USB_EP0CTRL_REG & USB_EP0CTRL_STL_BIT) )
            {
                udc_debug(" EP0 is stalled. \n");

                *p_stall = 1; 
            }
            else  /* un-STALL */ 
            {
                udc_debug(" EP0 is NOT stalled. \n");
                *p_stall = 0;  
            }
            break;

        /* EP1_RX  */
        case USB_EP1_RX:
            /* EP1 STALL  */
            if(USB_EPCTRL_OSTL_BIT == (USB_EP1CTRL_REG & USB_EPCTRL_OSTL_BIT) )
            {
                udc_debug(" EP1_RX is stalled. \n");
                *p_stall = 1;
            }
            else /* un-STALL */
            {
                udc_debug(" EP1_RX is NOT stalled. \n");
                *p_stall = 0; 
            }
            break;

        /* EP2_TX  */
        case USB_EP2_TX:
            /* EP2 STALL  */
            if(USB_EPCTRL_ISTL_BIT == (USB_EP2CTRL_REG & USB_EPCTRL_ISTL_BIT) )
            {
                udc_debug(" EP2_TX is stalled. \n");
                *p_stall = 1; 
            }
            else
            {
                udc_debug(" EP2_TX is NOT stalled. \n");
                *p_stall = 0; 
            }
            break;
            
        /*  */
        default:
            udc_error(" endpointIndex Error. \n");
            return -1;
            //break;
    }

    udc_debug("ep_idx=0x%02x: stall=%d.\n", endpointIndex, *p_stall); 
    
    return 0;
}



int usbFncIsBufferEmpty(int endpointIndex, int *p_buf_empty) 
{
    udc_debug("entry test, endpointIndex:%d. \n", endpointIndex); 
 
    /* attention: endpointIndex is NOT ep_num.  */
    if (!p_buf_empty) 
    {
        udc_error(" p_buf_empty is NULL, endpointIndex:0x%0x. \n", endpointIndex);
        return -1; 
    }

    switch (endpointIndex)
    {
        /* EP0_RX  */
        case USB_EP0_RX:
            /* EP0_RX  */
            if (USB_EP0STS_OUT_EMPTY_BIT == (USB_EP0STS_REG & USB_EP0STS_OUT_EMPTY_BIT))
            {
                *p_buf_empty = 1; 
            }
            else
            {
                *p_buf_empty = 0;  
            }
            
            udc_debug("EP0_RX --> Empty = %d. \n", *p_buf_empty);
            break;

        /* EP0_TX  */
        case USB_EP0_TX:
            /* EP0_TX  */
            if (USB_EP0STS_IN_EMPTY_BIT == (USB_EP0STS_REG & USB_EP0STS_IN_EMPTY_BIT))
            {
                *p_buf_empty = 1; 
            }
            else
            {
                *p_buf_empty = 0;   
            }
            
            udc_debug("EP0_TX --> Empty = %d. \n", *p_buf_empty);
            break;

        /* EP1_RX */
        case USB_EP1_RX:
            /* EP1  */
            if (USB_EPSTS_OUT_EMPTY_BIT == (USB_EP1STS_REG & USB_EPSTS_OUT_EMPTY_BIT))
            {
                *p_buf_empty = 1; 
            }
            else
            {
                *p_buf_empty = 0;  
            }
            
            udc_debug("EP1_RX --> Empty = %d. \n", *p_buf_empty);
            break;

        /* EP2_TX  */
        case USB_EP2_TX:
            /* EP2  */
            if (USB_EPSTS_IN_EMPTY_BIT == (USB_EP2STS_REG & USB_EPSTS_IN_EMPTY_BIT))
            {
                *p_buf_empty = 1; 
            } 
            else
            {
                *p_buf_empty = 0;  
            }

            udc_debug(" EP2_TX --> Empty = %d. \n", *p_buf_empty); 
            break;

        default:
            udc_error(" endpointIndex Error, endpointIndex:0x%0x. \n", endpointIndex);
            return -1; 
    }

    udc_debug("exit--OK, endpointIndex:0x%0x. \n", endpointIndex);
    
    return 0;
}


int usbFncCopyDataFromEpBuf(struct cxd9225gg_ep  *udc_ep, 
                                       struct cxd9225gg_req *udc_req
                                       ) 
{ 
    PUSBTARGET       pTarget         = NULL;  
    PUSBENDPOINT     pEndpointInfo   = NULL; 
    int                         ret_val         = -1; 
    int                         tmp             = 0; 
    unsigned char               endpointIndex = 0;

    udc_debug("entry test. \n"); 

    pTarget = udc_ep->udc->usb_target; 

    /* power state. */
    if ( USB_POW_ST_POWER_ON != pTarget->ulPowState )
    {
        /* ACTIVE */
        udc_error("[FATAL]PowState is NOT POWER_ON!! \n");
        return -1; /* error code */
    }
    
    /* attention.  */ 
    if (udc_ep->num)
    {
        pEndpointInfo = *(pTarget->pEndPointInfoTop + udc_ep->num + 1); 
    }
    else
    {
        //attention: only for ep0-rx. 
        pEndpointInfo = *(pTarget->pEndPointInfoTop + 0); 
    }
        
    /* get the sempahore for serialization. */
    UDC_STAT_QUE_STEP(__LINE__);
    ret_val = wait_for_completion_interruptible_timeout(&(pEndpointInfo->hSem4PipeSerialize), 
                                                        msecs_to_jiffies(USB_FUNC_TIMEOUT));
    UDC_STAT_QUE_STEP(__LINE__);
    if (!ret_val || (-ERESTARTSYS == ret_val))  
    {
        if (-ERESTARTSYS == ret_val)
        {
            udc_error("it means this kthread is killed. \n");
            return -1; 
        }
        
        udc_error("it means timeout--hSem4PipeSerialize. \n");   
        tmp = 0;
    }
    else
    {
        tmp = 1;  
    }

    endpointIndex = pEndpointInfo->endpointIndex;
    
    /* EP1_RX   */
    if ( USB_EP1_RX == endpointIndex )
    {
        /* DMA out. */
        ret_val = usbBulkOutDma(udc_ep, udc_req);
        
        udc_debug("BulkOutDma. \n"); 
    }
    else if (USB_EP0_RX == endpointIndex) /* EP0_RX  */
    {
        udc_error("----------  should NOT reach here now --------------\n");    
    }
    else
    {
        udc_error(" CopyFromEpBuff -- do nothing. \n");
    }
    
    if (tmp)
    {
        complete(&(pEndpointInfo->hSem4PipeSerialize));
    }
    
    return ret_val;
}


int usbFncCopyDataToEpBuf(struct cxd9225gg_ep  *udc_ep, 
                                     struct cxd9225gg_req *udc_req
                                     ) 
{
    PUSBTARGET       pTarget         = NULL;  
    PUSBENDPOINT     pEndpointInfo   = NULL; 
    int                         ret_val         = -1; 
    int                         tmp             = 0; 
    unsigned char               endpointIndex = 0;

    udc_debug("entry test. \n"); 

    pTarget = udc_ep->udc->usb_target; 
    
    /* power state. */
    if ( USB_POW_ST_POWER_ON != pTarget->ulPowState )
    {
        /* ACTIVE */
        udc_error("[FATAL]PowState is NOT POWER_ON!! \n");
        return -1; /* error code */
    } 

    if (udc_ep->num)
    {
        pEndpointInfo = *(pTarget->pEndPointInfoTop + udc_ep->num + 1);
    }
    else
    {
        //attention: only for ep0-tx.  
        pEndpointInfo = *(pTarget->pEndPointInfoTop + 1);  
    }

    /* get the sempahore for serialization. */
    UDC_STAT_QUE_STEP(__LINE__);
    ret_val = wait_for_completion_interruptible_timeout(&(pEndpointInfo->hSem4PipeSerialize), 
                                                        msecs_to_jiffies(USB_FUNC_TIMEOUT));
    UDC_STAT_QUE_STEP(__LINE__);
    if (!ret_val || (-ERESTARTSYS == ret_val))  
    {
        if (-ERESTARTSYS == ret_val)
        {
            udc_error("it means this kthread is killed. \n");
            return -1; 
        }
        
        udc_error("it means timeout--hSem4PipeSerialize. \n");   
        tmp = 0;
    }
    else
    {
        tmp = 1;  
    }

    endpointIndex = pEndpointInfo->endpointIndex;
    
    /* EP2_TX */
    if ( USB_EP2_TX == endpointIndex )
    {
        /* DMA IN */
        udc_debug(" call BulkInDma(). \n");
        
        ret_val = usbBulkInDma(udc_ep, udc_req);
    }
    else if (USB_EP0_TX == endpointIndex) /* EP0_TX */
    {
        /* PIO  */
        udc_debug(" call PioDataSend(). \n");

        ret_val = usbPioDataSend(udc_ep, udc_req);
    } 
    else
    {
        udc_error(" CopyToEpBuff -- do nothing. \n");
    }
    
    if (tmp && !ret_val)
    {
        complete(&(pEndpointInfo->hSem4PipeSerialize));
    }
    
    return ret_val; 
}

const char * _get_request_string(int req_code)
{
	char * str = "UNKNOWN_REQUEST";
	switch (req_code) {
	/* Standard requests */
	case USB_REQ_GET_STATUS:
		str="USB_REQ_GET_STATUS"; break;
	case USB_REQ_CLEAR_FEATURE:
		str="USB_REQ_CLEAR_FEATURE"; break;
	case USB_REQ_SET_FEATURE:
		str="USB_REQ_SET_FEATURE"; break;
	case USB_REQ_SET_ADDRESS:
		str="USB_REQ_SET_ADDRESS"; break;
	case USB_REQ_GET_DESCRIPTOR:
		str="USB_REQ_GET_DESCRIPTOR"; break;
	case USB_REQ_SET_DESCRIPTOR:
		str="USB_REQ_SET_DESCRIPTOR"; break;
	case USB_REQ_GET_CONFIGURATION:
		str="USB_REQ_GET_CONFIGURATION"; break;
	case USB_REQ_SET_CONFIGURATION:
		str="USB_REQ_SET_CONFIGURATION"; break;
	case USB_REQ_GET_INTERFACE:
		str="USB_REQ_GET_INTERFACE"; break;
	case USB_REQ_SET_INTERFACE:
		str="USB_REQ_SET_INTERFACE"; break;
	case USB_REQ_SYNCH_FRAME:
		str="USB_REQ_SYNCH_FRAME"; break;

	/* Bulk-only class specific requests */
	case USB_GET_MAX_LUN:
		str="USB_GET_MAX_LUN"; break;
	case USB_REQ_BULK_RESET:
		str="USB_REQ_BULK_RESET"; break;

	default: break;
	}
	return str;
}

int usbSetupPioDataRecieve(struct cxd9225gg_ep  *udc_ep, 
                                      struct cxd9225gg_req *udc_req 
                                      ) 
{
    PUSBTARGET       pTarget     = NULL;
    PUSBENDPOINT     pEndpointInfo = NULL ;
    unsigned long               ulLoopCount = 0;    
    unsigned long               ulValue = 0;       
    int                         bConfigHit = 0; 
    struct cxd9225gg_ep            *tmp_ep; 
    int                         status = 0; 

    union u 
    {
        u32 dword[2];
        struct usb_ctrlrequest  r;
    } volatile u; 

    if (!udc_ep)
    {
        udc_error("udc_ep is invalid. \n");
        return -1; 
    }

    pTarget = udc_ep->udc->usb_target; 

    pEndpointInfo = *(pTarget->pEndPointInfoTop + 0);

    /* SETUP */ 
    u.dword[0] = USB_SETUPDATA0_REG;
    u.dword[1] = USB_SETUPDATA1_REG;

#define w_value     le16_to_cpu(u.r.wValue)
#define w_index     le16_to_cpu(u.r.wIndex)
#define w_length    le16_to_cpu(u.r.wLength)


    /* Delegate almost all control requests to the gadget driver,
     * except for a handful of ch9 status/feature requests that
     * hardware doesn't autodecode _and_ the gadget API hides.
     */
    udc_ep->udc->ep0_in = (u.r.bRequestType & USB_DIR_IN) != 0;
    udc_ep->udc->ep0_set_config = 0;
    udc_ep->udc->ep0_pending = 1;
    udc_ep->stopped = 0;

    /* get w-value. */
    ulValue = w_value;

    udc_debug("---- Request: 0x%02x(%s)\n", u.r.bRequest, _get_request_string(u.r.bRequest));
    /**** REQUEST CODE  ****/
    switch (u.r.bRequest) 
    {
        /* No-Data Control Case */
        case USB_REQ_SET_CONFIGURATION: 
          pTarget->ulCtrlTransType =  USB_CTRL_TRANS_READ; //need ack... 
 
          if (u.r.bRequestType != USB_RECIP_DEVICE)
          {
            udc_error("SET_CONFIGURATION --- bRequestType != USB_RECIP_DEVICE. \n");
            //goto delegate;
          }
          
         if (w_length != 0)
          {
            udc_error("SET_CONFIGURATION---w_length != 0. \n"); 
            //goto do_stall;
          }
          
          udc_ep->udc->ep0_set_config = 1;
          udc_ep->udc->ep0_reset_config = (w_value == 0);
           
            if  (0 == (USB_USBADDR_REG & ~USB_USB_ADDRESS_MASK))
           {
                udc_debug(" SET_CONFIGURATION--Default state:do nothing. \n");

            }
            else if (USB_USBSTS_CONF_BIT !=  (USB_USBSTS_REG & USB_USBSTS_CONF_BIT))
            {
                /* wValue 0  */
                if (0 == ulValue)
                {
                    /* Null*/
                    udc_debug(" SET_CONFIGURATION--Address state:value = 0. \n");
                }
            #ifdef USB_SETCONFIG_VALUE_SUPPORT
                /* wValue  */
                /* Configuration Descriptor HookEnable */
                else
                {
                    udc_debug("SETCONFIG_VALUE_SUPPORT, ulValue:0x%0x .\n", (unsigned int)ulValue); 
                    ulValue = w_value;  //need NOT to read again....  
                    for (ulLoopCount = 0; ulLoopCount < USB_CONF_INDEX_SIZE; ulLoopCount++)
                    {
                        if ( ulValue == pTarget->ulConfigValue[ulLoopCount] )
                        {
                            udc_debug(" ConfigValue hit on index %lu. \n", ulLoopCount);
                            
                            bConfigHit = 1;
                            break;
                        }
                    }

                    if (1 == bConfigHit)
                    {
                        pTarget->ulCurrentConfigValue = ulValue;

                        /* (USB Control Register CONF 1b  */
                        USB_USBCTRL_REG |= USB_USBSTS_CONF_BIT;
                    
                        udc_debug(" SET_CONFIGURATION USB_USBSTS_REG = 0x%08x. \n", (unsigned int)USB_USBSTS_REG);
                    }
                    else
                    {
                        /*  STALL */
                        USB_EP0CTRL_REG |= USB_EP0CTRL_STL_BIT;
                    }
                }
            #else   /* USB_SETCONFIG_VALUE_SUPPORT */
                else if(1)
                {
                    /* (USB Control Register CONF 1b */
                    udc_debug("SET_CONFIGURATION-Address state:value = %d. \n", ulValue);

                    USB_USBCTRL_REG |= USB_USBSTS_CONF_BIT;

                    udc_debug("SET_CONFIGURATION USB_USBSTS_REG = 0x%08x. \n", USB_USBSTS_REG);
                }
                else
                {
                    /* STALL  */
                    USB_EP0CTRL_REG |= USB_EP0CTRL_STL_BIT;
                }
            #endif /* USB_SETCONFIG_VALUE_SUPPORT */
            } 
            else
            {
                /* wValue = 0  */
                if (0 == ulValue)
                {
                    /* Address  (USB Control Register CONF=0b  */
                    udc_debug(" SET_CONFIGURATION-config state:value = 0. \n");

                    USB_USBCTRL_REG &= ~USB_USBCTRL_CONF_BIT;
                }
                /* wValue  */
            #ifdef USB_SETCONFIG_VALUE_SUPPORT
                else if(pTarget->ulCurrentConfigValue == ulValue)
                {
                    /* Null  */
                    udc_debug(" wValue(0x%4x) is same as the current - do nothing....... \n", (unsigned int)ulValue);
                }
                /* wValue  */
                /* Configuration Descriptor HookEnable */
                else
                {
                    udc_debug(" wValue(0x%4x) is NOT same as the current - to find matched one. \n", (unsigned int)ulValue);
                    
                    ulValue = w_value;  //need NOT to read again. 
                    for (ulLoopCount = 0; ulLoopCount < USB_CONF_INDEX_SIZE; ulLoopCount++)
                    {
                        if ( ulValue == pTarget->ulConfigValue[ulLoopCount] )
                        {
                            udc_debug("ConfigValue hit on index %lu. \n", ulLoopCount);

                            bConfigHit = 1;
                            break;
                        }
                    }
                    
                    if (1 == bConfigHit)
                    {
                        pTarget->ulCurrentConfigValue = ulValue;

                        /*  (USB Control Register CONF 1b  */
                        USB_USBCTRL_REG |= USB_USBSTS_CONF_BIT;

                        udc_debug(" SET_CONFIGURATION USB_USBSTS_REG = 0x%08x. \n", (unsigned int)USB_USBSTS_REG);
                    }
                    else
                    {
                        /* STALL   */
                        USB_EP0CTRL_REG |= USB_EP0CTRL_STL_BIT;
                    }
                }
            #else   /* USB_DEBUG_CONTROL */
                /* wValue  bConfigurationValue  */ 
                else
                {
                    /* Null  */
                    udc_debug(" SET_CONFIGURATION-config state:value = %d. \n", pTrb->pBuffer[USB_WVALUE]);
                }
            #endif  /* USB_DEBUG_CONTROL */ 
            }

            //goto delegate;  //to delegate the gadget-driver and continue to receive.
            break;  /* SET_CONFIGURATION */

        case USB_REQ_SET_INTERFACE: 
            pTarget->ulCtrlTransType = USB_CTRL_TRANS_NO_DATA;

            if (u.r.bRequestType == USB_RECIP_INTERFACE) 
            {
                udc_debug("bRequestType == USB_RECIP_INTERFACE. \n");
                /* set ep0 out... */
            }
            else  
            {
                udc_debug("bRequestType == 0x%0x. \n", u.r.bRequestType); 
            }
            
            if(0 == (USB_USBADDR_REG & ~USB_USB_ADDRESS_MASK))
            {
                udc_error("usb-address is 0....error......\n");
               
                /* STALL  */
                USB_EP0CTRL_REG |= USB_EP0CTRL_STL_BIT;
            }
            else if (USB_USBSTS_CONF_BIT !=  (USB_USBSTS_REG & USB_USBSTS_CONF_BIT))
            {
                udc_error("error--------config is NOT 1..... \n");
                
                /* STALL  */
                USB_EP0CTRL_REG |= USB_EP0CTRL_STL_BIT;
            }
            else
            {
                udc_debug("wValue  Alternate Setting, should NOT reach here. \n"); 
            }
            
                //ack. 
                usbPioDataSend_ack(&pTarget->udc_dev->ep[0]); 
            
                return 0;

       /* set address. */    
        case USB_REQ_SET_ADDRESS:
            if (u.r.bRequestType == USB_RECIP_DEVICE) 
            {
                udc_debug("bRequestType == USB_RECIP_DEVICE, address:0x%0x. \n", (unsigned int)(ulValue&0x7F)); 

                /* set address. */
                (void)usbFncAddressSet((int)(ulValue&0x7F));
                
                /* set ep0 out. */ 
            }
            else 
            {
                udc_debug("bRequestType == 0x%0x. \n", (unsigned int)u.r.bRequestType); 
            } 
        
            pTarget->ulCtrlTransType = USB_CTRL_TRANS_NO_DATA;

            //ack. 
            usbPioDataSend_ack(&pTarget->udc_dev->ep[0]); 

            return 0;
        
        case USB_REQ_SET_FEATURE:
            pTarget->ulCtrlTransType = USB_CTRL_TRANS_NO_DATA;

            if(u.r.bRequestType == USB_RECIP_DEVICE)
            {
            	if(w_value == USB_FSEL_DEV_TEST_MODE)
         		{
            		usbFncDeviceFeatureSet(w_value, w_index); 
					return 0;
				}
				else if(w_value == USB_FSEL_DEV_REMOTE_WAKEUP)
				{
					break;
				}
				else
				{
					goto do_stall;
				}
					
            }
			else if (u.r.bRequestType == USB_RECIP_ENDPOINT)
            {
                udc_debug("bRequestType == USB_RECIP_ENDPOINT. \n");
				
				if (w_value != USB_FSEL_DEV_ENDPOINT_HALT
	                || w_length != 0)
	            {
	                udc_error("w_value != USB_FSEL_DEV_ENDPOINT_HALT or w-length != 0. \n");
	                goto do_stall;
	            }
				
	            tmp_ep = &udc_ep->udc->ep[w_index & 0xf];
				
	            if (tmp_ep->bmAttributes == USB_ENDPOINT_XFER_ISOC
	                || tmp_ep == &udc_ep->udc->ep[0] || !tmp_ep->desc)
	            {
	              udc_error("args are invalid. \n");
				  goto do_stall;
	            }

				udc_debug("[%d]%s halted by host\n", (w_index & 0xf), tmp_ep->ep.name);
				/*set endpoint halt*/
				status=usb_ep_set_halt(&tmp_ep->ep);

				//ack. 
				usbPioDataSend_ack(tmp_ep);

				return 0;
			}
			else
			{
				break;
			}
			
        case USB_REQ_GET_DESCRIPTOR:     
            pTarget->ulCtrlTransType = USB_CTRL_TRANS_READ;
            pTarget->ulGetDescriptorValue = ulValue;
            break;

        case USB_REQ_GET_INTERFACE: 
            pTarget->ulCtrlTransType = USB_CTRL_TRANS_READ;
            break;

        case USB_REQ_SYNCH_FRAME: 
            pTarget->ulCtrlTransType = USB_CTRL_TRANS_READ;
            break;
            
        case USB_REQ_SET_DESCRIPTOR:  
            pTarget->ulCtrlTransType = USB_CTRL_TRANS_WRITE;
            
            /* control write OUT NAK  */
            USB_EP0CTRL_REG &= ~USB_EP0CTRL_ONAK_BIT;
            break; 

        case USB_GET_MAX_LUN:
            pTarget->ulCtrlTransType = USB_CTRL_TRANS_READ;
            break;

        case USB_REQ_GET_CONFIGURATION: 
            pTarget->ulCtrlTransType = USB_CTRL_TRANS_READ;
            break;

        case USB_REQ_CLEAR_FEATURE: 
            pTarget->ulCtrlTransType = USB_CTRL_TRANS_NO_DATA;
		
            if(u.r.bRequestType == USB_RECIP_DEVICE)
            {
            	usbFncDeviceFeatureClear(w_value);  
            }            
            else if (u.r.bRequestType == USB_RECIP_ENDPOINT)
            {
            	udc_debug("bRequestType == USB_RECIP_ENDPOINT. \n");
				
			    if (w_value != USB_ENDPOINT_HALT
                	|| w_length != 0)
            	{
            		udc_error("w_value != USB_ENDPOINT_HALT or w_length != 0. \n"); 
            		goto do_stall;
            	}

				tmp_ep = &udc_ep->udc->ep[w_index & 0xf];
				
				if (tmp_ep->bmAttributes == USB_ENDPOINT_XFER_ISOC
	                || tmp_ep == &udc_ep->udc->ep[0] || !tmp_ep->desc)
	            {
	            	udc_error("args are invalid. \n");
					goto do_stall;
	            }

				udc_debug("[%d]%s halt cleared by host\n", (w_index & 0xf), tmp_ep->ep.name);
				/*clear endporint halt*/
				usb_ep_clear_halt(&tmp_ep->ep);

				//ack
			    usbPioDataSend_ack(tmp_ep);
				
			    return 0;
            }
			else
			{
				break;
			}
            /* NOTE:  assumes the host behaves sanely,
               * only clearing real halts.  Else we may
               * need to kill pending transfers and then
               * restart the queue... very messy for DMA!
               */

        case USB_REQ_GET_STATUS: 
            pTarget->ulCtrlTransType = USB_CTRL_TRANS_READ;

            /* USB_ENDPOINT_HALT status? */
            udc_debug("USB_REQ_GET_STATUS: bRequestType=0x%x\n", u.r.bRequestType);
            if((u.r.bRequestType == (USB_DIR_IN|USB_RECIP_DEVICE)) ||
				(u.r.bRequestType == (USB_DIR_IN|USB_RECIP_INTERFACE)) ||
				(u.r.bRequestType == (USB_DIR_IN|USB_RECIP_ENDPOINT)))
            {
	            //udc_debug("bRequestType != (USB_DIR_IN|USB_RECIP_ENDPOINT). \n"); 
			    usbFncStdGetstatus(udc_ep, (struct usb_ctrlrequest *)&u.r);
			    return 0;
              
            }
		    else
		    {
		        goto intf_status;
		    }
    
            /* ep0 never stalls */
            if (!(w_index & 0xf))
            {
                udc_debug("ep0 should never stall, so need NOT to get its status. \n"); 
                goto zero_status;
            }
    
            /* only active endpoints count */
            tmp_ep = &udc_ep->udc->ep[w_index & 0xf];
            if (w_index & USB_DIR_IN)
            {
                tmp_ep += 1;
            }
            
            if (!tmp_ep->desc)
            {
                udc_debug("tmp_ep->desc in NULL. \n"); 
                //goto do_stall;
            }
    
            /* iso never stalls */
            if (tmp_ep->bmAttributes == USB_ENDPOINT_XFER_ISOC)
            {
                udc_debug("bmAttributes == USB_ENDPOINT_XFER_ISOC. \n"); 
                goto zero_status;
            }
    
            /* FIXME don't assume non-halted endpoints!! */
            udc_error("%s status, can't report. \n", tmp_ep->ep.name);
            //goto do_stall;
    
      intf_status:
            /* return interface status.  if we were pedantic,
            * we'd detect non-existent interfaces, and stall.
            */
            if (u.r.bRequestType != (USB_DIR_IN|USB_RECIP_INTERFACE))
            {
                udc_debug("bRequestType != (USB_DIR_IN|USB_RECIP_INTERFACE). \n"); 
                goto delegate;
            }
    
      zero_status:
            /* return two zero bytes */
            status = 0;
            udc_debug("GET_STATUS, interface %d. \n", w_index);
            /* next, status stage */
            break;

		case USB_REQ_BULK_RESET:
            pTarget->ulCtrlTransType = USB_CTRL_TRANS_READ;
			udc_ep->udc->ep0_reset_config = (w_value == 0);
            break;

        default:
            /* STALL */
            pTarget->bIsCtrlTransOnGoing = 0;
            pTarget->ulCtrlTransType = USB_CTRL_TRANS_TYPE_UNDEF;
            break;
    }

delegate:
    
#undef  w_value
#undef  w_index
#undef  w_length

    /* The gadget driver may return an error here,
     * causing an immediate protocol stall.
     *
     * Else it must issue a response, either queueing a
     * response buffer for the DATA stage, or halting ep0
     * (causing a protocol stall, not a real halt).  A
     * zero length buffer means no DATA stage.
     *
     * It's fine to issue that response after the setup()
     * call returns, and this IRQ was handled.
     */
    udc_ep->udc->ep0_setup = 1;


    /* notify gadget-driver. */
    status = udc_ep->udc->driver->setup(&udc_ep->udc->gadget, 
                                        (const struct usb_ctrlrequest *)&u.r);
    if (status < 0)
    {
        udc_debug("udc->driver->setup() failed, ret=%d.\n", status);
do_stall:
        usb_ep_set_halt(&udc_ep->ep);
        //usbPioDataSend_ack(udc_ep);
    }
    
    //spin_lock(&udc_ep->udc->lock);  //attetion...... 
    udc_ep->udc->ep0_setup = 0;

    udc_debug("----exit----OK. \n");

    return 0;  
}



int usbPioDataRecieve(struct cxd9225gg_ep  *udc_ep, 
                                 struct cxd9225gg_req *udc_req 
                                 ) 
{
    PUSBTARGET   pTarget         = NULL;
    PUSBENDPOINT pEndpointInfo   = NULL;
    unsigned long           lpcnt           = 0;
    unsigned long           pioRcvSize      = 0;
    unsigned long           *pUserBuf       = NULL;
    unsigned long           ulEpIndex       = 0;
    unsigned long           ulRemainSize    = 0;
    unsigned char           *pucUserBuf     = NULL;
    int                     ret_val         = 0; 
    unsigned long           ulReqSize; 

    udc_debug("entry test. \n");

    /* internal function, no arg-checking. */
    if (!udc_ep || !udc_req)
    {
        udc_error("udc_ep or udc_req is invalid. \n");
        return -1; 
    }

    pTarget = udc_ep->udc->usb_target; 

    pEndpointInfo = *(pTarget->pEndPointInfoTop + 0); 

    /* attention to the max-packet-size. */
    ulReqSize = udc_req->req.length - udc_req->req.actual; 
    pucUserBuf = (unsigned char *)udc_req->req.buf + udc_req->req.actual;

    prefetch(pucUserBuf); 

    udc_debug("request start -- %u %u %lu %lu. \n", 
              udc_req->req.actual, udc_req->req.length, 
              ulReqSize, udc_req->req.actual + ulReqSize); 

    pUserBuf = (unsigned long *)pucUserBuf;

    ulEpIndex = ((pEndpointInfo->endpointIndex & USB_EP_INDEX_MASK) >> 4); /* endpointIndex  */

    /* PIO OUT */
   if( USB_TRANSFER_TYPE_NULL == pEndpointInfo->ulPipeTransferType )
    {
        /* do nothing */
        /* buffer clear? */ 
        udc_error("we should NOT reach here. \n"); 
    }
    else if( (USB_TRANSFER_TYPE_PIO == pEndpointInfo->ulPipeTransferType) 
             || (USB_CTRL_TRANS_WRITE == pTarget->ulCtrlTransType) /* control write  */  
            )
    {
        /* Data Length  */
        pioRcvSize = USB_EPLEN_REG(ulEpIndex) & USB_EPLEN_LENGTH_MASK;  /* EP0 bit10~7  0 */

        /* attention.. */
        if (pioRcvSize > ulReqSize)
        {
            udc_error("fatal error: pioRcvSize > ulReqSize. \n");
            pioRcvSize = ulReqSize; 
        } 

        ulRemainSize = pioRcvSize % USB_EP_READ_REG_SIZE_BYTE;

        for(lpcnt = 0; lpcnt < (pioRcvSize / USB_EP_READ_REG_SIZE_BYTE); lpcnt++)
        {
            *pUserBuf++ = USB_EPREAD_REG(ulEpIndex);
        }

        if(0 == ulRemainSize)
        {
            /* do nothing */
            udc_error("received data-size is NOT 4-byte aligned. \n");
        }
        else
        {
            pucUserBuf = (unsigned char *)pUserBuf;
            
            for (lpcnt = 0; lpcnt < ulRemainSize; lpcnt++)
            {
                *pucUserBuf++ = (UINT8)((USB_EPREAD_REG(ulEpIndex) >> (lpcnt*8)) & 0xff);
            }
        }

        udc_req->req.actual += pioRcvSize;

        udc_debug("ep0 receive data, size:%lu. \n", pioRcvSize);
    }
    else
    {
        udc_error("error: ulPipeTransferType:%d, ulCtrlTransType:%d. \n", 
                  pEndpointInfo->ulPipeTransferType, pTarget->ulCtrlTransType); 
    }

    if (!udc_ep->stopped)
        cxd9225gg_udc_done(udc_ep, udc_req, 0);  

    udc_debug("exit---with--ret:%d. \n", ret_val); 
    
    return ret_val;  
}


int usbPioDataSend(struct cxd9225gg_ep  *udc_ep, 
                              struct cxd9225gg_req *udc_req
                              ) 
{
    PUSBENDPOINT     pEndpointInfo   = NULL;
    PUSBTARGET       pTarget         = NULL;
    unsigned long               lpcnt           = 0;
    unsigned short              maxPacketSize   = 0;
    unsigned long               *pUserBuf       = NULL;
    unsigned long               ulEnIndex = 0;  /* Endpoint, Offset */
    unsigned long               ulAutoBitBackup = 0;
    unsigned long               ulTmp;
    unsigned long               ulAdr;
    unsigned long               ulRemain;
    unsigned long               ulReqSize; 
    int                         ret_val = 0;

    udc_debug("entry test. \n");

    /* internal function, no arg-checking. */
    if (!udc_ep || !udc_req)
    {
        udc_error("udc_ep or udc_req is invalid. \n");
        return -1; 
    }

    pTarget = udc_ep->udc->usb_target;     

    pEndpointInfo = *(pTarget->pEndPointInfoTop + 1); 

    /* attention. max-packet-size. */
    maxPacketSize    = pEndpointInfo->maxPacketSize;
    
    ulEnIndex= ((pEndpointInfo->endpointIndex & USB_EP_INDEX_MASK) >> 4);  /* endpointIndex  */

    /* attention to the max-packet-size. */
    ulReqSize = udc_req->req.length - udc_req->req.actual; 
    pUserBuf = (unsigned long *)((unsigned char *)udc_req->req.buf + udc_req->req.actual);

    prefetch(pUserBuf);

/* for 'setup' debugging. */

    udc_debug("request start -- %u %u %lu %lu. \n", 
              udc_req->req.actual, udc_req->req.length, 
              ulReqSize, udc_req->req.actual + ulReqSize); 

    /* Buffer Clear */
    USB_EP0CTRL_REG |= USB_EP0CTRL_BCLR_BIT;
    for (lpcnt = 0; lpcnt < 10; lpcnt++)
    {
        ulTmp = USB_EPSTS_REG(ulEnIndex);
        if ( USB_EP0STS_IN_EMPTY_BIT == ((ulTmp & (USB_EP0STS_IN_DATA_BIT | USB_EP0STS_IN_EMPTY_BIT))))
        {
            break;
        }
    }

    if (lpcnt == 10)
    {
        udc_error("data buffer is NOT empty. \n"); 
        return -1;
    }

    /* IN INAK  */
    ulTmp = USB_EP0CTRL_REG;
    ulTmp |= USB_EP0CTRL_INAK_EN_BIT;
    ulTmp &= ~USB_EP0CTRL_INAK_BIT;
    USB_EP0CTRL_REG = ulTmp;
    
    if ( ulReqSize >= maxPacketSize )     
    {
        udc_error("req-size > max-packet-size, only partially. \n");
        
        /* MaxPacketSize Yes */
        for (lpcnt = 0; lpcnt < (maxPacketSize/4); lpcnt++)
        {
            USB_EPWRITE_REG(ulEnIndex) = *pUserBuf++;
        }
        
        /* AUTO */ 
        udc_req->req.actual += maxPacketSize; 
    }
    else if (0 == ulReqSize)  
    {
        udc_debug("send a data-end out for NULL-data. \n");
        
        /* Data end. */
        ulTmp = USB_EPCTRL_REG_ADR(ulEnIndex);
        USB_EP0CTRL_REG |= USB_EP0CTRL_DEND_BIT; 
    }
    else
    {
        /* Get Descriptor  */
        if(USB_DESC_TYPE_CONFIG == (pTarget->ulGetDescriptorValue & USB_DESC_TYPE_MASK))
        {
            pTarget->ulConfigValue[pTarget->ulGetDescriptorValue & USB_DESC_INDEX_MASK] = *((unsigned char*)(pUserBuf)+5);  /* error */
            pTarget->ulGetDescriptorValue = 0;  /* flag */
        }
        
        /* AUTO */
        ulAutoBitBackup = USB_EPCTRL_REG(ulEnIndex) & USB_EP0CTRL_AUTO_BIT;
        USB_EPCTRL_REG(ulEnIndex) &= ~USB_EP0CTRL_AUTO_BIT;

        /* write data out...  */
        for (lpcnt = 0; lpcnt < ((ulReqSize -1)/4 +1); lpcnt++)
        {                                                            
            USB_EPWRITE_REG(ulEnIndex) = *pUserBuf++;
        }
        
        ulRemain = ulReqSize % 4;
        
        /* Data */
        ulAdr = USB_EPCTRL_REG_ADR(ulEnIndex);
        ulTmp = USB_EP0CTRL_REG;
        ulTmp &= ~USB_EPCTRL_DW_MASK;
        ulTmp |= ulRemain << 5;     /* DW0 bit5 */
        ulTmp |= USB_EP0CTRL_DEND_BIT;

        udc_debug(" before set DEND - USB_EP0STS_REG  = 0x%08x. \n", (unsigned int)USB_EP0STS_REG);
        udc_debug(" value to set USB_EP0CTRL_REG = 0x%08x. \n", (unsigned int)ulTmp);

        USB_EP0CTRL_REG = ulTmp;

        udc_debug(" after set DEND - USB_EP0CTRL_REG = 0x%08x. \n", (unsigned int)USB_EP0CTRL_REG);
        udc_debug(" after set DEND - USB_EP0STS_REG  = 0x%08x. \n", (unsigned int)USB_EP0STS_REG);

        udc_req->req.actual += ulReqSize;  
    }

    spin_lock_irqsave(&pTarget->udc_dev->lock, pTarget->udc_dev->m_flags); 

    /* complete to gadget-driver. */
    if (!udc_ep->stopped)
        cxd9225gg_udc_done(udc_ep, udc_req, 0); 

    spin_unlock_irqrestore(&pTarget->udc_dev->lock, pTarget->udc_dev->m_flags);

    udc_debug(" ulCtrlTransType = %d. \n", pTarget->ulCtrlTransType);  

    /* send OUT NAK */
    USB_EP0CTRL_REG &= ~USB_EP0CTRL_ONAK_BIT; 

    /* Auto bit  */
    USB_EPCTRL_REG(ulEnIndex) |= ulAutoBitBackup;

    return ret_val; 
}


int usbPioDataSend_ack(struct cxd9225gg_ep  *udc_ep) 
{
    PUSBENDPOINT     pEndpointInfo   = NULL;
    PUSBTARGET       pTarget         = NULL;
    unsigned long               lpcnt           = 0;
    unsigned long               ulEnIndex = 0;  /* Endpoint, Offset */
    unsigned long               ulTmp;
    int                         ret_val = 0;
    udc_debug("entry test. \n");

    pTarget = udc_ep->udc->usb_target;     

    pEndpointInfo = *(pTarget->pEndPointInfoTop + 1); 

    ulEnIndex= ((pEndpointInfo->endpointIndex & USB_EP_INDEX_MASK) >> 4);  /* endpointIndex  */

    /* Buffer Clear */
    USB_EP0CTRL_REG |= USB_EP0CTRL_BCLR_BIT;
    for (lpcnt = 0; lpcnt < 10; lpcnt++)
    {
        ulTmp = USB_EPSTS_REG(ulEnIndex);
        if ( USB_EP0STS_IN_EMPTY_BIT == ((ulTmp & (USB_EP0STS_IN_DATA_BIT | USB_EP0STS_IN_EMPTY_BIT))))
        {
            break;
        }
    }

    if (lpcnt == 10)
    {
        udc_error("data buffer is NOT empty. \n"); 
        return -1;
    }

    /* IN INAK  */
    ulTmp = USB_EP0CTRL_REG;
    ulTmp |= USB_EP0CTRL_INAK_EN_BIT;
    ulTmp &= ~USB_EP0CTRL_INAK_BIT;
    USB_EP0CTRL_REG = ulTmp;
    
    udc_debug("send a data-end out. \n");
    
    /* Data end. */
    ulTmp = USB_EPCTRL_REG_ADR(ulEnIndex);
    USB_EP0CTRL_REG |= USB_EP0CTRL_DEND_BIT;

    udc_debug(" ulCtrlTransType = %d. \n", pTarget->ulCtrlTransType); 

    return ret_val; 
}

 

static int usbBulkOutDma(struct cxd9225gg_ep  *udc_ep, 
                                    struct cxd9225gg_req *udc_req
                                    ) 
{
    PUSBENDPOINT   pEndpointInfo   = NULL;
    PUSBTARGET   pTarget       = NULL;
    unsigned long           uBufAddress       = 0;
    unsigned long           maxPacketSize     = 0;    /* maxPacketSize */
    unsigned long           receivePacketSize = 0;
    unsigned long           uActLength    = 0;
    unsigned long           uDmaLength    = 0;    /* DMA */
    unsigned long           ulDmaCount    = 0;    /* DMA  */
    int                     result        = -1;
    unsigned char           *pucBuf;           /*  while  */
    unsigned long           ulTransfered  = 0;
    unsigned long           ulRemain      = 0xffff;
    unsigned long           ulLoopCnt     = 0;
    unsigned long           ulErrCnt      = 0;
    unsigned long           ulRequestWords  = 0;
    unsigned long           ulReceivedWords = 0;
    int                     ret_val         = 0; 
    unsigned long           ulReqSize; 
    unsigned char           *tmp_buf;     /*  for debug.  */

    udc_debug("entry test. \n");

    /* internal function, no arg-checking. */
    if (!udc_ep || !udc_req)
    {
        udc_error("udc_ep or udc_req is invalid. \n");
        return -1; 
    }

    pTarget = udc_ep->udc->usb_target; 

    pEndpointInfo = *(pTarget->pEndPointInfoTop + 2); 

    /* attention to the max-packet-size. */
    ulReqSize = udc_req->req.length - udc_req->req.actual; 
    pucBuf = (unsigned char *)udc_req->req.buf + udc_req->req.actual;

    //only for checking temporarily. 
    if (__pa(pucBuf) != ((unsigned long)udc_req->req.dma + udc_req->req.actual))
    {
        udc_error("__pa(pucBuf) is NOT compliant with dma-address. \n"); 
    }

    //only for debug temporarily 
    tmp_buf = (unsigned char *)udc_req->req.buf + udc_req->req.actual; 
    
    udc_debug("request start -- %u %u %lu %lu. \n", 
              udc_req->req.actual, udc_req->req.length, 
              ulReqSize, udc_req->req.actual + ulReqSize); 

    if (USB_CONNECTED_STATUS != pTarget->usbConnectStatus)
    {
        udc_error("USB_CONNECTED_STATUS != usbConnectStatus. \n"); 
        return -1;
    }
    
    if ( (1 == pTarget->usbPlugFlg) && (1 != pTarget->bIsPllUnlocked) ) /*  bIsPllUnlocked */
    { 
    loop_again: 
        ulTransfered = 0;
         
        maxPacketSize = pEndpointInfo->maxPacketSize;
        uActLength = udc_req->req.length - udc_req->req.actual;
        ulRemain = uActLength; 
        
        /* DMA  */
        uDmaLength = ((uActLength / USB_EP_WRITE_REG_SIZE_BYTE) * USB_EP_WRITE_REG_SIZE_BYTE);
        ulDmaCount = ((uDmaLength - 1) / maxPacketSize) + 1;

        /* attention: get physical address. */
        uBufAddress = (unsigned long)udc_req->req.dma + udc_req->req.actual;   

        pucBuf = (unsigned char *)udc_req->req.buf + udc_req->req.actual; 

        if ( 0 != (uBufAddress % 8) )
        {
            udc_error(" [FATAL]uBufAddress(0x%08x) is NOT 8byte aligned. \n", (unsigned int)uBufAddress);
            return -2;
        }
        
        udc_debug(" uActLength:%lu. \n", uActLength );
        udc_debug(" maxPacketSize:%lu. \n", maxPacketSize );
        udc_debug(" uDmaLength:%lu. \n", uDmaLength );
        udc_debug(" ulDmaCount:%lu. \n", ulDmaCount );
        udc_debug(" pucBuf:0x%08x. \n", (unsigned int)pucBuf );   
    
        /* --- Bridge DMA  --- */
        /* EP1 DCR2 Reg */
        USB_EP1DCR2_REG = (USB_EP1DCR2_REG & USB_EPDCR1_EPDMACNT_MASK) | (maxPacketSize);

        /* EP1 TADR Reg  */
        USB_EP1TADR_REG = uBufAddress;

        /* EP1 DCR1 Reg  */
        USB_EP1DCR1_REG |= USB_EPDCR1_EPDMADIR0_BIT;

        /* --- EPC DMA  --- */
        /* EP1 DMA Control Reg   */
        USB_EP1DMACTRL_REG = USB_EP1DMACTRL_REG
                                  | (USB_EPDMACTRL_STOP_MODE_BIT
                                  | USB_EPDMACTRL_STOP_SET_BIT
                                  | USB_EPDMACTRL_DMAMODE0_BIT);
        
        USB_EP1DMACTRL_REG = USB_EP1DMACTRL_REG
                                  & ~(USB_EPDMACTRL_BUS_SEL_1_BIT
                                  | USB_EPDMACTRL_BUS_SEL_0_BIT
                                  | USB_EPDMACTRL_DEND_SET_BIT
                                  | USB_EPDMACTRL_DMAMODE2_BIT);
    
        for (ulLoopCnt = 0; 
             ((ulLoopCnt < 6) && ((uActLength > 0) && (0 != ulRemain))); 
             ulLoopCnt++)   
        { 
            udc_debug(" Loop=%lu : factor=%lu,uActLength=%lu, ulTransfered=%lu, ulRemain=%lu. \n",
                      pTarget->uEp1IntEnableSts, ulLoopCnt, uActLength, ulTransfered, ulRemain);

            /* --- EMPTY or NOT --- */
            if (0 == (USB_EP1STS_REG & USB_EPSTS_OUT_EMPTY_BIT))
            {
                udc_debug(" Step1:have data\n"); 

                receivePacketSize = USB_EP1LENDCNT_REG & ~USB_EPLENDCNT_EPLDATA_MASK;     /* bug fix */

                udc_debug("EP1LENDCNT_REG:0x%08x, receivePacketSize:(0x%0x). \n", 
                           (unsigned int)USB_EP1LENDCNT_REG, 
                           (unsigned int)receivePacketSize); 

                if (receivePacketSize > uActLength)
                {
                    udc_error(" RecvSize(%lu) > ReqSize(%lu), only part of it would be received. \n", 
                              receivePacketSize, uActLength);

                    /* DMA OK  */
                    ulRequestWords = uActLength / 4;
                    if ( 0 != (uActLength % 4) )
                    {
                        ulRequestWords++;
                    }
                    
                    ulReceivedWords = receivePacketSize / 4;
                    if( ulRequestWords == ulReceivedWords )
                    {
                        /* 0, do nothing */
                    }
                    else
                    {
                        udc_error("ulRequestWords(%lu) != ulReceivedWords(%lu)\n", 
                                  ulRequestWords, ulReceivedWords);
                        return -1; 
                    }
                }
                else  /* <= */
                {
                    uActLength = receivePacketSize; 

                    udc_debug("attention: actual received data-size:%lu. \n", uActLength); 
                }
                
                /*  maxPacketSize  */
                if (uActLength < maxPacketSize)
                {
                    /* MaxPacketSize -- ShotPacket  */
                    result = usbBulkOutDmaShortPacket2(uActLength, 
                                                                  pucBuf, 
                                                                  &ulTransfered, 
                                                                  &ulRemain, 
                                                                  pEndpointInfo);

                    /*  OK */
                    udc_debug(" receive short packet -- Request=%lu,Transferd=%lu, Remain=%lu. \n", 
                               uActLength, ulTransfered, ulRemain); 
                }
                else  /* MultiPacket  */
                { 
                    result = usbBulkOutDmaMultiPacket(udc_ep,  
                                                                 udc_req, 
                                                                 uActLength, 
                                                                 &ulTransfered, 
                                                                 &ulRemain);
                     
                    udc_debug(" MULTI packet -- Request=%lu,Transferd=%lu, Remain=%lu. \n", 
                               uActLength, ulTransfered, ulRemain);

                /* must be virtual address. 
                 * if cacheable, we must do invalidation,and then read it. 
                 */ 
                }
            }
            else /* -- no data ??? ---*/
            {
                udc_debug(" have no data.  \n" );
                udc_debug(" EP1DCR2_REG   :0x%08x\n", (unsigned int)USB_EP1DCR2_REG );
                udc_debug(" EP1TADR_REG   :0x%08x\n", (unsigned int)USB_EP1TADR_REG );
                udc_debug(" EP1DCR1_REG   :0x%08x\n", (unsigned int)USB_EP1DCR1_REG );
                udc_debug(" EP1DMACTRL_REG:0x%08x\n", (unsigned int)USB_EP1DMACTRL_REG );
                udc_debug(" EP1LENDCNT_REG:0x%08x\n", (unsigned int)USB_EP1LENDCNT_REG );
                udc_debug(" EP1STS_REG    :0x%08x\n", (unsigned int)USB_EP1STS_REG );
                udc_debug(" semtake \n");

                /* whether we received a NULL-OUT data. 
                 * in fact, USB_EP1STS_REG has been cleared by epc-isr-task....  
                 */
                 if (USB_EP1STS_REG & USB_EPSTS_OUT_NULL_INT_BIT) 
                 {
                     udc_error(" ep1-receive a NULL data, do we need to notify usb-stack??? \n");
                 }  

                /* DMA. */ 
                UDC_STAT_QUE_STEP(__LINE__);
                ret_val = wait_for_completion_interruptible_timeout(&pEndpointInfo->hSem4WaitForInterrupt, 
                                                                    msecs_to_jiffies(20)); 
                UDC_STAT_QUE_STEP(__LINE__);
                if (!ret_val || (-ERESTARTSYS == ret_val))  
               {
                    if (-ERESTARTSYS == ret_val)
                    {
                        udc_error("it means this thread is killed, for hSem4WaitForInterrupt. \n");
                        return -1; 
                    }

                    udc_error("NO DATA TIMEOUT -- hSem4WaitForInterrupt -> EP1STS=0x%08x, LDATA=0x%08x. \n",
                              (unsigned int)USB_EP1STS_REG, (unsigned int)USB_EP1LENDCNT_REG);

                    ulErrCnt++;
                    result = -1;   
                }
                else
                { 
                    if (0 != ulErrCnt) 
                    {
                        if (2 < ulLoopCnt) 
                        { 
                            udc_error("[FATAL]NO DATA TIMEOUT Recover by %lu times retry. \n", ulErrCnt);
                        }
                    }

                    udc_debug(" :hSem4WaitForInterrupt came. \n");   
                }  
            }

            uActLength -= ulTransfered;
            pucBuf = (unsigned char *)(((unsigned long)pucBuf) + ulTransfered);
            udc_req->req.actual += ulTransfered;

            // only for debuggiing temporarily. 
            tmp_buf = (unsigned char *)(((unsigned long)tmp_buf) + ulTransfered);
            
        }   /* for end */

        if ( ulLoopCnt >= 20 )
        {
            udc_error(" -- FATAL: LOOP--%lu COUNT OVER. \n", ulLoopCnt);
            result = -1;
        }

        if (0 != result)
        {
            udc_error(" EXIT by %d. \n", result );
        }
        
        /* bus error */
        if (1 == pTarget->bIsBusError )
        {
            udc_error("bIsBusError is 1, and then stop dma bulk-out. \n");
            
            /* DMA stop. */
            usbDmaStop(pEndpointInfo->endpointIndex);
            
            USB_AHBBINT_REG = USB_AHBBINT_BUSERR_MASK;    /* clear */
            USB_AHBBINTEN_REG |= USB_AHBBINT_BUSERR_MASK;

            pTarget->bIsBusError = 0; 
            
            result = -3; 
        }

        /* BURST_SET */
        USB_EP1DMACTRL_REG &= ~USB_EPDMACTRL_BURST_SET_BIT;
    }
    else
    {
        udc_error("no data received, the device-state is wrong. \n");
        
        result = 0;
        
        /* BURST_SET */
        USB_EP1DMACTRL_REG &= ~USB_EPDMACTRL_BURST_SET_BIT;
    }

    /* ugly code, should use OUT not IN... */
    pTarget->bBulkOutResult = result; 
    pTarget->uBulkOutRemainSize = (udc_req->req.length - udc_req->req.actual); 

    ulLoopCnt = 0; 

    if ( (udc_req->req.length - udc_req->req.actual) > 0 && (udc_req->req.actual >= udc_ep->maxpacket) )
    {
        for (ulLoopCnt = 0; ulLoopCnt < 12; ulLoopCnt++)
        {
            if ( USB_EPSTS_OUT_EMPTY_BIT != (USB_EP1STS_REG & USB_EPSTS_OUT_EMPTY_BIT) )
            {
                udc_debug("NotEmpty DMACTRL=0x%08x,DCR1=0x%08x,LENDCNT=0x%08x,STS=0x%08x. \n",
                          (unsigned int)USB_EP1DMACTRL_REG, 
                          (unsigned int)USB_EP1DCR1_REG, 
                          (unsigned int)USB_EP1LENDCNT_REG, 
                          (unsigned int)USB_EP1STS_REG);

                goto loop_again; 
            }
            else 
            {
                /* do we need to sleep sometime??? */
                udelay(100);
            } 
        }

        udc_error("error--ep1 request is NOT completed fully. req.length=0x%x, req.actual=0x%x\n", 
                udc_req->req.length, udc_req->req.actual); 
    }

    spin_lock_irqsave(&pTarget->udc_dev->lock, pTarget->udc_dev->m_flags); 
    pTarget->uInterruptAssign |= USB_SEQUENCE_CTRL_INT;
    pEndpointInfo->bBulkOutEnd = 1;

    /* complete -- done. */
    if (!udc_ep->stopped)
        cxd9225gg_udc_done(udc_ep, udc_req, (pTarget->bBulkOutResult ? -1 : 0));  
    spin_unlock_irqrestore(&pTarget->udc_dev->lock, pTarget->udc_dev->m_flags);  

    /* to notify epc-isr-thread. */
    UDC_STAT_OUT_DMA_INC();
    complete(&(pTarget->sUsbIntSemId)); 

    udc_debug("exit--with--ret=%d. \n", result);

    return result; 
} 



static int usbBulkOutDmaMultiPacket(struct cxd9225gg_ep  *udc_ep, 
                                               struct cxd9225gg_req *udc_req,
                                               unsigned long     ulBytes,
                                               unsigned long     *pulTransfered,
                                               unsigned long     *pulRemain 
                                               ) 
{
    PUSBENDPOINT pEndpointInfo = NULL;
    PUSBTARGET   pTarget = NULL;
    unsigned long       maxPacketSize = 0;
    unsigned long       uActLength = 0;
    unsigned long       uDmaLength = 0;
    unsigned long       ulDmaCount   = 0;
    int                 ret_val = 0;
    unsigned long       ulRemainDmaCnt = 0;
    int                 ret_code = 0;

    udc_debug("entry test. \n"); 

    pTarget = udc_ep->udc->usb_target;

    pEndpointInfo = *(pTarget->pEndPointInfoTop + 2);
 
    maxPacketSize = pEndpointInfo->maxPacketSize;
    uActLength = ulBytes;   /* pTrb->uActLength BulkDmaOut  */
    uDmaLength = ((uActLength / USB_EP_WRITE_REG_SIZE_BYTE) * USB_EP_WRITE_REG_SIZE_BYTE);
    ulDmaCount = ((uDmaLength - 1) / maxPacketSize) + 1;

    /* DMA COUNT 256 */
    if(ulDmaCount > 256)
    {
        ulDmaCount = 256;       /*  */
        uDmaLength = maxPacketSize * ulDmaCount;
    }
        
    udc_debug(" START---------. \n");

    /* --- Bridge DMA --- */
    /* EPn DCR1 Reg  */
    USB_EP1DCR1_REG = (USB_EP1DCR1_REG & ~USB_EPDCR1_EPDMACNT_MASK)
                            | ((ulDmaCount<<16) & 0x00ff0000);
    USB_EP1DCR1_REG |= USB_EPDCR1_EPDMAREQEN_BIT;
    
    /* --- EPC DMA  --- */
    /* EPn Length & DMA Count Reg */
    USB_EP1LENDCNT_REG = (USB_EP1LENDCNT_REG & USB_EPLENDCNT_EPDMACNT_MASK)
                                | (ulDmaCount<<16);
    
    udc_debug(" Before EP1INTEN_REG   :0x%08x\n", (unsigned int)USB_EP1INTEN_REG );
    udc_debug(" Before EP1STS_REG     :0x%08x\n ", (unsigned int)USB_EP1STS_REG );
    udc_debug(" Before USBINTSTS_REG  :0x%08x\n", (unsigned int)USB_USBINTSTS_REG );
    udc_debug(" Before USBINTEN_REG   :0x%08x\n", (unsigned int)USB_USBINTEN_REG );
    udc_debug(" Before EP1DCR1_REG    :0x%08x\n", (unsigned int)USB_EP1DCR1_REG );
    udc_debug(" Before EP1LENDCNT_REG :0x%08x\n", (unsigned int)USB_EP1LENDCNT_REG );
    udc_debug(" Before EP1DMACTRL_REG :0x%08x\n", (unsigned int)USB_EP1DMACTRL_REG );


    USB_EP1INTEN_REG |= USB_EPINTEN_OUT_END_EN_BIT;
    
    /* EPn DMA Control Reg */
    USB_EP1DMACTRL_REG = (USB_EP1DMACTRL_REG | USB_EPDMACTRL_BURST_SET_BIT)
                               | USB_EPDMACTRL_DMA_EN_BIT;
    
    udc_debug(" EP1INTEN_REG    :0x%08x\n", (unsigned int)USB_EP1INTEN_REG );
    udc_debug(" EP1STS_REG      :0x%08x\n", (unsigned int)USB_EP1STS_REG );
    udc_debug(" USBINTSTS_REG   :0x%08x\n", (unsigned int)USB_USBINTSTS_REG );
    udc_debug(" USBINTEN_REG    :0x%08x\n", (unsigned int)USB_USBINTEN_REG );
    udc_debug(" EP1DCR1_REG     :0x%08x\n", (unsigned int)USB_EP1DCR1_REG );
    udc_debug(" EP1LENDCNT_REG  :0x%08x\n", (unsigned int)USB_EP1LENDCNT_REG );
    udc_debug(" EP1DMACTRL_REG  :0x%08x\n", (unsigned int)USB_EP1DMACTRL_REG );


    /* DMA  */
    udc_debug(" semtake----hSemEpcDmaEndInt. \n"); 

    /* MaxPacketSize Bridge INT, EPC INT  */
    /* IsrExec,   EPC */ 
    UDC_STAT_QUE_STEP(__LINE__);
    ret_val = wait_for_completion_interruptible_timeout(&pEndpointInfo->hSemEpcDmaEndInt, 
                                                        msecs_to_jiffies(USB_FUNC_TIMEOUT));
    UDC_STAT_QUE_STEP(__LINE__);
    if (!ret_val || (-ERESTARTSYS == ret_val))  
    {
        if (-ERESTARTSYS == ret_val)
        {
            udc_error("it means this thread is killed, for hSemEpcDmaEndInt. \n");
            return -1; 
        } 

        ret_code = -1; 

        udc_error(" [FATAL]EPC DMA END INT timeout. \n");

        udc_debug(" after EP1INTEN_REG   :0x%08x\n", (unsigned int)USB_EP1INTEN_REG );
        udc_debug(" after EP1STS_REG     :0x%08x\n", (unsigned int)USB_EP1STS_REG );
        udc_debug(" after USBINTSTS_REG  :0x%08x\n", (unsigned int)USB_USBINTSTS_REG );
        udc_debug(" after USBINTEN_REG   :0x%08x\n", (unsigned int)USB_USBINTEN_REG );
        udc_debug(" after EP1DCR1_REG    :0x%08x\n", (unsigned int)USB_EP1DCR1_REG );
        udc_debug(" after EP1DCR2_REG    :0x%08x\n", (unsigned int)USB_EP1DCR2_REG );
        udc_debug(" after EP1LENDCNT_REG :0x%08x\n", (unsigned int)USB_EP1LENDCNT_REG );
        udc_debug(" after EP1DMACTRL_REG :0x%08x\n", (unsigned int)USB_EP1DMACTRL_REG );

    }
    else
    {
        UDC_STAT_EPC_WAIT_INC();
        udc_debug(" EPC DMA END INT occured. \n");
    } 
    
    /*  MaxPacketSize aligned ????  */
    if (0 == (uActLength % maxPacketSize))
    { 
        /* MaxPacketSize Bridge INT, EPC INT  */
        /* IsrExec,   EPC */ 
        UDC_STAT_QUE_STEP(__LINE__);
        ret_val = wait_for_completion_interruptible_timeout(&pEndpointInfo->hSemBridgeDmaEndInt, 
                                                            msecs_to_jiffies(USB_FUNC_TIMEOUT));
        UDC_STAT_QUE_STEP(__LINE__);
        if (!ret_val || (-ERESTARTSYS == ret_val))  
        {
            if (-ERESTARTSYS == ret_val)
            {
                udc_error("it means this thread is killed, for hSemBridgeDmaEndInt. \n");
                return -1; 
            } 
        
            ret_code = -1;
            
            udc_error(" [FATAL]BRDIGE DMA END INT timeout. \n"); 
        }
        else
        {
            UDC_STAT_BRG_WAIT_INC();
            udc_debug(" BRIDGE DMA END INT occured. \n"); 
        } 
    }
    else
    {
        /* MaxPacketSize -> */
        udc_debug("uActLength is NOT maxPacketSize aligned. \n"); 
    }

    USB_EP1INTEN_REG &= ~USB_EPINTEN_OUT_END_EN_BIT;

    ulRemainDmaCnt = (USB_EP1DCR1_REG >> 16) & 0xff;
    if ( ulRemainDmaCnt > ulDmaCount )
    {
        udc_error(" [FATAL]--invalid DMA COUNT. \n");
        ulRemainDmaCnt = ulDmaCount;
    }
    
    *pulTransfered = ((ulDmaCount - ulRemainDmaCnt) * maxPacketSize);
    *pulRemain = ulBytes - *pulTransfered;

    udc_debug("exit--with--ret=%d. \n", ret_code); 

    return ret_code; 
}



static int usbBulkOutDmaShortPacket2(unsigned long       ulBytes,            /*   */
                                                unsigned char       *pucStartPtr,       /*   */
                                                unsigned long       *pulTransfered,     /*   */
                                                unsigned long       *pulRemain,         /*   */
                                                PUSBENDPOINT pEndpointInfo  
                                                )
{
    unsigned char      *pUserBuf  = NULL;   /* PIO RAM  */
    unsigned char      *pDataBuf  = NULL;
    unsigned char      *pucBuf    = NULL;
    unsigned char      *pRegBuf   = NULL;
    unsigned long       ulCnt;
    unsigned long       ulBuf[1]; 
    int                 ret_val = 0; 
    unsigned long       uBufAddress = 0;    /* DMA  */

    /* internal functions -- no arg-check. */

    udc_debug(" START  pucStartPtr = 0x%08x. \n", (unsigned int)pucStartPtr);

    /* for debug, attention: physical address. */
    pucBuf = pucStartPtr;    

    /* 3byte  */
    if (ulBytes <= 3) 
    {
        udc_debug(" Less than 4byte. \n");

        /* PIO */
        pUserBuf = pucStartPtr;
        
        /* Read */
        ulBuf[0] = USB_EPREAD_REG(1);
        pDataBuf = (unsigned char *)(&ulBuf[0]);
        pRegBuf = pDataBuf;     /* for debug */
        
        /* User Buffer */
        for (ulCnt = 0; ulCnt<ulBytes; ulCnt++)
        {
            *pUserBuf++ = *pDataBuf++; 
        }
        
        *pulTransfered = ulBytes;
        *pulRemain = USB_EP1LENDCNT_REG & 0x7ff;
 
    }
    else /* ShortPacket  */
    {
        udc_debug(" ====SHORTPACKET====. \n"); 

        /* DMA, physical address. */
        uBufAddress = __pa(pucStartPtr);
        USB_EP1TADR_REG = uBufAddress;
        
        /* --- Bridge DMA DMA COUNT 1  --- */
        USB_EP1DCR1_REG = (USB_EP1DCR1_REG & ~USB_EPDCR1_EPDMACNT_MASK) | (0x1<<16);

        /* EPn DCR1 Reg  REQEN */
        USB_EP1DCR1_REG |= USB_EPDCR1_EPDMAREQEN_BIT;

        /* --- EPC DMA  --- */
        /* EPn Length & DMA Count Reg DMA COUNT 0 */
        USB_EP1LENDCNT_REG = (USB_EP1LENDCNT_REG & USB_EPLENDCNT_EPDMACNT_MASK)
                                   | (0x0 << 16); 
        
        USB_EP1INTEN_REG |= USB_EPINTEN_OUT_END_EN_BIT;

        /* EPn DMA Control Reg */
        USB_EP1DMACTRL_REG = (USB_EP1DMACTRL_REG & ~USB_EPDMACTRL_BURST_SET_BIT)
                                | USB_EPDMACTRL_DMA_EN_BIT;
        
        /* DMA */
        udc_debug(" wait for semaphore--hSemBridgeDmaEndInt and epc-dma-int. \n"); 

        /* IsrExec,   EPC */ 
        UDC_STAT_QUE_STEP(__LINE__);
        ret_val = wait_for_completion_interruptible_timeout(&pEndpointInfo->hSemBridgeDmaEndInt, 
                                                            msecs_to_jiffies(USB_FUNC_TIMEOUT));
        UDC_STAT_QUE_STEP(__LINE__);
        if (!ret_val || (-ERESTARTSYS == ret_val))  
        {
            if (-ERESTARTSYS == ret_val)
            {
                udc_error("it means this thread is killed, for hSemBridgeDmaEndInt. \n");
                return -1; 
            } 
       
            udc_error(" [FATAL]BRDIGE DMA END INT timeout. \n"); 
            ret_val = -1; 
        }
        else
        { 
            UDC_STAT_BRG_WAIT_INC();
            UDC_STAT_QUE_STEP(__LINE__);
            /* IsrExec,   EPC */ 
            ret_val = wait_for_completion_interruptible_timeout(&pEndpointInfo->hSemEpcDmaEndInt, 
                                                                msecs_to_jiffies(USB_FUNC_TIMEOUT));
            UDC_STAT_QUE_STEP(__LINE__);
            if (!ret_val || (-ERESTARTSYS == ret_val))  
            {
                if (-ERESTARTSYS == ret_val)
                {
                    udc_error("it means this thread is killed, for hSemEpcDmaEndInt. \n");
                    return -1; 
                } 
           
                udc_error(" [FATAL]EPC DMA END INT timeout. \n"); 
                ret_val = -1; 
            }
            else
            {
                /* do nothing */ 
                ret_val = 0;
                UDC_STAT_EPC_WAIT_INC();
            } 
        } 

        USB_EP1INTEN_REG &= ~USB_EPINTEN_OUT_END_EN_BIT;
        *pulTransfered = (USB_EP1DCR2_REG >> 16) & 0x7ff;
        *pulRemain = USB_EP1LENDCNT_REG & 0x7ff;
 
    }

    udc_debug("exit--with--ret:%d. \n", ret_val); 
    
    return ret_val;
}


static int usbBulkInDma(struct cxd9225gg_ep  *udc_ep, 
                                   struct cxd9225gg_req *udc_req
                                   ) 
{
    PUSBENDPOINT         pEndpointInfo   = NULL;
    PUSBTARGET           pTarget         = NULL;
    unsigned long                   uActLength      = 0;
    unsigned long                   lpcnt           = 0;
    unsigned long                   ulRemainSize    = 0;
    unsigned char                   *pucBuf;
    int                             ret_val         = 0;
    int                             result        = -1;
    unsigned long                   ulTransfered    = 0;
    unsigned long                   ulReqSize = 0;

    udc_debug("entry test. \n");

    /* internal function, no arg-checking. */
    if (!udc_ep || !udc_req)
    {
        udc_error("udc_ep or udc_req is invalid. \n");
        return -1; 
    }

    pTarget = udc_ep->udc->usb_target; 
    pEndpointInfo = *(pTarget->pEndPointInfoTop + 3); 

    /* attention to the max-packet-size. */
    ulReqSize = udc_req->req.length - udc_req->req.actual; 
    pucBuf = (unsigned char *)udc_req->req.buf + udc_req->req.actual;

    //only for checking temporarily. 
    if (__pa(pucBuf) != ((unsigned long)udc_req->req.dma + udc_req->req.actual))
    {
        udc_error("__pa(pucBuf) is NOT compliant with dma-address.\n"); 
    }

    udc_debug("request start -- act=%u, len=%u, req=%lu, (act+req)=%lu.\n", 
              udc_req->req.actual, udc_req->req.length, 
              ulReqSize, udc_req->req.actual + ulReqSize); 

    if ( (1 == pTarget->usbPlugFlg) && (1 != pTarget->bIsPllUnlocked) ) /* bIsPllUnlocked  */
    {
        uActLength = (udc_req->req.length - udc_req->req.actual);
        
        for (lpcnt = 0; lpcnt < 5; lpcnt++) 
        {
            if ( USB_CONNECTED_STATUS != pTarget->usbConnectStatus )
            {
                udc_debug("[BulkInDma] USB Disconnected. \n");
                
                /* DMA stop */
                usbDmaStop(pEndpointInfo->endpointIndex);
                
                ret_val = -2; 
                break;
            }

            /* attention: pucBuf is physical address. */
            result = usbBulkInDmaMax( ulReqSize, pucBuf, &ulTransfered, &ulRemainSize, pEndpointInfo->maxPacketSize, pEndpointInfo );
            
            /* bus error */
            if ( 1 == pTarget->bIsBusError )
            {
                udc_error("error ----- bIsBusError is 1. \n");
                
                /* DMA stop */
                usbDmaStop(pEndpointInfo->endpointIndex);

                /* clear */ 
                USB_AHBBINT_REG = USB_AHBBINT_BUSERR_MASK;    
                USB_AHBBINTEN_REG |= USB_AHBBINT_BUSERR_MASK;

                pTarget->bIsBusError = 0;
                ret_val = -3;
                break; 
            }
            else if (0 != result)
            {
                udc_error("error -- usbBulkInDmaMax failed. \n");
                
                /* DMA stop */
                usbDmaStop(pEndpointInfo->endpointIndex);
                
                ret_val = -4;
                
                udc_req->req.actual += ulTransfered;
                break;
            }
            else
            {
                udc_req->req.actual += ulTransfered;
                if ( udc_req->req.actual >= udc_req->req.length)     
                {
                    udc_debug(" packet is full--exit loop. \n"); 
                    break;
                }
                else
                {
                    ulReqSize = udc_req->req.length - udc_req->req.actual;
                    pucBuf += ulTransfered;
                }
            }
        }
    }
    else
    {
        udc_error("udc-device state is wrong. \n");
        
        ret_val = 0;
        udc_req->req.actual = udc_req->req.length;  //attention -- pseudo... 
    }

    if (ret_val)
        udc_error(" EXIT(%d), Trans=%d\n", ret_val, udc_req->req.actual);

    /* The UDC accept the case that req.length == 0. 
     * This means to send back the packet with last packet size. */
    if (!udc_req->req.length)
        udc_req->req.length = udc_req->req.actual;

    pTarget->bBulkInResult = result;
    pTarget->uBulkInRemainSize = (udc_req->req.length - udc_req->req.actual); 

    if (!result && (0 != pTarget->uBulkInRemainSize)) {
        udc_error("should NOT happen. result=%d, length=%d, actual=%d\n", 
            result, udc_req->req.length, udc_req->req.actual); 
    }

    spin_lock_irqsave(&pTarget->udc_dev->lock, pTarget->udc_dev->m_flags); 
    pTarget->uInterruptAssign |= USB_SEQUENCE_CTRL_INT;
    pEndpointInfo->bBulkInEnd = 1;

    /* complete -- done. */
    if (!udc_ep->stopped)
        cxd9225gg_udc_done(udc_ep, udc_req, (pTarget->bBulkInResult ? -1 : 0));  

    spin_unlock_irqrestore(&pTarget->udc_dev->lock, pTarget->udc_dev->m_flags);  

    /* to notify epc-isr-thread. */
    UDC_STAT_IN_DMA_INC();
    complete(&(pTarget->sUsbIntSemId));

    udc_debug("exit--with--ret=%d. \n", ret_val);

    return ret_val;
}



static int usbBulkInDmaMax(unsigned long     ulBytes,            /*   */
                                      unsigned char     *pucStartPtr,       /*   */
                                      unsigned long     *pulTransfered,     /*   */
                                      unsigned long     *pulRemain,         /*     */
                                      unsigned long     ulMaxPacketSize,    /*     */
                                      PUSBENDPOINT pEndpointInfo
                                      )
{ 
    unsigned long       maxPacketSize   = 0;
    unsigned long       uBufAddress     = 0;
    unsigned long       uRemainLength   = 0;
    unsigned long       uDmaLength      = 0;  /* 4byte  */
    unsigned long       ulDmaCount      = 0;  /* 4byte 0 */
    unsigned long       lpcnt           = 0;
    unsigned long       ulRemainSize    = 0;
    unsigned long       *pUserBuf       = NULL;
    unsigned long        ulTmp;
    unsigned char       *pucBuf; 
    int                 ret_val          = 0; 
    unsigned long       ulRemainDmaCount = 0;
    unsigned long       ulLastPacketSize = 0;


    /* internal function -- no arg-check. */
    udc_debug("entry test, ulMaxPacketSize:0x%0x. \n", (unsigned int)ulMaxPacketSize); 

    /* EP2  */

    /* NULL  */
    if (ulBytes == 0)
    {
        USB_EP2CTRL_REG |= USB_EPCTRL_BCLR_BIT;
        for (lpcnt = 0; lpcnt < USB_EP_BCLR_MAX_LOOP; lpcnt++)
        {
            if( USB_EPSTS_IN_EMPTY_BIT == ((USB_EP2STS_REG & 
                                                (USB_EPSTS_IN_DATA_BIT | USB_EPSTS_IN_EMPTY_BIT))))
            {
                break;
            }
        }
        
       if (lpcnt >= USB_EP_BCLR_MAX_LOOP)
        {
            udc_error("[FATAL]BUF cannot clear in NULL packet. \n");
            return -2;    /* ERROR CODE? */ 
        }
        
        /* Data end. */
        USB_EP2CTRL_REG |= USB_EPCTRL_DEND_BIT;
    }
    else if (pucStartPtr == NULL)
    {
        udc_error(" [FATAL]pucStartPtr is NULL....... \n"); 
        return -3; 
    } 
    else if (3 < ulBytes) /* 3byte Dma  */
    {
        /* DMA */
        uDmaLength = ((ulBytes / USB_EP_READ_REG_SIZE_BYTE) * USB_EP_READ_REG_SIZE_BYTE);
        ulDmaCount = ((uDmaLength - 1 ) / ulMaxPacketSize) + 1;
        
        /* DMA COUNT -- 256 */
        if (ulDmaCount > 256)
        {
            udc_debug("ulDmaCount > 256. \n"); 
            
            ulDmaCount = 256;       /*   */
            uDmaLength = ulMaxPacketSize * ulDmaCount;
        }
        
        /* RAM, physical addrress already. */
        uBufAddress = __pa(pucStartPtr);
        
        if ( 0 != (uBufAddress % 8) )
        {
            udc_error("[FATAL]uBufAddress(0x%08x) is NOT 8byte aligned. \n", (unsigned int)uBufAddress);
            return -2; 
        }
        
        pucBuf = (unsigned char *)pucStartPtr;
        uRemainLength = uDmaLength - (ulMaxPacketSize * (ulDmaCount - 1));

        /*  DMA  */
        USB_EP2DCR2_REG = ((uRemainLength) << 16 | (ulMaxPacketSize));

        /*  IO AUTO  */
        USB_EP2CTRL_REG |= USB_EPCTRL_AUTO_BIT;           

        /* DMA address set. */
        USB_EP2TADR_REG = uBufAddress;

        /* DMA count */
        USB_EP2DCR1_REG = ((ulDmaCount << 16) & 0x00ff0000);
        USB_EP2DCR1_REG |= USB_EPDCR1_EPDMAREQEN_BIT;

        /* DMA  */
        USB_EP2LENDCNT_REG = ulDmaCount << 16;

        /* 1Packet DMA  Burst */
        USB_EP2DMACTRL_REG = USB_EPDMACTRL_DMAMODE0_BIT;

        if ( ulBytes < maxPacketSize )
        {
            /* ShortPacket -> clear BUSRT_SET  */
            USB_EP2DMACTRL_REG &= ~USB_EPDMACTRL_BURST_SET_BIT;
        }
        else
        {
            /* MultiPacket -> BURST_SET  */
            USB_EP2DMACTRL_REG |= USB_EPDMACTRL_BURST_SET_BIT;
        }

        /* re-enable dma intr. */
        USB_EP2DMACTRL_REG |= USB_EPDMACTRL_DMA_EN_BIT;

        /**** DMA Kick  ****/
    
        udc_debug( " semtake----hSem4WaitForInterrupt.\n" );

        /* IsrExec,   EPC */ 
        UDC_STAT_QUE_STEP(__LINE__);
        ret_val = wait_for_completion_interruptible_timeout(&pEndpointInfo->hSem4WaitForInterrupt, 
                                                            msecs_to_jiffies(USB_FUNC_TIMEOUT));
        UDC_STAT_QUE_STEP(__LINE__);
        if (!ret_val || (-ERESTARTSYS == ret_val))  
        {
            if (-ERESTARTSYS == ret_val)
            {
                udc_error("it means this thread is killed, for hSem4WaitForInterrupt. \n");
                return -1; 
            } 
       
            udc_error(" hSem4WaitForInterrupt---timeout-----------------------. \n");
  
            udc_debug(" pEndpointInfo->ulDmaInPhase = %u. \n", pEndpointInfo->ulDmaInPhase);
            udc_debug(" EP2DCR2 Reg    =0x%08x\n", (unsigned int)USB_EP2DCR2_REG );
            udc_debug(" EP2TADR Reg    =0x%08x\n", (unsigned int)USB_EP2TADR_REG );
            udc_debug(" EP2DCR1 Reg    =0x%08x\n", (unsigned int)USB_EP2DCR1_REG );
            udc_debug(" EP2LENDCNT Reg =0x%08x\n", (unsigned int)USB_EP2LENDCNT_REG );
            udc_debug(" EP2DMACTRL Reg =0x%08x\n", (unsigned int)USB_EP2DMACTRL_REG );
            udc_debug(" EP2CTRL Reg    =0x%08x\n", (unsigned int)USB_EP2CTRL_REG );
        
            ret_val = -1; 
        }
        else
        {
            /* do nothing */ 
            ret_val = 0; 
            udc_debug(" get semaphore--hSem4WaitForInterrupt. \n"); 
        }  
    }  
    else /* DMA   */
    {
        /* 4byte DMA  */
        /*  PIO AUTO */ 
        USB_EP2DCR2_REG = ulMaxPacketSize; /*  LastPacketSize  0  */
    }
    
    if (0 != ret_val) { goto _out; }
    
    /* NULL */
    if (ulBytes == 0)
    {
        /* do nothing */
        udc_debug("ulBytes is 0, nothing should be done. \n"); 
    }
    else
    {
        /* DMA */
        if (0 != (ulBytes % USB_EP_READ_REG_SIZE_BYTE) )
        {
            udc_debug(" data--- Fragment - USB_EPCTRL_REG(2) = 0x%08x. \n", (unsigned int)USB_EPCTRL_REG(2) );

            USB_EP2CTRL_REG &= ~USB_EPCTRL_AUTO_BIT;
            
            ulRemainSize = ulBytes % USB_EP_READ_REG_SIZE_BYTE;

            /*  PIO */
            pUserBuf = (unsigned long *)((unsigned long)pucStartPtr + uDmaLength);

            //attention..........it's physical address. 
            USB_EP2WRITE_REG = *pUserBuf; //can NOT do it.....  
            //USB_EP2WRITE_REG = pUserBuf - udc_req->;  

            ulTmp = USB_EP2CTRL_REG;
            ulTmp &= ~USB_EPCTRL_DW_MASK;
            ulTmp |= ((ulRemainSize << 5) & USB_EPCTRL_DW_MASK);
            ulTmp |= USB_EPCTRL_DEND_BIT;

            udc_debug(" Fragment - additional data = 0x%08x\n", (unsigned int)*pUserBuf);
            udc_debug(" Fragment - data to ep2Ctrl = 0x%08x\n", (unsigned int)ulTmp);

            USB_EP2CTRL_REG = ulTmp;
        }
        else if (0 != (ulBytes % ulMaxPacketSize) )
        {
            udc_debug("ulBytes is NOT ulMaxPacketSize aligned. \n");
            
            /* MaxPacketSize  */
            USB_EP2CTRL_REG &= ~USB_EPCTRL_AUTO_BIT;

            USB_EPCTRL_REG(2) |= USB_EPCTRL_DEND_BIT; 
        }
        else
        { 
            udc_debug(" This transfer is MaxPacketSize(%lu) aligned. \n", ulMaxPacketSize);
        }
    }
    
    ulRemainDmaCount = (USB_EP2LENDCNT_REG >> 16) & 0x1ff;
    ulLastPacketSize = (USB_EP2DCR2_REG >> 16) & 0x7ff;
    if( ulDmaCount > ulRemainDmaCount )
    {
        if(ulRemainDmaCount == 0){
            *pulTransfered = ((ulDmaCount - 1                  ) * ulMaxPacketSize) + ulLastPacketSize + ulRemainSize;
        }else{
            *pulTransfered = ((ulDmaCount - ulRemainDmaCount   ) * ulMaxPacketSize) + ulLastPacketSize + ulRemainSize;
        }
    }
    else
    {
        *pulTransfered = ulLastPacketSize + ulRemainSize;
    }
    *pulRemain = ulBytes - *pulTransfered;
    
    if (0 != ret_val)
    {
        udc_error("---------error-----error-----. \n");
        
        udc_debug(" EP2DCR2_REG      = 0x%08x\n", (unsigned int)USB_EP2DCR2_REG);
        udc_debug(" EP2LENDCNT_REG   = 0x%08x\n", (unsigned int)USB_EP2LENDCNT_REG);
        udc_debug(" ulMaxPacketSize  = %lu\n", ulMaxPacketSize);
        udc_debug(" ulBytes          = %lu\n", ulBytes);
        udc_debug(" uDmaLength       = %lu\n", uDmaLength);
        udc_debug(" ulDmaCount       = %lu\n", ulDmaCount);
        udc_debug(" uBufAddress      = 0x%08x\n", (unsigned int)uBufAddress);
        udc_debug(" ulRemainSize     = %lu\n", ulRemainSize);
        udc_debug(" ulRemainDmaCount = %lu\n", ulRemainDmaCount);
        udc_debug(" ulLastPacketSize = %lu\n", ulLastPacketSize);
        udc_debug(" *pulTransfered   = %lu\n", *pulTransfered);
        udc_debug(" *pulRemain       = %lu\n", *pulRemain);
        udc_debug(" ulDmaInPhase     = %u\n", pEndpointInfo->ulDmaInPhase);
        udc_debug(" AHBBINT_REG      = 0x%08x\n", (unsigned int)USB_AHBBINT_REG);
        udc_debug(" USBINTSTS_REG    = 0x%08x\n", (unsigned int)USB_USBINTSTS_REG);
        udc_debug(" EP2STS_REG       = 0x%08x\n", (unsigned int)USB_EP2STS_REG);
        udc_debug(" USBINTEN_REG     = 0x%08x\n", (unsigned int)USB_USBINTEN_REG);
        udc_debug(" EP2INTEN_REG     = 0x%08x\n", (unsigned int)USB_EP2INTEN_REG);
    
        ret_val = -2;
    }

_out:
    /* set dma idle state. */
    pEndpointInfo->ulDmaInPhase = USB_DMA_IN_PH_IDLE;
    
    return ret_val;
}


int usbDmaStop(unsigned char uEpNo)
{
    if( !(USB_EPCTR_REG & USB_EPCTR_PLL_LOCK_BIT) ){
        udc_error("PLL is not locked\n");
        goto exit;
    }

    /* EP1-rx */
    if (USB_EP1_RX == uEpNo)
    {
        udc_debug("to stop EP1_RX--DMA. \n");
        
        /* AHB-EPC Bridge DMA    */ 
        USB_EP1DCR1_REG &= ~USB_EPDCR1_EPDMAREQEN_BIT;

        /* DMA  WAIT */
        msleep(1);
        
        /* EPC DMA  */
        USB_EP1DMACTRL_REG &= ~USB_EPDMACTRL_DMA_EN_BIT;
    }
    /* EP2-tx */
    else if (USB_EP2_TX == uEpNo)
    {
        udc_debug("to stop EP2_TX--DMA. \n");

        /* AHB-EPC Bridge DMA   */
        USB_EP2DCR1_REG &= ~USB_EPDCR1_EPDMAREQEN_BIT;

        /* DMA WAIT */
        msleep(1);

        /* EPC DMA  */
        USB_EP2DMACTRL_REG &= ~USB_EPDMACTRL_DMA_EN_BIT;
    }
    else
    {
        udc_error(" endpoint index=%u ERROR. \n", uEpNo);
        return -1;
    }

    udc_debug("exit---OK. \n");
    
exit:
    return 0;
}
