/*
 * Copyright (C) 2012 Sony Corporation 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License, version 2, as 
 * published by the Free Software Foundation. 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>

#include <linux/usb.h>
#include <linux/usb/gadget.h>
#include <linux/dma-mapping.h>

#include "cxd9225gg_udc.h"
#include "cxd9225gg_usbExt.h"
#include "cxd9225gg_usbEvalLib.h"


extern PUSBTARGET    g_UsbpTarget;

int usbDmaStop(unsigned char uEpNo);

/* forward declaration */
int usbEpcReset(void);
int usbBridgeInit(void);

int usbPowerStateControl(USB_POW_ST *pulPowState,
                                    USB_POW_EVENT ulPowEvent,
                                    int *pbPowerDownFlag);

int usbPowerResume(void);
int usbDirectPowerDown(void);
int usbReturnfromDirectPowerDown(void);

int usbFncAttach(struct cxd9225gg_udc *udc_dev);
int usbFncDetach(struct cxd9225gg_udc *udc_dev);
int usbFncEnable(PUSBTARGET pTarget);
int usbFncDisable(PUSBTARGET pTarget);

int usbEpcInit(void);
int usbIntcSetting(void);
int usbPlugIn(void);

int usbPlugOut(USB_POW_EVENT ulEvent);




int usbEpcReset(void)
{
    unsigned long  ulTmp = 0;

    /* EPC -- reset assert */
    if (USB_EPCTR_EPC_RST_BIT != (USB_EPCTR_REG & USB_EPCTR_EPC_RST_BIT))
    {
        USB_EPCTR_REG |= USB_EPCTR_EPC_RST_BIT;
        ulTmp = USB_EPCTR_REG;
        if(USB_EPCTR_EPC_RST_BIT != (ulTmp & USB_EPCTR_EPC_RST_BIT))
        {
            udc_error("[FATAL]EPC reset ERROR. \n");
            return -1;
        }
    }

    udelay(3);

    /* EPC -- reset de-assert. */
    USB_EPCTR_REG &= ~USB_EPCTR_EPC_RST_BIT;
    if(0 != (USB_EPCTR_REG & USB_EPCTR_EPC_RST_BIT))
    {
        udc_error("[FATAL]EPC un-reset ERROR. \n");
        return -1;
    }

    udc_debug("exit--0. \n");

    return 0;
}


int usbBridgeInit(void)
{
    unsigned long    ulTmp;

    /* AHB Slave (AHBSCTR Register) */
    USB_AHBSCTR_REG = USB_AHBSCTR_INIT;
    if(USB_AHBSCTR_INIT != (USB_AHBSCTR_REG & USB_AHBSCTR_INIT))
    {
        udc_error(" [FATAL] AHB Slave init error. \n");
        return -1;
    }

    /* AHB Master (AHBMCTR Register). */
    USB_AHBMCTR_REG = USB_AHBMCTR_INIT;
    if(USB_AHBMCTR_INIT != (USB_AHBMCTR_REG & USB_AHBMCTR_INIT))
    {
        udc_error("[FATAL] AHB Master init error. \n");
        return -1;
    }

    ulTmp = USB_AHBBINT_REG;
    USB_AHBBINT_REG = ulTmp;
    USB_AHBBINTEN_REG = USB_AHBBINTEN_INIT;
    if(USB_AHBBINTEN_INIT != (USB_AHBBINTEN_REG & USB_AHBBINTEN_INIT))
    {
        udc_error(" [FATAL]AHB Intr-Enable init error. \n");
        return -1;
    }

    udc_debug("exit--0. \n");

    return 0;
}


int usbPowerStateControl(USB_POW_ST *pulPowState,
                                    USB_POW_EVENT ulPowEvent,
                                    int *pbPowerDownFlag
                                    )
{
    USB_POW_ST ulCurrentState = *pulPowState;
    USB_POW_ST ulNextState = ulCurrentState;
    int      result = 0;

    switch (ulCurrentState)
    {
        case USB_POW_ST_DEFAULT:
          udc_debug("case--USB_POW_ST_DEFAULT, it's the initial-state. \n");
          /* initialize  */
          if(0 != usbEpcInit())
          {
            udc_error("usbEpcInit() failed in initialize.\n");
            return -1;
          }

          if(USB_POW_EVENT_INIT_ON == ulPowEvent)
          {
            /* suspend */
            USB_USBCTRL_REG |= USB_USBCTRL_SUSPEND_BIT;
            ulNextState = USB_POW_ST_POWER_DOWN;
            *pbPowerDownFlag = 1;
          }
          else if(USB_POW_EVENT_INIT_OFF == ulPowEvent)
          {
            /* direct power down */
            (void)usbDirectPowerDown();

            ulNextState = USB_POW_ST_DIRECT_PD;
            *pbPowerDownFlag = 1;
          }
          else
          {
            /* invalid */
            udc_error("USB_POW_ST_DEFAULT -- ulPowEvent is invalid. \n");
          }
          break;

        case USB_POW_ST_POWER_DOWN:
          udc_debug("case--USB_POW_ST_POWER_DOWN. \n");
          if(USB_POW_EVENT_ENABLE == ulPowEvent)
          {
            /* resume */
            (void)usbPowerResume();

            ulNextState = USB_POW_ST_POWER_ON;
            *pbPowerDownFlag = 0;
          }
          else if(USB_POW_EVENT_PLUGOUT == ulPowEvent)
          {
            /* direct power down */
            (void)usbDirectPowerDown();

            ulNextState = USB_POW_ST_DIRECT_PD;
            *pbPowerDownFlag = 1;
          }
          else if(USB_POW_EVENT_RESUME == ulPowEvent)
          {
            /* hardware automatically resume usb func */
            ulNextState = USB_POW_ST_POWER_ON;
            *pbPowerDownFlag = 0;
          }
          else
          {
            /* invalid */
            udc_error("USB_POW_ST_POWER_DOWN--ulPowEvent is invalid. \n");
          }
          break;

        case USB_POW_ST_DIRECT_PD:
          udc_debug("case--USB_POW_ST_DIRECT_PD. \n");
          if(USB_POW_EVENT_ENABLE == ulPowEvent)
          {
            ulNextState = USB_POW_ST_DPD_ENABLED;
            *pbPowerDownFlag = 1;
          }
          else if(USB_POW_EVENT_PLUGIN == ulPowEvent)
          {
            /* power on */
            (void)usbReturnfromDirectPowerDown();

            if(0 != usbEpcInit())
            {
              udc_error("[FATAL]EpcInit() Error in DPD to PowerOn. \n");
              return -1;
            }

            /* power down */
            USB_USBCTRL_REG |= USB_USBCTRL_SUSPEND_BIT;
            ulNextState = USB_POW_ST_POWER_DOWN;
            *pbPowerDownFlag = 1;
          }
          else
          {
            /* invalid */
            udc_error("USB_POW_ST_DIRECT_PD--ulPowEvent is invalid. \n");
          }
          break;

        case USB_POW_ST_DPD_ENABLED:
          udc_debug("case--USB_POW_ST_DPD_ENABLED. \n");
          if(USB_POW_EVENT_DISABLE == ulPowEvent)
          {
            ulNextState = USB_POW_ST_DIRECT_PD;
            *pbPowerDownFlag = 1;
          }
          else if(USB_POW_EVENT_PLUGIN == ulPowEvent)
          {
            /* power on */
            (void)usbReturnfromDirectPowerDown();

            /* DirectPowerDown, PowerOn */
            if(0 != usbEpcInit())
            {
              udc_error("[FATAL]EpcInit() Error in DPD to PowerOn. \n");
              return -1;
            }
            ulNextState = USB_POW_ST_POWER_ON;
            *pbPowerDownFlag = 0;
          }
          else
          {
            /* invalid */
            udc_error("USB_POW_ST_DPD_ENABLED--ulPowEvent is invalid. \n");
          }
          break;

        case USB_POW_ST_POWER_ON:
          udc_debug("case--USB_POW_ST_POWER_ON. \n");
          if((USB_POW_EVENT_DISABLE == ulPowEvent) || (USB_POW_EVENT_SUSPEND == ulPowEvent))
          {
            /* suspend */
            USB_USBCTRL_REG |= USB_USBCTRL_SUSPEND_BIT;
            ulNextState = USB_POW_ST_POWER_DOWN;
            *pbPowerDownFlag = 1;
          }
          else if(USB_POW_EVENT_PLUGOUT == ulPowEvent)
          {
            /* direct power down */
            (void)usbDirectPowerDown();

            ulNextState = USB_POW_ST_DPD_ENABLED;
            *pbPowerDownFlag = 1;
          }
          else if(USB_POW_EVENT_PLUGIN == ulPowEvent) {
            /* udc plug in, remain power state */
          }
          else if(USB_POW_EVENT_RESUME == ulPowEvent) {
            /* udc resume, remain power state */
          }
          else
          {
            /* invalid */
            udc_error("USB_POW_ST_POWER_ON--ulPowEvent is invalid. \n");
          }
          break;

        default:
          udc_error("case--default---error flow. \n");
          result = -1;
          break;
    }

    *pulPowState = ulNextState;

    udc_debug("cur(%u)->next(%u), DIRPD(%lu), PLL(%lu). \n",
              ulCurrentState,
              ulNextState,
              (USB_EPCTR_REG&USB_EPCTR_DIRPD_BIT)>>12,
              (USB_EPCTR_REG&USB_EPCTR_PLL_LOCK_BIT)>>4 );
    udc_debug("to msleep--1ms. \n");
    msleep(1);

    udc_debug("re cur(%u)->next(%u), DIRPD(%lu), PLL(%lu). \n",
              ulCurrentState,
              ulNextState,
              (USB_EPCTR_REG&USB_EPCTR_DIRPD_BIT)>>12,
              (USB_EPCTR_REG&USB_EPCTR_PLL_LOCK_BIT)>>4 );

    return result;
}


int usbPowerResume(void)
{
    unsigned long  ulCnt;

    /* pll resume. */
    USB_EPCTR_REG |= USB_EPCTR_PLL_RESUME_BIT;

    for (ulCnt = 0; ulCnt < 1000; ulCnt++)
    {
        if ( USB_EPCTR_PLL_LOCK_BIT == (USB_EPCTR_REG & USB_EPCTR_PLL_LOCK_BIT) )
        {
            USB_EPCTR_REG &= ~USB_EPCTR_PLL_RESUME_BIT;
            break;
        }
    }

    if (ulCnt >= 1000)
    {
        udc_error("exit --- ERROR. \n");
        return -1;
    }

    udc_debug("exit --- OK. \n");

    return 0;
}


int usbDirectPowerDown(void)
{
    if(USB_EPCTR_VBUS_LEVEL_BIT == (USB_EPCTR_REG & USB_EPCTR_VBUS_LEVEL_BIT))
    {
        udc_error("exit --- ERROR. \n");
        return -1;
    }
    else
    {
        /* USB direct power-down. */
        USB_EPCTR_REG |= USB_EPCTR_DIRPD_BIT;
    }

    udc_debug("exit --- OK. \n");

    return 0;
}



/* functions */
int usbReturnfromDirectPowerDown(void)
{
    USB_EPCTR_REG |= USB_EPCTR_EPC_RST_BIT;
    if(USB_EPCTR_EPC_RST_BIT != (USB_EPCTR_REG & USB_EPCTR_EPC_RST_BIT))
    {
        udc_error(" [FATAL]EPC reset ERROR. \n");
        return -1;
    }

    udelay(2);

    USB_EPCTR_REG &= ~USB_EPCTR_DIRPD_BIT;
    if(USB_EPCTR_DIRPD_BIT == (USB_EPCTR_REG & USB_EPCTR_DIRPD_BIT))
    {
        udc_error("[FATAL]DPD unreset ERROR. \n");
        return -1;
    }

    msleep(1);

    USB_EPCTR_REG &= ~USB_EPCTR_EPC_RST_BIT;
    if(USB_EPCTR_EPC_RST_BIT == (USB_EPCTR_REG & USB_EPCTR_EPC_RST_BIT) )
    {
        udc_error("[FATAL]EPC un-reset ERROR. \n");
        return -1;
    }

    if(USB_EPCTR_PLL_LOCK_BIT != (USB_EPCTR_REG & USB_EPCTR_PLL_LOCK_BIT) )
    {
        udc_error("[FATAL]Clock check ERROR. \n");
        return -1;
    }

    udc_debug("exit -- OK. \n");

    return 0;
}


int usbFncAttach(struct cxd9225gg_udc *udc_dev)
{
    int ret_val;
    unsigned long  ulTmp;
    PUSBTARGET  pTarget;

    if ( !udc_dev || (NULL == udc_dev->usb_target))
    {
        udc_error(" input arg -- udc_dev is NULL. \n");
        return -1;
    }

    pTarget = udc_dev->usb_target;

    udc_debug(" USB_AHBBINT_REG = 0x%04x\n", (unsigned int)USB_AHBBINT_REG);

    /* attention: clear all bits of AHB-intr-register,
     * write '1' to clean.
     * and some reserved bits must be 0.
     */
    ulTmp = USB_AHBBINT_REG;
    USB_AHBBINT_REG = ulTmp;
    ulTmp = USB_AHBBINT_REG;    /* Dummy read. */

    /* enable VBUS intr --- for ATTACH or detach. */
    USB_AHBBINTEN_REG = USB_AHBBINTEN_VBUS_INTEN_BIT;

    /* request INT_USBDEV interrupt. */
    ret_val = request_irq(INT_USBDEV_EPC, usbEpcIsr, IRQF_DISABLED, "usbEpcIsr", udc_dev);
    if (ret_val)
    {
        udc_error(" request irq -- usbEpcIsr failed. \n");

        free_irq(INT_USBDEV_EPC, udc_dev);

        return -EINVAL;
    }

    udc_debug("exit--OK. \n");

    return 0;
}



int usbFncDetach(struct cxd9225gg_udc *udc_dev)
{
    if (NULL == udc_dev)
    {
        udc_error(" input arg -- udc_dev is NULL. \n");
        return -1;
    }

    if (USB_EPCTR_PLL_LOCK_BIT == (USB_EPCTR_REG & USB_EPCTR_PLL_LOCK_BIT))
    {
        /* PLL Mask */
        USB_USBINTEN_REG = 0;  /* all mask EPC */
    }

    /* disconnect EPC interrupt. */
    free_irq(INT_USBDEV_EPC, udc_dev);

    return 0;
}



int usbFncEnable(PUSBTARGET pTarget)
{
    USB_CONNECT_STATUS     uStatus = 0;
    int ret_val;
    unsigned long flags;

    udc_debug(" entry test. \n");

//internal function, no arg-check.
    if (NULL == pTarget)
    {
        udc_error(" input arg -- pTarget is NULL. \n");
        return -1;
    }

    spin_lock_irqsave(&pTarget->lock, flags);
    pTarget->usbEnableFlg = 1;
    uStatus = pTarget->usbConnectStatus;
    spin_unlock_irqrestore(&pTarget->lock, flags);

    ret_val = wait_for_completion_interruptible_timeout(&pTarget->sUsbPlugInOutSemId,
                                                        msecs_to_jiffies(USB_FUNC_TIMEOUT));
    if (!ret_val || (-ERESTARTSYS == ret_val))
    {
        if (-ERESTARTSYS == ret_val)
        {
            udc_error("it means this thread is killed, for PlugInOutSemId. \n");
            return -1;
        }

        udc_error("it means timeout--PlugInOutSemId. \n");
        ret_val = -1;
        flags = 0;
    }
    else
    {
        flags = 1;
        ret_val = 0;
    }

    /* power enable. */
    ret_val = usbPowerStateControl(&pTarget->ulPowState,
                                              USB_POW_EVENT_ENABLE,
                                              &pTarget->bIsPllUnlocked);
    if (ret_val)
    {
        udc_error(" [FATAL] Power state change error. \n");

        if (flags)
        {
            complete(&(pTarget->sUsbPlugInOutSemId));
        }

        return -1;
    }

    /* USB: connected or not. */
    if ( uStatus == USB_CONNECTED_STATUS )
    {
        udc_debug(" usb is connected, and begin to handle plug-in. \n");

        ret_val = usbPlugIn();
    }

    /* release the mutex. */
    if (flags)
    {
        complete(&(pTarget->sUsbPlugInOutSemId));
    }

    if (ret_val)
    {
        udc_error(" usbPlugIn() failed. \n");
        return -1;
    }

    udc_debug(" exit OK. \n");

    return 0;
}



int usbFncDisable(PUSBTARGET pTarget)
{
    int          uStatus = 0;
    int          ret_val = 0;
    unsigned long flags;

    udc_debug("entry test. \n");

    if (NULL == pTarget)
    {
        udc_error(" input arg -- pTarget is NULL. \n");
        return -1;
    }

    spin_lock_irqsave(&pTarget->lock, flags);
    pTarget->usbEnableFlg = 0;
    uStatus = pTarget->usbConnectStatus;
    spin_unlock_irqrestore(&pTarget->lock, flags);

    /* USB */
    if (uStatus == USB_CONNECTED_STATUS)
    {
        ret_val = wait_for_completion_interruptible_timeout(&pTarget->sUsbPlugInOutSemId,
                                                            msecs_to_jiffies(USB_FUNC_TIMEOUT));
        if (!ret_val || (-ERESTARTSYS == ret_val))
        {
            if (-ERESTARTSYS == ret_val)
            {
                udc_error("it means this thread is killed, for PlugInOutSemId. \n");
                return -1;
            }

            udc_error("it means timeout. \n");
            flags = 0;
        }
        else
        {
            flags = 1;
        }

        /* to handle plug-out, and then power off. */
        ret_val = usbPlugOut(USB_POW_EVENT_DISABLE);

        if (flags)
        {
            complete(&(pTarget->sUsbPlugInOutSemId));
        }

        /* attention to its position. */
        if (ret_val)
        {
            udc_error(" usbPlugOut() failed. \n");
            return -1;
        }
    }
    else
    {
        ret_val = wait_for_completion_interruptible_timeout(&pTarget->sUsbPlugInOutSemId,
                                                            msecs_to_jiffies(USB_FUNC_TIMEOUT));
        if (!ret_val || (-ERESTARTSYS == ret_val))
        {
            if (-ERESTARTSYS == ret_val)
            {
                udc_error("it means this thread is killed, for PlugInOutSemId. \n");
                return -1;
            }

            udc_error("it means timeout. \n");
            flags = 0;
        }
        else
        {
            flags = 1;
        }

        /* USB disable. */
        ret_val = usbPowerStateControl(&pTarget->ulPowState,
                                                USB_POW_EVENT_DISABLE,
                                                &pTarget->bIsPllUnlocked);

        if (flags)
        {
            complete(&(pTarget->sUsbPlugInOutSemId));
        }

        if (ret_val)
        {
            udc_error(" [FATAL] Power state change error. \n");
            return -1;
        }
    }

    udc_debug("exit...OK \n");

    return 0;
}


int usbEpcInit(void)
{
    /*  unsigned long    ulTmp; */
    udc_debug("entry test. \n");

    /* SOF (USB Control Register). */
    udc_debug("before set USBCTRL=0x%08x. \n", (unsigned int)USB_USBCTRL_REG);

    USB_USBCTRL_REG |= USB_USBCTRL_INIT;

    udc_debug("after set USBCTRL=0x%08x. \n", (unsigned int)USB_USBCTRL_REG);

    if(USB_USBCTRL_INIT != (USB_USBCTRL_REG & USB_USBCTRL_INIT))
    {
        udc_error("[FATAL]USBCRL_INIT Set Error. \n");
        return -1;
    }

    /* EP1 (EP1 Control Register). */
    USB_EP1CTRL_REG |= (USB_EPCTRL_DIR0_OUT | USB_EPCTRL_AUTO_BIT);
    if( (USB_EPCTRL_DIR0_OUT | USB_EPCTRL_AUTO_BIT)
        != (USB_EP1CTRL_REG & (USB_EPCTRL_DIR0_OUT | USB_EPCTRL_AUTO_BIT))
       )
    {
        udc_error("[FATAL]EP1_CTRL Set Error. \n");
        return -1;
    }

    /* EP2 (EP2 Control Register). */
    USB_EP2CTRL_REG &= ~USB_EPCTRL_DIR0_OUT;
    USB_EP2CTRL_REG |= USB_EPCTRL_AUTO_BIT;
    if( USB_EPCTRL_AUTO_BIT
        != (USB_EP2CTRL_REG & (USB_EPCTRL_DIR0_OUT | USB_EPCTRL_AUTO_BIT))
       )
    {
        udc_error(" [FATAL]EP2_CTRL Set Error. \n");
        return -1;
    }

    /* attention: it's a 32-bit word address, NOT a byte-address. */
    USB_EP1PCKTADDR_REG = USB_EP1_BASEAD;
    USB_EP2PCKTADDR_REG = USB_EP2_BASEAD;

    udc_debug("exit--OK. \n");

    return 0;
}



int usbIntcSetting(void)
{
    /* ATTACH Connect Enable DETACH Disable Disconnect */
    int ret = 0;

  #ifdef USBFUNC_U2F_EPC_INT_EDGE
    udc_debug("intr trigger-type: rising-edge has been already configured. \n");
  #else
    udc_error("fatal error, you should re-configure the intr trigger type. \n");
    ret = -1;
  #endif

    return ret;
}


int usbPlugIn(void)
{
    int  ret = 0;
    unsigned long  tmpReg = 0;
    unsigned long  regIntDummy = 0;

    /* PlugIn, USB Enable, Active  */

    udc_debug("Before PlugIn USB_EPCTR_REG=0x%08x. \n", (unsigned int)USB_EPCTR_REG);

    /* check vbus level bit. */
    if( USB_EPCTR_VBUS_LEVEL_BIT
        != ( USB_EPCTR_REG & USB_EPCTR_VBUS_LEVEL_BIT )
       )
    {
        udc_error("[FATAL] VBUS Level Check Error. \n");
        return -1;
    }

    ret = usbPowerStateControl(&g_UsbpTarget->ulPowState,
                                          USB_POW_EVENT_PLUGIN,
                                          &g_UsbpTarget->bIsPllUnlocked);

    g_UsbpTarget->usbPlugFlg = 1;

    if ( USB_POW_ST_POWER_ON == g_UsbpTarget->ulPowState )
    {
        udc_debug("PHY connect!!\n");

        /* USB PHY  */
        tmpReg = USB_USBCTRL_REG;
        tmpReg &= ~USB_USBCTRL_CONNECTB_BIT;
        tmpReg |= USB_USBCTRL_PUE2_BIT;
        USB_USBCTRL_REG = tmpReg;

        USB_USBINTEN_REG = USB_USBINTEN_INIT;
        USB_USBINTSTS_REG = 0; /*   */
        regIntDummy = USB_USBINTSTS_REG;
    }

    udc_debug("After PlugIn USB_EPCTR_REG=0x%08x, and exit. \n",
              (unsigned int)USB_EPCTR_REG);

    return 0;
}


int usbPlugOut(USB_POW_EVENT ulEvent)
{
    unsigned long        ulTmp = 0;
    unsigned long        tmpReg = 0;

    udc_debug("Before PlugOut USB_EPCTR_REG=0x%08x. \n",
              (unsigned int)USB_EPCTR_REG);

    /* semaphore API take. */
    mutex_lock(&g_UsbpTarget->hSemApiExecuting);

    if ( USB_POW_ST_POWER_ON == g_UsbpTarget->ulPowState )
    {
        if ( 1 == g_UsbpTarget->bIsPllUnlocked )
        {
            /* PLL Unlocked, Resume,Suspend mass storage mode off. */
            udc_debug(" PlugOut(Disable) -> PLL resume. \n");

            usbPowerResume();

            g_UsbpTarget->bIsPllUnlocked = FALSE;
        }

        /* Bridge  */
        USB_AHBBINTEN_REG = USB_AHBBINTEN_ALL_DISABLE;
        if (USB_AHBBINTEN_ALL_DISABLE != USB_AHBBINTEN_REG)
        {
            mutex_unlock(&g_UsbpTarget->hSemApiExecuting);

            udc_error("AHBBINTEN Disable Error 0x%08x. \n",
                (unsigned int)USB_AHBBINTEN_REG);

            return -1;
        }

        /* EPC  */
        USB_USBINTEN_REG = USB_USBINTEN_ALL_DISABLE;
        if (USB_USBINTEN_ALL_DISABLE != USB_USBINTEN_REG)
        {
            mutex_unlock(&g_UsbpTarget->hSemApiExecuting);

            udc_error("USBINTEN Disable Error 0x%08x. \n",
                      (unsigned int)USB_USBINTEN_REG);

            return -1;
        }

        tmpReg &= ~(USB_USBSTS_CONF_BIT | USB_USBSTS_DEFAULT_BIT);
        USB_USBCTRL_REG = tmpReg;

        /* EP1 DMA --- DMA  */
        if (USB_EPDMACTRL_DMA_EN_BIT == (USB_EP1DMACTRL_REG & USB_EPDMACTRL_DMA_EN_BIT) )
        {
            usbDmaStop( USB_EP1_RX );
        }

        /* EP2 DMA --- DMA.*/
        if (USB_EPDMACTRL_DMA_EN_BIT == (USB_EP2DMACTRL_REG & USB_EPDMACTRL_DMA_EN_BIT) )
        {
            usbDmaStop( USB_EP2_TX );
        }

        /* USB PHY */
        tmpReg = USB_USBCTRL_REG;
        tmpReg |= USB_USBCTRL_CONNECTB_BIT;
        tmpReg &= ~USB_USBCTRL_PUE2_BIT;
        USB_USBCTRL_REG = tmpReg;
    }

    /* EPC */
    if (usbPowerStateControl(&g_UsbpTarget->ulPowState,
                                        ulEvent /* USB_POW_EVENT_PLUGOUT*/,
                                        &g_UsbpTarget->bIsPllUnlocked) != OK)
    {
        udc_error(" Power state change error!! \n");

        mutex_unlock(&g_UsbpTarget->hSemApiExecuting);

        return -1;
    }

    /* Bridge, VBUS  */
    USB_AHBBINTEN_REG = USB_AHBBINTEN_VBUS_INTEN_BIT;

    udc_debug("VBUS INT... \n");

    ulTmp = USB_AHBBINTEN_REG;
    if (USB_AHBBINTEN_VBUS_INTEN_BIT != (ulTmp | USB_AHBBINTEN_VBUS_INTEN_BIT) )
    {
        mutex_unlock(&g_UsbpTarget->hSemApiExecuting);

        udc_error(" [FATAL]VBUS Int Enable Error!! \n");
        return -1;
    }

    g_UsbpTarget->usbPlugFlg = 0;

    mutex_unlock(&g_UsbpTarget->hSemApiExecuting);

    udc_debug(" After PlugOut USB_EPCTR_REG=0x%08x. \n",
              (unsigned int)USB_EPCTR_REG);

    return 0;
}
