/*
 * Copyright (C) 2012 Sony Corporation
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/irq.h>

#include <linux/usb.h>
#include <linux/usb/gadget.h>
#include <linux/dma-mapping.h>

#include "cxd9225gg_udc.h"
#include "cxd9225gg_usbExt.h"
#include "cxd9225gg_usbEvalLib.h"

//#define UDC_DEAMON_WATCH_DOG

#ifdef WAIT_FOREVER
#undef WAIT_FOREVER
#endif
#define WAIT_FOREVER    LONG_MAX

extern void cxd9225gg_udc_quiesce(struct cxd9225gg_udc *udc);

extern PUSBTARGET    g_UsbpTarget;

extern int usbDmaStop(unsigned char uEpNo);

extern int g_bDebugBufWriteEnable;

#ifdef _USB_MEASUREMENT_INT
extern void usbTime ( int no );
#endif /* D_USB_MEASUREMENT_INT */

int usbIsrExec(void *data);
int usbFncInterruptStatusGet(PUSBTARGET pTarget);
int usbFncInterruptStatusClear(PUSBTARGET pTarget);
int usbFncEndpointIntStatusGet(PUSBTARGET pTarget, int ep_num);
int usbFncEndpointIntStatusClear(PUSBTARGET pTarget, int ep_num);
int usbFncHandleDisconnectInterrupt(PUSBTARGET pTarget);
int usbFncHandleResumeInterrupt(PUSBTARGET pTarget);
int usbFncHandleResetInterrupt(PUSBTARGET pTarget);
int usbFncHandleSuspendInterrupt(PUSBTARGET pTarget);

irqreturn_t usbBridgeIsr(int irq, void *dev_id)
{
    udc_debug("irq=%d\n", irq);

    /* backup BRIDGE intrrupt-enable-register. */
    g_UsbpTarget->ulAhbbIntEnBackup = USB_AHBBINTEN_REG;

    /* disable all BRIDGE interrupts. */
    USB_AHBBINTEN_REG = 0;

    if (0 != (USB_AHBBINT_REG & USB_AHBBINT_VBUS_INT_BIT))
    {
        /* attention: it is a bug,
         * since we should write '1' to clear it.
         */
        USB_AHBBINT_REG = USB_AHBBINT_VBUS_INT_BIT;

        g_UsbpTarget->bVbusIntOccured = 1;
        UDC_STAT_VBUS_INTR_INC();

        /* VBUS? */
        udc_debug("USB_AHBBINT_REG=0x%04x, send a sema to vbus-wait-task. \n",
                  (unsigned int)USB_AHBBINT_REG);

        /* notify vbus wait task. */
        complete(&(g_UsbpTarget->hSem4VbusWaitTask));
    }
    else
    {
        /* BRIDGE int. */
        g_UsbpTarget->uInterruptAssign |= USB_BRIDGE_INT;

        /* VBUS  */
        UDC_STAT_BRG_INTR_INC();
        complete(&(g_UsbpTarget->sUsbIntSemId));
    }

    return IRQ_HANDLED;
}


irqreturn_t usbEpcIsr(int irq, void *dev_id)
{
    struct cxd9225gg_udc       *udc_dev = (struct cxd9225gg_udc *)dev_id;

    udc_debug("irq=%d\n", irq);

    udc_dev = udc_dev;

    /* EPC intr flag. */
    g_UsbpTarget->uInterruptAssign |= USB_EPC_INT;

    /* backup EPC interrupt-enable-register.  */
    g_UsbpTarget->ulUsbIntEnBackup = USB_USBINTEN_REG;

    /* disable all EPC interrupts. */
    USB_USBINTEN_REG = 0;

    UDC_STAT_EPC_INTR_INC();
    complete(&(g_UsbpTarget->sUsbIntSemId));

    return IRQ_HANDLED;
}

#ifdef UDC_DEAMON_WATCH_DOG
void udc_deamon_timeout(unsigned long data)
{
	struct timer_list *tmr = (struct timer_list *)data;
	complete(&(g_UsbpTarget->sUsbIntSemId));
	mod_timer(tmr, jiffies + msecs_to_jiffies(USB_FUNC_TIMEOUT/2));
	UDC_STAT_WATCH_DOG_INC();
}
struct timer_list udc_timer = 
	TIMER_INITIALIZER(udc_deamon_timeout,0,(unsigned long)&udc_timer);
#endif /* UDC_DEAMON_WATCH_DOG */

int usbIsrExec(void *data)
{
    struct cxd9225gg_udc           *udc_dev;
    PUSBTARGET       pTarget    = NULL;     /* USBTARGET */
    int                         ret_val = -1;
    unsigned long               regIntStatus = 0;
    unsigned long               regIntEP0Status = 0;
    unsigned long               regIntEP1Status = 0;
    unsigned long               regIntEP2Status = 0;
    unsigned long               regIntBridgeStatus = 0;
    int                         bIsCallHalIsr = 0;
    PUSBENDPOINT     pEpInfo0Rcv;
    PUSBENDPOINT     pEpInfo0Send;
    PUSBENDPOINT     pEpInfo1Rcv;
    PUSBENDPOINT     pEpInfo2Send;

    int                         bIsCallPlugIn = 0;
    int                         bIsHandledEpcInt = 0;
    int                         bIsHandledBridgeInt = 0;
    int                         bReCheckFlag = 0;
    CXD9225GG_INT32                   lTimeout = WAIT_FOREVER;
    unsigned long               ulReInformCount = 0;
    unsigned long               regIntDummy = 0;
    unsigned long long loop = 0;

    pTarget = (PUSBTARGET)data;

    pEpInfo0Rcv = *pTarget->pEndPointInfoTop;
    pEpInfo0Send = *(pTarget->pEndPointInfoTop + 1);
    pEpInfo1Rcv = *(pTarget->pEndPointInfoTop + 2);
    pEpInfo2Send = *(pTarget->pEndPointInfoTop + 3);

#ifdef UDC_DEAMON_WATCH_DOG
    udc_deamon_timeout((unsigned long)&udc_timer);
#endif /* UDC_DEAMON_WATCH_DOG */

    /* main loop. */
    while (1)
    {
    	loop++;
        /*  WAIT (BridgeIsr/EpcIsr?semGive ) */
        UDC_STAT_ISR_STEP(__LINE__);
        if (1 != bReCheckFlag)
        {
            UDC_STAT_ISR_STEP(__LINE__);
            /* wait for a semaphore to contiue... */
            if (WAIT_FOREVER == lTimeout)
            {
                UDC_STAT_ISR_STEP(__LINE__);
                ret_val = wait_for_completion_interruptible(&(pTarget->sUsbIntSemId));
                UDC_STAT_ISR_STEP(__LINE__);
                if (-ERESTARTSYS == ret_val)
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    udc_error("it means this kthread--usb-isr is killed. \n");
                    return -1;
                }
            }
            else
            {
                UDC_STAT_ISR_STEP(__LINE__);
                ret_val = wait_for_completion_interruptible_timeout(
                              &(pTarget->sUsbIntSemId), lTimeout);
                UDC_STAT_ISR_STEP(__LINE__);
                if (!ret_val || (-ERESTARTSYS == ret_val))
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    if (-ERESTARTSYS == ret_val)
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        udc_error("it means this kthread--usb-isr is killed. \n");
                        return -1;
                    }
                }
            }
            UDC_STAT_ISR_STEP(__LINE__);
            UDC_STAT_ISR_WAKEUP_INC();

            /* consume all the pending completion. */
            UDC_STAT_ISR_STEP(__LINE__);
            while (try_wait_for_completion(&pTarget->sUsbIntSemId));
            UDC_STAT_ISR_STEP(__LINE__);
        }
        else {
            UDC_STAT_ISR_STEP(__LINE__);
            UDC_STAT_ISR_LOOP_INC();
        }

        UDC_STAT_ISR_STEP(__LINE__);

        bReCheckFlag = 0;

        lTimeout = WAIT_FOREVER;

        udc_dev = pTarget->udc_dev;
        if (!udc_dev)
        {
            UDC_STAT_ISR_STEP(__LINE__);
            udc_error("-------------udc_dev is NULL, to continue----------\n");
            continue;
        }

        regIntStatus = 0;
        regIntEP0Status = 0;
        regIntEP1Status = 0;
        regIntEP2Status = 0;
        regIntBridgeStatus = 0;
        bIsCallHalIsr = 0;
        bIsCallPlugIn = 0;

        UDC_STAT_ISR_STEP(__LINE__);
        if (0 != (pTarget->uInterruptAssign & USB_VBUS_TASK_INT))
        {
            UDC_STAT_ISR_STEP(__LINE__);
            udc_debug(" VBUS_TASK_INT DoPlugIn=%d,DoPlugOut=%d\n",
                      pTarget->bDoPlugIn, pTarget->bDoPlugOut);
            if (1 == pTarget->bDoPlugIn)
            {
                pTarget->bDoPlugIn = 0;
                udc_debug(" PlugIn Called. !!\n");

                /* lock. */
                UDC_STAT_ISR_STEP(__LINE__);
                ret_val = wait_for_completion_interruptible_timeout(&(pTarget->sUsbPlugInOutSemId),
                                                                    msecs_to_jiffies(USB_FUNC_TIMEOUT));
                UDC_STAT_ISR_STEP(__LINE__);
                if (!ret_val || (-ERESTARTSYS == ret_val))
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    if (-ERESTARTSYS == ret_val)
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        udc_error("it means this kthread--usb-isr is killed. \n");
                        return -1;
                    }

                    udc_error("it means timeout. \n");
                    regIntDummy= 0;
                }
                else
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    regIntDummy = 1;
                }

                /* handle plug-in. */
                UDC_STAT_ISR_STEP(__LINE__);
                ret_val = usbPlugIn();
                if (0 != ret_val)
                {
                    /* error */
                    udc_error(" [FATAL]usbPlugIn Fail !!\n");
                }

                /* to notify application -- usb plug-in. */
                UDC_STAT_ISR_STEP(__LINE__);
                usbConnectionNotify(1);

                /* unlock. */
                if (regIntDummy)
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    complete(&(pTarget->sUsbPlugInOutSemId));
                }
            }

            UDC_STAT_ISR_STEP(__LINE__);
            if (1 == pTarget->bDoPlugOut)
            {
                UDC_STAT_ISR_STEP(__LINE__);
                pTarget->bDoPlugOut = 0;

                /* Connect -> Disconnect */
                pTarget->uInterruptStatus |= USB_DISCONNECT_INTERRUPT_MASK;

                /* Interrupt Disable */
                spin_lock_irqsave(&udc_dev->lock, regIntDummy);

                pTarget->uInterruptAssign &= ~USB_EPC_INT;

                /* Interrupt Enable */
                spin_unlock_irqrestore(&udc_dev->lock, regIntDummy);
            }

            /* Interrupt Disable */
            UDC_STAT_ISR_STEP(__LINE__);
            spin_lock_irqsave(&udc_dev->lock, regIntDummy);

            pTarget->uInterruptAssign &= ~USB_VBUS_TASK_INT;

            /* Interrupt Enable */
            spin_unlock_irqrestore(&udc_dev->lock, regIntDummy);
        }
        else if (0 != (pTarget->uInterruptAssign & USB_EPC_INT))
        {
            UDC_STAT_ISR_STEP(__LINE__);
            /* read and clear all bits of usb intr-reg. */
            regIntStatus = USB_USBINTSTS_REG;
            USB_USBINTSTS_REG &= ~regIntStatus;
            regIntDummy = USB_USBINTSTS_REG;       /* Dummy Read */

            bIsHandledEpcInt = 1;

            udc_debug("USB_USBINTSTS_REG = 0x%08x. \n", (unsigned int)regIntStatus);

            /****** handle intr Request *******
             * disconnect interrupt
             * reset interrupt
             * suspend interrupt
             * resume interrupt
             */
            if( 0 != (regIntStatus & USB_USBINTSTS_RSUM_INT_BIT) )
            {
                UDC_STAT_ISR_STEP(__LINE__);
                pTarget->uInterruptStatus |= USB_RESUME_INTERRUPT_MASK;

                /* attention:
                 * we would make vbus-task to notify resume,
                 * also to notify queue-thread to continue....
                 */

                udc_debug(" handle intr-RESUME. \n");

                /* VBUS RESUME?PLL OFF? */
                pTarget->bResumeRequested = 1;
                pTarget->bIsPllUnlocked = 0;

                /* to notify vbus-task to handle 'resume'. */
                complete(&(pTarget->hSem4VbusWaitTask));
            }

            /* handle suspend-intr. */
            UDC_STAT_ISR_STEP(__LINE__);
            if ( 0 != (regIntStatus & USB_USBINTSTS_SPND_INT_BIT) )
            {
                UDC_STAT_ISR_STEP(__LINE__);
                pTarget->uInterruptStatus |= USB_SUSPEND_INTERRUPT_MASK;

                /* attention:
                 * we would make vbus-task to notify suspend,
                 * also to notify queue-thread to continue....
                 */

                udc_debug(" handle intr-SUSPEND. \n");

                /*  VBUS RESUME?PLL OFF. */
                pTarget->bSuspendRequested = 1;

                /* to notify vbus-task to handle 'suspend'. */
                complete(&(pTarget->hSem4VbusWaitTask));
            }

            /* handle reset-intr. */
            UDC_STAT_ISR_STEP(__LINE__);
            if( 0 != (regIntStatus & USB_USBINTSTS_USB_RST_INT_BIT) )
            {
                UDC_STAT_ISR_STEP(__LINE__);
                pTarget->ulDeviceState = USB_DEV_ST_DEFAULT;
                pTarget->uInterruptStatus |= USB_RESET_INTERRUPT_MASK;

                /* attention:
                 * we would make vbus-task to notify reset,
                 * ....
                 */

                udc_debug(" handle intr-RESET.\n");

                pTarget->bResetRequested = 1;
                pTarget->bIsPllUnlocked = 0;

                /* to notify vbus-task to handle 'reset'. */
                complete(&(pTarget->hSem4VbusWaitTask));

                usbFncHandleResetInterrupt(pTarget);
            }
        }

        /* whether is there a bridge intr.  */
        UDC_STAT_ISR_STEP(__LINE__);
        if (0 != (pTarget->uInterruptAssign & USB_BRIDGE_INT))
        {
            UDC_STAT_ISR_STEP(__LINE__);
            /* read and clear all bits of AHB-EPC register.
             * attention: write '1' to clear bit...
             */
            regIntBridgeStatus = USB_AHBBINT_REG;
            USB_AHBBINT_REG = regIntBridgeStatus;  /* bit = 1 to clear. */
            regIntDummy = USB_AHBBINT_REG;         /* Dummy Read */
            bIsHandledBridgeInt = 1;

            udc_debug(" USB_AHBBINT_REG = 0x%02x. \n", (unsigned int)regIntBridgeStatus);

            /* bus error happened just now. */
            if( 0 != (regIntBridgeStatus & USB_AHBBINT_BUSERR_MASK) )
            {
                UDC_STAT_ISR_STEP(__LINE__);
                /* BUS ERROR */
                udc_error(" AHB BUS ERROR occured, AHBBINT=0x%08x. \n",
                          (unsigned int)regIntBridgeStatus);

                /* Interrupt Disable */
                spin_lock_irqsave(&udc_dev->lock, regIntDummy);

                pTarget->ulAhbbIntEnBackup &= ~(regIntBridgeStatus & USB_AHBBINT_BUSERR_MASK);

                /* Interrupt Enable */
                spin_unlock_irqrestore(&udc_dev->lock, regIntDummy);

                pTarget->bIsBusError = 1;
            }
        }

        UDC_STAT_ISR_STEP(__LINE__);
        /* whether is the EPC power-on now.  */
        if ( 1 != pTarget->bIsPllUnlocked )
        {
            UDC_STAT_ISR_STEP(__LINE__);
            udc_debug("EPC-power ON now, so begin to handle EPC-interrupt. \n");

            /* EP0 */
            if ( 0 != (regIntStatus & USB_USBINTSTS_EP0_INT_BIT) )
            {
                UDC_STAT_ISR_STEP(__LINE__);
                /* read ep0 status reg, and then cleanup. */
                regIntEP0Status = USB_EP0STS_REG;
                USB_EP0STS_REG = ~regIntEP0Status;
                regIntDummy = USB_EP0STS_REG;          /* Dummy Read */

                udc_debug(" USB_EP0STS_REG = 0x%08x\n", (unsigned int)regIntEP0Status);
            }

            /* ep1 */
            if ( 0 != (regIntStatus & USB_USBINTSTS_EP1_INT_BIT) )
            {
                UDC_STAT_ISR_STEP(__LINE__);
                /* read ep1 status reg, and then cleanup. */
                regIntEP1Status = USB_EP1STS_REG;
                USB_EP1STS_REG = ~regIntEP1Status;
                regIntDummy = USB_EP1STS_REG;          /* Dummy Read */

                udc_debug(" USB_EP1STS_REG = 0x%08x\n", (unsigned int)regIntEP1Status);
            }

            /* ep2 */
            if ( 0 != (regIntStatus & USB_USBINTSTS_EP2_INT_BIT) )
            {
                UDC_STAT_ISR_STEP(__LINE__);
                /* read ep2 status reg, and then cleanup. */
                regIntEP2Status = USB_EP2STS_REG;
                USB_EP2STS_REG = ~regIntEP2Status;
                regIntDummy = USB_EP2STS_REG;          /* Dummy Read */

                udc_debug(" USB_EP2STS_REG = 0x%08x\n", (unsigned int)regIntEP2Status);
            }

            /* PIO OUT of ep0. */
            UDC_STAT_ISR_STEP(__LINE__);
            if ( 0 != pEpInfo0Rcv->bIsAssigned )
            {
                /* control ep0 out. */
                UDC_STAT_ISR_STEP(__LINE__);
                if ( 1 != pTarget->bIsCtrlTransOnGoing )
                {
                    /* ep0 NULL data received. */
                    UDC_STAT_ISR_STEP(__LINE__);
                    if ( 0 != ( regIntEP0Status & USB_EP0STS_OUT_NULL_INT_BIT) )
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        udc_error("ep0Rcv receive a null data... \n");

                        /* NULL data. */
                        if ( USB_CTRL_TRANS_READ == pTarget->ulCtrlTransType )
                        {
                            UDC_STAT_ISR_STEP(__LINE__);
                            /* do nothing */
                            udc_error("ulCtrlTransType--error, should be receive from host. \n");
                        }
                        else
                        {
                            UDC_STAT_ISR_STEP(__LINE__);
                            pTarget->uInterruptStatus |= USB_ENDPOINT_INTERRUPT_MASK;
                            pEpInfo0Rcv->uEndpointInterruptStatus |= USB_ENDPOINT_OUT_PID_MASK;

                            pEpInfo0Rcv->ulPipeTransferType = USB_TRANSFER_TYPE_NULL;

                            udc_error("---------bIsCallHalIsr=1: PIO OUT(NULL)----------. \n");
                        }
                    }

                    /* ep0 non-NULL data received. */
                    UDC_STAT_ISR_STEP(__LINE__);
                    if ( 0 != ( regIntEP0Status & USB_EP0STS_OUT_INT_BIT) )
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        udc_error("receive NON-null data on ep0. \n");

                        pTarget->uInterruptStatus |= USB_ENDPOINT_INTERRUPT_MASK;
                        pEpInfo0Rcv->uEndpointInterruptStatus |= USB_ENDPOINT_OUT_PID_MASK;

                        pEpInfo0Rcv->ulPipeTransferType = USB_TRANSFER_TYPE_PIO;

                        udc_debug(" bIsCallHalIsr=1: PIO OUT. \n");

                        {
                        struct cxd9225gg_req *udc_req;

                        UDC_STAT_ISR_STEP(__LINE__);

                        /* lock. */
                        spin_lock_irqsave(&udc_dev->lock, regIntDummy);

                        if (!list_empty(&udc_dev->ep[0].queue))
                        {
                            UDC_STAT_ISR_STEP(__LINE__);
                            udc_req = list_entry(udc_dev->ep[0].queue.next, struct cxd9225gg_req, queue);

                            udc_debug("begin to call PioDataRecieve. \n");
                            ret_val = usbPioDataRecieve(&udc_dev->ep[0], udc_req);
                            if (ret_val)
                            {
                                UDC_STAT_ISR_STEP(__LINE__);
                                udc_error("usbPioDataRecieve() failed. \n");
                            }
                            udc_debug("after of PioDataRecieve. \n");
                        }

                        /* unlock. */
                        spin_unlock_irqrestore(&udc_dev->lock, regIntDummy);
                        } //only to shut up warnings
                    }
                }
            } /* end of PIO OUT for EP0. */

            UDC_STAT_ISR_STEP(__LINE__);

            /* PIO IN of ep0. */
            if( 0 != pEpInfo0Send->bIsAssigned )
            {
                /* control ep0 in. */
                UDC_STAT_ISR_STEP(__LINE__);
                if ( 0 != ( regIntEP0Status & USB_EP0STS_IN_INT_BIT) )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    pTarget->uInterruptStatus |= USB_ENDPOINT_INTERRUPT_MASK;
                    pEpInfo0Send->uEndpointInterruptStatus |= USB_ENDPOINT_IN_PID_MASK;

                    pEpInfo0Send->ulPipeTransferType = USB_TRANSFER_TYPE_PIO;

                    udc_debug(" bIsCallHalIsr=1: PIO IN, that is, we need send some data to host later. \n");

                    /* notify usb-stack to do "usbFncCopyDataToEpBuf"  */
                    if ( USB_CTRL_TRANS_TYPE_UNDEF == pTarget->ulCtrlTransType )
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        udc_error("ulCtrlTransType is un-defined, and then switch from recv to send. \n");

                        /* control, bug fix 0Rcv -> 0Send. */
                        complete(&pEpInfo0Send->hSem4WaitForInterrupt);
                    }

                    {
                    struct cxd9225gg_req *udc_req;
                    
                    UDC_STAT_ISR_STEP(__LINE__);
                    
                    /* lock. */
                    spin_lock_irqsave(&udc_dev->lock, regIntDummy);

                    if (!list_empty(&udc_dev->ep[0].queue))
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        udc_req = list_entry(udc_dev->ep[0].queue.next, struct cxd9225gg_req, queue);
                        udc_debug("begin to call PioDataSend. \n");
                        ret_val = usbPioDataSend(&udc_dev->ep[0], udc_req);
                        if (ret_val)
                        {
                            UDC_STAT_ISR_STEP(__LINE__);
                            udc_error("usbPioDataSend() failed. \n");
                        }
                        udc_debug("after of PioDataSend. \n");
                     }

                    /* unlock. */
                    spin_unlock_irqrestore(&udc_dev->lock, regIntDummy);
                    } //only to shut up warnings.
                }

            } /* end of PIO IN for ep0. */

            /* DMA OUT: EpBuf <- Endpoint */
            UDC_STAT_ISR_STEP(__LINE__);
            if ( 0 != pEpInfo1Rcv->bIsAssigned )
            {
                UDC_STAT_ISR_STEP(__LINE__);
                if ( 1 == pEpInfo1Rcv->bBulkOutEnd )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    pEpInfo1Rcv->bBulkOutEnd = 0;

                    /* read ep1-status-reg and cleanup. */
                    USB_EP1STS_REG = ~USB_EPSTS_OUT_INT_BIT;
                    regIntDummy = USB_EP1STS_REG;

                    /* re-enable the ep1-out interrupt. */
                    USB_EP1INTEN_REG |= USB_EPINTEN_OUT_EN_BIT;

                    if ( USB_EPSTS_OUT_EMPTY_BIT != (USB_EP1STS_REG & USB_EPSTS_OUT_EMPTY_BIT) )
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        pEpInfo1Rcv->bIntInformPending = 1;

                        udc_debug("NotEmpty DMACTRL=0x%08x,DCR1=0x%08x,LENDCNT=0x%08x,STS=0x%08x. \n",
                                  (unsigned int)USB_EP1DMACTRL_REG,
                                  (unsigned int)USB_EP1DCR1_REG,
                                  (unsigned int)USB_EP1LENDCNT_REG,
                                  (unsigned int)USB_EP1STS_REG);
                    }
                }

                /* EndPoint 1 out interrupt? */
                UDC_STAT_ISR_STEP(__LINE__);
                if ( 0 != ( (regIntEP1Status & USB_EP1INTEN_REG) & USB_EPSTS_OUT_INT_BIT) )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    USB_EP1INTEN_REG &= ~USB_EPINTEN_OUT_EN_BIT;
                    USB_EP1STS_REG = ~USB_EPSTS_OUT_INT_BIT;
                    regIntDummy = USB_EP1STS_REG;                      /* Dummy Read */

                    udc_debug("------to notify the ep1-receive-data.................. \n");

                    /* notify ep1-receive dma out... */
                    complete(&pEpInfo1Rcv->hSem4WaitForInterrupt);
    
                    if ( 0 != pEpInfo1Rcv->bOutRequestIntFlag )
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                    }
                    else
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        pEpInfo1Rcv->bOutRequestIntFlag = USB_ENDPOINT_OUT_PID_MASK;
                        pTarget->uInterruptStatus |= USB_ENDPOINT_INTERRUPT_MASK;
                        pEpInfo1Rcv->uEndpointInterruptStatus |= USB_ENDPOINT_OUT_PID_MASK;

                        bIsCallHalIsr = 1;

                        /* lock, and modify... */
                        spin_lock_irqsave(&udc_dev->lock, udc_dev->m_flags);
                        udc_dev->ep_bulk_recv = 1;
                        spin_unlock_irqrestore(&udc_dev->lock, udc_dev->m_flags);

                        pTarget->uEp1IntEnableSts = 1;  /* for debug */
                    }
                }

                /* whether we received a NULL-OUT data. */
                UDC_STAT_ISR_STEP(__LINE__);
                if (regIntEP1Status & USB_EPSTS_OUT_NULL_INT_BIT)
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    udc_error(" ep1-receive a NULL data, do we need to notify usb-stack??? \n");
                }

                if ( 1 == pEpInfo1Rcv->bIntInformPending )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    if ( 0 == pEpInfo1Rcv->bOutRequestIntFlag )
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        if ( USB_EPSTS_OUT_EMPTY_BIT != (USB_EP1STS_REG & USB_EPSTS_OUT_EMPTY_BIT) )
                        {
                            UDC_STAT_ISR_STEP(__LINE__);
                            pEpInfo1Rcv->bOutRequestIntFlag = USB_ENDPOINT_OUT_PID_MASK;
                            pEpInfo1Rcv->bIntInformPending = 0;
                            pTarget->uInterruptStatus |= USB_ENDPOINT_INTERRUPT_MASK;
                            pEpInfo1Rcv->uEndpointInterruptStatus |= USB_ENDPOINT_OUT_PID_MASK;

                            bIsCallHalIsr = 1;

                            /* lock, and modify... */
                            spin_lock_irqsave(&udc_dev->lock, udc_dev->m_flags);
                            udc_dev->ep_bulk_recv = 1;
                            spin_unlock_irqrestore(&udc_dev->lock, udc_dev->m_flags);

                            pTarget->uEp1IntEnableSts = 2;  /* for debug */
                            udc_debug("EP1 OUT_INT defferred inform occured...... \n");
                        }
                        else
                        {
                            UDC_STAT_ISR_STEP(__LINE__);
                            udc_debug(" EP1STS_REG     :0x%08x\n", (unsigned int)USB_EP1STS_REG);
                            udc_debug(" EP1LENDCNT_REG :0x%08x\n", (unsigned int)USB_EP1LENDCNT_REG);
                            udc_error(" OUT_INT pend but no data!!\n");

                            pEpInfo1Rcv->bIntInformPending = 0;
                        }
                    }
                }

                UDC_STAT_ISR_STEP(__LINE__);

                do {
                int epc_is_end = 0, brg_is_end = 0;

                /* use spin_lock to save thread-switch, makeing the 
                 * following 2 complete() together. */
                spin_lock_irqsave(&udc_dev->lock, udc_dev->m_flags);

                /* EndPoint 1 DMA in end.  */
                if ( 0 != ( regIntEP1Status & USB_EPSTS_OUT_END_INT_BIT) )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    epc_is_end = 1;
                    UDC_STAT_EPC_CMPL_INC();
                    UDC_STAT_EPC_BACKUP();
                    pEpInfo2Send->bCswCompleteIntFlag = 0;
                    complete(&(pEpInfo1Rcv->hSemEpcDmaEndInt));
                }

                /* Bridge DMA end */
                if ( 0 != (regIntBridgeStatus & USB_AHBBINT_DMA_ENDINT_1_BIT ) )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    brg_is_end = 1;
                    UDC_STAT_BRG_CMPL_INC();
                    UDC_STAT_BRG_BACKUP();
                    pEpInfo1Rcv->bOutRequestIntFlag = 0;
                    complete(&(pEpInfo1Rcv->hSemBridgeDmaEndInt));
                }

                UDC_STAT_ISR_STEP(__LINE__);
                spin_unlock_irqrestore(&udc_dev->lock, udc_dev->m_flags);
                UDC_STAT_ISR_STEP(__LINE__);

                if (epc_is_end ^ brg_is_end) {
                    UDC_STAT_ISR_STEP(__LINE__);
                    udc_error("EpcDmaEnd=%d, BridgeDmaEnd=%d\n", epc_is_end, brg_is_end);
                }
                } while (0); //only for shut up warnings.

            } /* end of DMA OUT ep1-recv. */

            UDC_STAT_ISR_STEP(__LINE__);

            /* EP2 DMA IN  */
            if ( 0 != pEpInfo2Send->bIsAssigned )
            {
                UDC_STAT_ISR_STEP(__LINE__);
                if ( 0 != (pTarget->uInterruptAssign & USB_SEQUENCE_CTRL_INT) )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    /* clear ep-in NAK bit. */
                    USB_EP2STS_REG = ~USB_EPSTS_IN_NAK_ERR_INT_BIT;
                    regIntDummy = USB_EP2STS_REG;              /* Dummy Read */

                    USB_EP2INTEN_REG |= USB_EPINTEN_IN_NAK_ERR_EN_BIT;

                    if ( 1 == pEpInfo2Send->bIntInformPending )
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        pEpInfo2Send->bInRequestIntFlag = USB_ENDPOINT_IN_PID_MASK;
                        pEpInfo2Send->uEndpointInterruptStatus = USB_ENDPOINT_IN_PID_MASK;
                        pTarget->uInterruptStatus |= USB_ENDPOINT_INTERRUPT_MASK;

                        bIsCallHalIsr = 1;
                        lTimeout = 1;  /* 16ms. */

                        /* lock... */
                        spin_lock_irqsave(&udc_dev->lock, udc_dev->m_flags);
                        udc_dev->ep_bulk_send = 1;

                        if (udc_dev->ep_bulk_recv)
                        {
                            UDC_STAT_ISR_STEP(__LINE__);
                            udc_debug("error -- send buffer is ready to use, but recv buffer is also ready. \n");
                        }

                        /* unlock. */
                        spin_unlock_irqrestore(&udc_dev->lock, udc_dev->m_flags);

                        ulReInformCount++;
                        pEpInfo2Send->bIntInformPending = 0;

                        udc_debug(" EP2STS_IN_NAK_ERR_INT Exec!!\n");
                    }

                    UDC_STAT_ISR_STEP(__LINE__);

                    /* Interrupt Disable */
                    spin_lock_irqsave(&udc_dev->lock, regIntDummy);

                    pTarget->uInterruptAssign &= ~USB_SEQUENCE_CTRL_INT;

                    /* Interrupt Enable */
                    spin_unlock_irqrestore(&udc_dev->lock, regIntDummy);
                }

                UDC_STAT_ISR_STEP(__LINE__);

                if ( 0 != ( regIntEP2Status & USB_EPSTS_IN_INT_BIT) )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    udc_debug("ep2 buffer is ready to use, that is you can send next data out. \n");

                    /* lock... */
                    spin_lock_irqsave(&udc_dev->lock, udc_dev->m_flags);

                    udc_dev->ep_bulk_send = 1;

                    if (udc_dev->ep_bulk_recv)
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        udc_debug("error -- send buffer is ready to use, but recv buffer is also ready. \n");
                    }

                    /* unlock. */
                    spin_unlock_irqrestore(&udc_dev->lock, udc_dev->m_flags);
                }

                UDC_STAT_ISR_STEP(__LINE__);

                /* Bridge DMA IN end, EpBuff -> EP */
                if( 0 != (regIntBridgeStatus & USB_AHBBINT_DMA_ENDINT_2_BIT ) )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    udc_debug("DmaIn: AHBBINT_DMA_ENDINT_2. \n");

                    if( USB_DMA_IN_PH_EPCDMA_END == pEpInfo2Send->ulDmaInPhase )
                    {
                        pEpInfo2Send->ulDmaInPhase = USB_DMA_IN_PH_ALL_END;
                    }
                    else
                    {
                        pEpInfo2Send->ulDmaInPhase = USB_DMA_IN_PH_BRIDGEDMA_END;
                    }
                }

                UDC_STAT_ISR_STEP(__LINE__);

                /* EP2 IN  */
                if( 0 != ( regIntEP2Status & USB_EPSTS_IN_END_INT_BIT) )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    udc_debug("DmaIn: EPSTS_IN_END_INT. \n");

                    if( USB_DMA_IN_PH_BRIDGEDMA_END == pEpInfo2Send->ulDmaInPhase )
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        pEpInfo2Send->ulDmaInPhase = USB_DMA_IN_PH_ALL_END;
                    }
                    else
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        pEpInfo2Send->ulDmaInPhase = USB_DMA_IN_PH_EPCDMA_END;
                    }
                }

                UDC_STAT_ISR_STEP(__LINE__);

                if ( USB_DMA_IN_PH_ALL_END == pEpInfo2Send->ulDmaInPhase )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    udc_debug("DmaIn: PH_ALL_END, and notify DMA-In EP2 send. \n");

                    pEpInfo2Send->bInRequestIntFlag = 0;
                    pEpInfo2Send->bCswCompleteIntFlag = USB_ENDPOINT_IN_PID_MASK;

                    /* Semaphore?to notify usb-stack to do CopyDataToEpBuf. */
                    complete(&(pEpInfo2Send->hSem4WaitForInterrupt));

                    pTarget->uInterruptStatus |= USB_ENDPOINT_INTERRUPT_MASK;

                    pEpInfo2Send->ulDmaInPhase = USB_DMA_IN_PH_IDLE;
                }
                /* DMA IN  */

                UDC_STAT_ISR_STEP(__LINE__);

                /* check intr-ENABLE and assert for IN_NAK_ERR_INT.  */
                if ( 0 != ( (regIntEP2Status & USB_EP2INTEN_REG) & USB_EPSTS_IN_NAK_ERR_INT_BIT) )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    /* BulkInDm. */
                    USB_EP2INTEN_REG &= ~USB_EPINTEN_IN_NAK_ERR_EN_BIT;
                    udc_debug(" EP2STS_IN_NAK_ERR_INT occured\n");
                    ret_val = mutex_trylock(&(pTarget->hSemApiExecuting));
                    if (ret_val) //successfully.
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        pEpInfo2Send->bInRequestIntFlag = USB_ENDPOINT_IN_PID_MASK;
                        pEpInfo2Send->uEndpointInterruptStatus = USB_ENDPOINT_IN_PID_MASK;

                        /* release it. */
                        mutex_unlock(&(pTarget->hSemApiExecuting));

                        pTarget->uInterruptStatus |= USB_ENDPOINT_INTERRUPT_MASK;
                        bIsCallHalIsr = 1;
                        lTimeout = 1;   /* 16ms. */

                        /* lock... */
                        spin_lock_irqsave(&udc_dev->lock, udc_dev->m_flags);

                        udc_dev->ep_bulk_send = 1;
                        if (udc_dev->ep_bulk_recv)
                        {
                            UDC_STAT_ISR_STEP(__LINE__);
                            udc_debug("error -- send buffer is ready to use, but recv buffer is also ready. \n");
                        }

                        /* unlock. */
                        spin_unlock_irqrestore(&udc_dev->lock, udc_dev->m_flags);

                        ulReInformCount++;
                    }
                    else
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        /* BULK-IN or BULK-OUT pending. */
                        udc_error(" EP2STS_IN_NAK_ERR_INT Pend.....\n");
                        pEpInfo2Send->bIntInformPending = 1;
                    }
                }
            }

            UDC_STAT_ISR_STEP(__LINE__);

            /****** Control setup handling. ******/
            if ( (0 != pEpInfo0Rcv->bIsAssigned) && (0 != pEpInfo0Send->bIsAssigned) )
            {
                UDC_STAT_ISR_STEP(__LINE__);
                if( (USB_CTRL_TRANS_PH_SETUP_DATA_STAGE_PENDING == pTarget->ulCtrlTransPhase )
                    || (USB_CTRL_TRANS_PH_SETUP_STS_STAGE_PENDING == pTarget->ulCtrlTransPhase )
                   )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    udc_debug("CtrlTransPhase -- setup data/status STAGE_PENDING. \n");

                    if ( USB_ENDPOINT_SETUP_PID_MASK
                         != (pEpInfo0Rcv->uEndpointInterruptStatus & USB_ENDPOINT_SETUP_PID_MASK)
                        )
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        pTarget->ulCtrlTransPhase = USB_CTRL_TRANS_PH_SETUP_DATA_STAGE;
                        pTarget->uInterruptStatus |= USB_ENDPOINT_INTERRUPT_MASK;
                        pEpInfo0Rcv->uEndpointInterruptStatus |= USB_ENDPOINT_SETUP_PID_MASK;

                        /* Request setup pid.  */
                        udc_debug(" set SETUP_DATA_STAGE_PENDING -> DATA_STAGE. \n");
                    }
                    else
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        /* do nothing */
                    }
                }

                UDC_STAT_ISR_STEP(__LINE__);

                if ( 0 != ( regIntEP0Status & USB_EP0STS_SETUP_INT_BIT) )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    pTarget->bIsSetupRequested = 1;
                    pTarget->bIsCtrlTransOnGoing = 1;

                    if ( USB_CTRL_TRANS_PH_IDLE != pTarget->ulCtrlTransPhase )
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        pTarget->bIsCtrlTransFinished = 1;

                        /* notify setup.... */
                        complete(&(pTarget->hSem4SyncCtrlTransfer));

                        udc_debug(" SETUP Token Receive before STG END. \n");
                    }

                    UDC_STAT_ISR_STEP(__LINE__);

                    /*  SETUP  */
                    if( USB_ENDPOINT_SETUP_PID_MASK != (pEpInfo0Rcv->uEndpointInterruptStatus & USB_ENDPOINT_SETUP_PID_MASK) )
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        pTarget->ulCtrlTransPhase = USB_CTRL_TRANS_PH_SETUP_DATA_STAGE;
                        pTarget->uInterruptStatus |= USB_ENDPOINT_INTERRUPT_MASK;
                        pEpInfo0Rcv->uEndpointInterruptStatus |= USB_ENDPOINT_SETUP_PID_MASK;
                        udc_debug(" bIsCallHalIsr=1:SETUP, set SETUP_DATA_STAGE. \n");
                    }
                    else
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        pTarget->ulCtrlTransPhase = USB_CTRL_TRANS_PH_SETUP_DATA_STAGE_PENDING;
                        udc_debug(" set SETUP_DATA_STAGE_PENDING. \n");
                    }

                    udc_debug(" begin to call SetupPioDataRecieve. \n");

                    pTarget->bIsSetupRequested = 0;

                    /* PIO */
                    UDC_STAT_ISR_STEP(__LINE__);
                    ret_val = usbSetupPioDataRecieve(&udc_dev->ep[0], NULL);
                    udc_debug(" after of SetupPioDataRecieve. \n");

                }

                UDC_STAT_ISR_STEP(__LINE__);

                if ( 0 != ( regIntEP0Status & USB_EP0STS_STG_START_INT_BIT) )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    if( USB_CTRL_TRANS_PH_SETUP_DATA_STAGE_PENDING == pTarget->ulCtrlTransPhase )
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        pTarget->ulCtrlTransPhase = USB_CTRL_TRANS_PH_SETUP_STS_STAGE_PENDING;
                        udc_debug(" set SETUP_STS_STAGE_PENDING.\n");
                    }
                    else
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        pTarget->ulCtrlTransPhase = USB_CTRL_TRANS_PH_SETUP_STS_STAGE;
                        udc_debug("set SETUP_STS_STAGE. \n");
                    }

                    if( USB_CTRL_TRANS_READ != pTarget->ulCtrlTransType )
                    {
                        UDC_STAT_ISR_STEP(__LINE__);
                        udc_debug("ulCtrlTransType is NOT read, so notify receive. \n");
                        complete(&(pTarget->hSem4SyncCtrlTransfer));
                    }
                }

                UDC_STAT_ISR_STEP(__LINE__);

                if ( 0 != ( regIntEP0Status & USB_EP0STS_STG_END_INT_BIT) )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    pEpInfo0Send->uEndpointInterruptStatus |= USB_ENDPOINT_IN_PID_MASK;
                    pTarget->bIsCtrlTransOnGoing = 0;
                    pTarget->ulCtrlTransPhase = USB_CTRL_TRANS_PH_IDLE;
                    pTarget->bIsCtrlTransFinished = 1;

                    udc_debug("set SETUP_IDLE_STAGE. \n");
                }
            }

            UDC_STAT_ISR_STEP(__LINE__);

            if ( 0 != (regIntBridgeStatus & USB_AHBBINT_VBUS_INT_BIT ) )
            {
                UDC_STAT_ISR_STEP(__LINE__);
                udc_error(" invalid VBUS INT detection, regIntBridgeStatus:0x%0x. \n", (unsigned int)regIntBridgeStatus);
            }

            UDC_STAT_ISR_STEP(__LINE__);

            /* handle epc intr at last.  */
            if (1 == bIsHandledEpcInt)
            {
                UDC_STAT_ISR_STEP(__LINE__);
                bIsHandledEpcInt = 0;
                if( 0 != (USB_USBINTSTS_REG & USB_USBINTEN_INIT) )
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    bReCheckFlag = 1;
                    udc_debug("USBINTSTS_REG Re-Run (0x%08x). \n",
                              (unsigned int)USB_USBINTSTS_REG);
                }
                else
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    udc_debug("begin to restore usb-intr-enable register. \n");

                    /* Interrupt Disable */
                    spin_lock_irqsave(&udc_dev->lock, regIntDummy);

                    pTarget->uInterruptAssign &= ~USB_EPC_INT;
                    USB_USBINTEN_REG = pTarget->ulUsbIntEnBackup;

                    /* Interrupt Enable */
                    spin_unlock_irqrestore(&udc_dev->lock, regIntDummy);
                    udc_debug(" USBINTEN_REG 0x%08x <- 0x%08x. \n",
                              (unsigned int)USB_USBINTEN_REG,
                              (unsigned int)pTarget->ulUsbIntEnBackup);
                }
            }

            UDC_STAT_ISR_STEP(__LINE__);

            /* handle AHB-bridge intr at last.  */
            if (1 == bIsHandledBridgeInt)
            {
                UDC_STAT_ISR_STEP(__LINE__);
                bIsHandledBridgeInt = 0;

                if ( 0 != (USB_AHBBINT_REG & USB_AHBBINT_CHECK_MASK))
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    bReCheckFlag = 1;
                    udc_debug(" AHBBINT_REG Re-Run (0x%08x)\n", (unsigned int)USB_AHBBINT_REG);
                }
                else
                {
                    UDC_STAT_ISR_STEP(__LINE__);
                    udc_debug("begin to restore AHB-EPC-intr-enable register. \n");

                    /* Interrupt Disable */
                    spin_lock_irqsave(&udc_dev->lock, regIntDummy);

                    USB_AHBBINTEN_REG = pTarget->ulAhbbIntEnBackup;
                    pTarget->uInterruptAssign &= ~USB_BRIDGE_INT;

                    /* Interrupt Enable */
                    spin_unlock_irqrestore(&udc_dev->lock, regIntDummy);
                }
            }
        }   /* PlugFlg = 1 */
        else
        {
            UDC_STAT_ISR_STEP(__LINE__);
            if (pTarget->uInterruptAssign & USB_VBUS_TASK_INT) {
                udc_error("EPC Power State is OFF. Can NOT handle Interrupt.\n");
            }
            spin_lock_irqsave(&udc_dev->lock, regIntDummy);
            pTarget->uInterruptAssign = 0;
            spin_unlock_irqrestore(&udc_dev->lock, regIntDummy);
        }

        UDC_STAT_ISR_STEP(__LINE__);

        /* notify usb-stack to do something... */
        if ( 1 == bIsCallHalIsr )
        {
            UDC_STAT_ISR_STEP(__LINE__);
            udc_debug("notify usb-stack to do something. \n");

            notify_to_udc_daemon();
        }

    } /* end of main loop. */

    /* never reach here. */
    udc_error(" exit test. \n");

    return 0;
}


int usbVbusWaitTask(void *data)
{
    struct cxd9225gg_udc           *udc_dev;
    PUSBTARGET       pTarget    = NULL;     /* USBTARGET */
    int                         ret_val;
    unsigned long               ulTmp;
    unsigned long               ulCnt;
    PUSBENDPOINT     pEpInfo1Rcv;
    PUSBENDPOINT     pEpInfo2Send;
    long                        lTimeout = WAIT_FOREVER;
    int                         bSuspendChecking = 0;
    unsigned long               regIntDummy = 0;

    pTarget = (PUSBTARGET)data;

    pEpInfo1Rcv = *(pTarget->pEndPointInfoTop + 2);
    pEpInfo2Send = *(pTarget->pEndPointInfoTop + 3);

    /* main loop. */
    while (1)
    {
        /* wait for a semaphore to contiue... */
        if (WAIT_FOREVER == lTimeout)
        {
            UDC_STAT_ISR_STEP(__LINE__);
            ret_val = wait_for_completion_interruptible(&(pTarget->hSem4VbusWaitTask));
            UDC_STAT_ISR_STEP(__LINE__);
            if (-ERESTARTSYS == ret_val)
            {
                udc_error("it means this kthread--VbusWaitTask is killed. \n");
                return -1;
            }
        }
        else
        {
            UDC_STAT_ISR_STEP(__LINE__);
            ret_val = wait_for_completion_interruptible_timeout(
                          &(pTarget->hSem4VbusWaitTask), lTimeout);
            UDC_STAT_ISR_STEP(__LINE__);
            if (!ret_val || (-ERESTARTSYS == ret_val))
            {
                if (-ERESTARTSYS == ret_val)
                {
                    udc_error("it means this kthread--VbusWaitTask is killed--hSem4VbusWaitTask. \n");
                    return -1;
                }

                udc_error("it means timeout--hSem4VbusWaitTask. \n");
                ret_val = 1;
            }
            else
            {
                ret_val = 0;
            }
        }
        udc_debug("get a vbus semaphore. \n");

        lTimeout = WAIT_FOREVER;

        udc_dev = pTarget->udc_dev;
        if (!udc_dev)
        {
            udc_error("-------------udc_dev is NULL, to continue----------\n");
            continue;
        }

        /* whether was there a timeout just now. */
        if (0 != ret_val)
        {
            if ( 1 == bSuspendChecking )
            {
                udc_error("do suspend power-control now.\n");
                bSuspendChecking = 0;
#ifdef UDC_POWER_CTRL_ON_STANDBY
                /* I tried to open this block of power control logic, 
                 * but if it is open, when we remove this usb mass-storage 
                 * device from PC, our system will halt.
                 * And i do not know why...
                */
                mutex_lock(&g_UsbpTarget->hSemApiExecuting); 
                /* suspend power control */
                ret_val = usbPowerStateControl(&pTarget->ulPowState, 
                                                USB_POW_EVENT_SUSPEND, 
                                                &pTarget->bIsPllUnlocked); 
                mutex_unlock(&g_UsbpTarget->hSemApiExecuting);
#endif
            }
        }

        if (1 == pTarget->bSuspendRequested)
        {
            udc_error("handle a suspend-request. \n");

            pTarget->bSuspendRequested = 0;
            bSuspendChecking = 1;
            if ( (1 != pTarget->bVbusIntOccured) && (1 != pTarget->bResumeRequested) )
            {
                /* 166msec  Bus Reset?139.8msec No Bus Active.  */
                /* Disconnect?300msec?VBUS INT. */
                /*  500ms Resume/Reset/VBUS  Suspend  */
                lTimeout = msecs_to_jiffies(500);  /* 496msec */
            }
            usbFncHandleSuspendInterrupt(pTarget);
        }

        /* resume... */
        if ( 1 == pTarget->bResumeRequested )
        {
            udc_error("handle a resume-request. \n");

            /* semaphore PlugIn/PlugOut/Enable/Disable Resume */
            UDC_STAT_ISR_STEP(__LINE__);
            ret_val = wait_for_completion_interruptible_timeout(&pTarget->sUsbPlugInOutSemId,
                                                                msecs_to_jiffies(USB_FUNC_TIMEOUT));
            UDC_STAT_ISR_STEP(__LINE__);
            if (!ret_val || (-ERESTARTSYS == ret_val))
            {
                if (-ERESTARTSYS == ret_val)
                {
                    udc_error("it means this kthread--VbusWaitTask is killed, for PlugInOutSemId. \n");
                    return -1;
                }

                udc_error("it means timeout. \n");
                ulTmp= 0;
            }
            else
            {
                ulTmp = 1;
            }

            ret_val = usbPowerStateControl(&pTarget->ulPowState,
                                                USB_POW_EVENT_RESUME,
                                                &pTarget->bIsPllUnlocked);
            if (ret_val)
            {
                udc_error(" usbPowerStateControl()--resume failed. \n");
            }

            pTarget->bResumeRequested = 0;
            bSuspendChecking = 0;
            ret_val = usbFncEndpointAssign(udc_dev, 0);
            if (ret_val)
            {
                udc_error("usbFncEndpointAssign()--0 failed. \n");
            }

            ret_val = usbFncEndpointAssign(udc_dev, 1);
            if (ret_val)
            {
                udc_error("usbFncEndpointAssign()--1 failed. \n");
            }

            ret_val = usbFncEndpointAssign(udc_dev, 2);
            if (ret_val)
            {
                udc_error("usbFncEndpointAssign()--2 failed. \n");
            }
            usbFncHandleResumeInterrupt(pTarget);

            /* only if gotten just now, and then release it. */
            if (ulTmp)
            {
                complete(&pTarget->sUsbPlugInOutSemId);
            }
            /* attention:
             * we need also make udc-queue-thread to resume....
             */
        }
        /* reset request... */
        if ( 1 == pTarget->bResetRequested )
        {
            udc_debug("handle a reset-request. \n");

            pTarget->bResetRequested = 0;
            bSuspendChecking = 0;
            if ( 1 != pTarget->bIsPllUnlocked )
            {
                udc_debug(" bResetRequested is 1, to start... \n");
            }
        }

        /* whether is it a vbus-interrupt... */
        if ( 1 == pTarget->bVbusIntOccured )
        {
            pTarget->bVbusIntOccured = 0;
            bSuspendChecking = 0;

            udc_debug(" VbusInt Occured just now. \n");

            msleep(100);

            /* check with 20ms delay to get a stable status.  */
            ulTmp = USB_EPCTR_REG & USB_EPCTR_VBUS_LEVEL_BIT;
            for (ulCnt = 0; ulCnt < 3; ulCnt++)
            {
                msleep(20);

                if (ulTmp != (USB_EPCTR_REG & USB_EPCTR_VBUS_LEVEL_BIT))
                {
                    break;
                }
            }

            /* VBUS  */
            USB_AHBBINT_REG = USB_AHBBINT_VBUS_INT_BIT;
            regIntDummy = USB_AHBBINT_REG;     /* Dummy Read */
            USB_AHBBINTEN_REG |= USB_AHBBINTEN_VBUS_INTEN_BIT;

            if (ulCnt < 3)
            {
                udc_error(" VBUS Status is NOT stable now. \n");
            }
            else
            {
                /* VBUS  */
                if ( 0 != ulTmp )
                {
                    /* VBUS Level = 1 -> Connected */
                    if( USB_CONNECTED_STATUS != pTarget->usbConnectStatus )
                    {
                        /* Disconnect or UNDEF -> Connect */
                        udc_debug(" VBUS Status Disconnect --> Connect. \n");
                        pTarget->usbConnectStatus = USB_CONNECTED_STATUS;
                        pTarget->bDoPlugIn = 1;
                    }
                    else
                    {
                        /* Connect -> Connect */
                        udc_debug(" VBUS Status Connect --> Connect. \n");
                    }
                }
                else
                {
                    /* VBUS Level = 0 -> Disconnected */
                    if ( USB_DISCONNECTED_STATUS != pTarget->usbConnectStatus )
                    {
                        /* Connect or UNDEF -> Disconnect */
                        udc_debug(" VBUS Status Connect --> Disconnect. \n");
                        pTarget->bDoPlugOut = 1;
                        pTarget->usbConnectStatus = USB_DISCONNECTED_STATUS;
                        pTarget->ulUsbIntEnBackup = 0;
                        pTarget->ulAhbbIntEnBackup = USB_AHBBINTEN_VBUS_INTEN_BIT;
                        cxd9225gg_udc_set_mass_flag(0);
                        if ( 1 == pTarget->bIsPllUnlocked )
                        {
                            /* PLL?Unlocked?Resume?Suspend. */
                            udc_debug(" disconnect -> PLL resume\n");

                            usbPowerResume();
                            pTarget->bIsPllUnlocked = 0;
                        }
                    }
                    else
                    {
                        /* Disconnect -> Disconnect */
                        /* do nothing? */
                        udc_debug(" VBUS Status Disconnect --> Disconnect. \n");
                    }

                    usbFncHandleDisconnectInterrupt(pTarget);
                    ret_val = usbFncEndpointRelease(udc_dev, 0);
                    if (ret_val)
                    {
                        udc_error("usbFncEndpointRelease()--0 failed. \n");
                    }

                    ret_val = usbFncEndpointRelease(udc_dev, 1);
                    if (ret_val)
                    {
                        udc_error("usbFncEndpointRelease()--1 failed. \n");
                    }

                    ret_val = usbFncEndpointRelease(udc_dev, 2);
                    if (ret_val)
                    {
                        udc_error("usbFncEndpointRelease()--2 failed. \n");
                    }
                    notify_to_udc_daemon();
                    usbFncDisable(pTarget);
                    ret_val = usbHwReInit();
                    if (ret_val)
                    {
                        udc_error("usbHwReInit failed. \n");
                    }
                    /* reinitialize the software resources. */
                    ret_val = usbReInit(pTarget);
                    if (ret_val)
                    {
                        udc_error("usbReInit failed. \n");
                    }

                    /* to notify application -- usb plug-out. */
                    usbConnectionNotify(0);
                }
            }

            spin_lock_irqsave(&udc_dev->lock, regIntDummy);
            pTarget->uInterruptAssign |= USB_VBUS_TASK_INT;
            spin_unlock_irqrestore(&udc_dev->lock, regIntDummy);

            complete(&pTarget->sUsbIntSemId);
        }   /* VbusInt End */
    }   /* end of main loop. */

    /* never reach here. */
    udc_error(" exit test. \n");

    return 0;
}


int usbFncInterruptStatusGet(PUSBTARGET pTarget)
{
    /*
    bit 0 disconnect interrupt
    bit 1 reset interrupt
    bit 2 suspend interrupt
    bit 3 resume interrupt
    bit 4 endpoint interrupt
    */
    udc_debug("entry test. \n");
    udc_debug(" uInterruptStatus: 0x%08x. \n", (unsigned int)pTarget->uInterruptStatus);

    return 0;
}


int usbFncInterruptStatusClear(PUSBTARGET pTarget)
{
    udc_debug("entry test. \n");
    return 0;
}


int usbFncEndpointIntStatusGet(PUSBTARGET pTarget, int ep_num)
{
    PUSBENDPOINT pEndpointInfo = NULL; /* PUSB___ENDPOINT */

    /*
     bit zero   endpoint Interrupt
     bit one    setup Interrupt
     bit two    OUT interrupt
     bit three  IN interrupt
     bit11-4
         USB_ENDPOINT_TRANSFER_SUCCESS 0x0000
         USB_ENDPOINT_DATA_TOGGLE_ERROR 0x00E0
         USB_ENDPOINT_PID_MISMATCH 0x00F0
         USB_ENDPOINT_COMMUN_FAULT 0x0100
         USB_ENDPOINT_STALL_ERROR 0x0110
         USB_ENDPOINT_DATA_OVERRUN 0x0120
         USB_ENDPOINT_DATA_UNDERRUN 0x0130
         USB_ENDPOINT_CRC_ERROR 0x0140
         USB_ENDPOINT_TIMEOUT_ERROR 0x0150
         USB_ENDPOINT_BIT_STUFF_ERROR 0x0160
    */

    pEndpointInfo = *(pTarget->pEndPointInfoTop + ep_num);
    if ( USB_EP1_RX == pEndpointInfo->endpointIndex )
    {
        udc_debug(" using bOutRequestIntFlag!! \n");
        udc_debug(" EP1_RX IntStatus=0x%08x,LENDCNT=0x%08x,STS=0x%08x. \n",
                  (unsigned int)pEndpointInfo->bOutRequestIntFlag,
                  (unsigned int)USB_EP1LENDCNT_REG,
                  (unsigned int)USB_EP1STS_REG);
    }
    else if( USB_EP2_TX == pEndpointInfo->endpointIndex )
    {
        udc_debug("EP2_TX IntStatus=0x%08x,LENDCNT=0x%08x,STS=0x%08x. \n",
                  (unsigned int)(pEndpointInfo->bCswCompleteIntFlag | pEndpointInfo->bInRequestIntFlag),
                  (unsigned int)USB_EP2LENDCNT_REG,
                  (unsigned int)USB_EP2STS_REG);
        pEndpointInfo->uEndpointInterruptStatus = 0;
    }
    else
    {
        udc_error(" endpointIndex ERROR (%u)\n", pEndpointInfo->endpointIndex);
    }

    return 0;
}


int usbFncEndpointIntStatusClear(PUSBTARGET pTarget, int ep_num)
{
    PUSBENDPOINT pEndpointInfo = NULL; /* PUSB___ENDPOINT */

    udc_debug("entry test. \n");

    pEndpointInfo = *(pTarget->pEndPointInfoTop + ep_num);
    if ( USB_EP2_TX == pEndpointInfo->endpointIndex )
    {
        udc_debug("EP2_TX IntStatus=0x%08x,LENDCNT=0x%08x,STS=0x%08x\n",
                  (unsigned int)(pEndpointInfo->bCswCompleteIntFlag | pEndpointInfo->bInRequestIntFlag),
                  (unsigned int)USB_EP2LENDCNT_REG,
                  (unsigned int)USB_EP2STS_REG);
    }
    else
    {
        pEndpointInfo->uEndpointInterruptStatus = 0;
    }

    if ( USB_EP1_RX == pEndpointInfo->endpointIndex )
    {
        udc_debug("ER1_RX DMACTRL=0x%08x, DCR1=0x%08x, LENDCNT=0x%08x, STS=0x%08x. \n",
                   (unsigned int)USB_EP1DMACTRL_REG,
                   (unsigned int)USB_EP1DCR1_REG,
                   (unsigned int)USB_EP1LENDCNT_REG,
                   (unsigned int)USB_EP1STS_REG);
    }

    if ( (USB_EP0_TX == pEndpointInfo->endpointIndex)
          || (USB_EP0_RX == pEndpointInfo->endpointIndex)
        )
    {
        udc_debug(" for ep0 rx or tx, send a int sema to usb-isr task. \n");
        complete(&(pTarget->sUsbIntSemId));
    }

    udc_debug("exit test. \n");

    return 0;
}


int usbFncHandleDisconnectInterrupt(PUSBTARGET pTarget)
{
    int ret_val = -1;
    unsigned long ulTmp;

    udc_debug("entry, pTarget->usbConnectStatus = %u. \n", pTarget->usbConnectStatus);

    if (pTarget->udc_dev->driver
        && pTarget->udc_dev->driver->disconnect)
    {
        udc_debug("call the gadget-driver's disconnect. \n");
        pTarget->udc_dev->driver->disconnect(&pTarget->udc_dev->gadget);
    }
    else
    {
        udc_error(" some arguments are invalid. \n");
    }

    UDC_STAT_ISR_STEP(__LINE__);
    ret_val = wait_for_completion_interruptible_timeout(&pTarget->sUsbPlugInOutSemId,
                                                        msecs_to_jiffies(USB_FUNC_TIMEOUT));
    UDC_STAT_ISR_STEP(__LINE__);
    if (!ret_val || (-ERESTARTSYS == ret_val))
    {
        if (-ERESTARTSYS == ret_val)
        {
            udc_error("it means this thread is killed, for PlugInOutSemId. \n");
            return -1;
        }

        udc_error("it means timeout. \n");
        ulTmp= 0;
    }
    else
    {
        ulTmp = 1;
    }

    /* usbPlugOut */
    udc_debug(" PlugOut Called. \n");

    ret_val = usbPlugOut( USB_POW_EVENT_PLUGOUT );
    if (ret_val)
    {
        udc_error(" usbPlugOut() failed. \n");
    }

    if (ulTmp)
    {
        complete(&(pTarget->sUsbPlugInOutSemId));
    }

    return ret_val;
}


int usbFncHandleResumeInterrupt(PUSBTARGET pTarget)
{
    udc_debug("entry test. \n");

    if ( USB_EPCTR_PLL_LOCK_BIT != (USB_EPCTR_REG & USB_EPCTR_PLL_LOCK_BIT) )
    {
        udc_error(" PLL is unlocked. \n");
        return -1;
    }

    return 0;
}


#define DEF_USB_START_ENDPOINT     1
#define DEF_USB_LAST_ENDPOINT      15

int usbFncHandleResetInterrupt(PUSBTARGET pTarget)
{
    unsigned long                   loopCount       = 0;
    unsigned long                   ulPipeIndex;
    PUSBENDPOINT*        pPtrPipeInfo;
    PUSBENDPOINT         pPipeInfo;
    unsigned long                   ulEndpointNumber;
    unsigned long                   ulRegOffset = 0;
    unsigned long                   regIntDummy = 0;

    if (!pTarget->udc_dev || !pTarget->udc_dev->driver)
    {
        udc_error("udc or its gadget-driver is NULL. \n");
        return -1;
    }

    /* FS->HS?USB_SPMODE_POLLING_COUNT. */
    for ( loopCount = 0; loopCount < USB_SPMODE_POLLING_COUNT; loopCount++ )
    {
        if ( 0 != (USB_USBINTSTS_REG & USB_USBINTSTS_SPEED_MODE_INT_BIT) )
        {
            USB_USBINTSTS_REG &= USB_USBINTSTS_SPEED_MODE_INT_BIT;
            regIntDummy = USB_USBINTSTS_REG;   /* Dummy Read */

            udc_debug(" SPEED_MODE_INT occurred. \n");
            break;
        }
    }

    pPtrPipeInfo = pTarget->pEndPointInfoTop;

    /* MaxPacketSize?EPn */
    if ( USB_USBSTS_SPEED_MODE_BIT == (USB_USBSTS_REG & USB_USBSTS_SPEED_MODE_BIT) )
    {
        pTarget->udc_dev->driver->speed = USB_SPEED_HIGH;
        pTarget->udc_dev->gadget.speed = USB_SPEED_HIGH;
        udc_debug(" DETECT HIGH SPEED MODE. \n");
    }
    else
    {
        pTarget->udc_dev->driver->speed = USB_SPEED_FULL;
        pTarget->udc_dev->gadget.speed = USB_SPEED_FULL;
        udc_debug(" DETECT FULL SPEED MODE. \n");
    }

    /* EPn_EN */
    for ( ulPipeIndex = 0; ulPipeIndex < USB_NUM_ENDPOINTS; ulPipeIndex++ )
    {
        if (NULL == (pPtrPipeInfo + ulPipeIndex))
        {
            udc_error("pPtrPipeInfo + ulPileIndex is NULL. \n");
            break;
        }

        pPipeInfo = *(pPtrPipeInfo + ulPipeIndex);
        if(NULL == pPipeInfo)
        {
            udc_error(" pPipeInfo is NULL. \n");
            break;
        }

        pPipeInfo->uEndpointInterruptStatus = 0;

        if ( 1 == pPipeInfo->bIsAssigned )
        {
            udc_debug(" ulPipeIndex[%lu] is assigned. \n", ulPipeIndex);
            ulEndpointNumber = pPipeInfo->endpointIndex & USB_EP_INDEX_MASK;
            if ( 0 != ulEndpointNumber )
            {
                ulRegOffset = ((ulEndpointNumber>>4) - 1) << 5;    /* EP1?Reg  */

                udc_debug("ulEndpointNumber=%lu,  ulRegOffset=0x%08x. \n",
                          ulEndpointNumber>>4, (unsigned int)ulRegOffset);

                /* EPn_EN=0b */
                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0040 + ulRegOffset)) &= ~USB_EPCTRL_EN_BIT;
                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0050 + ulRegOffset)) &= ~USB_MAXPACKET_MASK;

                if (USB_SPEED_HIGH == pTarget->udc_dev->driver->speed)
                {
                    (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0050 + ulRegOffset)) |= USB_MPKT_HS;
                    pPipeInfo->maxPacketSize = USB_MPKT_HS;    /* pPipeInfo->maxPacketSize */
                }
                else
                {
                    (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0050 + ulRegOffset)) |= USB_MPKT_FS;
                    pPipeInfo->maxPacketSize = USB_MPKT_FS;    /* pPipeInfo->maxPacketSize  */
                }

                /* EPn_EN=1b */
                (*(volatile unsigned long *)(USB_TC_OFFSET + 0x0040 + ulRegOffset)) |= USB_EPCTRL_EN_BIT;
            }
        }
    }

    /* (USB Control Register CONF 1b  */
    USB_USBCTRL_REG |= USB_USBSTS_CONF_BIT;
    return 0;
}


int usbFncHandleSuspendInterrupt(PUSBTARGET pTarget)
{
    udc_debug("entry test. \n");
    udc_debug("*** GOTO SUSPEND (do not PLL unlock ) ***\n");
    return 0;
}
