/*
 * Copyright (C) 2012 Sony Corporation
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>

#include <linux/usb.h>
#include <linux/usb/gadget.h>
#include <linux/dma-mapping.h>

#include "cxd9225gg_udc.h"
#include "cxd9225gg_usbExt.h"
#include "cxd9225gg_usbEvalLib.h"

extern int usbEpcReset(void);

int usbDiagReg(void);
int usbDiagPower(void);


/**
*
* usbDiagReg - Register Read/Write
*
* When diagnosis ID is 0x6, register Read/Write of EPC is executed.
*
* RETURNS: OK or ERROR.
*
*/
int usbDiagReg(void)
{
    int ret;
    unsigned long  uTcReg = 0;

    /* EPC  */
    ret = usbEpcReset();
    if (ret)
    {
        udc_error("usbEpcReset() failed. \n");
        return -1;
    }

    uTcReg = USB_SCRATCHPAD_REG;
    if (USB_SCRATCH_PAD_RST != uTcReg)
    {
        udc_error("[FATAL]SCRATCHPAD_REG READ ERROR (0x%08x). \n", (unsigned int)uTcReg);
        return -1;
    }

    USB_SCRATCHPAD_REG = USB_SCRATCH_PAD_ALL_F;
    uTcReg = USB_SCRATCHPAD_REG;
    if(USB_SCRATCH_PAD_ALL_F != uTcReg)
    {
        udc_error("[FATAL]SCRATCHPAD_REG WRITE ERROR (0x%08x). \n", (unsigned int)uTcReg);
        return -1;
    }

    return 0;
}


/**
*
* usbDiagPower - Check the uPD720122 Power Switch
*
* When diagnosis ID is 0x14, uPD720122 is good at power supply Off/On.
*
* RETURNS: OK or ERROR.
*
*/
int usbDiagPower(void)
{
    int  ret = 0;

    if( USB_EPCTR_VBUS_LEVEL_BIT != (USB_EPCTR_REG & USB_EPCTR_VBUS_LEVEL_BIT) )
    {
        ret = usbReturnfromDirectPowerDown();
        if (ret)
        {
            udc_error("[FATAL]ReturnfromDirectPowerDown Error=%d. \n", ret);
            return -1;
        }
        usbDirectPowerDown();
    }
    else
    {
        ret = usbPowerResume();
        if (ret)
        {
            udc_error("[FATAL]usbPowerResume() failed. \n");
            return -1;
        }
        USB_USBCTRL_REG |= USB_USBCTRL_SUSPEND_BIT;
    }

    return 0;
}
