/*
 * Copyright (C) 2012 Sony Corporation
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <asm/uaccess.h> 
#include <asm/io.h> 

char __iomem *boottime_addr;
unsigned long long userspace_time[8];
extern unsigned long long boottime_tmp[4];

#define BOOTTIMELOGPHYADDR  0x1c760000
#define BOOTTIMELOGSIZE  4*1024
#define BOOTTIMEOFFSET 50

/*  this function is copied directly from uptime.c */
static int _proc_calc_metrics(char *page, char **start, off_t off,
				 int count, int *eof, int len)
{
	if (len <= off + count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;
	return len;
}

static int genesis_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	/*  measure elapsed time from the genesis of kernel */
	int len;
	unsigned long long elapsed;
	unsigned long long under_sec;

	/* 
	 * we return 0 immediately because we know the string will never 
	 * be longer than 3KB, which is the value of count. thus, the caller 
	 * never needs to call this read_func more than once.
	*/
	if (off>0) {return 0;}

	elapsed = sched_clock();
	under_sec = do_div(elapsed, 1000000000);

	len = sprintf(page,"%llu.%09llu\n", elapsed, under_sec);

	printk(KERN_DEBUG "%s(): %s", __func__, page);

	return _proc_calc_metrics(page, start, off, count, eof, len);
}

static int boottime_addr_init(void)
{	
	
	boottime_addr = ioremap_nocache(BOOTTIMELOGPHYADDR, BOOTTIMELOGSIZE);
	if(!boottime_addr)
	{
		printk(KERN_ERR "ioremap_nocache err!!\n");
		return -1;
	}
	
	memset(boottime_addr + sizeof(unsigned long)*2, 0x00, 1024);

	return 0;
}

static int boot_time_open_proc(struct inode *inode, // IN: Inode of the proc file
                                         struct file *file)   // IN: File for this process' proc file access
{
    return 0;
}


static int boot_time_read_proc(char *page, char **start, off_t off,int count, int *eof, void *data)
{

	unsigned long tk_start;
	unsigned long tk_end;
	unsigned long mLinux_start;
	unsigned long mLinux_end;
	unsigned long mMbpinit_start;
	unsigned long mMbpinit_end;
	unsigned long mMount_start;
	unsigned long mMount_end;
	unsigned long rec_ready;
	unsigned long rec_start;
	unsigned long sec;
	unsigned long usec;
	int len[10];
	int total_len;

	tk_start = *(unsigned long *)boottime_addr;
	tk_end = *((unsigned long *)boottime_addr+1);
	
	sec = tk_start/1000000;
	usec = tk_start%1000000;

	len[0] = sprintf(boottime_addr+BOOTTIMEOFFSET, "\tT-Kernel Start:\t\t%lu.%lu\t-\n",sec, usec);
	total_len = len[0];
	
	sec = tk_end/1000000;
	usec = tk_end%1000000;
	
	len[1] = sprintf(boottime_addr+BOOTTIMEOFFSET+total_len, "\tT-Kernel End:\t\t%lu.%lu\t%lu\n",
		             sec, usec,(tk_end-tk_start)/1000);
	total_len += len[1];

	mLinux_start = boottime_tmp[0]*1000000 + boottime_tmp[1]/1000;
	len[2] = sprintf(boottime_addr+BOOTTIMEOFFSET+total_len, "\tLinux_Kernel Start:\t%llu.%09llu\t%lu\n",
		             boottime_tmp[0],boottime_tmp[1],(mLinux_start-tk_end)/1000);
	total_len += len[2];

	mLinux_end = boottime_tmp[2]*1000000 + boottime_tmp[3]/1000;
	len[3] = sprintf(boottime_addr+BOOTTIMEOFFSET+total_len, "\tLinux_Kernel End:\t%llu.%09llu\t%lu\n",
		             boottime_tmp[2],boottime_tmp[3],(mLinux_end-mLinux_start)/1000);
	total_len += len[3];

	mMbpinit_start = userspace_time[0]*1000000 + userspace_time[1]/1000;
	len[4] = sprintf(boottime_addr+BOOTTIMEOFFSET+total_len, "\tsytmInitialize Start:\t%llu.%09llu\t%lu\n",
		             userspace_time[0],userspace_time[1],((mMbpinit_start == 0)? mMbpinit_start:(mMbpinit_start-mLinux_end)/1000));	
	total_len += len[4];

	mMbpinit_end = userspace_time[2]*1000000 + userspace_time[3]/1000;
	len[5] = sprintf(boottime_addr+BOOTTIMEOFFSET+total_len, "\tsytmInitialize End:\t%llu.%09llu\t%lu\n", 
		             userspace_time[2],userspace_time[3],(mMbpinit_end-mMbpinit_start)/1000);
	total_len += len[5];

	mMount_start = userspace_time[4]*1000000 + userspace_time[5]/1000;
	len[6] = sprintf(boottime_addr+BOOTTIMEOFFSET+total_len, "\tMount Start:\t%llu.%09llu\t%lu\n", 
		             userspace_time[4],userspace_time[5],((mMount_start == 0)? mMount_start:(mMount_start-mMbpinit_end)/1000));	
	total_len += len[6];

	mMount_end = userspace_time[6]*1000000 + userspace_time[7]/1000;
	len[7] = sprintf(boottime_addr+BOOTTIMEOFFSET+total_len, "\tMount End:\t%llu.%09llu\t%lu\n", 
		             userspace_time[6],userspace_time[7],(mMount_end-mMount_start)/1000);
	total_len += len[7];

	
	rec_ready = *((unsigned long *)boottime_addr+2);
	rec_start = *((unsigned long *)boottime_addr+3);

	sec = rec_ready/1000000;
	usec = rec_ready%1000000;

	len[8] = sprintf(boottime_addr+BOOTTIMEOFFSET+total_len, "\tRec Ready:\t\t%lu.%lu\t%lu\n",
		             sec, usec,(rec_ready == 0)? rec_ready : (rec_ready-mMount_end)/1000);
	total_len += len[8];
	
	sec = rec_start/1000000;
	usec = rec_start%1000000;

	len[9] = sprintf(boottime_addr+BOOTTIMEOFFSET+total_len, "\tRec Start:\t\t%lu.%lu\t%lu\n",sec, usec,(rec_start-rec_ready)/1000);
	total_len += len[9];
	
    return _proc_calc_metrics(boottime_addr+BOOTTIMEOFFSET, start, off, count, eof, total_len);
}

static ssize_t boot_time_write_proc(struct file *file, const char __user *buf, size_t size, loff_t *ppos)
{

	static int cnt = 0;
	unsigned long long boottime;
	unsigned long long _nsec;
	unsigned long long _under_sec;

	if (cnt >= 8) { return size; }

	boottime = sched_clock();
	_nsec = boottime;
	_under_sec = do_div(_nsec, 1000000000);

	userspace_time[cnt] = _nsec;
	userspace_time[cnt+1] = _under_sec;
	cnt += 2;
	
    return size;
}

struct file_operations fops = {
    .open = boot_time_open_proc,
    .write = boot_time_write_proc,
}; 

static int __init proc_genesis_init(void)
{
	create_proc_read_entry("genesis", 0, NULL, genesis_read_proc, NULL);
	proc_create("boot_time", 0, NULL, &fops);
	create_proc_read_entry("print_boot_time", 0, NULL, boot_time_read_proc, NULL);
	boottime_addr_init();
	return 0;
}
module_init(proc_genesis_init);

