/* 2014-08-28: File changed by Sony Corporation */
/*
 * ALTERNATIVELY, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") as published by the Free Software
 * Foundation, either version 2 of that License or (at your option) any
 * later version.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



#ifndef _USB_EXTCMD_H
#define _USB_EXTCMD_H

/**************/
/*   Macro    */
/**************/
#define EXTCMD_FSG_NUM_BUFFERS  2   // Caution : MUST be the sama value with FSG_NUM_BUFFERS in storage_common.c
#define EXTCMD_MAX_COMMAND_SIZE 16  // Caution : MUST be the sama value with MAX_COMMAND_SIZE in storage_common.c


/**************/
/*    Enum    */
/**************/
enum fsg_buffer_state {
    BUF_STATE_EMPTY = 0,
    BUF_STATE_FULL,
    BUF_STATE_BUSY
};

enum fsg_state {
    /* This one isn't used anywhere */
    FSG_STATE_COMMAND_PHASE = -10,
    FSG_STATE_DATA_PHASE,
    FSG_STATE_STATUS_PHASE,

    FSG_STATE_IDLE = 0,
    FSG_STATE_ABORT_BULK_OUT,
    FSG_STATE_RESET,
    FSG_STATE_INTERFACE_CHANGE,
    FSG_STATE_CONFIG_CHANGE,
    FSG_STATE_DISCONNECT,
    FSG_STATE_EXIT,
    FSG_STATE_TERMINATED
};

enum data_direction {
    DATA_DIR_UNKNOWN = 0,
    DATA_DIR_FROM_HOST,
    DATA_DIR_TO_HOST,
    DATA_DIR_NONE
};


/**************/
/*   Struct   */
/**************/
/* file storage lun info */
struct fsg_lun {
    struct file     *filp;
    loff_t          file_length;
    loff_t          num_sectors;

    unsigned int    initially_ro:1;
    unsigned int    ro:1;
    unsigned int    removable:1;
    unsigned int    cdrom:1;
    unsigned int    prevent_medium_removal:1;
    unsigned int    registered:1;
    unsigned int    info_valid:1;

#ifdef CONFIG_USB_GADGET_CXD9225GG
    /* the hard sector size of storage file. */
    unsigned int    hd_sec_size_shift;
#endif

    u32             sense_data;
    u32             sense_data_info;
    u32             unit_attention_data;

    struct device   dev;
};

/* file storage buffer info */
struct fsg_buffhd {
#ifdef FSG_BUFFHD_STATIC_BUFFER
    char                    buf[FSG_BUFLEN];
#else
    void                    *buf;
#endif
    enum fsg_buffer_state   state;
    struct fsg_buffhd       *next;

    /* The NetChip 2280 is faster, and handles some protocol faults
     * better, if we don't submit any short bulk-out read requests.
     * So we will record the intended request length here. */
    unsigned int            bulk_out_intended_length;

    struct usb_request      *inreq;
    int                     inreq_busy;
    struct usb_request      *outreq;
    int                     outreq_busy;
};

/* file storage device info */
struct fsg_dev {
    /* lock protects: state, all the req_busy's, and cbbuf_cmnd */
    spinlock_t          lock;
    struct usb_gadget   *gadget;

    /* filesem protects: backing files in use */
    struct rw_semaphore filesem;

    /* reference counting: wait until all LUNs are released */
    struct kref         ref;

    struct usb_ep       *ep0;           // Handy copy of gadget->ep0
    struct usb_request  *ep0req;        // For control responses
    unsigned int        ep0_req_tag;
    const char          *ep0req_name;

    struct usb_request  *intreq;        // For interrupt responses
    int                 intreq_busy;
    struct fsg_buffhd   *intr_buffhd;

    unsigned int        bulk_out_maxpacket;
    enum fsg_state      state;          // For exception handling
    unsigned int        exception_req_tag;

    u8                  config, new_config;

    unsigned int        running : 1;
    unsigned int        bulk_in_enabled : 1;
    unsigned int        bulk_out_enabled : 1;
    unsigned int        intr_in_enabled : 1;
    unsigned int        phase_error : 1;
    unsigned int        short_packet_received : 1;
    unsigned int        bad_lun_okay : 1;

    unsigned long       atomic_bitflags;
#define REGISTERED      0
#define IGNORE_BULK_OUT 1
#define SUSPENDED       2

    struct usb_ep       *bulk_in;
    struct usb_ep       *bulk_out;
    struct usb_ep       *intr_in;

    struct fsg_buffhd   *next_buffhd_to_fill;
    struct fsg_buffhd   *next_buffhd_to_drain;
    struct fsg_buffhd   buffhds[EXTCMD_FSG_NUM_BUFFERS];

    int                 thread_wakeup_needed;
    struct completion   thread_notifier;
    struct task_struct  *thread_task;

    int                 cmnd_size;
    u8                  cmnd[EXTCMD_MAX_COMMAND_SIZE];
    enum data_direction data_dir;
    u32                 data_size;
    u32                 data_size_from_cmnd;
    u32                 tag;
    unsigned int        lun;
    u32                 residue;
    u32                 usb_amount_left;

    /* The CB protocol offers no way for a host to know when a command
     * has completed.  As a result the next command may arrive early,
     * and we will still have to handle it.  For that reason we need
     * a buffer to store new commands when using CB (or CBI, which
     * does not oblige a host to wait for command completion either). */
    int                 cbbuf_cmnd_size;
    u8                  cbbuf_cmnd[EXTCMD_MAX_COMMAND_SIZE];

    unsigned int        nluns;
    struct fsg_lun      *luns;
    struct fsg_lun      *curlun;
};

/* ExtCmd driver I/F */
struct f_storage_ext_opcode_driver {
    u8  opcode;                                                 // opcode
    int (*receive)(struct fsg_dev *fsg, struct fsg_buffhd *bh); // function for opcode
};


/**************/
/*   Extern   */
/**************/
int register_ext_opcode_driver(struct f_storage_ext_opcode_driver * drv);
int unregister_ext_opcode_driver(struct f_storage_ext_opcode_driver * drv);

#endif  /* _USB_EXTCMD_H */
